/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An implementation of an <small>AVL</small> tree.  Since an <small>AVL</small> tree is an ordered binary
 *  tree it inherits from <code>OrderedBinaryTree</code>.
 *
 *  <p>An <code>AVLTree</code> is an <code>OrderedBinaryTree</code> that maintains a balance of the nodes so
 *  as to minimize the depth.  On each alteration of the <code>AVLTree</code>, there is a rebalancing
 *  action.  Because all the actual <code>Node</code> activity is in the <code>Node</code>s the
 *  implementation is relatively straightforward.</p>
 *
 *  <p>There is a slight change of strategy in <code>AVLTree</code> compared with
 *  <code>OrderedBinaryTree</code>.  Addition and removal of items in a <code>OrderedBinaryTree</code> is
 *  achieved by relying on the <code>SearchIterator</code> and then operation using the resultant iterator.
 *  Addition and removal in the <code>AVLTree</code> relies on putting recursive algorithms into the
 *  <code>AVLNode</code>.  This change is used to ensure that proper balancing happens.</p>
 *
 *  @author Russel Winder
 */
public class AVLTree<T> extends OrderedBinaryTree<T> {
  /**
   *  An <code>AVLNode</code> for an <code>AVLTree</code>.  Since such nodes are binary tree nodes we
   *  inherit from <code>AbstractBinaryTree</code>.<code>Node</code>.
   *
   *  <p> <code>AVLNode</code> are highly "intelligent".  All the work of inserting and deleting is handled
   *  in each node so that it can manage the balancing of the sub-tree rooted at that place.  Add the
   *  balance factor as part of the state.</p>
   *
   *  <p> The nodes hold a balance factor which indicates whether the left or right child is "heavier", -1
   *  => left, 0 => balanced, +1 => right.  There can be no other factors due to the balancing!</p>
   */
  //  This class is an inner class (unlike Node in AbstractBinaryTree) since the methods need access to the
  //  order relation and equality predicate in the enclosing AVLTree.
  protected class AVLNode extends Node<T> {
    /**
     *  The extra state we are adding to that that the <code>Node</code> already has.
     */
    protected int balance ;
    /**
     *  Constructor of <code>AVLNode</code>.
     *
     *  <p>Since the root item of the <code>OrderedBinaryTree</code> that is our superclass (actually it's
     *  the <code>AbstractBinaryTree</code>) is a <code>Node</code> and not an <code>AVLNode</code>, we must
     *  allow the parent value to be a <code>Node</code>.</p>
     *
     *  <p>Since the root item of the <code>Abstract BinaryTree</code> is a <code>Node</code> and not an
     *  <code>AVLNode</code>, we have to allow <code>Node</code> as the parent parameter rather than
     *  restricting to <code>AVLNode</code> as might be tried at first.</p>
     *
     *  @param o the datum for this node.
     *  @param l the left subtree of this node.
     *  @param r the right subtree of this node.
     *  @param p the parent node of this subtree.
     *  @param b the balance for this node.
     */
    public AVLNode ( final T o , final AVLNode l , final AVLNode r , final Node<T> p , final int b ) {
      super ( o , l , r , p ) ;
      balance = b ;
    }
    /**
     *  Insert a node into an <code>AVLTree</code>.
     *
     *  <p> Insertion of a node is seriously non-trivial.  We need to find the right place to do the
     *  insertion as ever but once the insertion is performed, the balance factor needs updating and then a
     *  check to see if a rebalance is required must be made.  The function is highly recursive.  Also note
     *  that the function returns a balance sub-tree for relinking into the tree!</p>
     *
     *  @param o the datum to add to the tree.
     *  @return the node that the datum was added in.
     */
    public final AVLNode add ( final T o ) {
      if ( order.relation ( o , datum ) ) {
        if ( left != null ) {
          //  The new item gets added into the left sub-tree.  Afterwards, check the balance factor to see
          //  if it needs adjusting.
          final int oldBalance = ( (AVLNode) left ).balance ;
          left = ( (AVLNode) left ).add ( o ) ;
          if ( ( ( (AVLNode) left ).balance != oldBalance ) && ( ( (AVLNode) left ).balance != 0 ) ) {
            --balance ;
          }
        } else {
          //  We found a null so add the item here and adjust the balance factor.
          left = new AVLNode ( o , null , null , this , 0 ) ;
          --balance ;
        }
      } else {
        if ( equal.relation ( o , datum ) ) {
          if ( ! duplicatesAllowed ) {
            //  It may seem strange to do this replacement given that the two elements are the same.
            //  The reason is that if the data are structured and equality is measured only on part of the
            //  data structure then the two could actually be different even though they are the same.  Our
            //  strategy is replacement to ensure updatedness of the value.
            datum = o ;
            return this ;
          }
        }
        if ( right != null ) {
          //  The new item gets added into the right sub-tree.  Afterwards, check the balance factor to see
          //  if it needs adjusting.
          final int oldBalance = ( (AVLNode) right ).balance ;
          right = ( (AVLNode) right ).add ( o ) ;
          if ( ( ( (AVLNode) right ).balance != oldBalance ) && ( ( (AVLNode) right ).balance != 0 ) ) {
            ++balance ;
          }
        } else {
          //  We found a null so add the item here and adjust the balance factor.
          right = new AVLNode ( o , null , null , this , 0 ) ;
          ++balance ;
        }
      }
      if ( ( balance < -1 ) || ( balance > 1 ) ) { return doBalance ( ) ; }
      return this ;
    }
    /**
     *  Remove an item from the <code>AVLTree</code>.
     *
     *  <p> Removal is non-trivial and done in a very recursive manner.  Search for the node to be deleted
     *  and return a sub-tree with the item deleted.</p>
     *
     *  <p> NB The use of a class scope variable to tell the <code>AVLTree</code> <code>remove</code> method
     *  when a removal actually happened.</p>
     *
     *  @param o the item to find and remove from the tree.
     *  @return the node that was removed.
     */
    public final AVLNode remove ( final T o ) {
      if ( equal.relation ( datum , o ) ) {
        //  The item required to be removed has been found, it is us.  We must return a sub-tree
        //  with ourself (the root node for the sub-tree) missing.  If there is only one child simply return
        //  it.  Otherwise find the successor of the current node (which is the left-most child of the right
        //  sub-tree) cut it out and use it to replace ourself.  Rebalance as required.
        if ( left == null ) {
          if ( right != null ) { right.parent = parent ; }
          return (AVLNode) right ;
        }
        if ( right == null ) {
          if ( left != null ) { left.parent = parent ; }
          return (AVLNode) left ;
        }
        final int oldBalance = ( (AVLNode) right ).balance ;
        final Pair<AVLNode, AVLNode> p = ( (AVLNode) right ).cutOutLeftmostNode ( ) ;
        final AVLNode leftmostNode = p.getB ( ) ;
        leftmostNode.left = left ;
        leftmostNode.left.parent = leftmostNode ;
        leftmostNode.right = p.getA ( ) ;
        leftmostNode.right.parent = leftmostNode ;
        leftmostNode.parent = parent ;
        leftmostNode.balance = balance ;
        return leftmostNode.rebalanceRight ( oldBalance ) ;
      } else if ( order.relation ( o , datum ) ) {
        //  The item is in our left child if it was anywhere, act accordingly.
        if ( left == null ) { return this ; }
        final int oldBalance = ( (AVLNode) left ).balance ;
        left = ( (AVLNode) left ).remove ( o ) ;
        return rebalanceLeft ( oldBalance ) ;
      } else {
        //  The item is in our right child if it was anywhere, act accordingly.
        if ( right == null ) { return this ; }
        final int oldBalance = ( (AVLNode) right ).balance ;
        right = ( (AVLNode) right ).remove ( o ) ;
        return rebalanceRight ( oldBalance ) ;
      }
    }
    /**
     *  Single rotate a sub-tree left.
     *
     *  <p> We are the current root and we are to be replaced by another so we return a new sub-tree to our
     *  parent for them to connect to.  The manipulations are actually relatively simple but the balance
     *  factor equations are magic numbers from the literature.  The a and b are introduced to avoid the
     *  equations getting bizarrely long.</p>
     *
     *  @return root of the rotated subtree.
     */
    protected final AVLNode rotateLeft ( ) {
      final AVLNode oldRoot = this ;
      final AVLNode newRoot = (AVLNode) right ;
      oldRoot.right = newRoot.left ;
      if ( oldRoot.right != null ) { oldRoot.right.parent = oldRoot ; }
      newRoot.left = oldRoot ;
      newRoot.parent = oldRoot.parent ;
      oldRoot.parent = newRoot ;
      final int a = oldRoot.balance ;
      final int b = newRoot.balance ;
      if ( b <= 0 ) {
        newRoot.balance = ( a >= 1 ) ? b - 1 : a + b - 2 ;
        oldRoot.balance = a - 1 ;
      } else {
        newRoot.balance = ( a <= b ) ? a - 2 : b - 1 ;
        oldRoot.balance = a - b - 1 ;
      }
      return newRoot ;
    }
    /**
     *  Single rotate a sub-tree right.
     *
     *  <p> We are the current root and we are to be replaced by another so we return a new sub-tree to our
     *  parent for them to connect to.  The manipulations are actually relatively simple but the balance
     *  factor equations are magic numbers from the literature.  The a and b are introduced to avoid the
     *  equations getting bizarrely long.</p>
     *
     *  @return root of the rotated subtree.
     */
    protected final AVLNode rotateRight ( ) {
      final AVLNode oldRoot = this ;
      final AVLNode newRoot = (AVLNode) left ;
      oldRoot.left = newRoot.right ;
      if ( oldRoot.left != null ) { oldRoot.left.parent = oldRoot ; }
      newRoot.right = oldRoot ;
      newRoot.parent = oldRoot.parent ;
      oldRoot.parent = newRoot ;
      final int a = oldRoot.balance ;
      final int b = newRoot.balance ;
      if ( b <= 0 ) {
        newRoot.balance = ( b > a ) ? b + 1 : a + 2 ;
        oldRoot.balance = a - b + 1 ;
      } else {
        newRoot.balance = (a < 0) ? b + 1 : a + b + 2 ;
        oldRoot.balance = a + 1 ;
      }
      return newRoot ;
    }
    /**
     *  Patch up the balance factors after a rotate operation and then check for and undertake if necessary
     *  a re-balancing of the tree rooted at us.  Return a new tree for relacment.
     *
     *  @param oldBalance the balance of the subtree as was.
     *  @return root of the rebalanced subtree.
     */
    protected final AVLNode rebalanceLeft ( final int oldBalance ) {
      if ( left == null ) { ++balance ; }
      else if ( ( ( (AVLNode) left ).balance == 0 ) && ( oldBalance != 0 ) ) { ++balance ; }
      return ( balance > 1 ) ? doBalance ( ) : this ;
    }
    /**
     *  Patch up the balance factors after a rotate operation and then check for and undertake if necessary
     *  a re-balancing of the tree rooted at us.  Return a new tree for relacment.
     *
     *  @param oldBalance the balance of the subtree as was.
     *  @return root of the rebalanced subtree.
     */
    protected final AVLNode rebalanceRight ( final int oldBalance ) {
      if ( right == null ) { -- balance ; }
      else if ( ( ( (AVLNode) right ).balance == 0 ) && ( oldBalance != 0 ) ) { --balance ; }
      return ( balance < -1 ) ? doBalance ( ) : this ;
    }
    /**
     *  Perform a rebalancing as needed.  Decides whether a single or double rotation is required and then
     *  does it.  This is only called if it is known that a rebalance must occur.
     *
     *  @return root of the rebalanced subtree.
     */
    protected final AVLNode doBalance ( ) {
      if ( balance < 0 ) {
        //  Left is too heavy.
        if ( ( (AVLNode) left ).balance > 0 ) {
          //  Must do double rotation.
          left = ( (AVLNode) left ).rotateLeft ( ) ;
        }
        return rotateRight ( ) ;
      } else {
        //  Right is too heavy.
        if ( ( (AVLNode) right ).balance < 0 ) {
          //  Must do double rotation.
          right = ( (AVLNode) right ).rotateRight ( ) ;
        }
        return rotateLeft ( ) ;
      }
    }
    /**
     *  Return two items: the sub-tree rooted at the <code>Node</code> that is ourself with its left-most
     *  node removed and a reference to the cut out <code>Node</code>.
     *
     *  <p>Care is taken to rebalance the edited tree so no further action is required.</p>
     *
     *  @return the pair of the sub-tree rooted at the <code>Node</code> that is ourself with its left-most
     *  node removed and a reference to the cut out <code>Node</code>.
     */
    protected final Pair<AVLNode, AVLNode> cutOutLeftmostNode ( ) {
      //  This is a recursive method which will eventually arrive at the position left == null so we will
      //  eventually have to execute this bit of code.
      if ( left == null ) { return new Pair<AVLNode, AVLNode> ( (AVLNode) right , this ) ; }
      final int oldBalance = ( (AVLNode) left ).balance ;
      final Pair<AVLNode, AVLNode> p = ( (AVLNode) left ).cutOutLeftmostNode ( ) ;
      left = p.getA ( ) ;
      return new Pair<AVLNode, AVLNode> ( rebalanceLeft ( oldBalance ) , p.getB ( ) ) ;
    }
  }
  /**
   *  Construct an <code>AVLTree</code>.  A <code>Comparator</code> must be provided to give the insert and
   *  remove policy.  The default equality test (<code>EqualToComparator</code> is assumed.  Default of
   *  allowing duplicate entries.
   *
   *  @param o the object to compare to this one.
   */
  public AVLTree ( final Comparator<? super T> o ) { this ( o , new EqualToComparator<T> ( ) , true ) ; }
  /**
   *  Construct an <code>AVLTree</code>.  A <code>Comparator</code> must be provided to give the insert and
   *  remove policy.  The default equality test (<code>EqualToCOmparator</code> is assumed.  Specify whether
   *  duplicates are allowed.
   *
   *  @param o the object to compare this one to.
   *  @param dups determines whether duplicates are acceptible or not.
   */
  public AVLTree ( final Comparator<? super T> o , final boolean dups ) { this ( o , new EqualToComparator<T> ( ) , dups ) ; }
  /**
   *  Construct an <code>AVLTree</code>.  A <code>Comparator</code> must be provided to give the
   *  insert and remove policy.  The default equality test (<code>EqualToCOmparator</code> is
   *  assumed.  Default of allowing duplicate entries.
   *
   *  @param o the object to compare this one to.
   *  @param e the comparator to use to compare objects.
   */
  public AVLTree ( final Comparator<? super T> o , final Comparator<? super T> e ) { this ( o , e , true ) ; }
  /**
   *  Construct an <code>AVLTree</code>.  A <code>Comparator</code> must be provided to give the insert and
   *  remove policy.  The default equality test (<code>EqualToCOmparator</code> is assumed.  Specify whether
   *  duplicates are allowed.
   *
   *  <p> We choose to use <code>Comparator</code>s to define equality for flexibility.  <code>equals</code>
   *  tests the whole value.  When we have data that has internal structure we may want to define equality
   *  as only a partial match.</p>
   *
   *  @param o the object to compare this one to.
   *  @param e the comparator to use to compare objects.
   *  @param dups determines whether duplicates are acceptible or not.
   */
  public AVLTree ( final Comparator<? super T> o , final Comparator<? super T> e , final boolean dups ) { super ( o , e , dups ) ; }
  /**
   *  Insert a value into the <code>AVLTree</code>.
   */
  @Override
  public void add ( final T o ) {
    root.left = isEmpty ( ) ? new AVLNode ( o , null , null , root , 0 ) : ( (AVLNode) root.left ).add ( o ) ;
  }
  /**
   *  Remove an item from the <code>AVLTree</code>.
   */
  @Override
  public boolean remove ( final T o ) {
    if ( ! isEmpty( ) ) {
      root.left = ( (AVLNode) root.left ).remove ( o ) ;
      return true ;
    }
    return false ;
  }
  /**
   *  Clone the whole tree.  NB <code>CloneNotSupportedException</code> is caught in
   *  <code>AbstractContainer</code> so we don't need to worry about it here.
   */
  @Override public AVLTree<T> clone ( ) {
    final AVLTree<T> t = (AVLTree<T>) super.clone ( ) ;
    if ( ! isEmpty ( ) ) { t.root.left = root.left.clone ( t.root ) ; }
    return t ;
  }
  /**
   *  The search iterator over <code>AVLTree</code>s.
   *
   *  <p>This subclass is required in order to ensure the correct node type is created when setting via the
   *  iterator, i.e. <code>AVLNode</code>s are created not <code>Node</code>s.</p>
   */
  protected class AVLSearchIterator extends SearchIterator {
    /**
     *  The constructor.  We must have an item to search for.  This constructor allows us to use an
     *  alternate equality relation.
     *
     *  @param o the object to search the tree for.
     */
    public AVLSearchIterator ( final T o ) { super ( o ) ; }
    /**
     *  Set the datum of the current <code>AVLNode</code> to be the proferred object.  If
     *  <code>theItem</code> is null then attach to the parent as a new <code>AVLNode</code>.
     */
    //  TODO : This operation could leave the tree in an inconsistent state since the new item could
    //  be out of sorted sequence.  There really ought to be some sort of protection so as to guarantee the
    //  order relation.  Of course the whole SearchIterator/searchFor mechanism is protected so they should
    //  not get used improperly, nothing wrong with a bit of paranoia though.
    @Override public void set ( final T o ) {
      if ( theParent == null ) { throw new InvalidIteratorOperationException ( ) ; }
      if ( theItem == null ) {
        final AVLNode n = new AVLNode ( o , null , null , theParent , 0 ) ;
        if ( isLeft ) { theParent.left = n ; }
        else { theParent.right = n ; }
      }
      else { theItem.datum = o ; }
    }
  }
  /**
   *  The method to find a particular node in the tree.
   */
  //  Override this method in order to get the correct SearchIterator type.
  @Override protected SearchIterator searchFor ( final T o ) { return new AVLSearchIterator ( o ) ; }
}
