/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010--2011 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  The abstract class representing a binary tree.
 *
 *  @author Russel Winder
 */
public abstract class AbstractBinaryTree<T> extends AbstractContainer<T> implements BinaryTree<T> {
  /**
   *  The datum that is the root for the <code>BinaryTree</code>.  Employ a dummy item rather than a null
   *  reference since it make many of the algorithms a easier.
   *
   *  <p>NB root.left refers to the tree of data itself.</p>
   */
  //  Can't make this variable final as it is altered in clone.
  protected Node<T> root = new Node<T> ( null , null , null , null ) ;
  /**
   *  The <code>Node</code> of an <code>AbstractBinaryTree</code>.  This may well be subclassed in
   *  subclasses of <code>AbstractBinaryTree</code>.
   *
   *  <p><code>Node</code>s know who their parent is as well as who their children are in order to make
   *  editing and traversal easier.  Clearly not all binary trees need a parent reference, traditionally
   *  trees and tree algorithms have had only uni-directional references from parent to children.  However,
   *  such trees only really support applicators (via recursive procedures) and not iterators.  Since ADS is
   *  iterator oriented, we need to properly support iterators.  Therefore, the links in the tree are made
   *  bi-directional by having a reference between child and parent.</p>
   *
   *  @author Russel Winder
   */
  protected static class Node<E> implements Cloneable {
    /**
     *  The datum that is stored in this <code>Node</code>.
     */
    protected E datum ;
    /**
     *  The left sub-tree of this <code>Node</code>.
     */
    protected Node<E> left ;
    /**
     *  The right sub-tree of this <code>Node</code>.
     */
    protected Node<E> right ;
    /**
     *  The parent of this <code>Node</code>.
     */
    protected Node<E> parent ;
    /**
     *  Constructor of a <code>Node</code>.
     *
     *  @param datum datum for this node.
     *  @param left left subtree of this node.
     *  @param right right subtree of this node.
     *  @param parent parent of this node.
     */
    public Node ( final E datum , final Node<E> left ,final Node<E> right , final Node<E> parent ) {
      this.datum = datum ;
      this.left = left ;
      this.right = right ;
      this.parent = parent ;
    }
    /**
     *  Remove ourself from the tree.
     */
    protected void remove ( ) {
      final Node<E> newThis = mergeChildren ( ) ;
      if ( newThis != null ) { newThis.parent = parent ; }
      if ( parent.left == this ) { parent.left = newThis ; }
      else if ( parent.right == this ) { parent.right = newThis ; }
      else { throw new InvalidOperationException ( ) ; }
    }
    /**
     *  Support method for the tree editing.  Usual algorithm: If there is only one sub-tree then deliver it
     *  up.  If there are two sub-trees but the right sub-tree has no left sub-tree, link the left sub-tree
     *  as the left child of the right sub-tree and deliver it up.  Otherwise find the left-most node of the
     *  right sub-tree and cut it out to replace the current node.
     *
     *  @return reference to the merged tree.
     */
    private Node<E> mergeChildren ( ) {
      if ( left == null ) { return right ; }
      if ( right == null ) { return left ; }
      if ( right.left == null ) {
        right.left = left ;
        left.parent = right ;
        return right ;
      }
      Node<E> newRootParent = right ;
      Node<E> newRoot = newRootParent.left ;
      while ( newRoot.left != null ) {
        newRootParent = newRoot ;
        newRoot = newRoot.left ;
      }
      newRootParent.left = newRoot.right ;
      if ( newRoot.right != null ) { newRoot.right.parent = newRootParent ; }
      newRoot.right = right ;
      newRoot.left = left ;
      left.parent = newRoot ;
      right.parent = newRoot ;
      return newRoot ;
    }
    /**
     *  Cloning means copying not just this item but both sub-trees rooted here.  This is a shallow copy in
     *  the sense that although new <code>Node</code>s are created the data items are not cloned.
     */
    @Override protected Node<E> clone ( ) {
      //  In supporting the OrderedBinaryTree<T> we need to provide a parent for the first Node<T> in the
      //  tree.  There are two options to implementing clone in this circumstance, make the parent attribute
      //  accessible to OrderedBinaryTree<T>, or provide a clone method that takes a parameter.  For safety's
      //  sake we do the latter.  This means we can implement clone without a parameter as a delegated call
      //  to this other clone method with a parent of null.
      //
      //  IntelliJ IDEA  insists on giving a warning that super.clone ( ) is not being called :-(
      return clone ( null ) ;
    }
    /**
     *  Cloning means copying not just this item but both sub-trees rooted here.  This is a shallow copy in
     *  the sense that although new <code>Node</code>s are created the data items are not cloned.
     *
     *  @param p the parent of the node being cloned.
     *  @return the new clone of this node.
     */
    @SuppressWarnings ( "unchecked" )
    protected Node<E> clone ( final Node<E> p ) {
      try {
        final Node<E> n = (Node<E>) super.clone ( ) ;
        n.parent = p ;
        if ( left != null ) { n.left = left.clone ( n ) ; }
        if ( right != null ) { n.right = right.clone ( n ) ; }
        return n ;
      } catch ( CloneNotSupportedException cnse ) { throw new InternalError ( ) ; }
    }
    /**
     *  Print out the tree on the standard output..  This method was used during debugging and it not really
     *  part of the API for this class.  Leave it in since it doesn't hurt and it may again prove useful.
     *
     *  @param level the current depth in the recursive descent.
     */
    protected void printTree ( final int level ) {
      if ( right != null ) { right.printTree ( level + 1 ) ; }
      for ( int i = 0 ; i <= level ; ++i ) { System.out.print ( "\t" ) ; }
      System.out.println(datum) ;
      if ( left != null ) { left.printTree ( level + 1 ) ; }
    }
  }
  @Override public void makeEmpty ( ) { root.left = null ; }
  @Override public boolean isEmpty ( ) { return root.left == null ; }
  @Override public abstract boolean equals ( final Object o ) ;
  @Override public AbstractBinaryTree<T> clone ( ){ return (AbstractBinaryTree<T>) super.clone ( ) ; }
  //  There are a number of different ways of traversing a BinaryTree<T>, provide an InputIterator for each
  //  of them.  Implement the necessary iterator creators with one of the variants.
  @Override public final InputIterator<T> begin ( ) { return inorderTraversal ( ) ; }
  @Override public final InputIterator<T> end ( ) {
    final InputIterator<T> i = inorderTraversal ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public InputIterator<T> iterator ( ) { return inorderTraversal ( ) ; }
  /**
   *  The basic iterator template for post-order, pre-order and in-order traversals.  This is an abstract
   *  class defining the common elements.  The parameter is the queueing method.
   *
   *  @author Russel Winder
   */
  protected abstract class AbstractBinaryTreeIterator extends AbstractIterator<T> {
    /**
     *  The item to be delivered up by the iterator.
     */
    protected Node<T> theItem = root.left ;
    @Override public T get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorException ( ) ; }
      return theItem.datum ;
    }
    @Override public abstract void advance ( ) ;
    @Override public void advance ( int increment ) {
      for ( ; increment > 0 ; --increment ) {
        if ( atEnd ( ) ) { break ; }
        advance ( ) ;
      }
    }
    @Override public boolean atBegin ( ) { return isEmpty() || ( theItem == root.left ) ; }
    @Override public boolean atEnd() { return isEmpty() || ( theItem == root ) ; }
    @Override public abstract boolean equals ( Object o ) ;
    @Override public BinaryTree<T> getContainer ( ) { return AbstractBinaryTree.this ; }
  }
  /**
   *  Post-order traversal iterator.  Post-order traversal of the <code>BinaryTree</code> visits the nodes
   *  in the order: left child, right child, current node.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public final InputIterator<T> postorderTraversal ( ) {
    class PostOrderIterator extends AbstractBinaryTreeIterator {
      public PostOrderIterator ( ) { slideLeftGoRightRepeatedly ( ) ; }
      @Override public void advance ( ) {
        final Node<T> child = theItem ;
        theItem = theItem.parent ;
        if ( ( theItem != null ) && ! atEnd ( ) && ( theItem.right != null ) && ( theItem.right != child ) ) {
          theItem = theItem.right ;
          slideLeftGoRightRepeatedly ( ) ;
        }
      }
      @Override public boolean equals ( final Object o ) {
        if ( o == this ) { return true ; }
        //  TODO: Eclipse cannot handle this use of instanceof.
        //if ( ( o == null ) || ! ( o instanceof PostOrderIterator ) ) { return false ; }
        //return theItem.equals (  ( (PostOrderIterator) o ).theItem ) ;
        //  TODO: IntelliJ IDEA complains that instanceof has not been used.
        try { return theItem.equals ( ( (PostOrderIterator) o ).theItem ) ; }
        catch ( final ClassCastException cce) { return false ; }
      }
      private void slideLeftGoRightRepeatedly ( ) {
        while ( ( theItem != null ) && ( ( theItem.left != null ) || ( theItem.right != null ) ) ) {
          while ( ( theItem != null ) && ( theItem.left != null ) ) { theItem = theItem.left ; }
          if ( ( theItem != null ) && ( theItem.right != null ) ) { theItem = theItem.right ; }
        }
      }
    }
    return new PostOrderIterator ( ) ;
  }
  /**
   *  Pre-order traversal iterator.  Preorder traversal means visit nodes in the order: current node, left
   *  child, right child.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public final InputIterator<T> preorderTraversal ( ) {
    class PreOrderIterator extends AbstractBinaryTreeIterator {
      @Override public void advance ( ) {
        if ( theItem.left != null ) { theItem = theItem.left ; }
        else if ( theItem.right != null ) { theItem = theItem.right ; }
        else {
          Node<T> child = theItem ;
          theItem = theItem.parent ;
          while ( ( theItem != null ) && ! atEnd ( ) && ( ( theItem.right == child ) || ( theItem.right == null ) ) ) {
            child = theItem ;
            theItem = theItem.parent ;
          }
          if ( ( theItem != null ) && ! atEnd ( ) && ( theItem.right != null ) ) { theItem = theItem.right ; }
        }
      }
      @Override public boolean equals ( final Object o ) {
        if ( o == this ) { return true ; }
        //  TODO: Eclipse cannot handle this use of instanceof.
        //if ( ( o == null ) || ! ( o instanceof PreOrderIterator ) ) { return false ; }
        //return theItem.equals (  ( (PreOrderIterator) o ).theItem ) ;
        //  TODO: IntelliJ IDEA complains that instanceof has not been used.
        try { return theItem.equals ( ( (PreOrderIterator) o ).theItem ) ; }
        catch ( final ClassCastException cce) { return false ; }
      }
    }
    return new PreOrderIterator ( ) ;
  }
  /**
   *  In-order traversal iterator.  In-order traversal visits the nodes in the order: left child, current
   *  node, right child.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public final InputIterator<T> inorderTraversal ( ) {
    class InOrderIterator extends AbstractBinaryTreeIterator {
      public InOrderIterator ( ) {
        while ( ( theItem != null ) && ( theItem.left != null ) ) { theItem = theItem.left ; }
      }
      @Override public void advance ( ) {
        if ( theItem.right != null ) {
          theItem = theItem.right ;
          while ( ( theItem != null ) && ( theItem.left != null ) ) { theItem = theItem.left ; }
        } else {
          Node<T> child = theItem ;
          theItem = theItem.parent ;
          while ( ( theItem != null ) && ! atEnd ( ) && ( theItem.right == child ) ) {
            child = theItem ;
            theItem = theItem.parent ;
          }
        }
      }
      @Override public boolean equals ( final Object o ) {
        if ( o == this ) { return true ; }
        //  TODO: Eclipse cannot handle this use of instanceof.
        //if ( ( o == null ) || ! ( o instanceof InOrderIterator ) ) { return false ; }
        //return theItem.equals (  ( (InOrderIterator) o ).theItem ) ;
        //  TODO: IntelliJ IDEA complains that instanceof has not been used.
        try { return theItem.equals ( ( (InOrderIterator) o ).theItem ) ; }
        catch ( final ClassCastException cce ) { return false ; }
      }
    }
    return new InOrderIterator ( ) ;
  }
  /**
   *  Level-order traversal iterator. Level-order traversal visits the nodes left-to-right, level-by-level.
   *  It is a breadth-first traversal rather than the depth-first traversal that the other basically are.
   *  For this reason we need the support of a queue.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public final InputIterator<T> levelorderTraversal ( ) {
    class LevelOrderIterator extends AbstractBinaryTreeIterator {
      protected Queue<Node<T>> itemQueue = new SLListQueue<Node<T>> ( ) ; {
        theItem = root.left ;
        if ( theItem != null )  {
          if ( theItem.left != null ) { itemQueue.add ( theItem.left ) ; }
          if ( theItem.right != null ) { itemQueue.add ( theItem.right ) ; }
        }
      }
      @Override public void advance ( ) {
        if ( itemQueue.isEmpty ( ) ) { theItem = root ; }
        else {
          theItem = itemQueue.remove ( ) ;
          if ( theItem != null ) {
            if ( theItem.left != null ) { itemQueue.add ( theItem.left ) ; }
            if ( theItem.right != null ) { itemQueue.add ( theItem.right ) ; }
          }
        }
      }
      @Override public boolean equals ( final Object o ) {
        if ( o == this ) { return true ; }
        //  TODO: Eclipse cannot handle this use of instanceof.
        //if ( ( o == null ) || ! ( o instanceof LevelOrderIterator ) ) { return false ; }
        //return theItem.equals (  ( (LevelOrderIterator) o ).theItem ) ;
        //  TODO: IntelliJ IDEA complains that instanceof has not been used.
        try { return theItem.equals ( ( (LevelOrderIterator) o ).theItem ) ; }
        catch ( final ClassCastException cce ) { return false ; }
      }
    }
    return new LevelOrderIterator ( ) ;
  }
}
