/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  This abstract class provides implementations of the methods in the interface <code>Container</code> that
 *  can be implemented using representation independent algorithms.
 *
 *  @author Russel Winder
 */
public abstract class AbstractContainer<T> implements Container<T> {
  @Override public abstract void makeEmpty ( ) ;
  @Override public boolean isEmpty ( ) { return size ( ) <= 0 ; }
  @SuppressWarnings ( "unused" )
  @Override public int size ( ) {
    int count = 0 ;  
    for ( final T item : this ) { ++count ; } // Generates an ignorable "never read" warning.
    return count ;
  }
  @Override public abstract boolean equals ( final Object o ) ;
  @Override public boolean contains ( final T o ) {
    //  This implements a simple linear search [ O ( n ) ] in the order determined by the
    //  iterator used to back the for loop.  This implementation is probably all that is needed
    //  for sequences and other structures with no relations defined on them.  Containers with relations
    //  (i.e. ordering properties really ought to override this method to make things more efficient.
    if ( ( o == null ) || isEmpty ( ) ) { return false ; }
    for ( final T item : this ) {
      if ( item.equals ( o ) ) { return true ; }
    }
    return false ;
  }
  @Override public abstract void add ( T o ) ;
  @Override public abstract boolean remove ( T o ) ;
  @SuppressWarnings ( "unchecked" )
  @Override public T[] toPrimitiveArray ( final Class<T> c ) {
    //  This implementation uses the iterator delivered by the iterator method to deliver a
    //  sequence of the values in the container.  Concrete data structures may wish to override this
    //  implementation for efficiency.  For example Array will definitely.
    //
    //  Cannot use "(T[]) new Object [size ( )]" since even though it seems type correct, it
    //  isn't due to erasure issues so in order to really deliver a T[] we must use reflection.  Hence the
    //  Class<T> parameter.
    final T[] data = (T[]) java.lang.reflect.Array.newInstance ( c , size ( ) ) ;
    int index = 0 ;
    for ( InputIterator<T> i = iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) , ++index ) { data[index] = i.get ( ) ; }
    return data ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public T[] toPrimitiveArray ( T[] a ) {
    //  This implementation uses the iterator delivered by the iterator method to deliver a
    //  sequence of the values in the container.  Concrete data structures may wish to override this
    //  implementation for efficiency.  For example Array will definitely.
    //
    //  Cannot use "(T[]) new Object [size ( )]" since even though it seems type correct, it isn't due to
    //  erasure issues so in order to really deliver a T[] we must use reflection.
    final int size = size ( ) ;
    if ( a.length < size ) {
      a = (T[]) java.lang.reflect.Array.newInstance ( a.getClass ( ).getComponentType ( ) , size ) ;
    }
    int index = 0 ;
    for ( InputIterator<T> i = iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) , ++index ) { a[index] = i.get ( ) ; }
    if ( a.length > size ) {
      for ( int i = a.length ; i < size ; ++i ) { a[i] = null ; }
    }
    return a ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public Container<T> clone ( ){
    //  Object.clone can throw CloneNotSupportedException, this is not allowed for the containers in ADS.
    try { return (Container<T>) super.clone ( ) ; }
    catch ( CloneNotSupportedException e ) { throw new InternalError ( ) ; }
  }
  @Override public String toString ( ) {
    final StringBuilder sb = new StringBuilder ( ) ;
    sb.append ( "[ " ) ;
    for ( final T item : this ) { sb.append ( item.toString ( ) ).append ( ' ' ) ; }
    sb.append ( ']' ) ;
    return sb.toString ( ) ;
  }
  @Override public abstract InputIterator<T> begin ( ) ;
  @Override public abstract InputIterator<T> end ( ) ;
  @Override public abstract InputIterator<T> iterator ( ) ;
  /**
   *  A method to support copy construction.  Copy all the elements in the parameter <code>Container</code>
   *  into this <code>Container</code>. This is a shallow copy operation in that we do not clone the data
   *  but just copy references to the data.
   *
   *  @param c the <code>Container</code> from which to copy all the elements into this
   *  <code>Container</code>.
   */
  public void copyIn ( final Container<T> c ) { for ( final T t : c ) { add ( t ) ; } }
  /**
   *  Copy some data items defined by two <code>InputIterator</code>s into this <code>Container</code>.
   *  This is a shallow copy operation in that we do not clone the data but just copy references to the
   *  data.
   *
   *  @param b the beginning iterator of the items to be copied.
   *  @param e the end iterator of the items to be copied.
   *  @exception InvalidIteratorException if the two <code>InputIterator</code>s are not working the same
   *  <code>Container</code>.
   *  @exception InvalidIteratorOperationException if the iterator <code>e</code> is not after
   *  <code>b</code>.
   */
  public final void copyIn ( final InputIterator<T> b , final InputIterator<T> e ) {
    //  b and e must be working the same Container<T> or they cannot be defining a valid sequence of items.
    if ( b.getContainer ( ) != e.getContainer ( ) ) { throw new InvalidIteratorException ( ) ; }
    //  The iterator e should be after b in the Container<T>.
    if ( b.distance ( e ) < 0 ) { throw new InvalidIteratorOperationException ( ) ; }
    //  Can now actually do the copying in the sure knowledge that there is definite work to be done.
    for ( InputIterator<T> i = b.clone ( ) ; ! i.atEnd ( ) && ! i.equals ( e ) ; i.advance ( ) ) {
      add ( i.get ( ) ) ;
    }
  }
}
