/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  All iterators over containers in ADS extend this class.  This class provides default implementations for
 *  those methods that can have them.  This class ensures that ADS iterators realize the necessary features
 *  of Collections iterators (providing implementations for the methods of the <code>Iterator</code>
 *  interface) so that the for statement over containers will work, i.e. the <code>for ( E e : c ) {
 *  . . . }</code> loops work for containers from ADS because all ADS containers conform to
 *  <code>Iterable</code> and all ADS iterators to <code>Iterator</code>.
 *
 *  @author Russel Winder
 */
public abstract class AbstractIterator<T> implements InputIterator<T> {
  @Override public abstract T get ( ) ;
  @Override public abstract void advance ( ) ;
  @Override public abstract void advance ( int increment ) ;
  @Override public abstract boolean atBegin ( ) ;
  @Override public abstract boolean atEnd ( ) ;
  @Override public abstract boolean equals ( Object o ) ;
  @Override public abstract Container<T> getContainer ( ) ;
  @Override public int distance ( final InputIterator<T> i ) {
    //  This is an O(n) operation since it moves iterators through the data structure trying to find the
    //  target, i.e. it is a linear search.  This is fine for iterators into list-like data structures but
    //  is a serious error for iterators over random access data structures.  So it is important to override
    //  this definition if at all possible.
    if ( equals ( i ) ) { return 0 ; }
    if ( getContainer ( ) != i.getContainer ( ) ) { throw new InvalidIteratorException ( ) ; }
    int count = 0 ;
    InputIterator<T> j = clone ( ) ;
    for ( ; ! j.atEnd ( ) ; ++count , j.advance ( ) ) { if ( j.equals ( i ) ) { return count ; } }
    //  Check to see if i.atEnd ( ).
    if ( j.equals ( i ) ) { return count ; }
    count = 0 ;
    j = i.clone ( ) ;
    for ( ; ! j.atEnd ( ) ; ++count , j.advance ( ) ) { if ( j.equals ( this ) ) { return -count ; } }
    //  Check to see if atEnd ( ).  Actually this is redundant as we already checked this case above.  But
    //  as we can never get here let us not worry about keeping the test in for symmetry and just in case
    //  :-)
    if ( j.equals ( this ) ) { return -count ; }
    throw new InternalError ( ) ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public AbstractIterator<T> clone ( ) {
    try { return (AbstractIterator<T>) super.clone ( ) ; }
    catch ( CloneNotSupportedException e ) { throw new InternalError ( ) ; }
  }
  /**
   *  Return true if there are more items in the iteration, otherwise return false. Implementation of the
   *  <code>hasNext</code> method from <code>Iterator</code>.
   *
   *  @return <code>true</code> if there are more items, <code>false</code> otherwise.
   */
  @Override public boolean hasNext ( ) { return ! atEnd ( ) ; }
  /**
   *  Implementation of the <code>next</code> method from <code>Iterator</code>.
   *
   *  @return the value the iterator is pointing at after it has moved on one.
   */
  @Override public T next ( ) {
    final T returnValue = get ( ) ;
    advance ( ) ;
    return returnValue ;
  }
  /**
   *  Implementation of the <code>remove</code> method from <code>Iterator</code>.  This implementation
   *  throws <code>RuntimeException</code> as base iterators are read-only.
   *
   *  @throws RuntimeException always.
   */
  @Override public void remove ( ) { throw new RuntimeException ( "Cannot apply remove with this iterator." ) ; }
}
