/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An abstract <code>Queue</code> to act as a superclass for all the concrete queues.  Provides all the
 *  default implementations that are representation independent for the abstraction by conformance
 *  heirarchy..
 *
 *  <p>Although the order of items in a queue is important, i.e. a queue is a sequence, we do not inherit
 *  from <code>Sequence</code> since a sequence is an open data structure whereas a queue is a closed data
 *  structue &mdash; we do not want client code to be able to deal with anything other than the two ends of
 *  the queue.  This does lead to what may seem to be code replication, cf. <code>equals</code>.  It is.</p>
 *
 *  @author Russel Winder
 */
public abstract class AbstractQueue<T> extends AbstractContainer<T> implements Queue<T> {
  @Override public abstract T peek ( ) ;
  @Override public abstract T remove ( ) ;
  /**
   *  Always causes and exception since it is a nonsense operation, <code>Queue</code>s are closed
   *  containers.
   *
   *  @exception InvalidOperationException always thrown since this is not a valid operation.
   */
  @Override public boolean remove ( final T o ) { throw new InvalidOperationException ( ) ; }
  @SuppressWarnings ( "unchecked" )
  @Override public boolean equals ( final Object o ) {
    if ( this == o ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof Queue ) || ( size ( ) != ((Queue<T>) o).size ( ) ) ) {
      return false ;
    }
    final InputIterator<T> i1 = iterator ( ) ;
    final InputIterator<T> i2 = ( (Queue<T>) o).iterator ( ) ;
    for ( ; ! i1.atEnd ( ) && ! i2.atEnd ( ) ; i1.advance ( ) , i2.advance ( ) ) {
      final T o1 = i1.get ( ) ;
      final T o2 = i2.get ( ) ;
      if ( ( o1 == null ) || ( o2 == null ) || ! o1.equals ( o2 ) ) { return false ; }
    }
    assert i1.atEnd ( ) && i2.atEnd ( )  ;
    return true ;
  }
  @Override public Queue<T> clone ( ) { return (Queue<T>) super.clone ( ) ; }
}
