/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An abstract <code>Sequence</code> to act as a superclass for all the concrete <code>Sequence</code>s.
 *  Provides realization for those methods that can implemented using representation independent algorithms.
 *
 *  @author Russel Winder
 */
public abstract class AbstractSequence<T> extends AbstractContainer<T> implements Sequence<T> {
  @Override public T first ( ) { return get ( 0 ) ; }
  @Override public T last ( ) { return get ( size ( ) - 1 ) ; }
  @Override public abstract T get ( int index ) ;
  @Override public abstract void set ( int index , T o ) ;
  @Override public abstract void remove ( int index ) ;
  @Override public abstract void insert ( int index , T o ) ;
  @Override public int indexOf ( final T o ) {
    validityCheck ( ) ;
    int count = 0 ;
    for ( final T item : this ) {
      if ( item.equals ( o ) ) { return count ; }
      ++count ;
    }
    return -1 ;
  }
  @Override public final void copyIn ( final Sequence<T> s , final int b , final int e ) {
    if ( ( b < 0 ) || ( e < 0 ) || ( b > e ) || ( b > s.size ( ) ) || ( e > s.size ( ) ) ) {
      throw new InvalidIteratorOperationException ( ) ;
    }
    //  Don't actually use indexing, use an iterator since indexing into a list is horrendously
    //  inefficient even arrays can handle iterators and indexing more or less equally efficiently.
    InputIterator<T> i = s.iterator ( ) ;
    i.advance ( b ) ;
    for ( int j = b ; j < e ; ++j , i.advance ( ) ) { add ( i.get ( ) ) ; }
  }
  @Override public void reverse ( ) { reverse ( 0 , size ( ) - 1 ) ; }
  @Override public void reverse ( final int b , final int e ) {
    if ( ( b < 0 ) || ( e < 0 ) || ( b > e ) || ( b > size ( ) ) || ( e > size ( ) ) ) {
      throw new InvalidIteratorOperationException ( ) ;
    }
    //  Don't actually use indexing to do the Sequence editing, use iterators since indexing into a list is
    //  horrendously inefficient whilst arrays can handle iterators and indexing more or less equally
    //  efficiently.
    final ForwardIterator<T> beginning = (ForwardIterator<T>) iterator ( ) ; beginning.advance ( b ) ;
    final ForwardIterator<T> end = (ForwardIterator<T>) iterator ( ) ; end.advance ( e ) ;
    reverse ( beginning , end ) ;
  }
   /**
   *  Reverse a slice of a <code>Sequence</code> defined using iterators.
   *
   *  @param b the beginning index of the slice to be reversed.
   *  @param e the end index of the slice to be reversed.
   *  @exception InvalidIteratorException if the iterators are not working the same container.
   *  @exception InvalidIteratorOperationException if the iterator <code>b</code> is not less than iterator
   *  <code>e</code> or either <code>b</code> or <code>e</code> are not valid indexes.
   */
  public static <E> void reverse ( final ForwardIterator<E> b , final ForwardIterator<E> e ) {
    if ( b.getContainer ( ) != e.getContainer ( ) ) { throw new InvalidIteratorException ( ) ; }
    final int distance = b.distance ( e ) ;
    if ( distance < 0 ) { throw new InvalidIteratorOperationException ( ) ; }
    //  We can do this operation neatly if we have bi-directional iterators.
    if ( ( b instanceof BiDirectionalIterator<?> ) && ( e instanceof BiDirectionalIterator<?> ) ) {
      //  Loop swapping the items pointed to by the two iterators and then advancing the earlier iterator
      //  and retreating the later iterator.
      for ( int i = 0 ; i < ( ( distance + 1 ) / 2 ) ; ++i , b.advance ( ) , ( (BiDirectionalIterator<E>) e ).retreat ( ) ) {
        final E temp = b.get ( ) ;
        b.set ( e.get ( ) ) ;
        e.set ( temp ) ;
      }
    } else {
      //  Loop twice from beginning to end.  On the first pass, put all the items into a stack.  On the
      //  second pass replace the Node datum with the next item from the stack.
      final Stack<E> s = new SequenceStack<E> ( ) ;
      for ( InputIterator<E> i = b.clone ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
        s.push ( i.get ( ) ) ;
        if ( i.equals ( e ) ) { break ; }
      }
      for (  ; ! b.atEnd ( ) ; b.advance ( ) ) {
        b.set ( s.pop ( ) ) ;
        if ( b.equals ( e ) ) { break ; }
      }
    }
  }
  @SuppressWarnings ( "unchecked" )
  @Override public boolean equals ( final Object o ) {
    if ( this == o ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof Sequence ) || ( size ( ) != ( (Sequence<T>) o ).size ( ) ) ) {
      return false ;
    }
    final InputIterator<T> i1 = iterator ( ) ;
    final InputIterator<T> i2 = ( (Sequence<T>) o ).iterator ( ) ;
    for ( ; ! i1.atEnd ( ) && ! i2.atEnd ( ) ; i1.advance ( ) , i2.advance ( ) ) {
      final T o1 = i1.get ( ) ;
      final T o2 = i2.get ( ) ;
      if ( ( o1 == null ) || ( o2 == null ) || ! o1.equals ( o2 ) ) { return false ; }
    }
    assert i1.atEnd ( ) && i2.atEnd ( )  ;
    return true ;
  }
  @Override public Sequence<T> clone ( ) { return (Sequence<T>) super.clone ( ) ; }
  /**
   *  Check for the correctness of the state.  This just checks that the <code>Sequence</code> actually has
   *  elements.
   *
   *  @throws AccessEmptyContainerException if the <code>Sequence</code> is empty.
   */
  protected void validityCheck ( ) {
    if ( size ( ) == 0 ) { throw new AccessEmptyContainerException ( ) ; }
  }
  /**
   *  Check for the correctness of the state.  This checks both that the <code>Sequence</code> actually has
   *  elements and that the proferred index is a legal one.
   *
   *  @param index the value to check that it is a valid index in this <code>Sequence</code>.
   *  @throws AccessEmptyContainerException  if the <code>Sequence</code> is empty.
   *  @throws IndexOutOfBoundsException if the <code>Sequence</code> is shorter than required by the index
   *  given.
  */
  protected void validityCheck ( final int index ) {
    if ( size ( ) == 0 ) { throw new AccessEmptyContainerException ( ) ; }
    if ( ( index < 0 ) || ( index >= size ( ) ) ) { throw new IndexOutOfBoundsException ( ) ; }
  }
}
