/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An <code>Array</code> is a <code>Sequence</code> where each element in the sequence can be accessed
 *  directly by indexing.  This implementation uses a primitive array as storage for the
 *  <code>Array</code>. The primitive array has an initial size which can be determined at construction time
 *  (default 10).  Whenever there is an attempt to add more items than the primitive array can hold then a
 *  resizing occurs with the increment being <code>incrementSize</code> (determined at construction time
 *  with a default of 10) or the number of items to be added whichever is the larger.  <code>Array</code>s
 *  never shrink.
 *
 *  @see Container
 *  @see Sequence
 *  @author Russel Winder
 */
public class Array<T> extends AbstractSequence<T> {
  //  NB We use Object[] here rather than T[] so as to avoid having to attempt to do things like "new
  //  T[...]" which is impossible or "(T[]) new Object [...]" which is unsafe.  We know that we are always
  //  dealing with T in the array but we never actually operate on the data we are just a container so we
  //  never use methods from T and as long as we always emit T values there is no problem using Object[]
  //  internally. (cf. Bug ID 6250148 at http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6250148.)
  private Object[] datum ;
  /**
   *  This is the current size of the filled portion of the array.  Any space in the array above this is not
   *  actually there as far as the application is concerned.
   */
  private int size = 0 ;
  /**
   *  The amount by which to increase the size of the backing primitive array when the size needs to be
   *  increased.
   */
  private final int increment ;
  /**
   *  Construct a zero size <code>Array</code> with increment value of 10.
   */
  public Array ( ) { this ( 0 , 10 ) ; }
  /**
   *  Construct an <code>Array</code> of size <code>size</code> with increment value of 10.
   *
   *  @param size the initial capacity of the array.
   */
  public Array ( final int s ) { this ( s , 10 ) ; }
  /**
   *  Construct an <code>Array</code> of size <code>size</code> with increment size of
   *  <code>increment</code>.
   *
   *  @param size the initial capacity of the array.
   *  @param increment the number of items to increment by when an increment is required.
   */
  public Array ( final int s , final int i ) {
    this.increment = i ;
    datum = new Object [s + i] ;
  }
  /**
   *  Construct an <code>Array</code> by copying in all the values from the
   *  <code>Container</code> parameter.
   *
   *  @param c the container to copy the items from.
   */
  public Array ( final Container<T> c ) {
    this ( c.size ( ) , 10 ) ;
    copyIn ( c ) ;
  }
  /**
   *  Construct an <code>Array</code> by copying the slice defined by two indexes into a
   *  <code>Sequence</code>s.
   *
   *  @param s the <code>Sequence</code> a slice of which is to be copied.
   *  @param b beginning index of the slice of <code>s</code>.
   *  @param e end index of the slice of <code>s</code>.
   */
  public Array ( final Sequence<T> s , final int b , final int e ) {
    this ( e - b + 1 , 10 ) ;
    copyIn ( s , b , e ) ;
  }
  /**
   *  Construct an <code>Array</code> by copying the sequence defined by two
   *  <code>InputIterator</code>s.
   *
   *  @param b beginning iterator of the sequence.
   *  @param e end iterator of the sequence.
   *  @exception InvalidIteratorException if the two <code>InputIterator</code>s are not working the
   *  same <code>Container</code> or e < b.
   */
  public Array ( final InputIterator<T> b , final InputIterator<T> e ) {
    this ( b.distance ( e ) , 10 ) ;
    copyIn ( b , e ) ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public final T get ( final int index ) { validityCheck ( index ) ; return (T) datum[index] ; }
  @Override public final void set ( final int index , final T o ) {
    try { validityCheck ( index ) ; }
    catch ( AccessEmptyContainerException aece ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw aece ; }
    } catch ( IndexOutOfBoundsException ioobe ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw ioobe ; }
    }
    datum[index] = o ;
  }
  @Override public final void remove ( final int index ) {
    validityCheck ( index ) ;
    System.arraycopy ( datum , index + 1 , datum , index , size - index ) ;
    --size ;
  }
  @Override public final void insert ( final int index , final T o ) {
    try { validityCheck ( index ) ; }
    catch (AccessEmptyContainerException aece) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw aece ; }
    }
    if ( size == datum.length ) {
      //  The array must be extended.  Create a new larger one.  Copy up to the insertion point insert the
      //  new element, copy the rest of the array.
      final Object[] newDatum = new Object [datum.length + increment] ;
      System.arraycopy ( datum , 0 , newDatum , 0 , index ) ;
      newDatum[index] = o ;
      System.arraycopy ( datum , index , newDatum , index + 1 , datum.length - index ) ;
      datum = newDatum ;
    } else {
      for ( int i = size ; i > index ; --i ) { datum[i] = datum[i-1] ; }
      datum[index] = o ;
    }
    ++size ;
  }
  @Override public final void makeEmpty ( ) {
    for ( int i = 0 ; i < size ; ++i ) { datum[i] = null ; }
    size = 0 ;
  }
  @Override public final boolean isEmpty ( ) { return size == 0 ; }
  @Override public final int size ( ) { return size ; }
  @Override public final void add ( final T o ) {
    if ( size == datum.length ) {
      final Object[] newDatum = new Object [datum.length + increment] ;
      System.arraycopy ( datum , 0 , newDatum , 0 , datum.length ) ;
      datum = newDatum ;
    }
    datum[size++] = o ;
  }
  @Override public final boolean remove ( final T o ) {
    validityCheck ( ) ;
    final int i = indexOf ( o ) ;
    if ( i < 0 ) { return false ; }
    remove ( i ) ;
    return true ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public T[] toPrimitiveArray ( final Class<T> c ) {
    //  Cannot use "(T[]) new Object [size]" since even though it seems type correct, it isn't due to
    //  erasure issues so in order to really deliver a T[] we must use reflection.  Hence the Class<T>
    //  parameter.
    final int s = size ( ) ;
    final T[] data = (T[]) java.lang.reflect.Array.newInstance ( c , s ) ;
    System.arraycopy ( datum , 0 , data , 0 , s ) ;
    return data ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public T[] toPrimitiveArray ( T[] a ) {
    final int s = size ( ) ;
    if ( a.length < s ) {
      a = (T[]) java.lang.reflect.Array.newInstance ( a.getClass ( ).getComponentType ( ) , s ) ;
    }
    System.arraycopy ( datum , 0 , a , 0 , size ) ;
    if ( a.length > s ) {
      for ( int i = a.length ; i < size ; ++i ) { a[i] = null ; }
    }
    return a ;
  }
  @Override public final Array<T> clone ( ) {
    final Array<T> newItem = (Array<T>) super.clone ( ) ;
    // Must clone the representation which is currently shared but do not need to clone the data objects
    // themselves since it is fine to share them.
    newItem.datum = datum.clone ( ) ;
    return newItem ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.position = size ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  The iterator over <code>Array</code>s.  This is an inner class (as opposed to being a nested top-level
   *  class) so that it automatically has a reference to it's parent object, i.e. the object it is an
   *  iterator for.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> implements RandomAccessIterator<T> {
    /**
     *  The current location of the iterator.  Since indexing is efficient for this data structure, we keep
     *  only the current index position.
     */
    private int position = 0 ;
    @SuppressWarnings ( "unchecked" )
    @Override public final T get ( ) { validityCheck ( ) ; return (T) datum[position] ; }
    @Override public final void set (final T o ) { validityCheck ( ) ; datum[position] = o ; }
    @Override public final void advance ( ) { if ( ! atEnd ( ) ) { ++position ; } }
    @Override public final void advance ( final int i ) {
      if ( ! atEnd ( ) ) {
        position += i ;
        if ( position > size ) { position = size ; }
        if ( position < 0 ) { position = 0 ; }
      }
    }
    @Override public final void retreat ( ) { if ( ! atBegin ( ) ) { --position ; } }
    @Override public final void retreat ( final int i ) {
      if ( ! atBegin ( ) ) {
        position -= i ;
        if ( position > size ) { position = size ; }
        if ( position < 0 ) { position = 0 ; }
      }
    }
    @Override public final int index ( ) { return position ; }
    @Override public final int distance ( final InputIterator<T> i ) {
      if ( getContainer ( ) != i.getContainer ( ) ) { throw new InvalidIteratorException ( ) ; }
      return ( (Iterator) i ).position - position ;
    }
    @Override public final boolean atBegin ( ) { return position == 0 ; }
    @Override public final boolean atEnd ( ) { return position == size ; }
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof Array.Iterator ) ) { return false ; } // IntelliJ IDEA incorrectly says the Array. is unnecessary.
      @SuppressWarnings ( "unchecked" )
      final Iterator i = (Iterator) o ;
      return ( getContainer ( ) == i.getContainer ( ) ) && ( position == i.position ) ;
    }
    @Override public Iterator clone ( ) { return (Iterator) super.clone ( ) ; }
    @Override public final Array<T> getContainer ( ) { return Array.this ; }
    /**
     *  Check for the correctness of the state.  This checks both that the <code>Array</code> actually has
     *  elements and that the index is a legal one.
     *
     *  @throws AccessEmptyContainerException if the <code>Array</code> is empty.
     *  @throws InvalidIteratorException if the iterator state means that it is not pointing at a valid item
     *  in the <code>Array</code>.
     */
    private void validityCheck ( ) {
      if ( size == 0 ) { throw new AccessEmptyContainerException ( ) ; }
      if ( (position < 0) || (position >= size) ) { throw new InvalidIteratorException ( ) ; }
    }
  }
}
