/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Queue</code> implemented using an associated <code>Array</code>.  This is abstraction by
 *  conformance.
 *
 *  <p> The head (the place from which items will be withdrawn) of the <code>Queue</code> is position 0 in
 *  the <code>Array</code>, the tail (the place to which new items will be appended) is the last
 *  element.</p>
 *
 *  @author Russel Winder
 */
public class ArrayQueue<T> extends AbstractQueue<T> {
  /**
   *  The underlying representation.
   */
  private Array<T> storage ;
  //
  //  Arrays can be constructed in three ways, specifying the base size and the extension size, just the
  //  base size using the default extension size or using the defaults for both. Reflect this in our
  //  constructors.
  /**
   *  Default constructor.
   */
  public ArrayQueue ( ) { this (10 , 10 ) ; }
  /**
   *  Constructor defining the initial size of the <code>Array</code> we are using to actually implement
   *  this form of <code>Queue</code>.
   *
   *  @param initialSize The initial size of the <code>Queue</code>.
   */
  public ArrayQueue ( final int initialSize ) { this ( initialSize , 10 ) ; }
  /**
   *  Constructor defining the initial size of the <code>Array</code> we are using to actually implement
   *  this form of <code>Queue</code> and also the increment whenever the <code>Array</code> needs
   *  extending.
   *
   *  @param initialSize The initial size of the <code>Queue</code>.
   *  @param incrementSize The size of the <code>Queue</code> space increment.
   */
  public ArrayQueue ( final int initialSize , final int incrementSize ) {
    storage = new Array<T> ( initialSize , incrementSize ) ;
  }
  /**
   *  Constructor that forces a <code>Queue</code> onto the backing <code>Array</code>.
   *
   *  @param a the <code>Array</code> to use as backing for the <code>ArrayQueue</code>
   */
  public ArrayQueue ( final Array<T> a ) { storage = a ; }
  @Override public final  T peek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return storage.first ( ) ; }
  @Override public final T remove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    final T o = storage.first ( ) ;
    storage.remove ( 0 ) ;
    return o ;
  }
  @Override public final void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public final int size ( ) { return storage.size ( ) ; }
  @Override public final void add ( final T o ) { storage.add ( o ) ; }
  @Override public final ArrayQueue<T> clone ( ) {
    final ArrayQueue<T> q = (ArrayQueue<T>) super.clone ( ) ;
    q.storage = storage.clone ( ) ;
    return q ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  An iterator over <code>ArrayQueue</code>s.
   *
   *  <p>This is just a proxy for an iterator from the underlying representation so as to stop the ability
   *  to write.  Implement only a forward iterator.</p>
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  An iterator from the underlying representation.
     */
    private InputIterator<T> current = storage.iterator ( ) ;
    @Override public final T get ( ) { return current.get ( ) ; }
    @Override public final void advance ( ) { current.advance ( ) ; }
    @Override public final void advance ( final int increment ) { current.advance ( increment ) ; }
    @Override public final boolean atBegin ( ) { return current.atBegin ( ) ; }
    @Override public final boolean atEnd ( ) { return current.atEnd ( ) ; }
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof ArrayQueue<?>.Iterator ) ) { return false ; } // IntelliJ IDEA incorrectly says the ArrayQueue. is unecessary.
      @SuppressWarnings ( "unchecked" )
      final Iterator i = (Iterator) o ;
      return current.equals ( i.current ) ;
    }
    @Override public final ArrayQueue<T> getContainer ( ) { return ArrayQueue.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.current = current.clone ( ) ;
      return i ;
    }
  }
}
