/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  The interface describing a bag.  A bag is like a set except that elements need not be unique, i.e. a bag
 *  can hold more than one instance of a value whereas a set cannot.  Technically a set is a bag where there
 *  is only ever a single instance of a given item and so it could be argued that <code>Set</code> should be
 *  a subclass of <code>Bag</code>.  However, reversing this relationship and making a <code>Bag</code> a
 *  subclass of <code>Set</code> as a pure implementation artefact leads to a more efficient implementation
 *  of the set idea and sets are more used than bags.  So in terms of classification the inheritance
 *  relationship is the wrong way round, in terms of implementation technique it is the right way round.
 *
 *  @author Russel Winder
 */
public interface Bag<T> extends Set<T> {
  /**
   *  Determine the number of times an item appears in the <code>Bag</code>.
   *
   *  @param o the object to search for and count.
   *  @return the number of times <code>o</code> is in this <code>Bag</code>.
   */
  int countOf ( T o ) ;
  /**
   *  Clone this <code>Bag</code>.
   *
   *  @return a clone of this <code>Bag</code>.
   */
  @Override Bag<T> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>Bag</code> for use in multi-threaded
   *  applications.  This is just a forwarding proxy with synchronization control and locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<E> extends Set.Synchronized<E> implements Bag<E> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param b the <code>Bag</code> to synchronize.
     */
    public Synchronized ( final Bag<E> b ) { super ( b ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param b the <code>Bag</code> to synchronize.
     *  @param o the object to use as a lock object for synchronizing.
     */
    public Synchronized ( final Bag<E> b , final Object o ) { super ( b , o ) ; }
    @Override public int countOf ( final E o ) {
      synchronized ( theLock ) { return ( (Bag<E>) theContainer ).countOf ( o ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  Bag.Synchronized, Set.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public Bag.Synchronized<E> clone ( ) { return (Bag.Synchronized<E>) super.clone ( ) ; }
  }
}
