/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  The abstraction of a binary tree.
 *
 *  <p> A <code>BinaryTree</code> is not a closed abstract data type but, like a linked list, is really more
 *  of a data structure implementation mechanism. </p>
 *
 *  @author Russel Winder
 */
public interface BinaryTree<T> extends Container<T> {
  /**
   *  The method to deliver a post-order traversal iterator.  Post-order traversal of the
   *  <code>BinaryTree</code> visits the nodes in the order: left child, right child, current node.
   *
   *  @return a post-order iterator over this <code>BinaryTree</code>
   */
  InputIterator<T> postorderTraversal ( ) ;
  /**
   *  The method to deliver a pre-order traversal.  Preorder traversal means visit nodes in the order:
   *  current node, left child, right child.
   *
   *  @return a pre-order iterator over this <code>BinaryTree</code>
   */
  InputIterator<T> preorderTraversal ( ) ;
  /**
   *  The method to deliver a in-order traversal In-order traversal visits the nodes in the order: left
   *  child, current node, right child.
   *
   *  @return a in-order iterator over this <code>BinaryTree</code>
   */
  InputIterator<T> inorderTraversal ( ) ;
  /**
   *  The method to deliver a level-order traversal. Level-order traversal visits the nodes left-to-right,
   *  level-by-level.  It is a breadth-first traversal rather than the depth-first traversal that the others
   *  basically are.
   *
   *  @return a level-order iterator over this <code>BinaryTree</code>
   */
  InputIterator<T> levelorderTraversal ( ) ;
  /**
   *  Clone this <code>BinaryTree</code>.
   *
   *  @return a clone of this <code>BinaryTree</code>.
   */
  @Override BinaryTree<T> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>BinaryTree</code> for use in
   *  multi-threaded applications. This is just a forwarding proxy with synchronization control and locking.
   *
   *  <p>NB The user is responsible for all synchronization issues associated with iterators.</p>
   *
   *  @author Russel Winder
   */
  class Synchronized<E> extends Container.Synchronized<E> implements BinaryTree<E> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param t the <code>BinaryTree</code> to synchronize.
     */
    public Synchronized ( final BinaryTree<E> t ) { super ( t ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param t the <code>BinaryTree</code> to synchronize.
     *  @param o the object to use as a lock object.
     */
    public Synchronized ( final BinaryTree<E> t , final Object o ) { super ( t , o ) ; }
    @Override public final InputIterator<E> postorderTraversal ( ) {
      return ( (BinaryTree<E>) theContainer ).postorderTraversal ( ) ;
    }
    @Override public final InputIterator<E> preorderTraversal ( ) {
      return ( (BinaryTree<E>) theContainer ).preorderTraversal ( ) ;
    }
    @Override public final InputIterator<E> inorderTraversal ( ) {
      return ( (BinaryTree<E>) theContainer ).inorderTraversal ( ) ;
    }
    @Override public final InputIterator<E> levelorderTraversal ( ) {
      return ( (BinaryTree<E>) theContainer ).levelorderTraversal ( ) ;
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  BinaryTree.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public  BinaryTree.Synchronized<E> clone ( ) { return (BinaryTree.Synchronized<E>) super.clone ( ) ; }
  }
}
