/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An hash table that uses chaining.  In this implementation, the chains of the hash table are
 *  <code>DLList&lt;T></code>s.  Use this rather than the simpler <code>SLList&lt;T></code> so that editing
 *  the list in the middle is easier.  We need this capability only to support removal of an item.
 *
 *  @author Russel Winder
 */
public class ChainedHashTable<K, V> extends AbstractHashTable<K, V> {
  /**
   *  The storage for the hash table.  Each entry in the table is a list of ( key , value ) pairs that
   *  deliver the same hash value.
   */
  protected DLList<Pair<K, V>>[] table ;
  /**
   *  Keep a count of the number of items to make things more efficient.
   */
  protected int pairCount = 0 ;
  /**
   *  The hash function method object defining our hash function.  The hash function guarantees to deliver a
   *  positive integer.  The table indexing operations are responsible for ensuring that the integer is a
   *  valid index.  On most occassions this simply means taking the modulus of the hash value against the
   *  size of the array being used as representation.
   */
  protected HashFunction<K> hashFunction ;
  /**
   *  The default size of the hash table.  Best if this is a prime number.
   */
  protected static final int DEFAULT_SIZE = 511 ;
  /**
   *  The default constructor. Use the default size of 511 and the default hash function as supplied by the
   *  class <code>HashFunction&lt;K></code>.
   */
  public ChainedHashTable ( ) { this ( DEFAULT_SIZE , new HashFunction<K> ( ) ) ; }
  /**
   *  The constructor specifying the size of the table but using the default hash function as supplied by
   *  the class <code>HashFunction&lt;K></code>.
   *
   *  @param tableSize initial siuze of the hash table.
   */
  public ChainedHashTable ( final int tableSize ) { this ( tableSize , new HashFunction<K> ( ) ) ; }
  /**
   *  The constructor that assumes the default size but provides an hash function..
   *
   *  @param f the hash function to use.
   */
  public ChainedHashTable ( final HashFunction<K> f ) { this ( DEFAULT_SIZE , f ) ; }
  /**
   *  The constructor specifying the size of the table and the hash function.
   *
   *  @param tableSize initial siuze of the hash table.
   *  @param f the hash function to use.
   */
  @SuppressWarnings ( "unchecked" )
  public ChainedHashTable ( final int tableSize , final HashFunction<K> f ) {
    table = (DLList<Pair<K, V>>[]) new Object [tableSize] ;
    hashFunction = f ;
  }
  /**
   *  Remove all the references from the hash table.
   */
  @Override public void makeEmpty ( ) {
    //  Leave the garbage collector to handle all the clean up by just releasing all the DLLists to be
    //  garbage.
    for ( int i = 0 ; i < table.length ; ++i ) { table[i] = null ; }
    pairCount = 0 ;
  }
  /**
   *  Determine whether the <code>ChainedHashTable&lt;K, V></code> is an empty one.
   */
  @Override public final boolean isEmpty ( ) { return pairCount <= 0 ; }
  /**
   *  Return the number of items in the <code>ChainedHashTable&lt;K, V></code>.
   */
  @Override public final int size ( ) { return pairCount ; }
  /**
   *  Insert a ( key , value ) pair into the <code>ChainedHashTable&lt;K, V></code>.  Since removal of items
   *  is allowed we do not just add the new item, we remove the old item associated with key -- or (key,
   *  value), depending on whether equality is key equality or <code>Pair&lt;K, V></code> equality --
   *  i.e. we do not use shadowing.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public void add ( final Pair<K, V> o ) {
    final int index = findIndex ( o.getA ( ) ) ;
    if ( table[index] == null ) { table[index] = (DLList<Pair<K, V>>) new Object ( ) ; }
    for ( DLList<Pair<K, V>>.Iterator i = table[index].iterator ( ) ; ! i.atEnd ( ) ; i.advance( ) ) {
      if ( i.get ( ).equals ( o ) ) {
        i.remove ( ) ;
        --pairCount ;
        break ;
      }
    }
    table[index].add ( o ) ;
    ++pairCount ;
  }
  /**
   *  Find a value for a given key.
   *
   *  @return value from (key, value) pair given a key or null if the key is not present in the
   *  <code>Table</code>.
   */
  @Override public V getValue ( final K key ) {
    for ( InputIterator<Pair<K, V>> i = table[findIndex ( key )].iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
      final Pair<K, V> p = i.get ( ) ;
      if ( p.getA ( ).equals ( key ) ) { return p.getB ( ) ; }
    }
    return null ;
  }
  /**
   *  Remove the ( key, value ) pair.
   */
  @Override public boolean remove ( final Pair<K, V> o ) {
    final DLList<Pair<K, V>> d = table[findIndex ( o.getA ( ) )] ;
    if ( (d != null) && ! d.isEmpty ( ) ) {
      if ( d.remove ( o ) ) {
        --pairCount ;
        return true ;
      }
    }
    return false ;
  }
  /**
   *  Compare this <code>OpenHashTable&lt;K, V></code> with another and determine whether they represent the
   *  same value, i.e. have the same item values in the same order.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public boolean equals (final Object o ) {
    if ( o == this ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof ChainedHashTable ) ) { return false ; }
    final ChainedHashTable<K, V> other = (ChainedHashTable<K, V>) o ;
    for ( int i = 0 ; i < table.length ; ++i ) {
      if ( table[i] == null ) {
        if ( other.table[i] != null ) { return false ; }
      }
      else {
        if ( ! table[i].equals ( other.table[i] ) ) { return false ; }
      }
    }
    return true ;
  }
  /**
   *  Deliver up a complete shallow copy of the <code>ChainedHashTable&lt;K, V></code>.  NB
   *  <code>CloneNotSupportedException</code> is caught in <code>AbstractContainer&lt;Pair&lt;K, V>></code>
   *  so we don't need to worry about it here.
   */
  @Override public ChainedHashTable<K, V> clone ( ) {
    final ChainedHashTable<K, V> h = (ChainedHashTable<K, V>) super.clone ( ) ;
    h.table = table.clone ( ) ;
    for ( int i = 0 ; i < table.length ; ++i ) {
      if ( table[i] != null ) { h.table[i] = table[i].clone ( ) ; }
    }
    return h ;
  }
  /**
   *  Deliver up an iterator over all the items in the <code>ChainedHashTable&lt;K, V></code>.
   */
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  /**
   *  Deliver up an iterator over all the items in the <code>ChainedHashTable&lt;K, V></code>.
   */
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  /**
   *  Deliver up an iterator over all the items in the <code>ChainedHashTable&lt;K, V></code>.
   */
  @Override public Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  The basic iterator.  Subclass this to get different behaviours as needed.
   *
   *  @author Russel Winder
   */
  public  class Iterator extends AbstractIterator<Pair<K, V>> {
    //
    //  Start the index at -1 since the first action of the instance initializer of cursor is to call the
    //  method to set up the iterator and its first action is to increment this index.  The first used value
    //  must be 0 so we initialize to -1.  NB This declaration must come before the setting up of cursor so
    //  that things get properly initialized: it is important that index has the value -1 before
    //  setUpNextIterator is called.  We need to remember that the initializations happen in the order they
    //  are written down!
    /**
     *  Keep note of which <code>DLList&lt;Pair&lt;K, T>></code> we are working on.
     */
    protected int index = -1 ;
    /**
     *  The index to the first item.  Cache this to support <code>atBegin</code>.
     */
    protected final int beginIndex ;
    /**
     *   A cursor is a <code>DLList&lt;Pair&lt;K, T>>.Iterator</code>.
     */
    protected DLList<Pair<K, V>>.Iterator cursor = null ; {
      setUpNextIterator ( ) ;
      beginIndex = index ;
    }
    /**
     *  Deliver the value we are currently referring to.
     */
    @Override public Pair<K, V> get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorOperationException ( ) ; }
      return cursor.get() ;
    }
    /**
     *  Move on to the next item in the container.
     */
    @Override public void advance ( ) {
      if ( ! atEnd ( ) ) {
        if ( cursor.atEnd ( ) ) { setUpNextIterator ( ) ; }
        else {
          cursor.advance ( ) ;
          if ( cursor.atEnd ( ) ) { setUpNextIterator ( ) ; }
        }
      }
    }
    /**
     *  Move on n items in the container.
     */
    @Override public void advance ( int increment ) {
      for ( ; increment > 0 ; --increment ) {
        if ( atEnd ( ) ) { break ; }
        advance ( ) ;
      }
    }
    /**
     *  Are we positioned at the beginning of the iteration?
     */
   @Override  public boolean atBegin ( ) { return ( index == beginIndex ) && cursor.atBegin ( ) ; }
    /**
     *  Have we reached the end of the iteration?
     */
    @Override public boolean atEnd ( ) { return ( index == -1 ) || ( cursor == null ) ; }
    /**
     *  Are two iterators equal, i.e. do two iterators refer to the same item in the same data structure.
     */
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof ChainedHashTable.Iterator ) ) { return false ; }
      final Iterator i = (Iterator) o ;
      return ( index == i.index ) && cursor.equals ( i.cursor ) ;
    }
    /**
     *  Return a reference to the <code>ChainedHashTable&lt;K, V></code> of this iterator.
     */
    @Override public ChainedHashTable<K, V> getContainer ( ) { return ChainedHashTable.this ; }
    /**
     *  Clone this iterator.
     */
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.cursor = cursor.clone ( ) ;
      return i ;
    }
    /**
     *  This is an editing iterator so we allow removal.
     */
    @Override public void remove ( ) {
      if ( ! atEnd ( ) ) {
        cursor.remove ( ) ;
        --pairCount ;
      }
    }
    /**
     *  Set up the cursor to refer to a valid <code>DLList&lt;Pair&lt;K, T>>.  Utility method used in the
     *  constructor and in advance.
     */
    protected void setUpNextIterator ( ) {
      while( ++index < table.length ) {
        if ( table[index] != null ) {
          if ( table[index].size ( ) > 0 ) {
            cursor = table[index].iterator ( ) ;
            return ;
          }
        }
      }
      cursor = null ;
      index = -1 ;
    }
  }
  /**
   *  Find the index of a given key, i.e. the position in the table of an entry with this key or to the
   *  first null entry.
   *
   *  <p> This is as near O(1) as we can get.
   *
   *  @param o the item to be found.
   *  @return the index into table of the value the entry is null is key is not in the table.
   */
  private int findIndex ( final K o ) { return hashFunction.hash ( o ) % table.length ; }
}
