/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A set implemented using an <code>ChainedHashTable</code>.
 *
 *  @author Russel Winder
 */
public class ChainedHashTableSet<T> extends AbstractSet<T> {
  /**
   *  The ChainedHashTable that acts as our storage.
   */
  protected ChainedHashTable<T, T> storage ;
  /**
   *  Construct an <code>ChainedHashTableSet</code>.
   */
  public ChainedHashTableSet ( ) { storage = new ChainedHashTable<T, T> ( ) ; }
  @Override public void makeEmpty ( ) { storage.makeEmpty() ; }
  @Override public int size ( ) { return storage.size() ; }
  @Override public void add ( final T o) { storage.add ( new Pair<T, T> ( o , null ) ) ; }
  @Override public boolean remove ( final T o ) { return storage.remove ( new Pair<T, T> ( o , null ) ) ; }
  @Override public ChainedHashTableSet<T> clone ( ) {
    final ChainedHashTableSet<T> t = (ChainedHashTableSet<T>) super.clone ( ) ;
    if ( ! isEmpty ( ) ) { t.storage = storage.clone ( ) ; }
    return t ;
  }
  @Override public boolean contains ( final T o ) { return storage.contains ( new Pair<T, T> ( o , null ) ) ; }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  Define the iterator over <code>ChainedHashTableSet</code>s.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  The cursor is just an iterator over the <code>ChainedHashTable&lt;T, T></code>. Use the default of
     *  in-order traversal.
     */
    protected InputIterator<Pair<T, T>> cursor = storage.iterator ( ) ;
    @Override public T get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorOperationException ( ) ; }
      return cursor.get ( ).getA ( ) ;
    }
    @Override public void advance ( ) { cursor.advance ( ) ; }
    @Override public void advance ( final int increment ) { cursor.advance ( increment ) ; }
    @Override public boolean atBegin ( ) { return cursor.atBegin ( ) ; }
    @Override public boolean atEnd ( ) { return cursor.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( this == o ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof ChainedHashTableSet.Iterator ) ) { return false ; }
      return cursor.equals ( ( (Iterator) o ).cursor ) ;
    }
    @Override public ChainedHashTableSet<T> getContainer ( ) { return ChainedHashTableSet.this ; }
    @Override public final Iterator clone() {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.cursor = cursor.clone ( ) ;
      return i ;
    }
  }
}
