/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  The interface that defines the necessary properties to be deemed to be a container in the ADS package.
 *
 *  <p>This interface inherits from <code>Iterable</code> as part of the infrastructure that allows ADS
 *  containers to be used with the <code>for ( E e : c ) { . . . }</code> loops.</p>
 *
 *  @author Russel Winder
 */
public interface Container<T> extends Cloneable, Iterable<T> {
  /**
   *  Remove all the items from this <code>Container</code>.
   */
  void makeEmpty ( ) ;
  /**
   *  Determine whether this <code>Container</code> is empty.
   *
   *  @return <code>true</code> if the <code>Container</code> is empty and <code>false</code> otherwise.
   */
  boolean isEmpty ( ) ;
  /**
   *  Return the number of items in this <code>Container</code>.
   *
   *  @return the number of items in this <code>Container</code>.
   */
  int size ( ) ;
  /**
   *  Compare this <code>Container</code> with another object and determine whether the two represent the
   *  same value, i.e. are of the same type and contain the same values: Implements value equality rather
   *  than object identity equality.
   *
   *  <p>The method <code>Object.equals</code> is defined to be object identity equality, i.e. two
   *  <code>Object</code>s are equal if and only if they are the same <code>Object</code>.  For
   *  <code>Container</code>s, we want to define equality to be that they contain the same values.</p>
   *
   *  <p>NB Must take <code>Object</code> as parameter to overload the definition in
   *  <code>Object</code>.</p>
   *
   *  @param o the object to be tested for equality.
   *  @return true if the containers contain the same values false otherwise.
   */
  @Override boolean equals ( Object o ) ;
  /**
   *  Determine whether this <code>Container</code> contains a given value.
   *
   *  @param o the object to be tested for being cintained in this <code>Container</code>.
   *  @return true if the value of the item referred to by the parameter is present in the
   *  <code>Container</code> false otherwise.
   */
  boolean contains ( T o ) ;
  /**
   *  Add an element to this <code>Container</code>.
   *
   *  @param o the object to be added to the <code>Container</code>.
   */
  void add ( T o ) ;
  /**
   *  Remove an element of a given value from this <code>Container</code> if it is in the
   *  <code>Container</code>.
   *
   *  @param o the value to remove from the <code>Container</code>.
   *  @return whether the item was actually in this <code>Container</code> and removed.
   *  @throws AccessEmptyContainerException if there are no items in the <code>SLList</code>.
   */
  boolean remove ( T o ) ;
  /**
   *  Create a primitive array from a <code>Container</code>.  The method is repsonsible for creating the
   *  array that is returned.
   *
   *  <p>It seems strange that this method has to take a parameter.  The problem stems from the need to
   *  create a primitive array within the method and the fact that, due to erasure, information about the
   *  type parameter is not available at runtime except via reflection.  We have to provide a vehicle for
   *  obtaining the correct type hence the @code{Class<T>} parameter.  Damn silly really but
   *  . . . </p>
   *
   *  <p>cf. Bugs numbered 5072831, 5094009, 5101892 and 6201178 at http://bugs.sun.com/bugdatabase.</p>
   *
   *  @param c the class object that provides information about <code>T</code> for the reflective
   *  construction of the array.
   *  @return reference to the created array.
   */
  T[] toPrimitiveArray ( final Class<T> c ) ;
  /**
   *  Create a primitive array from a <code>Container</code>.  The intention here is that the parameter is a
   *  space into which to put the values.  If it is bug enough then it is used otherwise a new array is
   *  created and returned.
   *
   *  <p>It seems strange that this method has to take a parameter.  The problem stems from the need to
   *  create a primitive array within the method and the fact that, due to erasure, information about the
   *  type parameter is not available at runtime except via reflection.  We have to provide a vehicle for
   *  obtaining the correct type hence the @code{Class<T>} parameter.  Damn silly really but
   *  . . . </p>
   *
   *  <p>cf. Bugs numbered 5072831, 5094009, 5101892 and 6201178 at http://bugs.sun.com/bugdatabase.</p>
   *
   *  @param a an array in which to put the values from the container if the array has enough space to hold
   *  all the items.  If the array is too small a new one is created.
   *  @return a reference to the array containing the data.
   */
  T[] toPrimitiveArray ( final T[] a ) ;
  /**
   *  Clone this <code>Container</code>.
   *
   *  @return a clone of this <code>Container</code>.
   */
  Container<T> clone ( ) ;
  /**
   *  Create a string representation of this <code>Container</code>.
   *
   *  @return a string representation of this <code>Container</code>.
   */
  @Override String toString ( ) ;
  /**
   *  Provide an iterator positioned at the beginning of the iteration sequence.
   *
   *  @return an iterator over the <code>Container</code> positioned at the beginning of the iteration
   *  sequence.
   */
  InputIterator<T> begin ( ) ;
  /**
   *  Provide an iterator positioned at the end of the iteration sequence.
   *
   *  @return an iterator over the <code>Container</code> positioned at the end of the iterations sequence,
   *  i.e. such that <code>atEnd() == true</code>.
   */
  InputIterator<T> end ( ) ;
  /**
   *  Provide an iterator positioned at the beginning of the iteration sequence.  This is invariably a
   *  synonym for <code>begin</code> and is provided really just for code readability where
   *  <em>iterator</em> is more the right concept than <em>begin</em>.
   *
   *  @return an iterator over the <code>Container</code>.
   */
  @Override InputIterator<T> iterator ( ) ;
  /**
   *  A proxy implementing synchronization for <code>Container</code>s .
   *
   *  <p>All <code>Container</code>s are unsynchronized, i.e. not thread safe.  This class is an adapter
   *  that provides the synchronization functionality required to use <code>Container</code>s in a
   *  multi-threaded context.  The class creates proxies that have a reference to the actual
   *  <code>Container</code> and a lock object which is used to implement the the semaphore required to
   *  serialize potentially concurrent access.</p>
   *
   *  <p>This is just a forwarding proxy with synchronization control and locking.</p>
   *
   *  @author Russel Winder
   */
  class Synchronized<E> implements Container<E> {
    //  Cannot be final due to clone needing to amend the values.
    /**
     *  Reference to the object we are the synchronization proxy for.
     */
    protected /*final*/ Container<E> theContainer ;
    /**
     *  Reference to the lock object for synchronization.
     */
    protected /*final*/ Object theLock ;
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param c the <code>Container</code> to be synchronized.
     */
    public Synchronized ( final Container<E> c ) {
      theContainer = c ;
      theLock = this ;
    }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param c the <code>Container</code> to be synchronized.
     *  @param o the object to use as the lock.
     */
    public Synchronized ( final Container<E> c , final Object o ) {
      theContainer = c ;
      theLock = o ;
    }
    //  Javadoc comments will be copied for these methods so no need to replicate, javadoc does it for us!
    @Override public final void makeEmpty ( ) {
      synchronized ( theLock ) { theContainer.makeEmpty ( ) ; }
    }
    @Override public final boolean isEmpty ( ) {
      synchronized ( theLock ) { return theContainer.isEmpty ( ) ; }
    }
    @Override public final int size ( ) {
      synchronized ( theLock ) { return theContainer.size ( ) ; }
    }
    @Override public final boolean equals ( final Object o ) {
      synchronized ( theLock ) { return theContainer.equals ( o ) ; }
    }
    @Override public final boolean contains ( final E o ) {
      synchronized ( theLock ) { return theContainer.contains ( o ) ; }
    }
    @Override public final void add ( final E o ) {
      synchronized ( theLock ) { theContainer.add ( o ) ; }
    }
    @Override public final boolean remove ( final E o ) {
      synchronized ( theLock ) { return theContainer.remove ( o ) ; }
    }
    @Override public final E[] toPrimitiveArray ( final Class<E> c ) {
      synchronized ( theLock ) { return theContainer.toPrimitiveArray ( c ) ; }
    }
    @Override public final E[] toPrimitiveArray ( final E[] a ) {
      synchronized ( theLock ) { return theContainer.toPrimitiveArray ( a ) ; }
    }
    @SuppressWarnings ( "unchecked" )
    @Override public Synchronized<E> clone ( ){
      try {
        final Synchronized<E> c = (Synchronized<E>) super.clone ( ) ;
        c.theContainer = theContainer.clone ( ) ;
        c.theLock = new Object ( ) ; // Can't do theLock.clone ( ) ; because clone is protected in Object.
        return c ;
      }
      catch ( CloneNotSupportedException cnse ) { throw new InternalError ( ) ; }
    }
    @Override public final String toString ( ) {
      synchronized ( theLock ) { return theContainer.toString ( ) ; }
    }
    @Override public InputIterator<E> begin ( ) { return theContainer.begin ( ) ; }
    @Override public InputIterator<E> end ( ) { return theContainer.end ( ) ; }
    @Override public InputIterator<E> iterator ( ) { return theContainer.iterator ( ) ; }
  }
}
