/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Sequence</code> implemented as a doubly-linked list.
 *
 *  <p> This implementation uses a dummy <code>Node</code> to act as both ends of the list of doubly-linked
 *  elements. For each <code>Node</code>, previous points from the tail to the head and next points from the
 *  head to the tail.</p>
 *
 *  @see Container
 *  @see Sequence
 *  @author Russel Winder
 */
public class DLList<T> extends AbstractSequence<T> {
  /**
   *  This is the class from which linked-list nodes are instantiated.  This node knows how to undertake
   *  various editing operations. This is possible because the doubly linked nature of the list means that
   *  any node has all the information to do the editing.  Although this is quite an active sort of node it
   *  is a nested top-level class and not an inner class since it has no need to ever know who its parent
   *  is.
   *
   *  @author Russel Winder
   */
  protected static class Node<E> {
    /**
     *  A <code>Node</code> has some datum that it is storing.
     */
    protected E datum ;
    /**
     *  A <code>Node</code> has a next, a reference to another <code>Node</code>.
     */
    protected Node<E> next ;
    /**
     *  A <code>Node</code> has a previous, a reference to another <code>Node</code>.
     */
    protected Node<E> previous ;
    /**
     *  Constructor for an unconnected <code>Node</code>.
     *
     *  @param o The datum for this <code>Node</code>.
     */
    public Node ( final E o ) { this ( o , null , null ) ; }
    /**
     *  Constructor for a <code>Node</code> with a given list of <code>Node</code>s to become the head of.
     *
     *  @param o The datum for this <code>Node</code>.
     *  @param n The <code>Node</code> to be the tail of this one.
     *  @param p The <code>Node</code> to be the previous of this one.
     */
    public Node ( final E o , final Node<E> n , final Node<E> p ) {
      datum = o ;
      next = n ;
      previous = p ;
    }
    /**
     *  Append an element after ourself.
     */
    public final void add ( final Node<E> n ) {
      n.previous = this ;
      n.next = next ;
      next.previous = n ;
      next = n ;
    }
    /**
     *  Excise ourself from the <code>DLList</code>.
     */
    public final void remove ( ) {
      next.previous = previous ;
      previous.next = next ;
      next = null ;
      previous = null ;
    }
    /**
     *  Insert an element before ourself.
     */
    public final void insert ( final Node<E> n ) {
      n.next = this ;
      n.previous = previous ;
      previous.next = n ;
      previous = n ;
    }
    /**
     *  Merge a <code>DLList</code> after ourself.
     */
    public final void merge ( final DLList<E> l ) {
      next.previous = l.endNode.previous ;
      l.endNode.previous.next = next ;
      next = l.endNode.next ;
      l.endNode.next.previous = this ;
    }
  }
  /**
   *  The <code>Node</code> that does not have any data but acts as both the head and the tail of the list
   *  of element.  Can't be final because of the <code>clone</code> operation.
   */
  protected /*final*/ Node<T> endNode = new Node<T> ( null ) ; { makeEmpty ( ) ; }
  /**
   *  The count of the number of elements in the <code>DLList</code>.  This is not absolutely necessary but
   *  having it makes a lot of operations a lot easier.  A good efficiency measure.
   */
  protected int elementCount = 0 ;
  /**
   *  Zero parameter constructor.
   */
  public DLList ( ) { }
  /**
   *  Construct a <code>DLList</code> by copying in the values from the the <code>Container</code> that is
   *  the parameter.
   */
  public DLList ( final Container<T> c ) { copyIn ( c ) ; }
  /**
   *  Construct a <code>DLList</code> by copying the sub-sequence defined by two indexes into a
   *  <code>Sequence</code>.
   *
   *  @param s the <code>Sequence</code> a slice of which is to be copied.
   *  @param b beginning index of the slice of <code>s</code>.
   *  @param e end index of the slice of <code>s</code>.
   */
  public DLList ( final Sequence<T> s , final int b , final int e ) { copyIn ( s , b , e ) ; }
  /**
   *  Construct an <code>DLList</code> by copying the sequence defined by two
   *  <code>InputIterator</code>s.
   *
   *  @param b beginning iterator of the sequence.
   *  @param e end iterator of the sequence.
   *  @exception InvalidIteratorException if the two <code>InputIterator</code>s are not working the same
   *  <code>DLList</code>.
   */
  public DLList ( final InputIterator<T> b , final InputIterator<T> e ) { copyIn ( b , e ) ; }
  @Override public final T first ( ) { validityCheck ( ) ; return endNode.next.datum ; }
  @Override public final T last ( ) { validityCheck ( ) ; return endNode.previous.datum ; }
  @Override public final T get ( final int index ) { validityCheck ( index ) ; return nodeAt ( index ).datum ; }
  @Override public /*final*/ void set ( final int index , final T o ) {
    try { validityCheck ( index ) ; }
    catch ( AccessEmptyContainerException aece) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw aece ; }
    } catch ( IndexOutOfBoundsException ioobe ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw ioobe ; }
    }
    nodeAt ( index ).datum = o ;
  }
 @Override public final void remove ( final int index ) {
    validityCheck ( index ) ;
    nodeAt ( index ).remove ( ) ;
    --elementCount ;
  }
  @Override public final void insert ( final int index , final T o ) {
    try { validityCheck ( index ) ; }
    catch ( AccessEmptyContainerException aece ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw aece ; }
    }
    nodeAt ( index ).insert ( new Node<T> ( o ) ) ;
    ++elementCount ;
  }
  @Override public final void makeEmpty ( ) {
    endNode.next = endNode ;
    endNode.previous =  endNode;
    elementCount = 0 ;
  }
  @Override public final int size ( ) { return elementCount ; }
  @Override public /*final*/ void add ( final T o ) {
    endNode.previous.add ( new Node<T> ( o ) ) ;
    ++elementCount ;
  }
  @Override public final boolean remove ( final T o ) {
    validityCheck ( ) ;
    for ( Iterator i = iterator ( ) ; ! i .atEnd ( ) ; i.advance ( ) ) {
      if ( i.get ( ).equals ( o ) ) { i.remove() ; return true ; }
    }
    return false ;
  }
  @Override public DLList<T> clone ( ) {
    DLList<T> l = (DLList<T>) super.clone ( ) ;
    //  Replace the representation hook in the new node and then use makeEmpty to initialize it.
    l.endNode = new Node<T> ( null ) ;
    l.makeEmpty ( ) ;
    //  Make copies of the data.
    for ( T t : this ) { l.add ( t ) ; }
    return l ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    Iterator i = new Iterator ( ) ;
    i.current = endNode ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  Merge the <code>DLList</code> <code>l</code>after a given index in this<code>DLList</code>.  This
   *  moves (not copies) all the <code>Nodes</code> from the <code>DLList</code> that is the parameter into
   *  us.
   *
   *  @param index location in this <code>DLList</code> after which to insert <code>l</code>.
   *  @param l the <code>DLList</code> to insert.
   */
  public final void merge ( final int index , final DLList<T> l ) {
    nodeAt ( index ).merge ( l ) ;
    elementCount += l.elementCount ;
    l.makeEmpty ( ) ;
  }
  /**
   *  Append a <code>DLList</code> to ourself.  The moves (not copies) all the <code>Node<T>s</code> from
   *  the <code>DLList</code> that is the parameter into us.
   *
   *  @param l the <code>DLList</code> to append.
   */
  public final void merge ( final DLList<T> l ) {
    endNode.previous.merge ( l ) ;
    elementCount += l.elementCount ;
    l.makeEmpty ( ) ;
  }
  /**
   *  Find a <code>Node</code> given an index.
   *
   *  @param index the index into the <code>DLList</code> to deliver up the <code>Node</code> of.
   *  @exception IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size
   *  of the <code>DLList</code>.
   */
  private final Node<T> nodeAt ( final int index ) {
    //  The index is already known to be reasonable, choose the end which is nearest the node we want and
    //  then walk the list to the appropriate Node<T>.
    Node<T> current = endNode.next ;
    if ( index >= size ( ) / 2 ) {
      current = endNode.previous ;
      for ( int i = size ( ) - 1 ; i > index ; --i ) { current = current.previous ; }
    } else {
      for ( int i = 0 ; i < index ; ++i ) { current = current.next ; }
    }
    return current ;
  }
  /**
   *  The iterator over <code>DLList</code>s.  This is an inner class (as opposed to being a nested
   *  top-level class) so that it automatically has a reference to it's parent object, i.e. the object it is
   *  an iterator for.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> implements BiDirectionalIterator<T> {
    private Node<T> current = endNode.next ;
    @Override public final T get ( ) { validityCheck ( ) ; return  current.datum; }
    @Override public final void set ( final T o ) { validityCheck() ; current.datum = o ; }
    @Override public final void advance ( ) { validityCheck ( ) ; current = current.next ; }
    @Override public final void advance ( int increment ) {
      if ( increment < 0 ) { retreat (-increment) ; }
      else {
        for ( ; increment > 0 ; --increment ) {
          if ( atEnd ( ) ) { break ; }
          advance ( ) ;
        }
      }
    }
    @Override public final void retreat ( ) { validityCheck ( ) ; current = current.previous ; }
    @Override public final void retreat ( int increment ) {
      if ( increment < 0 ) { advance(-increment) ; }
      else {
        for ( ; increment > 0 ; --increment ) {
          if ( atBegin ( ) ) { break ; }
          retreat ( ) ;
        }
      }
    }
    @Override public final boolean atBegin ( ) { return current == endNode.next ; }
    @Override public final boolean atEnd ( ) { return current == endNode ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( o == null ) { return false ; }
      try { return current.equals ( ( (Iterator) o ).current ) ; }
      catch ( ClassCastException cce ) { return false ; }
    }
    @Override public final DLList<T> getContainer ( ) { return DLList.this ; }
    @Override public Iterator clone ( ) { return (Iterator) super.clone ( ) ; }
    public final void insert ( final T o ) {
      validityCheck ( ) ;
      current.insert ( new Node<T> ( o ) ) ;
      ++elementCount ;
    }
   public final void add ( final T o ) {
      validityCheck ( ) ;
      current.add ( new Node<T> ( o ) ) ;
      ++elementCount ;
    }
   @Override public final void remove ( ) {
      validityCheck ( ) ;
      Node<T> n = current ;
      current = current.next ;
      n.remove ( ) ;
      --elementCount ;
    }
    public final void merge ( final DLList<T> l ) {
      validityCheck ( ) ;
      current.merge ( l ) ;
      elementCount += l.elementCount ;
      l.makeEmpty ( ) ;
    }
    /**
     *  Check to see if the iterator is valid.
     *
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    private final void validityCheck ( ) {
      if ( current == null ) { throw new InvalidIteratorException ( ) ; }
    }
  }
}
