/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2009 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Queue</code> implemented using an associated <code>DLList</code>. This is abstraction by
 *  conformance.
 *
 *  <p> The head (the place from which items will be withdrawn) of the <code>Queue</code> is position 0 in
 *  the <code>DLList</code>, the tail (the place to which new items will be appended) is the last
 *  element.</p>
 *
 *  @author Russel Winder
 */
public class DLListQueue<T> extends AbstractQueue<T> {
  /**
   *  The underlying representation.
   */
  private DLList<T> storage ;
  /**
   *  Constructor.
   */
  public DLListQueue ( ) { this ( new DLList<T> ( ) ) ; }
  /**
   *  Constructor that imposes a <code>Queue</code> onto an existing <code>DLList</code>.
   *
   *  @param l <code>DLList</code> to use to back the queue.
   */
  public DLListQueue ( final DLList<T> l ) { storage = l ; }
  @Override public final T peek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return storage.first ( ) ;
  }
  @Override public final T remove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    final T o = storage.first ( ) ;
    storage.remove ( 0 ) ;
    return o ;
  }
  @Override public final void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public final int size ( ) { return storage.size ( ) ; }
  @Override public final void add ( final T o ) { storage.add ( o ) ; }
  @Override public final DLListQueue<T> clone ( ) {
    final DLListQueue<T> q = (DLListQueue<T>) super.clone ( ) ;
    q.storage = storage.clone ( ) ;
    return q ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator() { return new Iterator ( ) ; }
  /**
   *  An iterator over <code>DLListQueue</code>s.
   *
   *  <p>This is just a proxy for an iterator from the underlying representation so as to stop the ability
   *  to write.  Implement only a forward iterator.</p>
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  Reference to the iterator for the underlying representation.
     */
    InputIterator<T> current = storage.iterator ( ) ;
    @Override public final T get ( ) { return current.get ( ) ; }
    @Override public final void advance ( ) { current.advance ( ) ; }
    @Override public void advance ( final int increment ) { current.advance ( increment ) ; }
    @Override public boolean atBegin ( ) { return current.atBegin ( ) ; }
    @Override public final boolean atEnd ( ) { return current.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof DLListQueue.Iterator ) ) { return false ; }
      return current.equals ( ( (Iterator) o ).current ) ;
    }
    @Override public final DLListQueue<T> getContainer ( ) { return DLListQueue.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.current = current.clone ( ) ;
      return i ;
    }
  }
}
