/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010--2011 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An interface defining the necessary properties to be deemed to be a dequeue &mdash a dequeue is a queue
 *  that can have items added or removed from both ends, i.e. it is a double-ended queue. A dequeue should
 *  not be confused with a deque which is a form of dequeue that supports random access.
 *
 *  @author Russel Winder
 */
public interface Dequeue<T> extends Queue<T> {
  /**
   *  Insert an item into the <code>Dequeue</code> at the head.
   *
   *  @param o the item to be inserted at the head of the <code>Dequeue</code>.
   */
  void headAdd ( T o ) ;
  /**
   *  Insert an item into the <code>Dequeue</code> at the tail.
   *
   *  @param o the item to be inserted at the tail of the <code>Dequeue</code>.
   */
  void tailAdd ( T o ) ;
  /**
   *  Remove an item from the head of the <code>Dequeue</code>.
   *
   *  @return the item at the head of the <code>Dequeue</code>.
   */
  T headRemove ( ) ;
  /**
   *  Remove an item from the tail of the <code>Dqueue</code>.
   *
   *  @return the item at the tail of the <code>Dequeue</code>.
   */
  T tailRemove ( ) ;
  /**
   *  Peek at the head of the <code>Dequeue</code> without removing it.
   *
   *  @return the item at the head of the <code>Dequeue</code>.
   */
  T headPeek ( ) ;
  /**
   *  Peek at the tail of the <code>Dequeue</code> without removing it.
   *
   *  @return the item at the tail of the <code>Dequeue</code>.
   */
  T tailPeek ( ) ;
  /**
   *  Clone this <code>Dequeue</code>.
   *
   *  @return a clone of this <code>Dequeue</code>.
   */
  @Override Dequeue<T> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>Dequeue</code> for use in multi-threaded
   *  applications.  This is just a forwarding proxy with synchronization control and locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<E> extends Container.Synchronized<E> implements Dequeue<E> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param q the <code>Dequeue</code> to synchronize.
     */
    public Synchronized ( final Dequeue<E> q ) { super ( q ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param  q the <code>Dequeue</code> to synchronize.
     *  @param o the lock object.
     */
    public Synchronized ( final Dequeue<E> q , final Object o ) { super ( q , o ) ; }
    @Override public final void headAdd ( final E o ) {
      synchronized ( theLock ) { ( (Dequeue<E>) theContainer ).headAdd(o) ; }
    }
    @Override public final void tailAdd ( final E o) {
      synchronized ( theLock ) {  ( (Dequeue<E>) theContainer ).tailAdd(o) ; }
    }
    @Override public final E headRemove ( ) {
      synchronized ( theLock ) { return ( (Dequeue<E>) theContainer ).headRemove ( ) ; }
    }
    @Override public final E tailRemove ( ) {
      synchronized ( theLock ) { return ( (Dequeue<E>) theContainer ).tailRemove ( ) ; }
    }
    @Override public final E headPeek ( ) {
      synchronized ( theLock ) { return ( (Dequeue<E>) theContainer ).headPeek ( ) ; }
    }
    @Override public final  E tailPeek ( ) {
      synchronized ( theLock ) { return ( (Dequeue<E>) theContainer ).tailPeek ( ) ; }
    }
    @Override public final E peek ( ) {
      synchronized ( theLock ) { return ( (Dequeue<E>) theContainer ).peek ( ) ; }
    }
    @Override public final E remove ( ) {
      synchronized ( theLock ) { return ( (Dequeue<E>) theContainer ).remove ( ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  Dequeue.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public final Dequeue.Synchronized<E> clone ( ) { return (Dequeue.Synchronized<E>) super.clone ( ) ; }
  }
}
