/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An hash table abstraction. This is intended to be an implementation tool for <code>Set</code>,
 *  <code>Relation</code> and <code>Mapping</code> rather than a publically used ADT.
 *
 *  @author Russel Winder
 */
public interface HashTable<K, V> extends Container<Pair<K, V>> {
  /**
   *  Add a ( key, value ) pair.
   *
   *  @param key the key value
   *  @param value the value to be associated with <code>key</code>in this <code>HashTable</code>.
   */
  void add ( K key , V value ) ;
  /**
   *  Find a key for a given value.
   *
   *  @param key the value to search on.
   *  @return the value associated with <code>key</code> in this <code>HashTable</code> if present,
   *  <code>null</code> otherwise.
   */
  V getValue ( K key ) ;
  /**
   *  Clone this <code>HashTable</code>.
   *
   *  @return a clone of this <code>HashTable</code>.
   */
  @Override HashTable<K, V> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>HashTable</code> for use in
   *  multi-threaded applications.  This is just a forwarding proxy with synchronization control and
   *  locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<SK, SV> extends Container.Synchronized<Pair<SK, SV>> implements HashTable<SK, SV> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param h the <code>HashTable</code> to be a synchronized.
     */
    public Synchronized ( final HashTable<SK, SV> h ) { super ( h ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param h the <code>HashTable</code> to be a synchronized.
     *  @param o the lock object to use.
     */
    public Synchronized ( final HashTable<SK, SV> h , final Object o ) { super ( h , o ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public final void add ( final SK key , final SV value ) {
      synchronized ( theLock ) { ( (HashTable<SK, SV>) theContainer ).add ( key , value ) ; }
    }
    @SuppressWarnings ( "unchecked" )
    @Override public final SV getValue ( final SK key ) {
      synchronized ( theLock ) { return ( (HashTable<SK, SV>) theContainer ).getValue ( key ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  HashTable.Synchronized and Container.Synchronized so this is a lookup conflict.
    @SuppressWarnings ( "unchecked" )
    @Override public final HashTable.Synchronized<SK, SV> clone ( ) { return (HashTable.Synchronized<SK, SV>) super.clone ( ) ; }
  }
}
