/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2009 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An implementation of heap as a data structure.
 *
 *  <p>A heap is a complete binary tree data structure in which there is no order relation between nodes at
 *  the same level but there is a rigorously controlled relationship between nodes at different levels,
 *  i.e. the heap implements a partial order.  There are two possibilities for the heap ordering relation:
 *  either all children of a node must have a value greater than or equal to that of the parent node or they
 *  must all be less than or equal to, with this relation holding for all nodes.  (NB Cannot mix these two
 *  in any given heap, must be one or the other!)  Because of this possible ordering difference we require
 *  all heaps to have an associated <code>Comparator</code> to define which of the two orderings a heap
 *  maintains.</p>
 *
 *  <p> Of course the problem this leaves is what happens if the heap is given an unreasonable comparator?
 *  We could have defined a subinterface of <code>Comparator</code>, <code>HeapOrderComparator</code> say,
 *  and required the application to supply one of these but we cannot actually guarantee that the order
 *  relation so provided is actually either "less than" or "greater than" so we are back where we started.
 *  So, whilst it might help the user get it right providing <code>HeapOrderComparator</code>, it adds
 *  unnecessary complexity so by Ockham's Razor we did not do it.  We hope then that the comparator provided
 *  to the heap is a reasonable one!</p>
 *
 *  <p> Insertion of a value into the heap can, and usually does, cause rearrangement of the nodes in the
 *  heap in order to preserve the heap property.</p>
 *
 *  <p> A heap could be implemented as a tree with nodes and links but because of the heap property the tree
 *  is always maximally complete for the number of nodes and is therefore not only as shallow as it is
 *  possible to make the tree for the number of elements, it can be represented very efficiently in an
 *  array.  In fact, we use a <code>Array</code>, making use of it's dynamic extensibility (as compared to a
 *  primitive array), so that we do not need to manage the changing size of the heap in our code.</p>
 *
 *  <p>NB Iterating over an <code>Heap</code> or transforming to a <code>String</code> uses the structure
 *  order and not any total order.  The store is in partial order, total order on the values is only imposed
 *  when values are removed.</p>
 *
 *  @see Comparator
 *  @see Array
 *  @author Russel Winder
 */
public class Heap<T> extends AbstractContainer<T> {
  /**
   *  The <code>Array</code> which actually stores all the values.  Initialize it in the constructors rather
   *  than here in case the user provides initialization arguments.
   */
  protected Array<T> storage ;
  /**
   *  The comparator which determines whether the heap property relation is "greater than" or "less than".
   */
  protected Comparator<? super T> order ;
  /**
   *  The constructor that requires only the order relation, everything else is defaulted.
   *
   *  @param c must be a <code>Comparator&lt;? super T></code> implementing either "less than" or "greater
   *  than".  Any other sort of relation will mean that the data structure is not necessarily a heap!
   */
  public Heap ( final Comparator<? super T> c ) {
    storage = new Array<T> ( ) ;
    order = c ;
  }
  /**
   *  The constructor that defines the initial size and the order relation, but defaults the amount by which
   *  the <code>Array</code> is extended when required to do so.
   *
   *  @param initialSize the initial size of the <code>Array</code> that implements the <code>Heap</code>
   *  @param c must be a <code>Comparator&lt;? super T></code> implementing either "less than" or "greater
   *  that".  Any other sort of relation will mean that the data structure is not necessarily a heap!
   */
  public Heap ( final int initialSize , final Comparator<? super T> c ) {
    storage = new Array<T> ( initialSize ) ;
    order = c ;
  }
  /**
   *  The constructor that defines all aspects of the behaviour, the initial size and expansion amount of
   *  the <code>Array</code> and the order relation.
   *
   *  @param initialSize the initial size of the <code>Array</code> that implements the <code>Heap</code>.
   *  @param resizeSize the size by which the <code>Array</code> is incremented when it is grown.
   *  @param c must be a <code>Comparator&lt;? super T></code> implementing either "less than" or "greater
   *  than".  Any other sort of relation will mean that the data structure is not necessarily a heap!
   */
  public Heap ( final int initialSize , final int resizeSize , final Comparator<? super T> c ) {
    storage = new Array<T> ( initialSize , resizeSize ) ;
    order = c ;
  }
  /**
   *  Create an <code>Heap</code> from the primitive array provided.
   *
   *  @param v the array that is used to initialize the <code>Heap</code>.
   *  @param c must be a <code>Comparator&lt;? super T></code> implementing either "less than" or "greater
   *  than".  Any other sort of relation will mean that the data structure is not necessarily a heap!
   */
  public Heap ( final T[] v , final Comparator<? super T> c ) {
    storage = new Array<T> ( v.length ) ;
    for ( final T item : v ) { storage .add ( item ) ; }
    order = c ;
    if ( ! storage.isEmpty ( ) ) { makeHeap ( storage , order ) ; }
  }
  /**
   *  Create an <code>Heap</code> using the <code>Sequence</code> provided.
   *
   *  @param s the <code>Sequence</code> providing the data for the <code>Heap</code>.
   *  @param c must be a <code>Comparator&lt;? super T></code> implementing either "less than" or "greater
   *  than".  Any other sort of relation will mean that the data structure is not necessarily a heap!
   */
  public Heap ( final Sequence<T> s , final Comparator<? super T> c ) {
    storage = new Array<T> ( s.size ( ) ) ;
    for ( InputIterator<T> i = s.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) { storage.add ( i.get ( ) ) ; }
    order = c ;
    if ( ! storage.isEmpty ( ) ) { makeHeap ( storage , order ) ; }
  }
  /**
   *  The method for inserting new items into the heap.
   *
   *  @param o the <code>T</code> to be inserted.
   */
  @Override public void add ( final T o ) {
    //  Add the item to the end ready for percolation.  Actually this is here just to make sure the Array<T>
    //  is the right size, where we have put the datum is actually treated as a gap.  Position is the imdex
    //  of this gap.
    int position = storage.size ( ) ;
    storage.add ( o ) ;
    //  Check to see that if the new element were really put in the empty slot it would be in a place that
    //  preserves the heap property.  If it is then there is nothing to do.  If it is not, then move the
    //  parent to the empty space and see if we can put the value in the space opened up.
    int parent = ( position - 1 ) / 2 ;
    while ( ( position > 0 ) && order.relation( o , storage.get ( parent ) ) ) {
      storage.set ( position , storage.get ( parent ) ) ;
      position = parent ;
      parent = ( position - 1 ) / 2 ;
    }
    //  At this point we know that the empty space is a valid space in which to put the new element ensuring
    //  preservation of the heap property.
    storage.set ( position , o ) ;
  }
  /**
   *  Peek at the item that is at the top of the heap.
   *
   *  @return the item from the top of the heap.
   *  @exception AccessEmptyContainerException if there are no items in the
   *  <code>Array</code>.
   */
  public T peek ( ) { return storage.get ( 0 ) ; }
  /**
   *  Remove the item that is at the top of the heap.
   *
   *  @return the item removed.
   *  @exception AccessEmptyContainerException if there are no items in the
   *  <code>Array</code>.
   */
  public T remove ( ) {
    final T returnValue = storage.get ( 0 ) ;
    remove ( 0 ) ;
    return returnValue ;
  }
  /**
   *  Remove a specified value from the <code>Heap</code>.  Requires <code>equals</code> to be defined
   *  appropriately for all obejcts in the <code>Heap</code>.  Uses a linear search to find the element.
   *
   *  @param o the value to be searched for.
   */
  @Override public boolean remove ( final T o ) {
    for ( int i = 0 ; i < storage.size ( ) ; ++i ) {
      if ( storage.get ( i ).equals ( o ) ) {
        remove ( i ) ;
        return true ;
      }
    }
    return false ;
  }
  /**
   *  Remove a specified value from the <code>Heap</code>.  Uses <code>equality.relation</code> to compare
   *  items for equality.  Uses a linear search to find the element.
   *
   *  @param o the value to be searched for.
   *  @param equality a <code>Comparator&lt;? super T></code> to test for equality of <code>o</code> with
   *  the items in the <code>Heap</code>.
   *  @return true if the value was successfully removed, false otherwise.
   */
  public boolean remove ( final T o , final Comparator<? super T> equality ) {
    for ( int i = 0 ; i < storage.size ( ) ; ++i ) {
      if ( equality.relation ( o , storage.get ( i ) ) ) {
        remove ( i ) ;
        return true ;
      }
    }
    return false ;
  }
  /**
   *  Ask whether the <code>Heap</code> contains a specific value.  Uses <code>equality.relation</code> to
   *  compare items for equality.
   *
   *  @param o the value to be searched for.
   *  @param equality a <code>Comparator&lt;? super T></code> to test for equality of <code>o</code> with
   *  the items in the <code>Heap</code>.
   */
  public boolean contains ( final T o , final Comparator<? super T> equality ) {
    for ( int i = 0 ; i < storage.size ( ) ; ++i ) {
      if ( equality.relation ( o , storage.get ( i ) ) ) { return true ; }
    }
    return false ;
  }
  /**
   *  Determine the size of the <code>Heap</code>.
   */
  @Override public int size ( ) { return storage.size ( ) ; }
  /**
   *  Empty out the heap.
   */
  @Override public void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  /**
   *  Redefine <code>equals</code> so that it is a semantic equality, i.e. the two <code>Heap</code>s
   *  contain the same values in the same heap ordering, rather than being object equality.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public boolean equals ( final Object o ) {
    if ( this == o ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof Heap ) || ( size ( ) != ( (Heap<T>) o).size ( ) ) ) {
      return false ;
    }
    return storage.equals ( ( (Heap<T>) o).storage ) ;
  }
  /**
   *  Cloning is not only allowed it is required.
   */
  @Override public Heap<T> clone ( ) {
    final Heap<T> h = (Heap<T>) super.clone ( ) ;
    h.storage = storage.clone ( ) ;
    return h ;
  }
  /**
   *  Create a string representation of the <code>Heap</code>.
   *
   *  <p>NB The order of items is the order in the data structure not in the value order of contained
   *  values.  The data structure is partially ordered the final total order is only applied when values are
   *  removed from the <code>Heap</code>.</p>
   */
  @Override public String toString ( ) { return storage.toString ( ) ; }
  /**
   *  Even though we are a closed container, conforming to <code>Container</code> requires that we provide a
   *  public iterator.  However, we do ensure that there is no possibility of updates.
   */
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  /**
   *  Even though we are a closed container, conforming to <code>Container</code> requires that we provide a
   *  public iterator.  However, we do ensure that there is no possibility of updates.
   */
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  /**
   *  Even though we are a closed container, conforming to <code>Container</code> requires that we provide a
   *  public iterator.  However, we do ensure that there is no possibility of updates.
   */
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  An iterator over all the values in the <code>Heap</code>.
   *
   *  <p>NB An iterator does not return the values in value order determined by the comparator it returns
   *  the values in underlying structure order, i.e. it just does a breadth first tour through the values in
   *  the partially ordered tree.  Obtaining the values in value order requires the values to be removed
   *  from the <code>Heap</code>.</p>
   *
   *  <p>There is an assumption that the heap remains unaltered during use of the iterator.  If changes are
   *  made then there are no gurantees as to what the iterator will do.</p>
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  The current cursor is represented by an index into the <code>Array</code>.
     */
    private int currentIndex = 0 ;
    /**
     *  Deliver the value we are currently referring to.
     */
    @Override public T get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorException () ; }
      return storage.get ( currentIndex ) ;
    }
    /**
     *  Move on to the next element in the container.
     */
    @Override public void advance ( ) {
      //  Perform the action anyway since if we are atEnd() increment will leave us atEnd() and it save the
      //  time of the test for the normal case.
      ++currentIndex ;
    }
    /**
     *  Move on n items in the container.
     *
     *  @exception InvalidIteratorOperationException if the incrment in negative.
     */
    @Override public void advance ( final int increment ) {
      if ( increment < 0 ) { throw new InvalidIteratorOperationException () ; }
      int putativeIndex = currentIndex + increment ;
      if ( putativeIndex >= storage.size ( ) ) { putativeIndex = storage.size ( ) ; }
      currentIndex = putativeIndex ;
    }
    /**
     *  Are we positioned at the beginning of the iteration?
     */
    @Override public boolean atBegin ( ) { return currentIndex == 0 ; }
    /**
     *  Have we reached the end of the iteration?
     */
    @Override public boolean atEnd ( ) { return ( currentIndex < 0 ) || ( currentIndex >= storage.size ( ) ) ; }
    /**
     *  Are two iterators equal, i.e. do two iterators refer to the same item in the same data structure.
     */
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof Heap.Iterator ) ) { return false ; }
      final Iterator i = (Iterator) o ;
      return ( getContainer ( ) == i.getContainer ( ) ) && ( currentIndex == i.currentIndex ) ;
    }
    @Override public Iterator clone ( ) { return (Iterator) super.clone ( ) ; }
    /**
     *  Return a reference to the <code>Container</code> of this <code>InputIterator</code>.
     */
    @Override public Heap<T> getContainer ( ) { return Heap.this ; }
  }
  /**
   *  The class that provides a synchronization adapter to a <code>Heap</code> for use in multi-threaded
   *  applications. This is just a forwarding proxy with synchronization control and locking. See
   *  <code>Heap</code> for all comments on the methods.
   *
   *  @author Russel Winder
   */
  public static class Synchronized<E> extends Heap<E> {
    /**
     *  The <code>Object</code> on which to do all locking.
     */
    private final Object theLock;
    /**
     *  The default constructor.  Gives an <code>Heap</code> with default initial size and increment.
     *
     *  @param h the <code>Heap</code> top synchronize.
     */
    public Synchronized ( final Heap<E> h ) {
      //  Cannot use a this call since we would have to have "this ( h , this )" and you can't use the this
      //  pointer till after the superclass construction.
      super ( h.order ) ;
      storage = h.storage ;
      theLock = this ;
    }
    /**
     *  The constructor that supplies an object to synchronize on.
     *
     *  @param h the <code>Heap</code> top synchronize.
     *  @param o the lock object.
     */
    public Synchronized ( final Heap<E> h , final Object o ) {
      super ( h.order ) ;
      storage = h.storage ;
      theLock = o ;
    }
    @Override public final E remove ( ) {
      synchronized ( theLock ) { return super.remove ( ) ; }
    }
    @Override public final void makeEmpty ( ) {
      synchronized ( theLock ) { super.makeEmpty ( ) ; }
    }
   @Override  public final boolean isEmpty ( ) {
      synchronized ( theLock ) { return super.isEmpty ( ) ; }
    }
    @Override public final int size ( ) {
      synchronized ( theLock ) { return super.size ( ) ; }
    }
    @Override public final boolean equals ( final Object o ) {
      synchronized ( theLock ) { return super.equals ( o ) ; }
    }
    @Override public final boolean contains(final E o) {
      synchronized ( theLock ) { return super.contains ( o ) ; }
    }
    @Override public final boolean contains ( final E o, final Comparator<? super E> c ){
      synchronized ( theLock ) { return super.contains ( o , c ) ; }
    }
    @Override public final void add ( final E o ) {
      synchronized ( theLock ) {  super.add ( o ) ; }
    }
    @Override public final boolean remove ( final E o ) {
      synchronized ( theLock ) { return super.remove ( o ) ; }
    }
    @Override public final boolean remove ( final E o , final Comparator<? super E> c ) {
      synchronized ( theLock ) { return super.remove ( o , c ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  Heap.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public final Heap.Synchronized<E> clone ( ) { return (Heap.Synchronized<E>) super.clone ( ) ; }
    @Override public final String toString ( ) {
      synchronized ( theLock ) { return super.toString ( ) ; }
    }
  }
  /**
   *  Make a <code>Array</code> into one that has the heap properties.
   *
   *  <p> This is static simply so that we can make use of it in <code>HeapSortArray</code> so that we can
   *  heap sort a <code>Array</code> without creating a copy of the data in a <code>Heap</code>.</p>
   *
   *  @param a the array to heapify.
   *  @param c the comparator to use during heapification.
   *
   *  @see org.devjavasoft.ads.sorting.HeapSortArray
   */
  public static <E> void makeHeap ( final Array<E> a , final Comparator<? super E> c ) {
    final int length = a.size ( ) ;
    if ( length > 1 ) {
      for ( int i = length / 2 ; i >= 0 ; --i ) { percolateDown ( a , length , i , c ) ; }
    }
  }
  /**
   *  Enforce the heap properties on the not-necessarily-an-heap rooted at start index <code>index</code>.
   *
   *  <p> The sub-trees of the root are both <code>Heap</code>s but the root node may break the heap
   *  properties.  Percolate the datum down the tree until heapness is restored.</p>
   *
   *  <p> This is public static simply so that we can make use of it in <code>HeapSortArray</code> so that
   *  we can heap sort a <code>Array</code> without creating a copy of the data in a <code>Heap</code>.</p>
   *
   *  @param a the <code>Array</code> that is to have heap properties re-imposed on it.
   *  @param maximum defines the end of the sub-<code>Heap</code> to be manipulated.
   *  @param index the index of the root of a sub-<code>Heap</code> which is potentially violating the heap
   *  order property.
   *  @param c the <code>Comparator</code> defining the order relation imposing heap order.
   *  @exception IndexOutOfBoundsException when either the desired maximum range or the desired index are
   *  larger that the size of the <code>Array</code>.
   *
   *  @see org.devjavasoft.ads.sorting.HeapSortArray
   */
  public static <E> void percolateDown ( final Array<E> a , final int maximum , int index , final Comparator<? super E> c ) {
    final int length = a.size ( ) ;
    if ( maximum > length ) { throw new IndexOutOfBoundsException ( "maximum > length" ) ; }
    if ( index > length ) { throw new IndexOutOfBoundsException ("index > length") ; }
    if ( length <= 1 ) { return ; }
    //  Take the root value, creating the 'gap' at the root.
    final E o = a.get(index) ;
    //  Loop, moving the 'gap' until we find the right situation for the value.
    for ( int child =  2 * index + 1 ; child < maximum ; index = child , child = 2 * child + 1 ) {
      //  We have assumed that it is the left child that we need to deal with.  Check to see if it is
      //  actually the right child that we need.  We must ensure we preserve the heap properties whenever an
      //  item is moved!
      final int right = child + 1 ;
      if ( ( right < maximum ) && c.relation ( a.get ( right ) , a.get ( child ) ) ) { child = right ; }
      //  Check to see if the current position is the place we need to put the item to create the heap.  If
      //  it is finish iterating, otherwise set up the next iteration by moving the child value to the
      //  current node.
      if ( c.relation ( o , a.get ( child ) ) ) { break ; }
      else { a.set ( index , a.get ( child ) ) ; }
    }
    //  The 'gap' is now in the right place so fill it.  Of course, if the gap was already in the right
    //  place we are just reassigning but this is probably very rare.
    a.set ( index, o ) ;
  }
  /**
   *  Remove the item at the index given.
   *
   *  @param index the index of the item to remove.
   */
  private void remove ( final int index ) {
    final int newSize = storage.size ( ) - 1 ;
    storage.set ( index, storage.get ( newSize ) ) ;
    storage.remove ( newSize ) ;
    if ( newSize > 1 ) { percolateDown ( storage , newSize , index , order ) ; }
  }
}
