/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>PriorityQueue</code> represented as an <code>Heap&lt;Pair&lt;T, Integer>></code> i.e. a heap of
 *  pairs ( datum, priority) -- the first item of the stored pair is the value and the second item is the
 *  priority.  It is assumed that smaller numbers are higher priority.  Default insert is at priority 10.
 *  Removal always takes the highest priority item. Because priorities are integers the application has no
 *  need to determine the number of different priorities at creation time.
 *
 *  <p>NB The removal order is priority order with no regard to insertion order which means that items of
 *  the same priority can be extracted in a different order to that which they were inserted.  If removal
 *  order needs to be insertion order (at a given priority level) then use a <code>SortedListQueue</code> or
 *  a <code>SequenceQueuePriorityQueue</code><p>
 *
 *  @author Russel Winder
 */
public class HeapPriorityQueue<T> extends AbstractPriorityQueue<T> {
  /**
   *  The <code>Heap</code> which is the representation.  Should be final but the way clone is implemented
   *  means it cannot be.
   */
  private /*final*/ Heap<Pair<T, Integer>> storage =
   new Heap<Pair<T, Integer>> ( new PairBComparator<T, Integer, Pair<T, Integer>> ( new LessThanComparator<Integer> ( ) ) ) ;
  @Override public final void add ( final T o ) { add ( o , 10 ) ; }
  @Override public final void add ( final T o , final int priority ) { storage.add ( new Pair<T, Integer> ( o , priority ) ) ; }
  @Override public final T peek ( ) { return storage.peek ( ).getA ( ) ; }
  @Override public final T remove ( ) { return storage.remove ( ).getA ( ) ; }
  @Override public final T remove ( final int priority ) { throw new InvalidOperationException ( ) ; }
  @Override public final void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public final boolean isEmpty ( ) { return storage.isEmpty ( ) ; }
  @Override public final int size ( ) { return storage.size ( ) ; }
  @Override public final HeapPriorityQueue<T> clone ( ) {
    final HeapPriorityQueue<T> p = (HeapPriorityQueue<T>) super.clone ( ) ;
    p.storage = storage.clone ( ) ;
    return p ;
  }
  @Override public final Iterator begin() { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.currentItem = storage.end() ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  Allow iteration over elements in the <code>PriorityQueue</code>.  This should only really available to
   *  <code>PriorityQueue</code> but we can only have public methods in interfaces so we have to put up with
   *  the publicness of the method.  This is not too bad since it does not allow update of the
   *  <code>PriorityQueue</code>.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    private InputIterator<Pair<T, Integer>> currentItem = storage.iterator ( ) ;
    @Override public final T get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorException  ( ) ; }
      return currentItem.get ( ).getA ( ) ;
    }
    @Override public final void advance ( ) { currentItem.advance ( ) ; }
    @Override public final void advance ( final int increment ) { currentItem.advance ( increment ) ; }
    @Override public final boolean atBegin ( ) { return currentItem.atBegin ( ) ; }
    @Override public final boolean atEnd ( ) { return currentItem.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof HeapPriorityQueue.Iterator ) ) { return false ; }
      final Iterator i = (Iterator) o ;
      return atEnd ( ) ? i.atEnd ( ) : currentItem.equals ( i.currentItem ) ;
    }
    @Override public final HeapPriorityQueue<T> getContainer ( ) { return HeapPriorityQueue.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.currentItem = currentItem.clone ( ) ;
      return i ;
    }
  }
}
