/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
import java.util.Iterator ;
/**
 *  All iterators over containers in ADS conform to this interface.
 *
 *  <p>This interface inherits from <code>java.util.Iterator</code> to allow ADS containers to be used with
 *  the foreach loop, i.e. <code>for ( E e : c ) { . . . }</code> loops.</p>
 *
 *  @author Russel Winder
 */
public interface InputIterator<T> extends Cloneable, Iterator<T> {
  /**
   *  Deliver the item in the container that this <code>InputIterator</code> currently refers to.
   *
   *  @return the value this <code>InputIterator</code> currently refers to.
   *  @exception AccessEmptyContainerException if there are no items in the container.
   *  @exception InvalidIteratorException if the iterator is not referring to a valid element.
   */
  T get ( ) ;
  /**
   *  Move on to the next item in iteration over the container.
   */
  void advance ( ) ;
  /**
   *  Move forward to the <code>increment</code>th item after the current one in the iteration over the
   *  container.  If <code>increment</code> is larger than the number of items left in the container the
   *  iterator is positioned at the end.
   *
   *  @param increment the number of items to advance by.
   *  @exception InvalidIteratorOperationException if the increment in negative.
   */
  void advance ( int increment ) ;
  /**
   *  Returns a boolean value depending on whether this <code>InputIterator</code> is positioned at the
   *  beginning of the iteration sequence for this container or not.
   *
   *  @return true if this iterator positioned at the first element in the iteration over the container,
   *  false otherwise.
   */
  boolean atBegin ( ) ;
  /**
   *  Returns a boolean value depending on whether this <code>InputIterator</code> is positioned at the end
   *  of the iteration sequence for this container or not.
   *
   *  @return true if this iterator is positioned after all elements (i.e. at the end) in the iteration over
   *  the container, false otherwise.
   */
  boolean atEnd ( ) ;
  /**
   *  Decide whether two <code>InputIterator</code>s are position in the same place in the same container.
   *
   *  @param o the <code>InputIterator</code> to be compared to.
   *  @return true if this object and the parameter are at the same position of an iteration over the same
   *  container and false otherwise.
   */
  @Override boolean equals ( Object o ) ;
  /**
   *  Return a reference to the container being worked by this <code>InputIterator</code>.
   *
   *  @return a reference to the container of this <code>InputIterator</code>.
   */
  Container<T> getContainer ( ) ;
  /**
   *  Calculate the distance between this iterator and the parameter iterator which must, of course, be an
   *  iterator into the same container.
   *
   *  @param i the iterator to measure the distance to.
   *  @return the number of items between this iterator and the parameter iterator.
   *  @exception InvalidIteratorException if the parameter iterator is not in the same container as this
   *  iterator.
   */
  int distance ( InputIterator<T> i ) ;
  /**
   *  Deliver a clone of this <code>InputIterator</code>.
   *
   *  @return a reference to a clone of this <code>InputIterator</code>.
   */
  InputIterator<T> clone ( ) ;
}
