/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  The interface describing a mapping.  A mapping is a relation in which each element of the domain maps to
 *  a single element of the co-domain.  This is a weaker restriction on relation than a function which
 *  requires the additional restriction that each element of the range is mapped to by a single element of
 *  the domain.
 *
 *  @author Russel Winder
 */
public interface Mapping<D, R> extends Relation<D, R> {
  /**
   *  Find the value associated with a given value from the domain.
   *
   *  @param d a value to search on.
   *  @return r from (d, r) pair if it is in the mapping or null if it is not.
   */
  R getValue ( D d ) ;
  /**
   *  Clone this <code>Mapping</code>.
   *
   *  @return a clone of this <code>Mapping</code>.
   */
  @Override Mapping<D, R> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>Mapping</code> for use in multi-threaded
   *  applications. This is just a forwarding proxy with synchronization control and locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<SD, SR> extends Relation.Synchronized<SD, SR> implements Mapping<SD, SR> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param m <code>Mapping</code> to synchronize.
     */
    public Synchronized ( final Mapping<SD, SR> m ) { super ( m ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param m <code>Mapping</code> to synchronize.
     *  @param o teh lock object.
     */
    public Synchronized ( final Mapping<SD, SR> m , final Object o ) { super(m, o) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public final SR getValue ( final SD d ) {
      synchronized ( theLock ) { return ( (Mapping<SD, SR>) theContainer ).getValue ( d ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  Mapping.Synchronized, Set.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public Mapping.Synchronized<SD, SR> clone ( ) { return (Mapping.Synchronized<SD, SR>) super.clone ( ) ; }
  }
}
