/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An hash table that uses open hashing.  In this implementation, a primitive array of references to
 *  <code>Pair&lt;K, V></code>s which are the data items is used.  The clash algorithm is to work through
 *  the table linearly step 1 to find the entry or the next empty place.
 *
 *  @author Russel Winder
 */
public class OpenHashTable<K, V> extends AbstractHashTable<K, V> {
  /**
   *  The storage for the hash table.  Each entry is just a pair ( key , value ).
   */
  protected Pair<K, V>[] table ;
  /**
   *  Keep a count of the number of items to make things more efficient.
   */
  protected int pairCount = 0 ;
  /**
   *  The hash function method object defining our hash function.  The hash function guarantees to deliver a
   *  positive integer.  The table indexing operations are responsible for ensuring that the integer is a
   *  valid index.  On most occassions this simply means taking the modulus of the hash value against the
   *  size of the array being used as representation.
   */
  protected HashFunction<K> hashFunction;
  /**
   *  The default size of the hash table.  Best if this is a prime number.
   */
  protected static final int DEFAULT_SIZE = 499 ;
  /**
   *  The default constructor. Use the default size of 511 and the default hash function as supplied by the
   *  class <code>HashFunction&lt;K></code>.
   */
  public OpenHashTable ( ) { this ( DEFAULT_SIZE , new HashFunction<K> ( ) ) ; }
  /**
   *  The constructor specifying the size of the table but using the default hash function as supplied by
   *  the class <code>HashFunction&lt;K></code>.
   *
   *  @param tableSize the size of the hash table.
   */
  public OpenHashTable ( final int tableSize ) { this ( tableSize , new HashFunction<K> ( ) ) ; }
  /**
   *  The constructor that assumes the default size but provides an hash function.
   *
   *  @param f the hash function.
   */
  public OpenHashTable ( final HashFunction<K> f ) { this ( DEFAULT_SIZE , f ) ; }
  /**
   *  The constructor specifying the size of the table and the hash function.
   *
   *   @param tableSize the size of the hash table.
   *   @param f the hash function.
   */
  @SuppressWarnings ( "unchecked" )
  public OpenHashTable ( final int tableSize , final HashFunction<K> f ) {
    table = (Pair<K, V>[]) new Object [tableSize] ;
    hashFunction = f ;
  }
  @Override public void makeEmpty ( ) {
    //  Leave the garbage collector to handle all the clean up by just releasing all the Pair<K, V> to be
    //  garbage.
    for ( int i = 0 ; i < table.length ; ++i ) { table[i] = null ; }
    pairCount = 0 ;
  }
  @Override  public final boolean isEmpty ( ) { return pairCount <= 0 ; }
  @Override public final int size ( ) { return pairCount ; }
  @Override public void add ( final Pair<K, V> o ) {
    table [ findIndex ( o.getA ( ) ) ] = o ;
    ++pairCount ;
  }
  /**
   *  Find a value for a given key.
   *
   *  @return value from ( key , value ) pair given a key or null if the key is not present in the hash
   *  table.
   */
  @Override public V getValue ( final K key ) {
    final int index = findIndex ( key ) ;
    return table[index] == null ? null : table[index].getB ( ) ;
  }
  /**
   *  Remove the pair with a given ( key , value ) pair.
   */
  @Override public boolean remove ( final Pair<K, V> o ) {
    final int index = findIndex ( o.getA ( ) ) ;
    if ( ( table[index] != null ) && table[index].equals ( o ) ) {
      table[index] = null ;
      --pairCount ;
      return true ;
    }
    return false ;
  }
  /**
   *  Compare this <code>OpenHashTable&lt;K, V></code> with another and determine whether they represent the
   *  same value, i.e. have the same item values in the same order.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public boolean equals ( final Object o ) {
    if ( o == this ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof OpenHashTable ) ) { return false ; }
    final OpenHashTable<K, V> other = (OpenHashTable<K, V>) o ;
    for ( int i = 0 ; i < table.length ; ++i ) {
      if ( table[i] == null ) {
        if ( other.table[i] != null ) { return false ; }
      }
      else {
        if ( ! table[i].equals ( other.table[i] ) ) { return false ; }
      }
    }
    return true ;
  }
  /**
   *  Deliver up a complete shallow copy of the <code>OpenHashTable&lt;K, V></code>.  NB
   *  <code>CloneNotSupportedException</code> is caught in <code>AbstractContainer&lt;Pai&lt;K, V>></code>
   *  so we don't need to worry about it here.
   */
  @Override  public OpenHashTable<K, V> clone ( ) {
    final OpenHashTable<K, V> h = (OpenHashTable<K, V>) super.clone ( ) ;
    h.table = table.clone ( ) ;
    return h ;
  }
  /**
   *  Deliver up an iterator over all the items in the <code>OpenHashTable&lt;K, V></code>.
   */
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  /**
   *  Deliver up an iterator over all the items in the <code>OpenHashTable&lt;K, V></code>.
   */
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  /**
   *  Deliver up an iterator over all the items in the <code>OpenHashTable&lt;K, V></code>.
   */
  @Override public Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  The basic iterator.  This iterates over all elements.  By subclassing we can change behaviours.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<Pair<K, V>> {
    /**
     *  The initial index of the iteration.
     */
    protected int startIndex ;
    /**
     *  Current index of the next array item to deliver.
     */
    protected int current = -1 ; {
      advance ( ) ;
      startIndex = current ;
    }
    @Override public Pair<K, V> get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorOperationException ( ) ; }
      return table[current] ;
    }
    @Override public void advance ( ) {
      while ( ( ++current < table.length ) && ( table[current] == null ) ) { }
    }
    @Override  public void advance ( int increment ) {
      for ( ; increment > 0 ; --increment ) {
        if ( atEnd ( ) ) { break ; }
        advance ( ) ;
      }
    }
    @Override  public boolean atBegin ( ) { return current == startIndex ; }
    @Override public boolean atEnd ( ) { return current >= table.length ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
    	if ( o == this ) { return true ; }
    	if ( ( o == null ) || ! ( o instanceof OpenHashTable.Iterator ) ) { return false ; }
      return current == ( (Iterator) o ).current ;
    }
    @Override  public OpenHashTable<K, V> getContainer ( ) { return OpenHashTable.this ; }
    /**
     *  Clone this iterator.
     */
    @Override  public final Iterator clone ( ) { return (Iterator) super.clone ( ) ; }
    /**
     *  This is an editing iterator so we allow removal.
     */
    @Override public void remove ( ) {
      if ( ! atEnd ( ) ) {
        table[current] = null ;
        --pairCount ;
      }
    }
  }
  /**
   *  Find the index of a given key, i.e. the position in the table of an entry with this key or to the
   *  first null entry.
   *
   *  <p> This is as near O(1) as we can get.</p>
   *
   *  @param o the item to find the index of.
   *  @return the index into table of the value the entry is null is key is not in the table.
   *  @exception HashTableFullException is the hash table has no more empty space and the value is
   *  not in the hash table.
   */
  private int findIndex ( final K o ) {
    final int hashValue = hashFunction.hash ( o ) % table.length ;
    int i = hashValue ;
    while ( ( table[i] != null ) && ! table[i].getA ( ).equals ( o ) ){
      i = ( i + 1 ) % table.length ;
      if ( i == hashValue ) { throw new HashTableFullException ( ) ; }
    }
    return i ;
  }
}
