/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  The simplest instantiatable <code>BinaryTree</code>.
 *
 *  <p>The insertion and deletion policy is determined by a <code>Comparator</code> method object realizing
 *  an order relation along with a <code>Comparator</code> realizing the equality relation.  We use this
 *  approach of using a <code>Comparator</code> to define equality rather than using the <code>equals</code>
 *  method directly so that clients of this class can define equality however they want.</p>
 *
 *  <p>Trees of this class can accept duplicate entries (the default) or not.  The ability to not have
 *  duplicates means that instances of this class can be used as representations of <code>Set</code>s and
 *  the ability to have duplicates means that objects of this class can be used as representations for
 *  <code>Bag</code>s .</p>
 *
 *  @author Russel Winder
 */
public class OrderedBinaryTree<T> extends AbstractBinaryTree<T> {
  protected final Comparator<? super T> order ;
  protected final Comparator<? super T> equal ;
  protected final boolean duplicatesAllowed ;
  /**
   *  Construct an <code>OrderedBinaryTree</code> with duplicate entries allowed using
   *  <code>EqualToComparator</code> as the equality relation.
   *
   *  @param o the order relation.
   */
  public OrderedBinaryTree ( final Comparator<? super T> o ) {
    this ( o , new EqualToComparator<T> ( ) , true ) ;
  }
  /**
   *  Construct an <code>OrderedBinaryTree</code> with explicit specification of whether duplicates are
   *  allowed and using <code>EqualToComparator</code> as the equality relation.
   *
   *  @param o the order relation.
   *  @param dups <code>true</code> means that duplicate entries are allowed.
   */
  public OrderedBinaryTree ( final Comparator<? super T> o , final boolean dups ) {
    this ( o , new EqualToComparator<T> ( ) , dups ) ;
  }
  /**
   *  Construct an <code>OrderedBinaryTree</code> with duplicate entries allowed using an explicitly
   *  specified equality relation.
   *
   *  @param o the order relation.
   *  @param e the equality relation.
   */
  public OrderedBinaryTree ( final Comparator<? super T> o , final Comparator<? super T> e ) {
    this ( o , e , true ) ;
  }
  /**
   *  Construct an <code>OrderedBinaryTree</code> with an explicitly specified equality relation and
   *  explicit specification of whether duplicates are allowed.
   *
   *  @param o the order relation.
   *  @param e the equality relation.
   *  @param dups <code>true</code> means that duplicate entries are allowed.
   */
  public OrderedBinaryTree ( final Comparator<? super T> o , final Comparator<? super T> e , final boolean dups ) {
    order = o ;
    equal = e ;
    duplicatesAllowed = dups ;
  }
  @Override public void add ( final T o ) {
    final SearchIterator si = searchFor ( o ) ;
    if ( si.atEnd ( ) ) { si.set ( o ) ; }
    else { if ( duplicatesAllowed ) { si.add(o) ; } }
  }
  @Override public boolean remove ( final T o) {
    final SearchIterator si = searchFor ( o ) ;
    if ( ! si.atEnd ( ) ) { si.remove() ; return true ; }
    return false ;
  }
  @Override public final boolean contains ( final T o ) { return ! searchFor ( o ).atEnd ( ) ; }
  @Override public OrderedBinaryTree<T> clone ( ) {
    final OrderedBinaryTree<T> t = (OrderedBinaryTree<T>) super.clone ( ) ;
    //  Have to clone the dummy node as well as the actual data nodes otherwise we end up with an element of
    //  sharing that ruins the cloning.
    t.root = new Node<T> ( null , null , null , null ) ;
    if ( ! isEmpty ( ) ) { t.root.left = root.left.clone ( t.root ) ; }
    return t ;
  }
@SuppressWarnings ( "unchecked" )
@Override public boolean equals ( final Object o ) {
    if ( this == o ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof OrderedBinaryTree<?> ) ) { return false ; }
    final InputIterator<T> i1 = iterator ( ) ;
    final InputIterator<T> i2 = ( (OrderedBinaryTree<T>) o ).iterator ( ) ;
    for ( ; ! i1.atEnd ( ) && ! i2.atEnd ( ) ; i1.advance ( ) , i2.advance ( ) ) {
      final T o1 = i1.get ( ) ;
      final T o2 = i2.get ( ) ;
      if ( ( o1 == null ) || ( o2 == null ) || ! o1.equals ( o2 ) ) { return false ; }
    }
    assert i1.atEnd ( ) && i2.atEnd ( )  ;
    return true ;
  }
  /**
   *  A special iterator to handle searching through the tree (usually for adding or removing an item).
   *  This iterator is needed so as to ensure O(log(n)) behaviour -- the default search behaviour is linear
   *  using the inorder traversal iterator.  Note that this is an iterator over the <code>Node&lt:T></code>s
   *  not over the <code>T</code>s so as to enable direct editing of the <code>Node</code> found.
   *
   *  @author Russel Winder
   */
  //  This is an inner class because it needs access to the data items held in the enclosing class.
  protected class SearchIterator extends AbstractIterator<Node<T>> {
    /**
     *  The item to be delivered up by the iterator.
     */
    protected Node<T> theItem = root.left ;
    /**
     *  The parent of the item to be delivered up by the iterator.
     */
    protected Node<T> theParent = root ;
    /**
     *  A variable to keep track of whether the child is the left or right child of the parent.  Cannot
     *  avoid use of this variable since an integral part of the algorithm is ending up with the child
     *  reference being null at which point left and right can be indistinguishable.
     *
     *  <p> The default stems from the root node of the tree being the left child of the dummy root
     *  hook.</p>
     */
    protected boolean isLeft = true ;
    /**
     *  The value we are given at the outset.
     */
    protected final T theValue ;
    /**
     *  The constructor.  We must have an item to search for.  This constructor allows us to use an
     *  alternate equality relation.
     *
     *  @param o the item to search for.
     */
    public SearchIterator ( final T o ) {
      theValue = o ;
      search ( theValue ) ;
    }
    /**
     *  Deliver up the next <code>Node</code> in the iteration.  NB This is a different approach to the
     *  normal iterators since it may, correctly, deliver null.
     */
    @Override public Node<T> get ( ) {
      if ( theParent == null ) { throw new InvalidIteratorOperationException ( ) ; }
      return theItem ;
    }
    /**
     *  Set the datum of the current <code>Node</code> to be the proferred object.  If <code>theItem</code>
     *  is null then attach to the parent as a new <code>Node</code>.
     *
     *  @param o the object to assign to teh slot.
     */
    //  TODO : This operation could leave the tree in an inconsistent state since the new item could be out
    //  of sorted sequence.  There really ought to be some sort of protection so as to guarantee the order
    //  relation.  Of course the whole SearchIterator/searchFor mechanism is protected so they should not
    //  get used improperly, nothing wrong with a bit of paranoia though.
    public void set ( final T o ) {
      if ( theParent == null ) { throw new InvalidIteratorOperationException ( ) ; }
      if ( theItem == null ) {
        final Node<T> n = new Node<T> (o, null, null, theParent) ;
        if ( isLeft ) { theParent.left = n ; }
        else { theParent.right = n ; }
      } else { theItem.datum = o ; }
    }
    @Override public void advance ( ) {
      if ( theParent == null ) { throw new InvalidIteratorOperationException ( ) ; }
      theParent = theItem ;
      if ( theItem != null ) {
        //  If there are duplicates then they are to be found in the left child.  See add.
        theItem = theItem.left ;
        isLeft = true ;
        search ( theValue ) ;
      }
    }
    @Override public void advance ( int increment ) {
      if ( increment < 0 ) { throw new InvalidIteratorOperationException ( ) ; }
      for ( ; increment > 0 ; --increment ) {
        if ( atEnd ( ) ) { break ; }
        advance ( ) ;
      }
    }
    @Override public boolean atBegin ( ) { return theItem == root.left ; }
    @Override public boolean atEnd ( ) { return theItem == null ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof OrderedBinaryTree.SearchIterator ) ) { return false ; }
      return theItem.equals ( ( (SearchIterator) o ).theItem ) ;
    }
    //  There is a type problem here: we must return a Container<Node<T>> but OrderedBinaryTree.this is a
    //  Container<T> and so there is no compatibility.  The only solution is to ensure that this method is
    //  not used in a program.
    @Override public Container<Node<T>> getContainer( ) { throw new InvalidIteratorOperationException ( ) ; }
    /**
     *  Add an item.
     *
     *  @param o the item to add to the tree.
     */
    public void add ( final T o ) {
      while ( theItem != null ) {
        search ( o ) ;
        if ( ( theItem != null ) && ! duplicatesAllowed ) { throw new InvalidIteratorOperationException ( ) ; }
        if ( theItem == null ) { break ; }
        //  If there are duplicates then they are to be found in the left child.  See advance.
        theParent = theItem ;
        theItem = theItem.left ;
        isLeft = true ;
      }
      set ( o ) ;
    }
    /**
     *  Remove this item.  Advance the iterator first.
     */
    @Override public void remove ( ) {
      if ( theItem != null ) {
        final Node<T> toGo = theItem ;
        advance ( ) ;
        toGo.remove ( ) ;
      }
    }
    /**
     *  Go searching for an item.
     *
     *  @param o the value to search for.
     */
    private void search ( final T o ) {
      while ( theItem != null ) {
        if ( equal.relation ( o , theItem.datum ) ) { return ; }
        theParent = theItem ;
        if ( order.relation ( o , theItem.datum ) ) {
          theItem = theItem.left ;
          isLeft = true ;
        } else {
          theItem = theItem.right ;
          isLeft = false ;
        }
      }
    }
  }
  /**
   *  Find a particular node in the tree using an O(log(n)) algorithm.  If this mechanism is not made
   *  available then all searching is linear since only the traversal iterators are available.  Unlike the
   *  other four iterators, this one is intended only for internal purposes since it has to provide access
   *  to the internal representation.</p>
   *
   *  @param o the value to search for.
   *  @return an iterator positioned at the value.
   */
  protected SearchIterator searchFor ( final T o ) { return new SearchIterator ( o ) ; }
}
