/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An implementation of <code>Mapping</code> using an <code>OrderedBinaryTree</code>.
 *
 *  @author Russel Winder
 */
public class OrderedBinaryTreeMapping<D extends Comparable<D>, R> extends OrderedBinaryTreeRelation<D, R> implements Mapping<D, R> {
  /**
   *  Construct an <code>OrderedBinaryTreeMapping</code> using the default equality relation.  We must
   *  insist on being given a valid <code>Comparator</code> defining the order relation on the domain so as
   *  to implement the insertion policy.
   *
   *  @param d the comparator providing an order.
   */
  public OrderedBinaryTreeMapping ( final Comparator<D> d ) { super ( d ) ; }
  /**
   *  Insert a (d, r) into the <code>OrderedBinaryTreeMapping</code>.  Ensure that we enforce the property
   *  that there must be a single item in the range related to a given item from the domain, i.e. do not
   *  just add pairs in but check if a pair for the domain value already exosts. The parameter has to be
   *  <code>Pair</code> here so as to conform correctly to <code>AbstractRelation</code> which works in
   *  terms of <code>Pair</code>s and not <code>ComparablePair</code>s.
   *
   *  @throws InvalidIteratorOperationException if the <code>Pair</code> parameter is not
   *  in fact a <code>ComparablePair</code>.
   */
  @Override public void add ( final Pair<D, R> o ) {
    if ( ! (o instanceof ComparablePair<?,?>) ) { throw new InvalidIteratorOperationException ( ) ; }
    final ComparablePair<D, R> oo = (ComparablePair<D, R>) o ;
    final OrderedBinaryTree<ComparablePair<D, R>>.SearchIterator si = storage.searchFor ( oo ) ;
    if ( si.atEnd ( ) ) { si.set ( oo ) ; }
    else {
      si.remove ( ) ;
      storage.add ( oo ) ;
    }
  }
  @Override public final R getValue ( final D d ) {
    final InputIterator<R> i = range ( d ) ;
    return i.atEnd ( ) ? null : i.get ( ) ;
  }
  @Override public boolean equals ( final Object o ) {
    if ( o == this ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof OrderedBinaryTreeMapping<?,?> ) ) { return false ; }
    return super.equals ( o ) ;
  }
  @Override public OrderedBinaryTreeMapping<D, R> clone ( ) { return (OrderedBinaryTreeMapping<D, R>) super.clone ( ) ; }
}
