/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2009,2011 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A relation of type ( D, R ) implemented using an <code>OrderedBinaryTree&lt;ComparablePair&lt;D,
 *  C>></code>.  <code>D</code> must conform to <code>Comparable&lt;D></code> so as to provide the order
 *  relation required for an ordered binary tree of pairs.
 *
 *  @author Russel Winder
 */
public class OrderedBinaryTreeRelation<D extends Comparable<D>, R> extends AbstractRelation<D, R> {
  /**
   *  The <code>OrderedBinaryTree&lt;ComparablePair&lt;D, R>></code> that acts as our storage.
   */
  protected OrderedBinaryTree<ComparablePair<D, R>> storage ;
  /**
   *  Construct an <code>OrderedBinaryTreeRelation&lt;D, R></code> using the default equality relation.  We
   *  must insist on being given a valid <code>Comparator&lt;ComparablePair&lt;D, R>></code> defining the
   *  order relation to implement the insertion policy.
   *
   *  @param c the comparator that provides the order.
   */
  public OrderedBinaryTreeRelation ( final Comparator<D> c ) {
    //  NB We can use a direct order relation on ComparablePair<D, R> for equality but must use an adapter
    //  to properly wrap the order relation on the first elements of the pairs.
    storage = new OrderedBinaryTree<ComparablePair<D, R>> (
                                                 new PairAComparator<D, R, ComparablePair<D, R>> ( c ),
                                                 new EqualToComparator<ComparablePair<D, R>> ( ) ) ;
  }
  @Override public void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public int size ( ) { return storage.size ( ) ; }
  @Override public void add ( final D d , final R r ) { add ( new ComparablePair<D, R> ( d , r ) ) ; }
  //  The parameter has to be <code>Pair<D, R></code> in add and remove as AbstractRelation<D,R> works in
  //  terms of Pair<D, R> and not ComparablePair<D, R> and we have to ensure overriding.
  @Override public void add ( final Pair<D, R> o ) {
    if ( ! ( o instanceof ComparablePair<?,?> ) ) { throw new InvalidIteratorOperationException ( ) ; }
    storage.add ( (ComparablePair<D, R>) o ) ;
  }
  @Override public boolean remove ( final Pair<D, R> o ) {
    if ( ! ( o instanceof ComparablePair<?,?> ) ) { throw new InvalidIteratorOperationException ( ) ; }
    return storage.remove ( (ComparablePair<D, R>) o ) ;
  }
  @Override public OrderedBinaryTreeRelation<D, R> clone ( ) {
    final OrderedBinaryTreeRelation<D, R> t = (OrderedBinaryTreeRelation<D, R>) super.clone ( ) ;
    if ( ! isEmpty( ) ) { t.storage = storage.clone ( ) ; }
    return t ;
  }
  public boolean contains ( final ComparablePair<D, R> o ) { return storage.contains ( o ) ; }
  @Override public final Iterator<Pair<D, R>> begin ( ) { return iterator ( ) ; }
  @Override public final Iterator<Pair<D, R>> end ( ) {
    final Iterator<Pair<D, R>> i = iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public Iterator<Pair<D, R>> iterator ( ) {
    return new Iterator<Pair<D, R>> ( ) {
      @Override public Pair<D, R> get ( ) { return cursor.get ( ) ; }
      @Override public OrderedBinaryTreeRelation<D, R> getContainer ( ) { return OrderedBinaryTreeRelation.this ; }
    } ;
  }
  /**
   *  The basic abstract iterator over <code>OrderedBinaryTreeRelation&lt;D, R></code>s.  This base is used
   *  to construct the main iterator over the pairs in the tree and also the projection iterators that work
   *  over the firsts of each pair and the seconds of each pair.
   *
   *  <p>This works fine except for the <code>getContainer</code> method.  The type system gets in the way
   *  of doing what was straight forward without generics.</p>
   *
   *  @author Russel Winder
   */
  public abstract class Iterator<T> extends AbstractIterator<T> {
    /**
     *  The cursor is just an iterator over the <code>OrderedBinaryTreeRelation&lt;D, R></code>. Use the
     *  default of in-order traversal.
     */
    protected InputIterator<ComparablePair<D, R>> cursor = storage.iterator ( ) ;
    @Override public abstract T get ( ) ;
    @Override public void advance ( ) { cursor.advance ( ) ; }
    @Override public void advance ( final int increment ) { cursor.advance ( increment ) ; }
    @Override public boolean atBegin ( ) { return cursor.atBegin ( ) ; }
    @Override public boolean atEnd ( ) { return cursor.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof Iterator ) ) { return false ; }
      return cursor.equals ( ( (Iterator<T>) o ).cursor ) ;
    }
     //  When the type parameter T is consistent with ComparablePair<D, R> this method is type correct.  If
     //  the T has any other value then we cannot return the correct type.  We could without generics but
     //  this wasn't really type correct though in the context of typeless containers it was fine.  The
     //  upshot is that we now enforce type correctness and so we throw an exception here and let the
     //  sub-classes do the right thing if they can.
    @Override public Container<T> getContainer ( ) { throw new InvalidIteratorOperationException ( ) ; }
    @Override public final Iterator<T> clone ( ) {
      final Iterator<T> i = (Iterator<T>) super.clone ( ) ;
      i.cursor = cursor.clone ( ) ;
      return i ;
    }
  }
  @Override public InputIterator<D> domain ( ) {
    return new Iterator<D> ( ) { @Override public D get ( ) { return cursor.get ( ).getA ( ) ; } } ;
  }
  // TODO: Eclipse marks this warning suppression as unneeded, but javac complains.
  @SuppressWarnings ( "unchecked" )
  @Override public InputIterator<D> domain ( final R r ) {
    class DomainIterator extends Iterator<D> {
      public DomainIterator ( ) { if ( ! cursor.get ( ).getB ( ).equals ( r ) ) { moveOn ( ) ; } }
      @Override public D get ( ) { return cursor.get ( ).getA ( ) ; }
      @Override public void advance ( ) { moveOn ( ) ; }
      private void moveOn ( ) {
        while ( ! atEnd ( ) ) {
          super.advance ( ) ;
          if ( atEnd ( ) || cursor.get ( ).getB ( ).equals ( r ) ) { break ; }
        }
      }
    }
    return new DomainIterator ( ) ;
  }
  @Override  public InputIterator<R> range ( ) {
    return new Iterator<R> ( ) { @Override public R get ( ) { return cursor.get ( ).getB ( ) ; } } ;
  }
  // TODO: Eclipse marks this warning suppression as unneeded, but javac complains.
  @SuppressWarnings ( "unchecked" )
  @Override public InputIterator<R> range ( final D d ) {
    class RangeIterator extends Iterator<R> {
      public RangeIterator ( ) { if ( ! cursor.get ( ).getA ( ).equals ( d ) ) { moveOn ( ) ; } }
      @Override public R get ( ) { return cursor.get ( ).getB ( ) ; }
      @Override public void advance ( ) { moveOn ( ) ; }
      private void moveOn ( ) {
        while ( ! atEnd ( ) ) {
          super.advance ( ) ;
          if ( atEnd ( ) || cursor.get ( ).getA ( ).equals ( d ) ) { break ; }
        }
      }
    }
    return new RangeIterator ( ) ;
  }
}
