/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A set implemented using an <code>OrderedBinaryTree</code>.
 *
 *  @author Russel Winder
 */
public class OrderedBinaryTreeSet<T> extends AbstractSet<T> {
  /**
   *  The <code>OrderedBinaryTree</code> that acts as our storage.
   */
  protected OrderedBinaryTree<T> storage ;
  /**
   *  Construct an <code>OrderedBinaryTreeSet</code> using the equality relation
   *  (<code>EqualToComparator</code>) as a method object encapsulating the <code>equals</code> method.  The
   *  less than order relation must be provided by the instantiator.
   *
   *  @param c a comparator to provide an order.
   */
  public OrderedBinaryTreeSet ( final Comparator<? super T> c ) {
    // NB Must ensure that the OrderedBinaryTree<T> does not allow duplicates.
    storage = new OrderedBinaryTree<T> ( c , new EqualToComparator<T> ( ) , false ) ;
  }
  /**
   *  Construct an <code>OrderedBinaryTreeSet</code> using the equality relation
   *  (<code>EqualToComparator</code>) as a method object encapsulating the <code>equals</code> method and
   *  copying in the values from the <code>Container</code>.
   *
   *  @param cp a comparator to provide an order.
   *  @param ct the container from which to shallow copy items into this set.
   */
  public OrderedBinaryTreeSet ( final Comparator<? super T> cp , final Container<T> ct ) {
    this ( cp ) ;
    copyIn ( ct ) ;
  }
  @Override public void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public int size ( ) { return storage.size ( ) ; }
  @Override public void add ( final T o ) { storage.add ( o ) ; }
  @Override public boolean remove ( final T o ) { return storage.remove ( o ) ; }
  @Override public OrderedBinaryTreeSet<T> clone ( ) {
    final OrderedBinaryTreeSet<T> t = (OrderedBinaryTreeSet<T>) super.clone ( ) ;
    if ( ! isEmpty ( ) ) { t.storage = storage.clone ( ) ; }
    return t ;
  }
  @Override public boolean contains ( final T o ) { return storage.contains ( o ) ; }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  Define the iterator over <code>OrderedBinaryTreeSet</code>s.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  The cursor is just an iterator over the <code>OrderedBinaryTreeSet</code>. Use the default of
     *  in-order traversal.
     */
    protected InputIterator<T> cursor = storage.iterator ( ) ;
    @Override public T get ( ) { return cursor.get ( ) ; }
    @Override public void advance ( ) { cursor.advance ( ) ; }
    @Override public void advance ( final int increment ) { cursor.advance ( increment ) ; }
    @Override public boolean atBegin ( ) { return cursor.atBegin ( ) ; }
    @Override public boolean atEnd ( ) { return cursor.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof OrderedBinaryTreeSet.Iterator ) ) { return false ; }
      return cursor.equals ( ( (Iterator) o ).cursor ) ;
    }
    @Override public OrderedBinaryTreeSet<T> getContainer ( ) { return OrderedBinaryTreeSet.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.cursor = cursor.clone ( ) ;
      return i ;
    }
  }
}
