/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Dequeue</code> implemented using a circular array.
 *
 *  <p> The head (which is the "smaller" index trailing behind the tail!)  and tail (which is the "larger"
 *  index leading the head!)  locations are wrapped around in the array as needed creating a circular array.
 *  Use the index value -1 (which is illegal of course) to represent an empty <code>Dequeue</code>.
 *
 *  @author Russel Winder
 */
public class PrimitiveArrayDequeue<T> extends AbstractDequeue<T> {
  /**
   *  The index of the head element in the array.
   */
  private int head ;
  /**
   *  The index of the tail element in the array.
   */
  private int tail ;
  /**
   *  The array that stores all the data.
   */
  private Object[] storage ;
  /**
   *  An <code>PrimitiveArrayQueue</code> constructor.  The array size is unspecified, a default size (100)
   *  is used.
   */
  public PrimitiveArrayDequeue ( ) { this ( 100 ) ; }
  /**
   *  An <code>PrimitiveArrayQueue</code> constructor.  The array size is specified and used.
   *
   *  @param initialSize The initial size of the <code>Queue</code>.
   */
  public PrimitiveArrayDequeue ( final int initialSize ) {
    storage = new Object [initialSize] ;
    makeEmpty ( ) ;
  }
  @Override public final void headAdd ( final T o ) {
    if ( isEmpty ( ) ) {
      head = 0 ;
      tail = 0 ;
    } else if ( head == 0 ) {
      if ( tail == storage.length - 1 ) { throw new ContainerCapacityExceededException ( ) ; }
      head = storage.length - 1 ;
    } else { head-- ; }
    storage[head] = o ;
  }
  @Override public final void tailAdd ( final T o ) {
    if ( isEmpty ( ) ) {
      head = 0 ;
      tail = 0 ;
    } else if ( tail == storage.length - 1 ) {
      if ( head == 0 ) { throw new ContainerCapacityExceededException() ; }
      tail = 0 ;
    } else { tail++ ; }
    storage[tail] = o ;
  }
  @SuppressWarnings ( "unchecked" ) @Override public final T headRemove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    final T returnValue = (T) storage[head] ;
    if ( head == tail ) { makeEmpty ( ) ; }
    else { head = ( head + 1 ) % storage.length ; }
    return returnValue ;
  }
  @SuppressWarnings ( "unchecked" ) @Override public final T tailRemove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    final T returnValue = (T) storage[tail] ;
    if ( head == tail ) { makeEmpty() ; }
    else { tail = ( tail == 0 ) ? storage.length - 1 : tail - 1 ; }
    return returnValue ;
  }
  @SuppressWarnings ( "unchecked" ) @Override public final T headPeek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException() ; }
    return (T) storage[head] ;
  }
  @SuppressWarnings ( "unchecked" ) @Override public final T tailPeek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return (T) storage[tail] ;
  }
  @Override public final void makeEmpty ( ) {
    head = -1 ;
    tail = -1 ;
    for ( int i = 0 ; i < storage.length ; ++i ) { storage[i] = null ; }
  }
  @Override public final boolean isEmpty ( ) { return head == -1 ; }
  @Override public final int size ( ) {
    if ( ( head == -1 ) || ( tail == -1 ) ) { return 0 ; }
    return ( head > tail ) ? ( storage.length - head + tail + 1 ) : ( tail - head + 1) ;
  }
  @Override public final PrimitiveArrayDequeue<T> clone ( ) {
    final PrimitiveArrayDequeue<T> q = (PrimitiveArrayDequeue<T>) super.clone ( ) ;
    q.storage = storage.clone ( ) ;
    return q ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  An iterator over <code>PrimitiveArrayDequeue</code>s.
   *
   *  This is just a proxy for an iterator from the underlying representation so as to stop the ability to
   *  write.  Implement only a forward iterator.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  We just need to have an index as a cursor.
     */
    private int current = head ;
    @SuppressWarnings ( "unchecked" ) @Override public final T get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorException ( ) ; }
      return (T) storage[current] ;
    }
    @Override public final void advance ( ) {
      if ( ! atEnd ( ) ) { current = (current == tail) ? -1 : ( (current == storage.length - 1) ? 0 : current + 1 ) ;
      }
    }
    @Override public void advance ( int increment ) {
      if ( increment < 0 ) { throw new InvalidIteratorOperationException ( ) ; }
      for ( ; increment > 0 ; --increment ) {
        if ( atEnd ( ) ) { break ; }
        advance ( ) ;
      }
    }
    @Override public boolean atBegin ( ) { return current == head ; }
    @Override public final boolean atEnd ( ) { return current == -1 ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof PrimitiveArrayDequeue.Iterator ) ) { return false ; }
      return current == ( (Iterator) o ).current ;
    }
    @Override public final PrimitiveArrayDequeue<T> getContainer ( ) { return PrimitiveArrayDequeue.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.current = current ;
      return i ;
    }
  }
}
