/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Stack</code> using a primitive array as underlying storage mechanism.
 *
 *  <p> The top of stack is the element at the end, this way we have O(1) insert and delete.  An array index
 *  of -1 indicates an empty stack.</p>
 *
 *  @author Russel Winder
 */
public class PrimitiveArrayStack<T> extends AbstractStack<T> {
  /**
   *  The top of the stack index.
   */
  private int topOfStack ;
  /**
   *  The data storage area.
   */
  private Object[] storage ;
  /**
   *  A <code>PrimitiveArrayStack</code> constructor.  The array size is unspecified, a default size of 100
   *  is used.
   */
  public PrimitiveArrayStack ( ) { this ( 100 ) ; }
  /**
   *  An <code>PrimitiveArrayStack</code> constructor.  The array size is specified and used.
   *
   *  @param initialSize The initial size of the <code>Stack</code>.
   */
  public PrimitiveArrayStack(final int initialSize) {
    storage = new Object [initialSize] ;
    makeEmpty ( ) ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public final T remove() {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return (T) storage[topOfStack--] ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public final T peek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return (T) storage[topOfStack] ;
  }
  @Override public final void makeEmpty ( ) { topOfStack = -1 ; }
  @Override public final int size ( ) { return topOfStack + 1 ; }
  @Override public final void add ( final T o ) {
    if ( topOfStack == storage.length - 1 ) { throw new ContainerCapacityExceededException ( ) ; }
    storage[++topOfStack] = o ;
  }
  @Override public final PrimitiveArrayStack<T> clone ( ) {
    final PrimitiveArrayStack<T> s = (PrimitiveArrayStack<T>) super.clone ( ) ;
    s.storage = storage.clone ( ) ;
    return s ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  An iterator over <code>PrimitiveArrayStack</code>s.
   *
   *  <p>This is just a proxy for an iterator from the underlying representation so as to stop the ability
   *  to write.  Implement only a forward iterator.</p>
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  We just need to have an index as a cursor.
     */
    private int current = 0 ;
    @SuppressWarnings ( "unchecked" )
    @Override public final T get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorException ( ) ; }
      return (T) storage[current] ;
    }
    @Override public final void advance ( ) { ++current ; }
    @Override public void advance ( final int increment ) {
      if ( increment < 0 ) { throw new InvalidIteratorOperationException ( ) ; }
      int putativeIndex = current + increment ;
      if ( putativeIndex >= size ( ) ) { putativeIndex = size ( ) ; }
      current = putativeIndex ;
    }
    @Override public boolean atBegin ( ) { return current == 0 ; }
    @Override public final boolean atEnd ( ) { return current >= size ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof PrimitiveArrayStack.Iterator ) ) { return false ; }
      return current == ( (Iterator) o ).current ;
    }
    @Override public final PrimitiveArrayStack<T> getContainer ( ) { return PrimitiveArrayStack.this ; }
  }
}
