/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An interface defining a priority queue.  A priority queue is a type of queue where items have a priority
 *  that imposes a second order on the queue.
 *
 *  @author Russel Winder
 */
public interface PriorityQueue<T> extends Queue<T> {
  /**
   *  Insert a datum into ourself specifying a priority.  If the priority is outside the range of
   *  possibilities it is assigned to the appropriate bound.
   *
   *  @param o the item to add to this <code>PriorityQueue</code>.
   *  @param priority the priority of this item in this <code>PriorityQueue</code>.
   */
  void add ( T o , int priority ) ;
  /**
   *  Remove the next item at a given priority.
   *
   *  @param priority the priority to search for an item to remove.
   *  @return a reference to the first item at the specified priority or null if there is no such item.
   */
  T remove ( int priority ) ;
  /**
   *  Clone this <code>PriorityQueue</code>.
   *
   *  @return a clone of this <code>PriorityQueue</code>.
   */
  @Override PriorityQueue<T> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>PriorityQueue</code> for use in
   *  multi-threaded applications.  This is just a forwarding proxy with synchronization control and
   *  locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<E> extends Queue.Synchronized<E> implements PriorityQueue<E> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param q the queue to synchronize.
     */
    public Synchronized ( final PriorityQueue<E> q ) { super ( q ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param q the queue to synchronize.
     *  @param o the lock object to use.
     */
    public Synchronized ( final PriorityQueue<E> q , final Object o ) { super ( q , o ) ; }
    @Override public final void add ( final E o , final int priority ) {
      synchronized ( theLock ) { ( (PriorityQueue<E>) theContainer ).add ( o , priority ) ; }
    }
    @Override public final E remove ( final int priority ) {
      synchronized ( theLock ) { return ( (PriorityQueue<E>) theContainer ).remove ( priority ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  PriorityQueue.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public PriorityQueue.Synchronized<E> clone ( ) { return (PriorityQueue.Synchronized<E>) super.clone ( ) ; }
  }
}
