/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  An interface defining the necessary properties to be deemed to be a queue.
 *
 *  <p> We define a queue to be any data structure that can have items inserted and removed and
 *  which guarantees a First in &ndash; First out ordering policy.</p>
 *
 *  @author Russel Winder
 */
public interface Queue<T> extends Container<T> {
  /**
   *  Peek at the head of the <code>Queue</code> but do not remove it.
   *
   *  @return the item at the head of this <code>Queue</code>.
   *  @exception AccessEmptyContainerException if we attempt to peek on an empty <code>Queue</code>.
   */
  T peek ( ) ;
  /**
   *  Remove an item from the <code>Queue</code> and deliver it up.
   *
   *  @return the item at the head of this <code>Queue</code>.
   *  @exception AccessEmptyContainerException if we attempt to remove from an empty <code>Queue</code>.
   */
  T remove ( ) ;
  /**
   *  Clone this <code>Queue</code>.
   *
   *  @return a clone of this <code>Queue</code>.
   */
  @Override Queue<T> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>Queue</code> for use in multi-threaded
   *  applications.This is just a forwarding proxy with synchronization control and locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<E> extends Container.Synchronized<E> implements Queue<E> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param q the <code>Queue</code> to be synchronized.
     */
    public Synchronized ( final Queue<E> q ) { super ( q ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param q the <code>Queue</code> to be synchronized.
     *  @param o the lock object.
     */
    public Synchronized ( final Queue<E> q , final Object o ) { super ( q , o ) ; }
    @Override public final E peek ( ) { synchronized ( theLock ) { return ( (Queue<E>) theContainer ).peek ( ) ; } }
    @Override public final E remove ( ) { synchronized ( theLock ) { return ( (Queue<E>) theContainer ).remove ( ) ; } }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  Queue.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public Queue.Synchronized<E> clone ( ) { return (Queue.Synchronized<E>) super.clone ( ) ; }
  }
}
