/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  The interface describing a relation.
 *
 *  @author Russel Winder
 */
public interface Relation<D, R> extends Set<Pair<D, R>> {
  /**
   *  Insert a (d, r) pair into the <code>Relation</code>.
   *
   *  @param d value from the domain
   *  @param r value from the range
   */
  void add ( D d , R r ) ;
  /**
   *  Remove the ( d , r ) pair.
   *
   *  @param d value from the domain
   *  @param r value from the range
   *  @return whether the element was found and deleted.
   */
  boolean remove ( D d , R r ) ;
  /**
   *  Create an iterator over all the sub-set of the domain that is used in this relation.
   *
   *  @return an iterator over the domain of this <code>Relation</code>.
   */
  InputIterator<D> domain ( ) ;
  /**
   *  Create an iterator over the sub-set of the domain that is used in this relation which relate
   *  to a given value in the range.
   *
   *  @param r value from the range
   *  @return an iterator over the items in the domain of this <code>Relation</code> that map to the
   *  specified item from the range.
   */
  InputIterator<D> domain ( R r ) ;
  /**
   *  Create an iterator over all the values range.
   *
   *  @return an iterator over the range of this <code>Relation</code>.
   */
  InputIterator<R> range ( ) ;
  /**
   *  Create an iterator over the sub-set of the range that relate to a given item from the domain.
   *
   *  @param d value from the domain
   *  @return an iterator over the items in the range of this <code>Relation</code> that are mapped to by
   *  the specified item from the domain.
   */
  InputIterator<R> range ( D d ) ;
  /**
   *  Clone this <code>Relation</code>.
   *
   *  @return a clone of this <code>Relation</code>.
   */
  @Override Relation<D, R> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>Relation</code> for use in multi-threaded
   *  applications.  This is just a forwarding proxy with synchronization control and locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<SD, SR> extends Set.Synchronized<Pair<SD, SR>> implements Relation<SD, SR> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param r the relation to synchronize.
     */
    public Synchronized ( final Relation<SD, SR> r ) { super ( r ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param r the relation to synchronize.
     *  @param o the lock object to use.
     */
    public Synchronized ( final Relation<SD, SR> r , final Object o ) { super ( r , o ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public final void add ( final SD d , final SR r ) {
      synchronized ( theLock ) { ( (Relation<SD, SR>) theContainer).add ( d , r ) ; }
    }
    @SuppressWarnings ( "unchecked" )
    @Override public final boolean remove ( final SD d , final SR r ) {
      synchronized ( theLock ) { return ( (Relation<SD, SR>) theContainer ).remove ( d , r ) ; }
    }
    @SuppressWarnings ( "unchecked" )
    @Override public final InputIterator<SD> domain ( ) {
      synchronized ( theLock ) { return ( (Relation<SD, SR>) theContainer ).domain ( ) ; }
    }
    @SuppressWarnings ( "unchecked" )
    @Override public final InputIterator<SD> domain ( final SR r ) {
      synchronized ( theLock ) { return ( (Relation<SD, SR>) theContainer ).domain ( r ) ; }
    }
    @SuppressWarnings ( "unchecked" )
    @Override public final InputIterator<SR> range ( ) {
      synchronized ( theLock ) { return ( (Relation<SD, SR>) theContainer ).range ( ) ; }
    }
    @SuppressWarnings ( "unchecked" )
    @Override public final InputIterator<SR> range ( final SD d ) {
      synchronized ( theLock ) { return ( (Relation<SD, SR>) theContainer ).range ( d ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  Relation.Synchronized, Set.Synchronized and Container.Synchronized so this is a lookup conflict.
    @SuppressWarnings ( "unchecked" )
    @Override public  Relation.Synchronized<SD, SR> clone ( ) { return (Relation.Synchronized<SD, SR>) super.clone ( ) ; }
  }
}
