/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A class providing reverse iteration capability.  This is an adapter on
 *  <code>BiDirectionalIterator</code> which reverses the meaning of the methods of
 *  <code>BiDirectionalIterator</code>.
 *
 *  <p>There is, of course, one major problem with <code>ReverseIterator</code>s, the semantics are subtly
 *  different for other <code>BiDirectionalIterator</code>s and this is to do with the reversal of the
 *  predicates <code>atBegin</code> and <code>atEnd</code> and the fact that the beginning of a sequence is
 *  the first element but the end of a sequence is the item after the last one.  Since
 *  <code>ReverseIterator</code> is actually not going to be used on a regular basis, it is here really for
 *  completeness, we do not attempt to fix the problem.  Instead we show the normal idiomatic use of
 *  <code>BiDirectionalIterator</code>s and the normal idiomatic use of <code>ReverseIterator</code>s.</p>
 *
 *  <p>First <code>BiDirectionalIterator</code>:
 *
 *  <pre>
 *    for ( BiDirectionalIterator i = datum.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
 *        //  Code implementing whatever algorithm using i.get ( )
 *    }
 *  </pre>
 *
 *   and now <code>ReverseIterator</code>:
 *
 *  <pre>
 *    ReverseIterator r = new ReverseIterator ( (BiDirectionalIterator)datum.end ( ) ) ;
 *    do {
 *        r.advance ( ) ;
 *        //  Code implementing whatever algorithm using r.get ( )
 *    } while ( ! r.atEnd ( ) ) ;
 *  </pre>
 *
 *  @author Russel Winder
 */
public class ReverseIterator<T> extends AbstractIterator<T> implements BiDirectionalIterator<T> {
  /**
   *  The <code>BiDirectionalIterator</code> we are working with.
   */
  protected BiDirectionalIterator<T> theIterator ;
  /**
   *  Constructor.  We can only work if we are given a <code>BiDirectionalIterator</code> to work with.
   *
   *  @param i the iterator to work in reverse.
   */
  public ReverseIterator ( final BiDirectionalIterator<T> i ) { theIterator = i ; }
  @Override public T get ( ) { return theIterator.get ( ) ; }
  @Override public void advance ( ) { theIterator.retreat ( ) ; }
  @Override public void advance ( final int increment ) { theIterator.retreat ( increment ) ; }
  @Override public boolean atBegin ( ) { return theIterator.atEnd ( ) ; }
  @Override public boolean atEnd ( ) { return theIterator.atBegin ( ) ; }
  @SuppressWarnings ( "unchecked" ) @Override public boolean equals ( final Object o ) {
    if ( o == this ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof ReverseIterator ) ) { return false ; }
    return theIterator.equals ( ( (ReverseIterator<T>) o ).theIterator ) ;
  }
  @Override public Container<T> getContainer ( ) { return theIterator.getContainer ( ) ; }
  @Override public ReverseIterator<T> clone ( ) {
    final ReverseIterator<T> newOne = (ReverseIterator<T>) super.clone ( ) ;
    newOne.theIterator = theIterator.clone ( ) ;
    return newOne ;
  }
  @Override public void set ( final T o ) { theIterator.set ( o ) ; }
  @Override public int distance ( final InputIterator<T> i ) { return theIterator.distance ( i ) ; }
  @Override public void retreat ( ) { theIterator.advance ( ) ; }
  @Override public void retreat ( final int increment ) { theIterator.advance ( increment ) ; }
}
