/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Sequence</code> implemented as a singly-linked list.
 *
 *  <p> This implementation uses a head and tail reference with the singly-linked list strung between
 *  them.</p>
 *
 *  @see Container
 *  @see Sequence
 *  @author Russel Winder
 */
public class SLList<T> extends AbstractSequence<T> {
  /**
   *  This is the class from which linked-list nodes are instantiated.  This is a very passive sort of node
   *  so it is a nested top-level class and not an inner class since it has no need to ever know who its
   *  parent is.
   *
   *  @author Russel Winder
   */
  private static class Node<E> {
    /**
     *  A <code>Node</code> has some datum that it is storing.
     */
    protected E datum ;
    /**
     *  A <code>Node</code> has a next which a reference to another <code>Node</code>.
     */
    protected Node<E> next ;
    /**
     *  Constructor for an unconnected <code>Node</code>.
     *
     *  @param o The datum for this <code>Node</code>.
     */
    public Node ( final E o ) { this ( o , null ) ; }
    /**
     *  Constructor for a <code>Node</code> with a given list of <code>Node</code>s that the newly
     *  constructed <code>Node</code> is to become the head of.
     *
     *  @param o The datum for this <code>Node</code>.
     *  @param n The <code>Node</code> to be the next of this one, i.e. the head of the extant list.
     */
    public Node ( final E o , final Node<E> n ) {
      datum = o ;
      next = n ;
    }
  }
  /**
   *  The head of the <code>SLList</code>.  This is the first <code>Node</code> in the list.
   */
  private Node<T> head = null ;
  /**
   *  The last <code>Node</code> in the list.  This is not absolutely necessary but having it makes
   *  appending <code>Node</code>s at the end of the list a lot easier.
   */
  private Node<T> tail = null ;
  /**
   *  The count of the number of elements in the <code>SLList</code>.  This is not absolutely necessary but
   *  having it makes a lot of operations a lot easier.  A good efficiency measure.
   */
  private int elementCount = 0 ;
  public SLList ( ) { }
  /**
   *  Construct an <code>SLList</code> by copying in the values from the the <code>Container</code> that is
   *  the parameter.
   *
   *  @param c the <code>Container</code> to copy the values from.
   */
  public SLList ( final Container<T> c ) { copyIn ( c ) ; }
  /**
   *  Construct an <code>SLList</code> by copying the slice of a <code>Sequence</code>defined by two
   *  indexes.
   *
   *  @param s the <code>Sequence</code> a slice of which is to be copied.
   *  @param b beginning index of the slice of <code>s</code>.
   *  @param e end index of the slice of <code>s</code>.
   */
  public SLList ( final Sequence<T> s , final int b , final int e ) { copyIn ( s , b , e ) ; }
  /**
   *  Construct an <code>SLList</code> by copying the sequence defined by two <code>InputIterator</code>s.
   *
   *  @param b beginning index of the slice of <code>s</code>.
   *  @param e end index of the slice of <code>s</code>.
   *  @exception InvalidIteratorException if the two <code>InputIterator</code>s are not working the same
   *  <code>SLList</code>.
   */
  public SLList ( final InputIterator<T> b , final InputIterator<T> e ) { copyIn ( b , e ) ; }
  @Override public final T first ( ) { validityCheck ( ) ; return head.datum ; }
  @Override public final T last ( ) { validityCheck ( ) ; return tail.datum ; }
  @Override public final T get ( final int index ) { validityCheck ( index ) ; return nodeAt ( index ).datum ; }
  @Override public final void set ( final int index , final T o ) {
    try { validityCheck ( index ) ; }
    catch ( AccessEmptyContainerException aece ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw aece ; }
    } catch ( IndexOutOfBoundsException ioobe ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw ioobe ; }
    }
    nodeAt ( index ).datum = o ;
  }
  @Override public final void remove ( final int index ) {
    validityCheck ( index ) ;
    Node<T> toGo = head ;
    if ( index == 0 ) {
      //  Deal with the case where it is the first Node<T> in the list that has to go as a special case.  If
      //  we removed the last item, make tail sensible.
      head = head.next ;
      if ( head == null ) { tail = null ; }
    } else {
      //  It is not the first item in the list that has to go and the index is known to be valid.  Find the
      //  element before the one we need to remove so that we can excise the one that has to go.
      final Node<T> previous = nodeAt ( index - 1 ) ;
      toGo = previous.next ;
      previous.next = toGo.next ;
      //  If we are removing the last element, we need to reset the tail.
      if ( toGo == tail ) { tail = previous ; }
    }
    //  Nullify the next of the garbage item just in case.  This is probably not required since the garbage
    //  collector will get things right but as a safety measure let's just do it.  Also ensure we decrement
    //  the element count to the proper value.
    toGo.next = null ;
    --elementCount ;
  }
  @Override public final void insert ( final int index , final T o ) {
    try { validityCheck ( index ) ; }
    catch (AccessEmptyContainerException aece) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else {  throw aece ; }
    }
    final Node<T> n = new Node<T> ( o ) ;
    if ( index == 0 ) {
      //  Handle insertion at the head of the SLList<T> as a special case: make the new item's next refer to
      //  the same object that head used to and then make head refer to the new item.
      n.next = head ;
      head = n ;
    }
    else {
      //  Find the correct place in the SLList<T> to do the insertion and then do the insertion.
      final Node<T> current = nodeAt ( index - 1 ) ;
      n.next = current.next ;
      current.next = n ;
    }
    ++elementCount ;
  }
  @Override public final void makeEmpty ( ) {
    head = null ;
    tail = null ;
    elementCount = 0 ;
  }
  @Override public final int size ( ) { return elementCount ; }
  @Override public final void add ( final T o ) {
    final Node<T> n = new Node<T> ( o ) ;
    if ( isEmpty ( ) ) { head = n ; tail = n ; }
    else { tail.next = n ; tail = n ; }
    ++elementCount ;
  }
  @Override public final boolean remove ( final T o ) {
    validityCheck ( ) ;
    Node<T> toGo = head ;
    if ( head.datum.equals ( o ) ) {
      //  Get rid of the head element.  Also if this was the last element make sure that tail gets reset.
      head = head.next ;
      if ( head == null ) { tail = null ; }
    }
    else {
      //  Walk down the list looking for the item to remove.  Cut it out only if it was found! Can do this
      //  in two ways:
      //
      //  1.  Use two references, one referring to the element being tested (current) and one to the
      //  previous element.  Previous is used to actually do the editing once current refers to the right
      //  Node<T>.
      //
      //  2.  Use a single reference to do the walking, performing all activities from the "previous"
      //  Node<T> so that we do not go past the place at which the editing has to be done.
      //
      //  The following is an implementation of algorithm 1 using two references.
      //
      //    Node<T> previous = head ;
      //    toGo = toGo.next ;
      //    while ( ( toGo != null ) && ! toGo.datum.equals ( o ) ) {
      //        previous = toGo ; toGo = toGo.next ;
      //    }
      //    if ( toGo == null ) { return false ; }
      //    previous.next = toGo.next ;
      //    if ( toGo == tail ) { tail = previous ; }
      //
      //  The following is an implementation of algorithm 2 using the previous reference for everything:
      //
      Node<T> previous = head ;
      while ( ( previous.next != null ) && ! previous.next.datum.equals ( o ) ) {
        previous = previous.next ;
      }
      if ( previous.next == null ) { return false ; }
      toGo = previous.next ;
      previous.next = toGo.next ;
      if ( toGo == tail ) { tail = previous ; }
    }
    toGo.next = null ;
    --elementCount ;
    return true ;
  }
  @Override public final SLList<T> clone ( ) {
    final SLList<T> l = (SLList<T>) super.clone ( ) ;
    //  We are sharing all the Node<T>s at this point which is not what is required.  We can share the data
    //  attached to the Node<T>s but we must make a new linked list.  Can use makeEmpty since it does not
    //  touch the linked list data.
    l.makeEmpty ( ) ;
    for ( final T t : this ) { l.add ( t ) ; }
    return l ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.current = null ;
    return i ;
  }
 @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  Merge the <code>SLList</code> <code>l</code>after a given index in this<code>SLList</code>.  This
   *  moves (not copies) all the <code>Nodes</code> from the <code>SLList</code> that is the parameter into
   *  us.
   *
   *  @param index location in this <code>SLList</code> after which to insert <code>l</code>.
   *  @param l the <code>SLList</code> to insert.
   */
  public final void merge ( final int index , final SLList<T> l ) {
    if ( isEmpty ( ) && ( index == 0 ) ) {
      head = l.head ;
      tail = l.tail ;
    }
    else {
      validityCheck ( index ) ;
      if ( index == elementCount - 1 ) {
        merge ( l ) ;
        return ;
      }
      else {
        final Node<T> n = nodeAt ( index ) ;
        l.tail.next = n.next ;
        n.next = l.head ;
      }
    }
    elementCount += l.elementCount ;
    l.makeEmpty ( ) ;
  }
  /**
   *  Append a <code>SLList</code> to ourself.  This moves (not copies) all the <code>Nodes</code> from the
   *  <code>SLList</code> that is the parameter into us.
   *
   *  @param l the <code>DLList</code> to insert.
   */
  public final void merge ( final SLList<T> l ) {
    if ( isEmpty ( ) ) { head = l.head ; }
    else { tail.next = l.head ; }
    tail = l.tail ;
    elementCount += l.elementCount ;
    l.makeEmpty ( ) ;
  }
  /**
   *  Find a <code>Node</code> given an index.
   *
   *  @param index the index into the <code>SLList</code> to deliver up the <code>Node</code> of.
   *  @return the node being searched for.
   *  @exception IndexOutOfBoundsException if the requested <code>index</code> is not valid due to
   *  the size of the <code>SLList</code>.
   */
  private Node<T> nodeAt ( final int index ) {
    //  A little optimization for access at the tail.
    if ( index == size ( ) - 1 ) { return tail ; }
    //  We know that the index is valid so walk the SLList<T> until we get to the right place and then
    //  return the value at that place.
    Node<T> n = head ;
    for ( int i = 0 ; i < index ; ++i ) { n = n.next ; }
    return n ;
  }
  /**
   *  The iterator over <code>SLList</code>s.  This is an inner class (as opposed to being a nested
   *  top-level class) so that it automatically has a reference to its parent object, i.e. the object it is
   *  an iterator for.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> implements ForwardIterator<T> {
    private Node<T> current = head ;
    @Override public final T get ( ) { validityCheck ( ) ; return current.datum; }
    @Override public final void set ( final T o ) { validityCheck ( ) ; current.datum = o ; }
    @Override public final void advance ( ) { validityCheck ( ) ; current = current.next ; }
    @Override public final void advance ( int increment ) {
      for ( ; increment > 0 ; --increment ) {
        if ( atEnd ( ) ) { break ; }
        advance ( ) ;
      }
    }
    @Override public final boolean atBegin ( ) { return current == head ; }
    @Override public final boolean atEnd ( ) { return current == null ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof SLList.Iterator ) ) { return false ; }
      final Iterator i = (Iterator) o ;
      return atEnd ( ) ? i.atEnd ( ) : current.equals ( i.current ) ;
    }
    @Override public final SLList<T> getContainer ( ) { return SLList.this ; }
    @Override public Iterator clone ( ) { return (Iterator) super.clone ( ) ; }
    /**
     *  Append a new item after the current one.
     *
     *  @param o the object to add.
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    public final void add ( final T o ) {
      validityCheck ( ) ;
      final Node<T> n = new Node<T> ( o , current.next ) ;
      current.next = n ;
      if ( tail == current ) { tail = n ; }
      ++elementCount ;
    }
    /**
     *  Merge an <code>SLList</code> after the current item.
     *
     *  @param l the <code>SLList</code> to merge in at this point in the <code>SLList</code> this iterator
     *  is pointing into.
     *  @exception InvalidIteratorException if the iterator is not valid.
     */
    public final void merge ( final SLList<T> l ) {
      validityCheck ( ) ;
      if ( l.isEmpty ( ) ) { return ; }
      l.tail.next = current.next ;
      current.next = l.head ;
      elementCount += l.elementCount ;
      l.makeEmpty ( ) ;
    }
    /**
     *  Check to see if the iterator is valid.
     *
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    private void validityCheck ( ) {
      if ( current == null ) { throw new InvalidIteratorException ( ) ; }
    }
  }
}
