/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Queue</code> implemented using an associated <code>SLList</code>.
 *
 *  <p> The head (the place from which items will be withdrawn) of the <code>Queue</code> is position 0 in
 *  the <code>SLList</code>, the tail (the place to which new items will be appended) is the last
 *  element.</p>
 *
 *  @author Russel Winder
 */
public class SLListQueue<T> extends AbstractQueue<T> {
  /**
   *  The underlying representation.
   */
  private SLList<T> storage ;
  /**
   *  Constructor.
   */
  public SLListQueue ( ) { this ( new SLList<T> ( ) ) ; }
  /**
   *  Constructor that forces a particular <code>SLList</code>.
   *
   *  @param l the list to use to back the queue.
   */
  public SLListQueue ( final SLList<T> l ) { storage = l ; }
  @Override public final T peek() {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return storage.first ( ) ;
  }
  @Override public final T remove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException () ; }
    final T o = storage.first ( ) ;
    storage.remove ( 0 ) ;
    return o ;
  }
  @Override public final void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public final int size ( ) { return storage.size() ; }
  @Override public final void add ( final T o ) { storage.add(o) ; }
  @Override public final SLListQueue<T> clone ( ) {
    final SLListQueue<T> s = (SLListQueue<T>) super.clone ( ) ;
    s.storage = storage.clone ( ) ;
    return s ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  An iterator over <code>SLListQueue</code>s.
   *
   *  <p>This is just a proxy for an iterator from the underlying representation so as to stop the ability
   *  to write.  Implement only a forward iterator.</p>
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  The iterator from the underlying representation.
     */
    private InputIterator<T> current = storage.iterator ( ) ;
    @Override public final T get ( ) { return current.get ( ) ; }
    @Override public final void advance ( ) { current.advance ( ) ; }
    @Override public final void advance(final int increment) { current.advance ( increment ) ; }
    @Override public final boolean atBegin() { return current.atBegin ( ) ; }
    @Override public final boolean atEnd() { return current.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( this == o ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof SLListQueue.Iterator ) ) { return false ; }
      return current.equals ( ( (Iterator) o ).current ) ;
    }
    @Override public final SLListQueue<T> getContainer ( ) { return SLListQueue.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.current = current.clone ( ) ;
      return i ;
    }
  }
}
