/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Sequence</code> is a <code>Container</code> where there is a total order on the data that is
 *  contained.  Each item in the <code>Sequence</code> has an index defining it's position in the
 *  <code>Sequence</code>.
 *
 *  @author Russel Winder
 */
public interface Sequence<T> extends Container<T> {
  /**
   *  Deliver the first item in this <code>Sequence</code>.
   *
   *  @return the first item in this <code>Sequence</code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>Sequence</code>.
   */
  T first ( ) ;
  /**
   *  Deliver the last item in this <code>Sequence</code>.
   *
   *  @return the last item in this <code>Sequence</code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>Sequence</code>.
   */
  T last ( ) ;
  /**
   *  Get the item at position <code>index</code> from this <codeSequence</code>.
   *
   *  @param index the position in the sequence of the item to be retrieved.
   *  @return the <code>index</code>th item in the <code>Sequence</code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>Sequence</code>.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size of
   *  the <code>Sequence</code>.
   */
  T get ( int index ) ;
  /**
   *  Assign a value to a position in the <code>Sequence</code>.  This overwrites whatever value was there.
   *
   *  @param index the position in the sequence of the item to be replaced.
   *  @param o the object to be used to replace the <code>index</code>th item in the <code>Sequence</code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>Sequence</code>.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size of
   *  the <code>Sequence</code>.
   */
  void set ( int index , T o ) ;
  /**
   *  Excise an element from the <code>Sequence</code>.
   *
   *  @param index the position in the sequence of the item to be removed.
   *  @throws AccessEmptyContainerException if there are no items in the <code>Sequence</code>.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size of
   *  the <code>Sequence</code>.
   */
  void remove ( int index ) ;
  /**
   *  Insert a new item in the <code>Sequence</code>.  Insertion happens before, i.e. the new item becomes
   *  the index'th item with the old index'th item becoming the index+1'th item, etc.
   *
   *  @param index the position in the sequence of the item that is to be after the item to be inserted.
   *  @param o the object to be inserted before the <code>index</code>th item in the <code>Sequence</code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>Sequence</code>.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size of
   *  the <code>Sequence</code>.
   */
  void insert ( int index , T o ) ;
  /**
   *  Find  the position of an item in the <code>Sequence</code>.
   *
   *  @param o the value to be searched for in the <code>Sequence</code>.
   *  @return the "index" of the sought item.  Return <code>-1</code> (which being negative is an illegal
   *  index) if the value <code>o</code> is not part of the <code>Sequence</code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>Sequence</code>.
   */
  int indexOf ( T o ) ;
  /**
   *  Copy a slice of the parameter <code>Sequence</code> into this
   *  <code>Sequence</code> with the slice being defined by indexes.  This is a shallow
   *  operation in that we do not clone the data but just copy references to the data.
   *
   *  @param s the <code>Sequence</code> from which to copy items.
   *  @param b the beginning index of the slice to be copied.
   *  @param e the end index of the slice to be copied.
   *  @exception InvalidIteratorOperationException if the value <code>b</code> is not less than parameter
   *  <code>e</code> or either <code>b</code> or <code>e</code> are not valid indexes.
   */
  void copyIn ( Sequence<T> s , int b , int e ) ;
  /**
   *  Reverse the whole <code>Sequence</code>.
   */
  void reverse ( ) ;
  /**
   *  Reverse a slice of this <code>Sequence</code> defined by two indexes.
   *
   *  @param b the beginning index of the slice to be reversed.
   *  @param e the end index of the slice to be reversed.
   *  @exception InvalidIteratorOperationException if the value <code>b</code> is not less than parameter
   *  <code>e</code> or either <code>b</code> or <code>e</code> are not valid indexes.
   */
  void reverse ( int b , int e ) ;
  /**
   *  Clone this <code>Sequence</code>.
   *
   *  @return a clone of this <code>Sequence</code>.
   */
  @Override Sequence<T> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>Sequence</code> for use in multi-threaded
   *  applications.  This is just a forwarding proxy with synchronization control and locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<E> extends Container.Synchronized<E> implements Sequence<E> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param s the <code>Sequence</code> to synchronize.
     */
    public Synchronized ( final Sequence<E> s ) { super ( s ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param s the <code>Sequence</code> to synchronize.
     *  @param o the lock object.
     */
    public Synchronized ( final Sequence<E> s , final Object o ) { super ( s , o ) ; }
    @Override public final E first ( ) {
      synchronized ( theLock ) { return ( (Sequence<E>) theContainer ).first ( ) ; }
    }
    @Override public final E last ( ) {
      synchronized ( theLock ) { return ( (Sequence<E>) theContainer ).last ( ) ; }
    }
    @Override public final E get ( final int index ) {
      synchronized ( theLock ) { return ( (Sequence<E>) theContainer ).get ( index ) ; }
    }
    @Override public final void set ( final int index , final E o ) {
      synchronized ( theLock ) { ( (Sequence<E>) theContainer ).set ( index , o ) ; }
    }
    @Override public final void remove ( final int index ) {
      synchronized ( theLock ) { ( (Sequence<E>) theContainer ).remove ( index ) ; }
    }
    @Override public final void insert ( final int index , final E o ) {
      synchronized ( theLock ) { ( (Sequence<E>) theContainer ).insert ( index, o ) ; }
    }
    @Override public final int indexOf ( final E o ) {
      synchronized ( theLock ) { return ( (Sequence<E>) theContainer ).indexOf ( o ) ; }
    }
    @Override public final void copyIn ( final Sequence<E> s , final int b , final int e ) {
      synchronized ( theLock ) { ( (Sequence<E>) theContainer ).copyIn ( s , b , e ) ; }
    }
    @Override public final void reverse ( ) {
      synchronized ( theLock ) { ( (Sequence<E>) theContainer ).reverse ( ) ; }
    }
    @Override public final void reverse ( final int b , final int e ){
      synchronized ( theLock ) { ( (Sequence<E>) theContainer ).reverse ( b , e ) ; }
    }
    @Override public final InputIterator<E> begin ( ) {
      synchronized ( theLock ) { return ( (Sequence<E>) theContainer ).begin ( ) ; }
    }
    @Override public final InputIterator<E> end ( ) {
      synchronized ( theLock ) { return ( (Sequence<E>) theContainer ).end ( ) ; }
    }
    @Override public final InputIterator<E> iterator ( ) {
      synchronized ( theLock ) { return ( (Sequence<E>) theContainer ).iterator ( ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  Sequence.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public Sequence.Synchronized<E> clone ( ){ return (Sequence.Synchronized<E>) super.clone ( ) ; }
  }
}
