/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Dequeue</code> using a <code>Sequence</code> as underlying storage mechanism.  The particular
 *  <code>Sequence</code> is chosen at <code>Dequeue</code> creation time, with <code>DLList</code> being
 *  the default.
 *
 *  <p> The head of the <code>Dequeue</code> is position 0 in the <code>Sequence</code>, the tail is the
 *  last element -- use the size() method to discover this: the <code>Sequence</code> is always exactly the
 *  size of the <code>Dequeue</code>.  <code>Sequence</code> itself keeps track of the distinction between
 *  the number of items in the <code>Sequence</code> and the storage used to store it.
 *
 *  Using an <code>DLList</code> addition and removal is O(1) at both ends.  Using an <code>Array</code>
 *  head addition and head removal are O(n) whereas tail addition and tail removal are O(1).  Using an
 *  <code>SLList</code>, head addition and head removal are O(1) as is tail addition, tail removal is O(n).
 *  It seems then that <code>DLList</code> is the only really sensible storage mechanism.
 *
 *  @author Russel Winder
 */
public class SequenceDequeue<T> extends AbstractDequeue<T> {
  /**
   *  The underlying representation.
   */
  private Sequence<T> storage ;
  /**
   *  Default constructor.
   *
   *  <p>Use a <code>DLList</code> as the storage medium as it is the most flexible <code>Sequence</code>
   *  that we have.</p>
   */
  public SequenceDequeue ( ) { storage = new DLList<T> ( ) ; }
  /**
   *  Constructor that uses the <code>Sequence</code> given as parameter as the data storage.
   *
   *  @param s the sequence to use to back the dequeue.
   */
  public SequenceDequeue ( final Sequence<T> s ) { storage = s ; }
  @Override public final void headAdd ( final T o ) { storage.insert ( 0 , o ) ; }
  @Override public final void tailAdd ( final T o ) { storage.add ( o ) ; }
  @Override public final T headRemove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    final T o = storage.get ( 0 ) ;
    storage.remove ( 0 ) ;
    return o ;
  }
  @Override public final T tailRemove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    final int i = storage.size ( ) - 1 ;
    final T o = storage.get ( i ) ;
    storage.remove ( i ) ;
    return o ;
  }
  @Override public final T headPeek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
   return storage.get ( 0 ) ;
  }
 @Override  public final T tailPeek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return storage.get ( storage.size ( ) - 1 ) ;
  }
  @Override public final void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public final  int size ( ) { return storage.size ( ) ; }
  @Override public final SequenceDequeue<T> clone ( ) {
    final SequenceDequeue<T> d = (SequenceDequeue<T>) super.clone ( ) ;
    d.storage = storage.clone ( ) ;
    return d ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator  ( ) ; }
  /**
   *  An iterator over <code>SequenceQueue</code>s.
   *
   *  This is just a proxy for an iterator from the underlying representation so as to stop the ability to
   *  write.  Implement only a forward iterator.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  We just need to act as a proxy for an iterator from the underlying representation.
     */
    private InputIterator<T> current = storage.iterator ( ) ;
    @Override public final T get ( ) { return current.get ( ) ; }
    @Override public final void advance ( ) { current.advance ( ) ; }
    @Override public void advance ( final int increment ) { current.advance ( increment ) ; }
    @Override public boolean atBegin ( ) { return current.atBegin ( ) ; }
    @Override public final boolean atEnd ( ) { return current.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof SequenceDequeue.Iterator ) ) { return false ; }
      return current.equals ( ( (Iterator) o ).current ) ;
    }
    @Override public final SequenceDequeue<T> getContainer ( ) { return SequenceDequeue.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.current = current.clone ( ) ;
      return i ;
    }
  }
}
