/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>PriorityQueue&ltT></code> implemented as a<code>Sequence&ltT></code> of
 *  <code>Queue&ltT></code>s.  This implementation allows the priority management strategy to be determined
 *  by the application.
 *
 *  <p> Priorities are integers but since the priority management strategy is application determined we have
 *  very little to do.  It is assumed that smaller numbers are higher priority.  Default insert is at lowest
 *  priority and default remove is at the highest priority.</p>
 *
 *  <p> The application is expected to supply the minimum number of different priority level at creation
 *  time.  We allow a default of 10 though.  New lower priority levels can be added as required.</p>
 *
 *  @author Russel Winder
 */
public class SequenceQueuePriorityQueue<T> extends AbstractPriorityQueue<T> {
  /**
   *  The underlying representation ia an array of queues.
   */
  private Array<Queue<T>> storage ;
  /**
   *  The constructor defaulting the number of different priority levels to 10.
   *
   *  @exception AccessEmptyContainerException if an attempt is made to start with 0 or less priority
   *  levels.
   */
  public SequenceQueuePriorityQueue ( )  { this ( 10 ) ; }
  /**
   *  The constructor specifying the number of different priority levels.
   *
   *  @param n The number of different priorities.
   */
  public SequenceQueuePriorityQueue ( final int n ) {
    storage = new Array<Queue<T>> ( n ) ;
    for ( int i = 0 ; i < n ; ++i ) { storage.add ( new SLListQueue<T>  ( ) ) ; }
  }
  /**
   *  Add one new priority level.
   */
  public final void addPriorityLevel ( )  { storage.add ( new SLListQueue<T>  ( ) ) ; }
  @Override public final void add ( final T o ) { add ( o , storage.size ( )  - 1 ) ; }
  @Override public final void add ( final T o , int priority ) {
    priority = ( priority < 0 ) ? 0 : ( priority >= storage.size ( ) ) ? storage.size ( )  -1 : priority ;
    storage.get ( priority ).add ( o ) ;
  }
  @Override public final T peek ( )  {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    for ( int i = 0 ; i < storage.size ( )  ; ++i ) {
      final Queue<T> q = storage.get ( i ) ;
      if ( ! q.isEmpty ( ) ) { return q.peek ( )  ; }
    }
    return null ;
  }
  @Override public final T remove ( )  {
    for ( int i = 0 ; i < storage.size ( )  ; ++i ) {
      final Queue<T> q = storage.get ( i ) ;
      if ( ! q.isEmpty ( ) ) { return q.remove ( )  ; }
    }
    return null ;
  }
  @Override public final T remove ( final int priority ) {
    final Queue<T> q = storage.get ( priority ) ;
    return q.isEmpty ( )  ? null : q.remove ( )  ;
  }
  @Override public final void makeEmpty ( )  {
    for ( int i = 0 ; i < storage.size ( )  ; ++i ) {
      storage.get ( i ).makeEmpty ( )  ;
    }
  }
  @Override public final boolean isEmpty ( )  {
    for ( int i = 0 ; i < storage.size ( )  ; ++i ) {
      if ( ! storage.get ( i ).isEmpty ( )  ) { return false ; }
    }
    return true ;
  }
  public final boolean isEmpty(final int priority) { return storage.get ( priority ).isEmpty ( )  ; }
  @Override public final int size ( )  {
    int count = 0 ;
    for ( int i = 0 ; i < storage.size ( )  ; ++i ) { count += storage.get ( i ).size ( )  ; }
    return count ;
  }
  public final int size ( final int priority ) { return storage.get ( priority ).size ( )  ; }
  @Override public final SequenceQueuePriorityQueue<T> clone ( )  {
    final SequenceQueuePriorityQueue<T> p = (SequenceQueuePriorityQueue<T>) super.clone ( )  ;
    for ( int i = 0 ; i < storage.size ( )  ; ++i ) { p.storage.set ( i , storage.get ( i ).clone ( ) ) ; }
    return p ;
  }
  @Override public final String toString ( )  {
    final StringBuilder sb = new StringBuilder ( )  ;
    sb.append ( "<<\n" ) ;
    for ( int i = 0 ; i < storage.size ( )  ; ++i ) {
      sb.append ( '\t' ) ;
      sb.append ( storage.get ( i ).toString ( ) ) ;
      sb.append ( '\n' ) ;
    }
    sb.append ( ">>" ) ;
    return sb.toString ( )  ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator ( )  { return new Iterator  ( )  ; }
  /**
   *  An iterator over <code>SequenceQueuePriorityQueue</code>s.
   *
   *  <p>This iterator assumes that no insertions or removals to the queue will be made during the
   *  traversal.  If items of lower priority than the iterator currently references are inserted or removed
   *  during the traversal then things should be OK.  However, any activity at priorities higher than the
   *  priority of the currently referred to item will not be observed by the iterator.</p>
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    //  Represent this iterator by an iterator over the contained queues and an iterator into a given queue.
    /**
     *  The iterator over the contained <code>Queue&ltT></code>s.
     */
    private InputIterator<Queue<T>> currentQueue = storage.iterator ( )  ;
    /**
     *  The iterator over the referred to <code>Queue&ltT></code>.
     */
    private InputIterator<T> currentItem ; {
      while ( ! currentQueue.atEnd ( ) && currentQueue.get ( ).isEmpty ( ) ) { currentQueue.advance ( ) ; }
      if ( ! currentQueue.atEnd ( )  ) { currentItem = currentQueue.get ( ) .iterator ( )  ; }
    }
    @Override public final T get ( )  {
      if ( atEnd ( )  ) { throw new InvalidIteratorException  ( )  ; }
      return currentItem.get ( )  ;
    }
    @Override public final void advance ( )  {
      currentItem.advance ( )  ;
      if ( currentItem.atEnd ( )  ) {
        currentQueue.advance ( )  ;
        if ( ! currentQueue.atEnd ( )  ) {
          currentItem = currentQueue.get ( ) .iterator ( )  ;
        }
      }
    }
    @Override public void advance ( int increment ) {
      for ( ; increment > 0 ; --increment ) {
        if ( atEnd ( )  ) { break ; }
        advance ( )  ;
      }
    }
    @Override public boolean atBegin ( )  { return currentQueue.atBegin ( )  && currentItem.atBegin ( )  ; }
    @Override public final boolean atEnd ( )  { return currentQueue.atEnd ( )  && currentItem.atEnd ( )  ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof SequenceQueuePriorityQueue.Iterator ) ) { return false ; }
      final Iterator i = (Iterator) o ;
      return currentItem.equals ( i.currentItem ) && currentQueue.equals ( i.currentQueue ) ;
    }
    @Override public final SequenceQueuePriorityQueue<T> getContainer ( )  { return SequenceQueuePriorityQueue.this ; }
    @Override public final Iterator clone ( )  {
      final Iterator i = (Iterator) super.clone ( )  ;
      i.currentQueue = currentQueue.clone ( )  ;
      i.currentItem = currentItem.clone ( )  ;
      return i ;
    }
  }
}
