/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Stack</code> using a <code>Sequence</code> as underlying storage mechanism. The particular
 *  <code>Sequence</code> is chosen at <code>Stack</code> creation time, with <code>DLList</code> being the
 *  default.
 *
 *  <p> The top of stack is the element at the end of the <code>Sequence</code>, i.e. largest index entry is
 *  the top of stack.  For an <code>DLList</code> it makes no odds whether the beginning or the end is used
 *  for the top of stack since addition and removal are O(1) at both ends.  For an <code>Array</code> we
 *  must use the end to get O(1) behaviour since to use the beginning would make addition and removal O(n)
 *  operations (all the data would have to be moved by one for every push or pop).  For an
 *  <code>SLList</code>, because we have used a tail pointer in the implementation, addition at either the
 *  beginning or the end is O(1) but removal is O(1) at the beginning and O(n) at the end.  Thus there is a
 *  conflict of performance between <code>Array</code> and <code>SLList</code> whether the top of stack is
 *  the beginning or the end.  We make the decision to support <code>DLList</code> and <code>Array</code> as
 *  the usual storage mechanisms and hence top of stack is at the end.</p>
 *
 *  @author Russel Winder
 */
public class SequenceStack<T> extends AbstractStack<T> {
  /**
   *  The data store.
   */
  private Sequence<T> storage ;
  /**
   *  Default constructor.
   *
   *  <p>Use a <code>DLList</code> as the storage medium as it is the most flexible <code>Sequence</code>
   *  that we have.</p>
   */
  public SequenceStack ( ) { storage = new DLList<T> ( ) ; }
  /**
   *  Constructor that uses the <code>Sequence</code> given as parameter as the data storage.
   *
   *  @param s the sequence used to back the stack.
   */
  public SequenceStack ( final Sequence<T> s ) { storage = s ; }
  @Override public final T remove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    final T o = storage.last ( ) ;
    storage.remove ( storage.size ( ) - 1 ) ;
    return o ;
  }
  @Override public final T peek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return storage.last ( ) ;
  }
  @Override public final void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public final int size ( ) { return storage.size() ; }
  @Override public final void add ( final T o ) { storage.add ( o ) ; }
  @Override public final SequenceStack<T> clone ( ) {
    final SequenceStack<T> s = (SequenceStack<T>) super.clone ( ) ;
    s.storage = storage.clone ( ) ;
    return s ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  An iterator over <code>SequenceStack</code>s.
   *
   *  <p>This is just a proxy for an iterator from the underlying representation so as to stop the ability
   *  to write.  Implement only an input iterator.</p>
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  The cursor is just an iterator.
     */
    private InputIterator<T> current = storage.iterator ( ) ;
    @Override public final T get ( ) { return current.get ( ) ; }
    @Override public final void advance ( ) { current.advance ( ) ; }
    @Override public void advance( final int increment ) { current.advance(increment) ; }
    @Override public boolean atBegin ( ) { return current.atBegin ( ) ; }
    @Override public final boolean atEnd ( ) { return current.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof SequenceStack.Iterator ) ) { return false ; }
      return current.equals ( ( (Iterator) o ).current ) ;
    }
    @Override public final SequenceStack<T> getContainer ( ) { return SequenceStack.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.current = current.clone ( ) ;
      return i ;
    }
  }
}
