/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  The interface defining a set.
 *
 *  @author Russel Winder
 */
public interface Set<T> extends Container<T> {
  /**
   *  Deliver the set union of this <code>Set</code> and the parameter <code>Set</code>.
   *
   *  @param s a set
   *  @return the set union of this <code>Set</code> and the parameter <code>Set</code>.
   */
  Set<T> union ( Set<T> s ) ;
  /**
   *  Deliver the set intersection of this <code>Set</code> and the parameter <code>Set</code>.
   *
   *  @param s a set
   *  @return the set intersection of this <code>Set</code> and the parameter <code>Set</code>.
   */
  Set<T> intersection ( Set<T> s ) ;
  /**
   *  Deliver the set difference of this <code>Set</code> and the parameter <code>Set</code>.
   *
   *  @param s a set
   *  @return the set difference of this <code>Set</code> and the parameter <code>Set</code>.
   */
  Set<T> difference ( Set<T> s ) ;
  /**
   *  Clone this <code>Set</code>.
   *
   *  @return a clone of this <code>Set</code>.
   */
  @Override Set<T> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>Set</code> for use in multi-threaded
   *  applications.  This is just a forwarding proxy with synchronization control and locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<E> extends Container.Synchronized<E> implements Set<E> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param s the set to synchronize.
     */
    public Synchronized ( final Set<E> s ) { super ( s ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param s the set to synchronize.
     *  @param o the object to use as the lock.
     */
    public Synchronized ( final Set<E> s , final Object o ) { super ( s , o ) ; }
    @Override public final Set<E> union ( final Set<E> s ) {
      synchronized ( theLock ) { return ( (Set<E>) theContainer ).union ( s ) ; }
    }
    @Override public final Set<E> intersection ( final Set<E> s ) {
      synchronized ( theLock ) { return ( (Set<E>) theContainer ).intersection ( s ) ; }
    }
    @Override public final Set<E> difference ( final Set<E> s ) {
      synchronized ( theLock ) { return ( (Set<E>) theContainer ).difference ( s ) ; }
    }
    //  Without qualification on Synchronized, the compiler appears not to be able to distinguish
    //  Set.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public Set.Synchronized<E> clone ( ) { return (Set.Synchronized<E>) super.clone ( ) ; }
  }
}
