/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  This class encapsulates a doubly-linked list which orders entries according to a key field.
 *
 *  <p> The list is always in sorted order, adding an item causes the list to shuffle the new item to the
 *  correct place.  The application must provide a <code>Comparator</code> to handle the ordering.</p>
 *
 *  <p> The semantics of the insertion functions have been changed over those of the superclass
 *  <code>DLList</code>.  In particular, attempting to insert at a given point in the list will not succeed,
 *  insertion is always by appending at the end and then percolating up, no matter what the interface is!
 *  Some might consider this bad form.</p>
 *
 *  @author Russel Winder
 */
public class SortedList<T> extends DLList<T> {
  /**
   *  The <code>Comparator</code> that defines the sort order.
   */
  private Comparator<? super T> order ;
  /**
   *  The constructor forces the application to define a <code>Comparator</code>
   *
   *  @param c the comparator used to provide the order.
   */
  public SortedList ( final Comparator<? super T> c ) { order = c ; }
  /**
   *  Assign a value to a position in the <code>SortedList</code>.  Actually does not do this at all.  It
   *  simply appends at the end and lets the percolation find the right spot.
   *
   *  @param index this parameter is ignored.
   *  @param o the item to add.
   */
  @Override public final void set ( final int index , final T o ) { add ( o ) ; }
  /**
   *  Insert a new item in the <code>SortedList</code>.  Actually does not do this at all.  It simply
   *  appends at the end and lets the percolation find the right spot.
   *
   *  @param index this parameter is ignored.
   *  @param o the item to add.
   */
  public final void add ( final int index , final T o ) { add ( o ) ; }
  /**
   *  Add a new item to the <code>SortedList</code>.
   *
   *  @param o the item to add.
   */
  @Override public final void add ( final T o ) {
    //  Insert the new item: Because we are using a Node as the end point the algorithms is always the same,
    //  we never worry about an empty list!
    final Node<T> n = new Node<T> ( o , endNode , endNode.previous ) ;
    endNode.previous.next = n ;
    endNode.previous = n ;
    ++elementCount ;
    //  Now we must swap this element we have just inserted with the one before it in the list until either
    //  it is at the end of the sub-list it is supposed to arrive at according to the Comparator<? super T>
    //  or until it reaches the head of the entire list.  We have reached the head of the queue if previous
    //  refers to the dummy node used to hold the list together.
    while ( n.previous != endNode ) {
      if ( order.relation ( n.datum , n.previous.datum ) ) {
        //  The sort order was violated so we need to edit.  Because the list is doubly linked we can do all
        //  the necessary editing from the node that needs to be moved one space forward.
        n.next.previous = n.previous ;
        n.previous.next = n.next ;
        n.next = n.previous ;
        n.previous = n.previous.previous ;
        n.next.previous = n ;
        n.previous.next = n ;
      } else {
        //  The order is now as we need it so we have finished.
        break ;
      }
    }
  }
  /**
   *  Redefinition of reverse to throw an exception since reversing is not allowed on
   *  <code>SortedList</code>s.
   *
   *  @exception InvalidOperationException always thrown.
   */
  @Override public void reverse ( ) { throw new InvalidOperationException ( ) ; }
  /**
   *  Redefinition of reverse to throw an exception since reversing is not allowed on
   *  <code>SortedList</code>s.
   *
   *  @param b the beginning of the slice to reverse.
   *  @param e the end of the slice to reverse.
   *  @exception InvalidOperationException always thrown.
   */
  @Override public void reverse ( final int b , final int e ) { throw new InvalidOperationException ( ) ; }
}
