/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>PriorityQueue</code> represented as a <code>SortedList</code> of <code>Pair&lt;T,
 *  Integer></code>s (datum, priority).
 *
 *  <p>Priorities are integers.  It is assumed that smaller numbers are higher priority.  Default insert is
 *  at priority 10.  Removal always takes the highest priority item. Because priorities are integers the
 *  application has no need to determine the number of different priorities at creation time.</p>
 *
 *  @author Russel Winder
 */
public class SortedListPriorityQueue<T> extends AbstractPriorityQueue<T> {
  /**
   *  The <code>SortedList</code> which is the representation.
   */
  //  We sort on the priority, which is the second item of the Pair which is an Integer, into the order
  //  where the lower the priority is the closer the item is to the head of the queue.
  private SortedList<Pair<T, Integer>> storage =
   new SortedList<Pair<T, Integer>> ( new PairBComparator<T, Integer, Pair<T, Integer>> ( new LessThanComparator<Integer> ( ) ) ) ;
  @Override public final void add ( final T o ) { add ( o , 10 ) ; }
  @Override public final void add ( final T o, final int priority ) { storage.add ( new Pair<T, Integer> ( o , new Integer ( priority ) ) ) ; }
  @Override public final T peek ( ) { return storage.first ( ).getA ( ) ; }
  @Override public final T remove ( ) {
    final Pair<T, Integer> p = storage.first ( ) ;
    storage.remove ( 0 ) ;
    return p.getA ( ) ;
  }
  @Override public final T remove ( final int priority ) { throw new InvalidOperationException ( ) ; }
  @Override public final void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public final boolean isEmpty ( ) { return storage.isEmpty ( ) ; }
  @Override public final int size ( ) { return storage.size ( ) ; }
  @Override public final SortedListPriorityQueue<T> clone ( ) {
    final SortedListPriorityQueue<T> p = (SortedListPriorityQueue<T>) super.clone ( ) ;
    p.storage = (SortedList<Pair<T, Integer>>) storage.clone ( ) ;
    return p ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final InputIterator<T> iterator ( ) { return new Iterator ( ) ; }
  /**
   *  An iterator over <code>SequenceQueuePriorityQueue</code>s.
   *
   *  <p>This is just a proxy for an iterator from the underlying representation so as to stop the ability
   *  to write.  Implement only an input iterator.</p>
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  The iterator cursor.
     */
    private InputIterator<Pair<T, Integer>> currentItem = storage.iterator ( ) ;
    @Override public final T get ( ) {
      if ( atEnd ( ) ) { throw new InvalidIteratorException  ( ) ; }
      return currentItem.get ( ).getA ( ) ;
    }
    @Override public final void advance ( ) { currentItem.advance ( ) ; }
    @Override public final void advance ( final int increment ) { currentItem.advance(increment) ; }
    @Override public final boolean atBegin ( ) { return currentItem.atBegin ( ) ; }
    @Override public final boolean atEnd ( ) { return currentItem.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof SortedListPriorityQueue.Iterator ) ) { return false ; }
      return currentItem.equals ( ( (Iterator) o ).currentItem ) ;
    }
    @Override public final SortedListPriorityQueue<T> getContainer ( ) { return SortedListPriorityQueue.this ; }
    @Override public final Iterator clone ( ) {
      final Iterator i = (Iterator) super.clone ( ) ;
      i.currentItem = currentItem.clone ( ) ;
      return i ;
    }
  }
}
