/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  This class implements a sparse array abstraction using a <code>Mapping</code>.
 *
 *  <p>The default representation is a <code>ChainedHashTableMapping&lt;Integer, T></code> but other options
 *  are <code>OrderedBinaryTreeMapping&lt;Integer, T></code>) or <code>OpenHashTableMapping&lt;Integer,
 *  T></code>.</p>
 *
 *  @author Russel Winder
 */
public class SparseArray<T> extends AbstractSequence<T>  {
  /**
   *  The storage medium.
   */
  protected Mapping<Integer, T> storage ;
  /**
   *  The default constructor.  Use a <code>ChainedHashHashMapping&lt;Integer, T></code> as representation.
   */
  public SparseArray ( ) { storage = new ChainedHashTableMapping<Integer, T> ( ) ; }
  /**
   *  The constructor which allows the representation to be specified explicitly.
   *
   *  @param m the mapping to use for the sparse array.
   */
  public SparseArray ( final Mapping<Integer, T> m ) { storage = m ; }
  @Override public int size ( ) { return storage.size ( ) ; }
  @Override public void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public boolean isEmpty ( ) { return storage.isEmpty ( ) ; }
  @SuppressWarnings ( "unchecked" )
  @Override public boolean equals ( final Object o ) {
    if ( o == this ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof SparseArray ) ) { return false ; }
    return storage.equals ( ( (SparseArray<T>) o ).storage ) ;
  }
  @Override public SparseArray<T> clone ( ) {
    final SparseArray<T> s = (SparseArray<T>) super.clone ( ) ;
    s.storage = storage.clone ( ) ;
    return s ;
  }
  @Override public Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  The iterator that we deliver up.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> {
    /**
     *  The iterator on a <code>Mapping&lt;Integer, T></code> that we employ to do the work.
     */
    private InputIterator<Pair<Integer, T>> cursor = storage.iterator ( ) ;
    @Override public T get ( ) { return cursor.get ( ).getB ( ) ; }
    @Override public void advance ( ) { cursor.advance ( ) ; }
    @Override public void advance ( final int increment ) { cursor.advance ( increment ) ; }
    @Override public boolean atBegin ( ) { return cursor.atBegin ( ) ; }
    @Override public boolean atEnd ( ) { return cursor.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof SparseArray.Iterator ) ) { return false ; }
      return cursor.equals ( ( (Iterator) o ).cursor ) ;
    }
    @Override public SparseArray<T> getContainer ( ) { return SparseArray.this ; }
    @Override public final Iterator clone ( ) {
    	final Iterator i = (Iterator) super.clone ( ) ;
    	i.cursor = cursor.clone ( ) ;
    	return i ;
    }
  }
  @Override public boolean contains ( final T o ) {
    for ( InputIterator<T> i = storage.range ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
      if ( i.get ( ).equals ( o ) ) { return true ; }
    }
    return false ;
  }
  @Override public int indexOf ( final T o ) {
    final Integer index = storage.domain ( o ).get ( ) ;
    if ( index == null ) { return -1 ; }
    return index ;
  }
  @Override public T get ( final int index ) { return storage.getValue ( index ) ; }
  @Override public T last ( ) { return get ( upperBound ( ) ) ; }
  /**
   *  Assign a value to a position in the <code>Sequence&lt;T></code>.  This overwrites whatever value was
   *  there.  Checks for a sensible index, throws an exception if this is not the case.
   *
   *  <p>NB Unlike primitive arrays and <code>Array&lt;T></code> there is no test for an upper bound, we can
   *  set any index we like.
   *
   *  @exception IndexOutOfBoundsException if the requested <code>index</code> is negative.
   */
  @Override public void set ( final int index , final T o ) {
    if ( index < 0 ) { throw new IndexOutOfBoundsException ( ) ; }
    storage.add ( index , o ) ;
  }
  @Override public void remove ( final int index ) {
    if ( index < 0 ) { throw new IndexOutOfBoundsException ( ) ; }
    final T o = storage.getValue ( index ) ;
    if ( o != null ) { storage.remove ( index , o ) ; }
  }
  @Override public void insert ( final int index , final T o ) {
    for ( InputIterator<Pair<Integer, T>> i = storage.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
      final Pair<Integer, T> p = i.get ( ) ;
      final int pIndex = p.getA ( ) ;
      if ( pIndex >= index ) {
        //  Removal here is horrendously inefficient, O(log(n)) instead of O(1), because there is no
        //  non-searching editing iterator defined on <code>Mapping</code>.  Clearly this needs to be added
        //  and used here.
        storage.remove ( p ) ;
        storage.add ( new Pair<Integer, T> ( pIndex + 1 , p.getB ( ) ) ) ;
      }
    }
    set ( index , o ) ;
  }
  @Override public void add ( final T o ) {
    int index = 0 ;
    if ( ! isEmpty ( ) ) { index = upperBound ( ) + 1 ; }
    storage.add ( index , o ) ;
  }
  @Override public boolean remove ( final T o ) {
    for ( InputIterator<Pair<Integer, T>> i = storage.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
      final Pair<Integer, T> p = i.get ( ) ;
      final int index = p.getA ( ) ;
      final T item = p.getB ( ) ;
      if ( item.equals ( o ) ) {
        storage.remove ( index , item ) ;
        return true ;
      }
    }
    return false ;
  }
  @Override public String toString ( ) { return storage.toString ( ) ; }
  /**
   *  A support function for finding the largest index used.  Returns -1 if the
   *  <code>SparseArray&lt;T></code> is empty.
   *
   *  @return the upper bound.
   */
  private int upperBound ( ) {
    if ( isEmpty ( ) ) { return -1 ; }
    int index = 0 ;
    for ( InputIterator<Integer> i = storage.domain ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
      final int key = i.get ( ) ;
      if ( index < key ) { index = key ; }
    }
    return index ;
  }
}
