/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A <code>Stack</code> &mdash; a container that imposes a first-in, last-out access policy.
 *
 *  @see Sequence
 *  @author Russel Winder
 */
public interface Stack<T> extends Container<T> {
  /**
   *  Remove the item from the top of this <code>Stack</code> and deliver it as the return value.
   *
   *  @return the value from the top of this <code>Stack</code>.
   *  @exception AccessEmptyContainerException when any attempt is made to remove data from an empty
   *  <code>Stack</code>.
   */
  T remove ( ) ;
  //
  //  Traditionally, stacks have been used with terms like push (for insert), pop (for remove) and top (for
  //  looking at the element on top of the stack) so we provide these in addition to add and remove.
  //
  /**
   *  Push an element onto the <code>Stack</code>.
   *
   *  @param o the item to be pushed onto this <code>Stack</code>.
   */
  void push ( T o ) ;
  /**
   *  Pop an element of the <code>Stack</code>.
   *
   *  @return the value from the top of this <code>Stack</code>.
   *  @exception AccessEmptyContainerException when any attempt is made to remove data from an empty
   *  <code>Stack</code>.
   */
  T pop ( ) ;
  /**
   *  Discover the value of the item on top of the <code>Stack</code>.
   *
   *  @return the value from the top of this <code>Stack</code>.
   *  @exception AccessEmptyContainerException when any attempt is made to remove data from an empty
   *  <code>Stack</code>.
   */
  T peek ( ) ;
  /**
   *  Discover the value of the item on top of the <code>Stack</code>.
   *
   *  @return the value from the top of this <code>Stack</code>.
   *  @exception AccessEmptyContainerException when any attempt is made to remove data from an empty
   *  <code>Stack</code>.
   */
  T top ( ) ;
  /**
   *  Clone this <code>Stack</code>.
   *
   *  @return a clone of this <code>Stack</code>.
   */
  @Override Stack<T> clone ( ) ;
  /**
   *  The class that provides a synchronization adapter to a <code>Stack</code> for use in multi-threaded
   *  applications.  This is just a forwarding proxy with synchronization control and locking.
   *
   *  @author Russel Winder
   */
  class Synchronized<E> extends Container.Synchronized<E> implements Stack<E> {
    /**
     *  The constructor that uses the proxy itself as the lock object.
     *
     *  @param s the stack to synchronize.
     */
    public Synchronized ( final Stack<E> s ) { super ( s ) ; }
    /**
     *  The constructor that specifies a lock object.
     *
     *  @param s the stack to synchronize.
     *  @param o teh object to use for the lock.
     */
    public Synchronized ( final Stack<E> s , final Object o ) { super ( s , o ) ; }
    @Override public final E remove ( ) {
      synchronized ( theLock ) { return ( (Stack<E>) theContainer ).remove ( ) ; }
    }
    @Override public final void push ( final E o ) {
      synchronized ( theLock ) { ( (Stack<E>) theContainer ).push ( o ) ; }
    }
    @Override public final E pop ( ) {
      synchronized ( theLock ) { return ( (Stack<E>) theContainer ).pop ( ) ; }
    }
    @Override public final E peek ( ) {
      synchronized ( theLock ) { return ( (Stack<E>) theContainer ).peek ( ) ; }
    }
    @Override public final E top ( ) {
      synchronized ( theLock ) { return ( (Stack<E>) theContainer ).top ( ) ; }
    }
    //  Without qualification on Synchronized the compiler appears not to be able to distinguish
    //  Stack.Synchronized and Container.Synchronized so this is a lookup conflict.
    @Override public final Stack.Synchronized<E> clone ( ) { return (Stack.Synchronized<E>) super.clone ( ) ; }
  }
}
