/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2011 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.alt ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.ComparatorRelation ;
/**
 *  A concrete class to represent an order relation.  This is a single class realization.
 *
 *  @author Russel Winder
 */
public class ConcreteComparator<T extends Comparable<T>> implements Comparator<T> {
  /**
   *  The relation that this <code>ConcreteComparator&lt;T></code> represents is determined by this stored
   *  datum.  This variable is final so that it can only be set by initialization.  Since there is no
   *  initializer here, it must be assigned in the constructor.  NB <code>ConcreteComparator&lt;T></code>s
   *  are immutable.
   */
  private final ComparatorRelation theRelation ;
  /**
   *  @param r A <code>ComparatorRelation</code> representing the desired relation.
   */
  public ConcreteComparator ( final ComparatorRelation r ) { theRelation = r ; }
  /**
   *  The relation that this <code>ConcreteComparator&lt;T></code> represents.  Compares the two parameters
   *  according to the relation.
   */
  @Override public final boolean relation ( final T a , final T b ) {
    switch ( theRelation ) {
    case LESS_THAN : return isLessThan ( a , b ) ;
    case LESS_THAN_OR_EQUAL_TO : return isLessThanOrEqualTo ( a , b ) ;
    case EQUAL_TO : return isEqualTo ( a , b ) ;
    case NOT_EQUAL_TO : return isNotEqualTo ( a, b ) ;
    case GREATER_THAN : return isGreaterThan ( a , b ) ;
    case GREATER_THAN_OR_EQUAL_TO : return isGreaterThanOrEqualTo  ( a , b ) ;
    }
    //  The compiler does not recognize that we have enumerated all possible options and says we have a
    //  missing return statement.
    throw new InternalError ( ) ;
  }
  public static <E extends Comparable<E>> boolean isLessThan ( final E a , final E b ) { return a.compareTo ( b ) < 0 ; }
  public static <E extends Comparable<E>> boolean isLessThanOrEqualTo ( final E a , final E b ) { return a.compareTo ( b ) <= 0 ; }
  public static <E extends Comparable<E>> boolean isEqualTo ( final E a , final E b ) { return a.equals ( b ) ; }
  public static <E extends Comparable<E>> boolean isNotEqualTo ( final E a , final E b ) { return ! a.equals ( b ) ; }
  public static <E extends Comparable<E>> boolean isGreaterThan ( final E a , final E b ) { return a.compareTo ( b ) > 0 ; }
  public static <E extends Comparable<E>> boolean isGreaterThanOrEqualTo ( final E a , final E b ) { return a.compareTo ( b ) >= 0 ; }
  /**
   *  The equals method to override that inherited from <code>Object</code> in order to ensure that equality
   *  is value equality rather than object equality.
   *
   *  @param o The object to compare ourself with.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public final boolean equals ( final Object o ) {
    if ( o == this ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof ConcreteComparator ) ) { return false ; }
    return theRelation == ( (ConcreteComparator<T>) o ).theRelation ;
  }
}
