/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2011 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.alt ;
import org.devjavasoft.ads.AccessEmptyContainerException ;
import org.devjavasoft.ads.DLList ;
import org.devjavasoft.ads.Sequence ;
/**
 *  An implementation of a <code>Dequeue&lt;T></code> (not to be confused with <code>Deque&ltT></code> in
 *  the <small>JGL</small>) which is a double-ended queue.
 *
 *  @see org.devjavasoft.ads.Container
 *  @see Sequence
 *  @author Russel Winder
 */
public class Dequeue<T> extends Queue<T> {
  /**
   *  The default constructor.  Use a <code>DLList&lt;T></code> as the default storage implementation since
   *  this give O(1) operations (both insert and remove) at both ends of the <code>Dequeue&lt;T></code>
   */
  public Dequeue ( ) { this ( new DLList<T> ( ) ) ; }
  /**
   *  The constructor that allows specific allocation of storage.  Must be a <code>Sequence&lt;T></code> of
   *  some sort.
   *
   *  <p> We know that the superclass (<code>Queue&lt;T></code>) constructor will make a shallow copy of the
   *  preferred container so as to avoid potential sharing problems.
   *
   *  @param sequence the <code>Sequence&lt;T></code> that is to be used as storage.
   */
  public Dequeue ( final Sequence<T> sequence ) { super ( sequence ) ; }
  /**
   *  Insert a new item at our head.
   *
   *  @param o Object to add.
   */
  public void headAdd ( final T o ) { storage.insert ( 0 , o ) ; }
  /**
   *  Insert a new item at our tail.
   *
   *  @param o Object to add.
   */
  public void tailAdd(final T o) { add ( o ) ; }
  /**
   *  Remove an item from our head.
   *
   *  @return A reference to the removed object.
   *  @exception AccessEmptyContainerException if we attempt to remove from an empty
   *  <code>Dequeue&lt;T></code>.
   */
  public T headRemove ( ) { return remove ( ) ; }
  /**
   *  Remove an item from the our tail.
   *
   *  @return A reference to the removed object.
   *  @exception AccessEmptyContainerException if we attempt to remove from an empty
   *  <code>Dequeue&lt;T></code>.
   */
  public T tailRemove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    else {
      final int i = storage.size ( ) - 1 ;
      final T o = storage.get ( i ) ;
      storage.remove ( i ) ;
      return o ;
    }
  }
  /**
   *  Peek at the head of the <code>Dequeue&lt;T></code>.
   *
   *  @exception AccessEmptyContainerException if we attempt to remove from an empty
   *  <code>Dequeue&lt;T></code>.
   */
  @Override public final T peek ( ) { return headPeek ( ) ; }
  /**
   *  Peek at the head of the <code>Dequeue</code>.
   *
   *  @return reference to the head object.
   *  @exception AccessEmptyContainerException if we attempt to remove from an empty
   *  <code>Dequeue&lt;T></code>.
   */
  public final T headPeek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return storage.get ( 0 ) ;
  }
  /**
   *  Peek at the tail of the <code>Dequeue&lt;T></code>.
   *
   *  @return reference to the tail object.
   *  @exception AccessEmptyContainerException if we attempt to remove from an empty
   *  <code>Dequeue&lt;T></code>.
   */
  public final T tailPeek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return storage.get ( storage.size ( )-1 ) ;
  }
  @SuppressWarnings ( "unchecked" )
  @Override public boolean equals ( final Object o ) {
    if ( this == o ) { return true ; }
    if ( ( o == null ) || ! ( o instanceof Dequeue ) || ( size ( ) != ( (Dequeue<T>) o ).size ( ) ) ) {
      return false ;
    }
    return super.equals ( o ) ;
  }
  /**
   *  The class that provides a synchronized interface for use in multi-threaded applications.
   *
   *  @author Russel Winder
   */
  //  Cannot extend both Dequeue and Queue.Synchronized so we have to list all the methods.
  public static class Synchronized<E> extends Dequeue<E> {
    private final Object theLock;
    public Synchronized ( ) { theLock = this ; }
    public Synchronized ( final Object o ) { theLock = o ; }
    @Override public final void headAdd ( final E o ) {
      synchronized ( theLock ) { super.headAdd ( o ) ; }
    }
    @Override public final void tailAdd ( final E o ) {
      synchronized ( theLock ) { super.tailAdd ( o ) ; }
    }
    @Override public final E headRemove ( ) {
      synchronized ( theLock ) { return super.headRemove ( ) ; }
    }
    @Override public final E  tailRemove ( ) {
      synchronized ( theLock ) { return super.tailRemove ( ) ; }
    }
    @Override public final E remove ( ) {
      synchronized ( theLock ) { return super.remove ( ) ; }
    }
    @Override public final void makeEmpty ( ) {
      synchronized ( theLock ) { super.makeEmpty ( ) ; }
    }
    @Override public final boolean isEmpty ( ) {
      synchronized ( theLock ) {  return super.isEmpty ( ) ; }
    }
    @Override public final int size ( ) {
      synchronized ( theLock ) { return super.size ( ) ; }
    }
    @Override public final boolean equals ( final Object o ) {
      synchronized ( theLock ) { return super.equals ( o ) ; }
    }
    @Override public final boolean contains ( final E o ) {
      synchronized ( theLock ) { return super.contains ( o ) ; }
    }
    @Override public final void add ( final E o ) {
      synchronized ( theLock ) { super.add ( o ) ; }
    }
    @Override public final boolean remove ( final E o ) {
      synchronized ( theLock ) {  return super.remove ( o ) ; }
    }
    @Override public final Synchronized<E> clone ( ) {
      synchronized ( theLock ) { return (Synchronized<E>) super.clone ( ) ; }
    }
    @Override public final String toString ( ) {
      synchronized ( theLock ) { return super.toString ( ) ; }
    }
  }
}
