/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2011 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.alt ;
import org.devjavasoft.ads.AccessEmptyContainerException ;
import org.devjavasoft.ads.Container ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.InvalidIteratorException ;
import org.devjavasoft.ads.InvalidOperationException ;
import org.devjavasoft.ads.Sequence ;
import org.devjavasoft.ads.SLList ;
/**
 *  A <code>Queue&lt;T></code> implemented as an adapter on a <code>Sequence&lt;T></code>.  The default
 *  storage mechanism is an <code>SLList&lt;T></code> but any <code>Sequence&lt;T></code> may be used.  The
 *  choice is made at creation time.
 *
 *  @see Sequence
 *  @author Russel Winder
 */
public class Queue<T> implements Container<T> {
  /**
   *  The place to store some data.  Should perhaps be final, but the implementation of clone means it
   *  cannot be.
   */
  protected /*final*/ Sequence<T> storage ;
  public Queue ( ) { this ( new SLList<T> ( ) ) ; }
  /**
   *  Constructor allowing a choice of storage representation.
   *
   *  @param sequence The <code>Sequence</code> object to be used to store data.
   */
  public Queue ( final Sequence<T> sequence ) { storage = sequence ; }
  /**
   *  Peek at the head of the <code>Queue&lt;T></code>.
   *
   *  @return the value at teh head of the queue.
   *  @exception AccessEmptyContainerException if we attempt to remove from an empty
   *  <code>Queue&lt;T></code>.
   */
  T peek ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    return storage.get ( 0 ) ;
  }
   /**
   *  Remove a datum from ourself.
   *
   *  @return the value removed.
   *  @exception AccessEmptyContainerException if we attempt to remove from an empty <code>Queue</code>.
   */
  public T remove ( ) {
    if ( isEmpty ( ) ) { throw new AccessEmptyContainerException ( ) ; }
    final T o = storage.get ( 0 ) ;
    storage.remove ( 0 ) ;
    return o ;
  }
  /**
   *  Cause ourself to be emptied.
   */
  @Override public void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  /**
   *  Check to see whether we are empty or not.
   */
  @Override public boolean isEmpty ( ) { return storage.isEmpty ( ) ; }
  /**
   *  Tell the world what size we are, i.e. how many <code>T</code>s we have stored in ourself.
   */
  @Override public int size ( ) { return storage.size ( ) ; }
  /**
   *  Test something against ourself for equality.
   *
   *  @param o the object to compare to.
   *  @return @{code true} if the items are equal, @{cade false} otherwise.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public boolean equals ( final Object o ) {
    if ( this == o ) { return true ; }
    if ( ( o == null ) ||! ( o instanceof Queue ) ) { return false ; }
    return storage.equals ( ( (Queue<T>) o ).storage ) ;
  }
  /**
   *  Determine whether we have the value represented by the parameter <code>T</code> in us.
   *
   *  @exception AccessEmptyContainerException if there are no items in the <code>Container&lt;T></code>.
   */
  @Override public boolean contains ( final T o ) { return storage.contains ( o ) ; }
  /**
   *  Add an element to the <code>Container&lt;T></code>.
   */
  @Override public void add ( final T o) { storage.add ( o ) ; }
  /**
   *  Remove an element of a given value from the <code>Container&lt;T></code> if it is in the
   *  <code>Container&lt;T></code>.
   *
   *  @exception InvalidOperationException is always thrown since this is an invalid operation,
   *  <code>Queue&lt;T></code>s are closed containers.
   */
  @Override public boolean remove ( final T o ) { throw new InvalidOperationException ( ) ; }
  /**
   *  Create a primitive array of <code>T</code> from the <code>Stack&lt;T></code>.
   *
   *  <p>Cannot use "(T[]) new Object [size ( )]" to create the array result since even though it seems type
   *  correct, it isn't because of erasure.  So, in order to really deliver a T[], we must use reflection.
   *  Hence the need for the Class<T> parameter.</p>
   *
   *  @return a reference to the newly created array.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public T[] toPrimitiveArray ( final Class<T> c ) {
    final T[] data = (T[]) java.lang.reflect.Array.newInstance ( c , size ( ) ) ;
    int index = 0 ;
    for ( InputIterator<T> i = iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) , ++index ) { data[index] = i.get ( ) ; }
    return data ;
  }
  /**
   *  Create a primitive array of <code>T</code> from the <code>Stack&lt;T></code>.
   *
   *  @param a reference to an array in which to put the data from the <code>Stack&lt;T></code> if it has
   *  enough space in which to put all the items.
   *  @return a reference to the filled array.  If the array passed as a parameter was big enough then the
   *  return value is a reference to that array otherwise a new array is created and a reference returned.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public T[] toPrimitiveArray ( T[] a ) {
    final int size = size ( ) ;
    if ( a.length < size ) {
      a = (T[]) java.lang.reflect.Array.newInstance ( a.getClass ( ).getComponentType ( ) , size ) ;
    }
    int index = 0 ;
    for ( InputIterator<T> i = iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) , ++index ) { a[index] = i.get ( ) ; }
    return a ;
  }
  /**
   *  We have to clone ourself properly as a <code>Queue&lt;T></code>.
   */
  @SuppressWarnings ( "unchecked" )
  @Override public Queue<T> clone ( ) {
    try {
      final Queue<T> q = (Queue<T>) super.clone ( ) ;
      q.storage = storage.clone ( ) ;
      return q ;
    }
    catch ( CloneNotSupportedException e ) { throw new InternalError ( ) ; }
  }
  /**
   *  It is always useful to be able to print out the <code>Queue&lt;T></code>, mostly to make debugging
   *  easy.
   */
  @Override public String toString ( ) { return storage.toString ( ) ; }
  /**
   *  Even though we are a closed container, conforming to <code>Container&lt;T></code> requires that we
   *  provide this iterator.  However, we do ensure that there is no possibility of updates.
   */
  @Override public Iterator begin ( ) { return new Iterator ( ) ; }
  /**
   *  Even though we are a closed container, conforming to <code>Container&lt;T></code> requires that we
   *  provide this iterator.  However, we do ensure that there is no possibility of updates.
   */
  @Override public Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  /**
   *  Even though we are a closed container, conforming to <code>Container&lt;T></code> requires that we
   *  provide a public iterator.  However, we do ensure that there is no possibility of updates.
   */
  @Override public Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  An iterator over <code>Queue&lt;T></code>s.
   *
   *  @author Russel Winder
   */
  public class Iterator implements InputIterator<T> {
    /**
     *  We just need to act as a proxy for an iterator from the underlying representation.
     */
    private InputIterator<T> cursor = storage.iterator ( ) ;
    /**
     *  Deliver the value we are currently referring to.
     */
    @Override public final T get ( ) { return cursor.get ( ) ; }
    /**
     *  Move on to the next item in the container.
     */
    @Override public final void advance ( ) { cursor.advance ( ) ; }
    /**
     *  Move on n items in the container.
     */
    @Override public final void advance ( final int increment ) { cursor.advance ( increment ) ; }
    /**
     *  Are we positioned at the beginning of the iteration?
     */
    @Override public final boolean atBegin ( ) { return cursor.atBegin ( ) ; }
    /**
     *  Have we reached the end of the iteration?
     */
    @Override public final boolean atEnd ( ) { return cursor.atEnd ( ) ; }
    /**
     *  Are two iterators equal, i.e. do two iterators refer to the same item in the same data structure.
     *
     *  @param o the object to compare to.
     *  @return @{code true} if the items are equal, @{cade false} otherwise.
     */
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof Queue.Iterator ) ) { return false ; }
      return cursor.equals ( ( (Iterator) o ).cursor ) ;
    }
    /**
     *  Return a reference to the <code>Container&lt;T></code> of this <code>Iterator</code>.
     */
    @Override public final Container<T> getContainer ( ) { return Queue.this ; }
    /**
     *  Calculate the distance between this iterator and the parameter iterator which must, of course, be an
     *  iterator into the same container.
     *
     *  @param i the iterator to measure the distance to.
     *  @return the number of items between this iterator and the parameter iterator.
     *  @exception InvalidIteratorException if the parameter iterator is not in the same container as this
     *  iterator.
     */
    @Override public int distance ( final InputIterator<T> i ) {
      if ( equals ( i ) ) { return 0 ; }
      if ( getContainer ( ) != i.getContainer ( ) ) { throw new InvalidIteratorException ( ) ; }
      int count = 0 ;
      InputIterator<T> j = clone ( ) ;
      for ( ; ! j.atEnd ( ) ; ++count , j.advance ( ) ) { if ( j.equals ( i ) ) { return count ; } }
      //  Check to see if i.atEnd ( ).
      if ( j.equals ( i ) ) { return count ; }
      count = 0 ;
      j = i.clone ( ) ;
      for ( ; ! j.atEnd ( ) ; ++count , j.advance ( ) ) { if ( j.equals ( this ) ) { return -count ; } }
      //  Check to see if atEnd ( ).
      if ( j.equals ( this ) ) { return -count ; }
      throw new InternalError ( ) ;
    }
    /**
     *  Clone this <code>Iterator</code>.
     */
    @Override public final Iterator clone ( ) {
      try {
        @SuppressWarnings ( "unchecked" )
        final Iterator i = (Iterator) super.clone ( ) ;
        i.cursor = cursor.clone ( ) ;
        return i ;
      }
      catch ( CloneNotSupportedException e ) { throw new InternalError ( ) ; }
    }
    /**
     *  Implementation of the <code>hasNext</code> method from <code>Iterator&lt;T></code>.
     */
    @Override public boolean hasNext ( ) { return ! atEnd ( ) ; }
    /**
     *  Implementation of the <code>next</code> method from <code>Iterator&lt;T></code>.
     *
     *  @return the value the iterator is pointing at after it has moved on one.
     */
    @Override public T next ( ) {
      final T returnValue = get ( ) ;
      advance ( ) ;
      return returnValue ;
    }
    /**
     *  Implementation of the <code>remove</code>  method from <code>Iterator&lt;T></code>.
     */
    @Override public void remove ( ) { throw new UnsupportedOperationException ( "Cannot apply remove with this iterator." ) ; }
  }
  /**
   *  The class that provides a synchronized interface for use in multi-threaded applications.  This is just
   *  a forwarding proxy with synchronization control and locking. See <code>Queue&lt;E></code> for all
   *  comments on the methods.
   *
   *  @author Russel Winder
   *  @version
   *    $Revision: 549 $
   *    $Date: 2006-09-01 19:19:38 +0100 (Fri, 01 Sep 2006) $
   */
  public static class Synchronized<E> extends Queue<E> {
    /**
     *  The <code>Object</code> on which to do all locking.
     */
    private final Object theLock;
    /**
     *  The default constructor.  Gives an <code>Queue</code> with default initial size and increment.
     */
    public Synchronized ( ) { theLock = this ; }
    /**
     *  The constructor that supplies an object to synchronize on.
     *
     *  @param o the object to lock on.
     */
    public Synchronized ( final Object o ) { theLock = o ; }
    @Override public final E remove ( ) {
      synchronized ( theLock ) { return super.remove ( ) ; }
    }
    @Override public final void makeEmpty ( ) {
      synchronized ( theLock ) { super.makeEmpty ( ) ; }
    }
    @Override public final boolean isEmpty ( ) {
      synchronized ( theLock ) { return super.isEmpty ( ) ; }
    }
    @Override public final int size ( ) {
      synchronized ( theLock ) { return super.size ( ) ; }
    }
    @Override public final boolean equals ( final Object o ) {
      synchronized ( theLock ) { return super.equals ( o ) ; }
    }
    @Override public final boolean contains ( final E o ) {
      synchronized ( theLock ) { return super.contains ( o ) ; }
    }
    @Override public final void add ( final E o ) {
      synchronized ( theLock ) { super.add ( o ) ; }
    }
    @Override public final boolean remove ( final E o ) {
      synchronized ( theLock ) { return super.remove ( o ) ; }
    }
    @Override public final E[] toPrimitiveArray ( final Class<E> c ) {
      synchronized ( theLock ) { return super.toPrimitiveArray ( c ) ; }
    }
    @Override public final Synchronized<E> clone ( ) {
      synchronized ( theLock ) { return (Synchronized<E>) super.clone ( ) ; }
    }
    @Override public final String toString ( ) {
      synchronized ( theLock ) { return super.toString ( ) ; }
    }
  }
}
