/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.alt ;
import org.devjavasoft.ads.AbstractSequence ;
import org.devjavasoft.ads.AbstractIterator ;
import org.devjavasoft.ads.AccessEmptyContainerException ;
import org.devjavasoft.ads.Container ;
import org.devjavasoft.ads.ForwardIterator ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.InvalidIteratorException ;
import org.devjavasoft.ads.Sequence ;
/**
 *  A <code>Sequence&lt;T></code> implemented as a singly-linked list.
 *
 *  <p> This implementation uses a circular singly-linked list (cf DLList) rather than head and tail
 *  reference with the singly-linked list strung between them.</p>
 *
 *  @see Container
 *  @see Sequence
 *  @author Russel Winder
 */
public class SLList<T> extends AbstractSequence<T> {
  /**
   *  This is the class from which linked-list nodes are instantiated.  This is a very passive sort of Node
   *  so it is a nested top-level class and not an inner class since it has no need to ever know who its
   *  parent is.
   *
   *  @author Russel Winder
   */
  private static class Node<E> {
    /**
     *  A <code>Node&lt;E></code> has some datum that it is storing.
     */
    protected E datum ;
    /**
     *  A <code>Node&lt;E></code> has a tail, a reference to another <code>Node&lt;E></code>.
     */
    protected Node<E> next ;
    /**
     *  Constructor for an unconnected <code>Node&lt;E></code>.
     *
     *  @param o The datum for this <code>Node&lt;E></code>.
     */
    @SuppressWarnings ( "unused" )
    public Node ( final E o) { this ( o , null ) ; } 
    /**
     *  Constructor for a <code>Node&lt;E></code> with a given list of <code>Node&lt;E></code>s to become
     *  the head of.
     *
     *  @param o The datum for this <code>Node&lt;E></code>.
     *  @param n The <code>Node&lt;E></code> to be the tail of this one.
     */
    public Node ( final E o , final Node<E> n ) {
      datum = o ;
      next = n ;
    }
  }
  /**
   *  The last <code>Node&lt;T></code> in the list.  This not only gives the tail but using tail.next we
   *  have the head of the list.
   */
  private Node<T> tail = null ;
  /**
   *  The count of the number of elements in the <code>SLList&lt;T></code>.  This is not absolutely
   *  necessary but having it makes a lot of operations a lot easier.  A good efficiency measure.
   */
  private int elementCount = 0 ;
  public SLList ( ) { }
  /**
   *  Construct an <code>SLList&lt;T></code> by copying in the values from the the
   *  <code>Container&lt;T></code>.
   *
   *  @param c the <code>Container&lt;T></code> to copy the values from.
   */
  public SLList ( final Container<T> c ) { copyIn ( c ) ; }
  /**
   *  Construct an <code>SLList&lt;T></code> by copying the slice of a <code>Sequence&lt;T></code>defined by
   *  two indexes.
   *
   *  @param s the <code>Sequence&lt;T></code> a slice of which is to be copied.
   *  @param b beginning index of the slice of <code>s</code>.
   *  @param e end index of the slice of <code>s</code>.
   */
  public SLList ( final Sequence<T> s , final int b , final int e ) { copyIn ( s , b , e ) ; }
  /**
   *  Construct an <code>SLList&lt;T></code> by copying the sequence defined by two
   *  <code>InputIterator&lt;T></code>s.
   *
   *  @param b the beginning of a slice to copy in to this new <code>SLList</code>.
   *  @param e the end of the slice to copy in to this new <code>SLList</code>.
   *  @throws InvalidIteratorException if the two <code>InputIterator&lt;T></code>s are not working the same
   *  <code>SLList&lt;T></code>.
   */
  public SLList ( final InputIterator<T> b , final InputIterator<T> e ) { copyIn ( b , e ) ; }
  /**
   *  @return the first item in the <code>SLList&lt;T></code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>SLList&lt;T></code>.
   */
  @Override public final T first ( ) { validityCheck ( ) ; return tail.next.datum ; }
  /**
   *  @return the last element in the <code>SLList&lt;T></code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>SLList&lt;T></code>.
   */
  @Override public final T last ( ) { validityCheck ( ) ; return tail.datum ; }
  /**
   *  @return the element at a given position in the <code>SLList&lt;T></code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>SLList&lt;T></code>.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size of
   *  the <code>SLList&lt;T></code>.
   */
  @Override public final T get ( final int index ) { validityCheck ( index ) ; return nodeAt ( index ).datum ; }
  /**
   *  Assign a value to a position in the <code>SLList&lt;T></code>.  This overwrites whatever value was
   *  there.
   *
   *  @throws AccessEmptyContainerException if there are no items in the <code>SLList&lt;T></code> and the
   *  index is not 0.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size of
   *  the <code>SLList&lt;T></code>.
   */
  @Override public final void set ( final int index , final T o ) {
    try { validityCheck ( index ) ; }
    catch ( AccessEmptyContainerException aece ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw aece ; }
    }
    catch ( IndexOutOfBoundsException ioobe ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw ioobe ; }
    }
    nodeAt ( index ).datum = o ;
  }
  /**
   *  Excise an element from the <code>SLList&lt;T></code>.
   *
   *  @param index the index of the item in the list to remove, 0 being the head element.
   *  @throws AccessEmptyContainerException if there are no items in the <code>SLList&lt;T></code>.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size of
   *  the <code>SLList&lt;T></code>.
   */
  @Override public final void remove ( final int index ) {
    validityCheck ( index ) ;
    Node<T> toGo = tail.next ;
    //  We could use either the expression "tail == tail.next" or "elementCount ==1" to test to see if we
    //  are removing the last item from the list.  The former is rather more safe in some ways since it deal
    //  with the data structure itself but if the element count is not correct we have a problem anyway.
    if ( tail == tail.next ) {
      //  We are removing the last item and so must take care to nullify the tail pointer.
      tail = null ;
    } else {
      //  We are not removing the last item and the index is known to be valid.  Find the element before the
      //  one we need to remove so that we can excise the one that has to go.
      final Node<T> previous = nodeAt ( index - 1 ) ;
      toGo = previous.next ;
      previous.next = toGo.next ;
    }
    //  Nullify the next of the garbage item just in case.  This is probably not required since the garbage
    //  collector will get things right but as a safety measure let's just do it.  Also ensure we decrement
    //  the element count to the proper value.
    toGo.next = null ;
    --elementCount ;
  }
  /**
   *  Insert a new item at the given index.  Insertion happens before, i.e. the new item becomes the
   *  index'th item with the old index'th item becoming the index+1'th item, etc.
   *
   *  @throws AccessEmptyContainerException if there are no items in the <code>SLList&lt;T></code> and the
   *  index is not 0.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the size of
   *  the <code>SLList&lt;T></code>.
   */
  @Override public final void insert ( final int index , final T o ) {
    try { validityCheck(index) ; }
    catch ( AccessEmptyContainerException aece ) {
      if ( index == 0 ) { add ( o ) ; return ; }
      else { throw aece ; }
    }
    final Node<T> current = nodeAt ( index - 1 ) ;
    current.next = new Node<T> ( o , current.next ) ;
    ++elementCount ;
  }
  /**
   *  Make the <code>DLList&lt;T></code> empty by disconnecting the <code>Node&lt;T></code>s that comprise
   *  the internal data structure.  Assume that garbage collection will tidy up whenever necessary.
   *
   *  <p>NB The clone and merge methods rely on the fact that this method does not actually do anything with
   *  the <code>Node&lt;T></code>s.</p>
   */
  @Override public final void makeEmpty ( ) {
    tail = null ;
    elementCount = 0 ;
  }
  /**
   *  @return the number of items in the <code>SLList&lt;T></code>.
   */
  @Override public final int size ( ) { return elementCount ; }
  /**
   *  Append an element to the end of the <code>SLList&lt;T></code>.
   */
  @Override public final void add ( final T o ) {
    final Node<T> n = new Node<T> ( o , tail.next ) ;
    tail.next = n ;
    tail = n ;
    ++elementCount ;
  }
  /**
   *  Remove an element of a given value from the <code>SLList&lt;T></code> if it is in the
   *  <code>SLList&lt;T></code>.
   *
   *  @return whether the item was actually in the <code>SLList&lt;T></code>.
   *  @throws AccessEmptyContainerException if there are no items in the <code>SLList&lt;T></code>.
   */
  @Override public final boolean remove ( final T o ) {
    validityCheck ( ) ;
    //  Work down the list checking to see if the element is there.  If it is break out of the loop and do
    //  the deletion.  Be careful to ensure we don't go round the circular list more than once.
    Node<T> previous= tail ;
    Node<T> toGo = tail.next ;
    while ( ! toGo.datum.equals ( o ) ) {
      previous = toGo ;
      toGo = toGo.next ;
      if ( toGo == tail.next ) { return false ; }
    }
    previous.next = toGo.next ;
    toGo.next = null ;
    --elementCount ;
    return true ;
  }
  @Override public final SLList<T> clone ( ) {
    final SLList<T> l = (SLList<T>) super.clone ( ) ;
    //  We are sharing all the Node<T>s at this point which is not what is required.  We can share the data
    //  attached to the Node<T>s but we must make a new linked list.  Can use makeEmpty since it does not
    //  touch the linked list data.
    l.makeEmpty ( ) ;
    for ( final T t : this ) { l.add ( t ) ; }
    return l ;
  }
  /**
   *  @return an iterator positioned at the beginning of the <code>SLList&lt;T></code>.
   */
  @Override public final Iterator begin() { return new Iterator ( ) ; }
  /**
   *  @return an iterator positioned at the end of the <code>SLList&lt;T></code>.
   */
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.current = tail ;
    return i ;
  }
  /**
   *  @return an iterator positioned at the beginning of the <code>SLList&lt;T></code>.
   */
  @Override public final Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  Merge the <code>SLList&lt;T></code> <code>l</code>after a given index in
   *  this<code>SLList&lt;T></code>.  This moves (not copies) all the <code>Node&lt;T>s</code> from the
   *  <code>SLList&lt;T></code> that is the parameter into us.
   *
   *  @param index location in this <code>SLList&lt;T></code> after which to insert <code>l</code>.
   *  @param l the <code>SLList&lt;T></code> to insert.
   */
  public final void merge ( final int index , final SLList<T> l ) {
    if ( isEmpty ( ) && ( index == 0 ) ) { tail = l.tail ; }
    else {
      validityCheck ( index ) ;
      final Node<T> n = nodeAt ( index ) ;
      final Node<T> t_n = n.next ;
      n.next = l.tail.next ;
      l.tail.next = t_n ;
    }
    elementCount += l.elementCount ;
    l.makeEmpty ( ) ;
  }
  /**
   *  Append a <code>SLList&lt;T></code> to ourself.  The moves (not copies) all the
   *  <code>Nodes&lt;T></code> from the <code>SLList&lt;T></code> that is the parameter into us.
   *
   *  @param l the <code>DLList&lt;T></code> to insert.
   */
  public final void merge ( final SLList<T> l ) {
    if ( isEmpty ( ) ) { tail = l.tail ; }
    else {
      final Node<T> t_n = tail.next ;
      tail.next = l.tail.next ;
      l.tail.next = t_n ;
    }
    elementCount += l.elementCount ;
    l.makeEmpty ( ) ;
  }
  /**
   *  Find a <code>Node&lt;T></code> given an index.
   *
   *  @param index the index into the <code>SLList&lt;T></code> to deliver up the <code>Node&lt;T></code>
   *  of.
   *  @return the node.
   *  @throws IndexOutOfBoundsException if the requested <code>index</code> is not valid due to the * size
   *  of the <code>SLList&lt;T></code>.
   */
  private Node<T> nodeAt ( final int index ) {
    //  A little optimization for access at the tail.
    if ( index == size ( ) - 1 ) { return tail ; }
    //  We know that the index is valid so walk the SLList until we get to the right place and then return
    //  the value at that place.
    Node<T> n = tail.next ;
    for ( int i = 0 ; i < index ; ++i ) { n = n.next ; }
    return n ;
  }
  /**
   *  The iterator over <code>SLList&lt;T></code>s.  This is an inner class (as opposed to being a nested
   *  top-level class) so that it automatically has a reference to its parent object, i.e. the object it is
   *  an iterator for.
   *
   *  @author Russel Winder
   */
  public class Iterator extends AbstractIterator<T> implements ForwardIterator<T> {
    private Node<T> current = tail.next ;
    /**
     *  @return the value we are currently referring to.
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    @Override public final T get ( ) {
      validityCheck ( ) ;
      return current.datum;
    }
    /**
     *  Amend the value of the element we are currently referring to.
     *
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    @Override public final void set ( final T o ) {
      validityCheck ( ) ;
      current.datum = o ;
    }
    /**
     *  Move on to the next element in the container.
     *
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    @Override public final void advance ( ) {
      validityCheck ( ) ;
      current = current.next ;
    }
    /**
     *  Move on n items in the container.
     *
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    @Override public final void advance ( int increment ) {
      for ( ; increment > 0 ; --increment ) {
        if ( atEnd ( ) ) { break ; }
        advance ( ) ;
      }
    }
    /**
     *  @return true if we are positioned at the beginning of the iteration or flase otherwise.
     */
    @Override public final boolean atBegin ( ) { return current == tail.next ; }
    /**
     *  @return true if we are positioned at the end of the iteration or false otherwise.
     */
    @Override public final boolean atEnd ( ) { return current == null ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( o == this ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof SLList.Iterator ) ) { return false ; }
      final Iterator i = (Iterator) o ;
      return atEnd ( ) ? i.atEnd ( ) : current.equals ( i.current ) ;
    }
    /**
     *  @return a reference to the <code>Container&lt;T></code> of this <code>InputIterator</code>
     */
    @Override public final Container<T> getContainer ( ) { return SLList.this ; }
    @Override public final Iterator clone ( ) { return (Iterator) super.clone ( ) ; }
    /**
     *  Append a new item after the current one.
     *
     *  @param o the object to add to the list.
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    public final void add ( final T o ) {
      validityCheck ( ) ;
      final Node<T> n = new Node<T> ( o , current.next ) ;
      current.next = n ;
      if ( tail == current ) { tail = n ; }
      ++elementCount ;
    }
    /**
     *  Merge an <code>SLList&lt;T></code> after the current item.
     *
     *  @param l the <code>SLList&lt;T></code> to merge in at this point in the <code>SLList&lt;T></code>
     *  this iterator is pointing into.
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    public final void merge ( final SLList<T> l ) {
      validityCheck ( ) ;
      if ( l.isEmpty ( ) ) { return ; }
      final Node<T> t_n = l.tail.next ;
      l.tail.next = current.next ;
      current.next = t_n ;
      elementCount += l.elementCount ;
      l.makeEmpty ( ) ;
    }
    /**
     *  Check to see if the iterator is valid.
     *
     *  @throws InvalidIteratorException if the iterator is not valid.
     */
    private void validityCheck ( ) {
      if ( current == null ) { throw new InvalidIteratorException ( ) ; }
    }
  }
}
