/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.alt.searching ;
import org.devjavasoft.ads.Comparator ;
/**
 *  Search a sorted array using binary chop search.  This is an O(log(n)) search.
 *
 *  @author Russel Winder
 */
public final class BinaryChopSearch {
  private BinaryChopSearch ( ) { }
  /**
   *  The search operation.  This uses a recursive algorithm.
   *
   *  @param a the sorted array to be searched.
   *  @param o the <code>Object</code> to be searched for.
   *  @param ce the <code>Comparator</code> defining value equality.
   *  @param cr this <code>Comparator</code> must either be "less than" or "greater than" and the
   *  same comparator that defines the order on the array.
   *  @return index of the item or -1 if it is not there.
   */
  public static <T> int search ( final T[] a, final T o, final Comparator<T> ce, final Comparator<T> cr) {
    return binaryChop ( a , 0 , a.length - 1 , o , ce , cr ) ;
  }
  /**
   *  The (tail) recursive support function that actually does the hard work.
   *
   *  <p>Since Java does not support tail recursion, this function uses a stack frame for each call.
   *  It should therefore be treated as in need of refactoring, i.e. to be replaced by an iterative
   *  implementation.</p>
   *
   *  @param a the sorted array to be searched.
   *  @param lo the low index of the slice to search.
   *  @param hi the high index of the slice to search.
   *  @param o the <code>Object</code> to be searched for.
   *  @param ce the <code>Comparator</code> defining value equality.
   *  @param cr this <code>Comparator</code> must either be "less than" or "greater than" and the
   *  same comparator that defines the order on the array.
   *  @return  index of the item or -1 if it is not there.
   */
  private static <T> int binaryChop ( final T[] a , final int lo , final int hi , final T o , final Comparator<T> ce , final Comparator<T> cr ) {
    if ( hi == lo ) { return  ce.relation(a[hi], o) ? hi : -1 ; }
    final int centre = ( hi + lo ) / 2 ;
    if ( cr.relation ( a[centre] , o ) ) { return binaryChop ( a , centre , hi , o , ce , cr ) ; }
    else if ( cr.relation ( o , a[centre] ) ) { return binaryChop ( a , lo , centre , o , ce , cr ) ; }
    else { return centre ; }
  }
}
