/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.searching ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.RandomAccessIterator ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.InvalidIteratorException ;
/**
 *  Search an ordered sequence using binary chop search.  This is an O(log(n)) search.
 *
 *  <p>Only support primitive arrays, <code>Array</code>s and sequences defined by
 *  <code>RandomAccessIterator</code>s.  Although it would seem we could support binary chop search on a
 *  <code>DLList</code> using <code>BiDirectionalIterator</code>s this would in fact be an O(n) operation
 *  because the <code>BiDirectionalIterator</code>s are O(n) for an <code>advance(n)</code> or
 *  <code>retreat(n)</code> and binary chop search is only O(log(n)) if the iterator operations are O(1).
 *  <code>SLList</code>s cannot be binary chop searched as they cannot support
 *  <code>BiDirectionalIterator</code>s.</p>
 *
 *  @author Russel Winder
 */
public final class BinaryChopSearch {
  private BinaryChopSearch ( ) { }
  /**
   *  Search a primitive array.  It is assumed that the primitive array is sorted.
   *
   *  @param a the sorted array to be searched.
   *  @param o the <code>T</code> to be searched for.
   *  @param ce the <code>Comparator</code> defining value equality.
   *  @param cr this <code>Comparator</code> must either be "less than" or "greater than" and the same
   *  comparator that defines the order on the array.
   *  @return index of the item or -1 if it is not there.
   */
  public static <T> int search ( final T[] a , final T o , final Comparator<? super T> ce , final Comparator<? super T> cr ) {
    int hi = a.length ;
    int lo = 0 ;
    while ( true ) {
      final int centre = ( hi + lo ) / 2 ;
      if ( centre == lo ) {
        return ( ce.relation ( a[centre] , o ) ? centre :
                 ( ce.relation ( a[centre + 1] , o ) ? centre + 1
                     : -1 ) ) ;
      }
      if ( cr.relation ( a[centre] , o ) ) { lo = centre ; }
      else if ( cr.relation ( o , a[centre] ) ) { hi = centre ; }
      else { return centre ; }
    }
  }
  /**
   *  Search a sorted <code>Array</code>.
   *
   *  @param a the sorted <code>Array</code> of to be searched.
   *  @param o the <code>T</code> to be searched for.
   *  @param ce the <code>Comparator</code> defining value equality.
   *  @param cr this <code>Comparator</code> must either be "less than" or "greater than" and the same
   *  comparator that defines the order on the array.
   *  @return iterator positioned at the item or null if the item is not in <code>a</code>.
   */
  public static <T> RandomAccessIterator<T> search ( final Array<T> a , final T o , final Comparator<? super T> ce , final Comparator<? super T> cr ) {
    int hi = a.size ( ) ;
    int lo = 0 ;
    int rv_index = 0 ;
    while ( true ) {
      final int centre = ( hi + lo )  /  2 ;
      if ( centre == lo ) {
        rv_index =  ( ce.relation ( a.get ( centre ) , o ) ? centre :
                ( ce.relation ( a.get ( centre + 1 ) , o ) ? centre + 1 :
                  -1 ) ) ;
        break ;
      }
      if ( cr.relation ( a.get ( centre ) , o ) ) { lo = centre ; }
      else if ( cr.relation ( o , a.get ( centre ) ) ) { hi = centre ; }
      else { rv_index = centre ; break ; }
    }
    final RandomAccessIterator<T> rv = a.begin ( ) ;
    rv.advance ( rv_index ) ;
    return rv ;
  }
  /**
   *  Binary chop search a sequence (assumed sorted) defined by two random access iterators.  NB The
   *  iterator operations MUST be O(1).
   *
   *  @param begin the beginning of the iterator defined sequence to be searched.
   *  @param end the end of the iterator defined sequence to be searched.
   *  @param o the <code>T</code> to be searched for.
   *  @param ce the <code>Comparator</code> defining value equality.
   *  @param cr this <code>Comparator</code> must either be "less than" or "greater than" and the same
   *  comparator that defines the order on the underlying sequence.
   *  @return iterator positioned at the item or null if the item is not in <code>s</code>.
   *  @throws InvalidIteratorException if <code>begin</code> and <code>end</code> are not working the same
   *  data structure or <code>end</code> is before <code>begin</code>.
   */
  public static <T> RandomAccessIterator<T> search ( final RandomAccessIterator<T> begin , final RandomAccessIterator<T> end , final T o , final Comparator<? super T> ce , final Comparator<? super T> cr ) {
    //  NB distance throws InvalidIteratorException if the iterators are not working the same container.
    int hi = begin.distance ( end ) ;
    if ( hi < 0 ) { throw new InvalidIteratorException ( ) ; }
    int lo = 0 ;
    boolean goUp = true ;
    final RandomAccessIterator<T> i = begin.clone ( ) ;
    while ( true ) {
      final int centre = ( hi + lo ) / 2 ;
      if ( centre == lo ) {
        if ( ce.relation ( i.get ( ) , o ) ) { return i ; }
        i.advance ( ) ;
        if ( ce.relation ( i.get ( ) , o ) ) { return i ; }
        return null ;
      }
      if ( goUp ) { i.advance ( centre - lo ) ; }
      else { i.retreat ( hi - centre ) ; }
      if ( cr.relation ( i.get ( ) , o ) ) {
        lo = centre ;
        goUp = true ;
      } else if ( cr.relation ( o , i.get ( ) ) ) {
        hi = centre ;
        goUp = false ;
      } else { return i ; }
    }
    //  Cannot get here.
  }
}
