/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.searching ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.Container ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.InvalidIteratorException ;
/**
 *  Search using linear search, an O(n) search.
 *
 *  @author Russel Winder
 */
public final class LinearSearch {
  private LinearSearch ( ) { }
  /**
   *  Search a primitive array.
   *
   *  @param a the array to be searched.
   *  @param o the <code>T</code> to be searched for.
   *  @param c the <code>Comparator</code> defining value equality.
   *  @return index of the item or -1 if it is not there.
   */
  public static <T> int search ( final T[] a , final T o , final Comparator<? super T> c ) {
    for ( int i = 0 ; i < a.length ; ++i ) {
      if ( c.relation ( a[i] , o ) ) { return i ; }
    }
    return -1 ;
  }
  /**
   *  Search a <code>Container</code>
   *
   *  @param s the <code>Container</code> to be searched.
   *  @param o the <code>T</code> to be searched for.
   *  @param c the <code>Comparator</code> defining value equality.
   *  @return iterator positioned at the item or null if the item is not in <code>s</code>.
   */
  public static <T> InputIterator<T> search ( final Container<T> s , final T o , final Comparator<? super T> c ) {
    for ( InputIterator<T> i = s.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
      if ( c.relation ( i.get ( ) , o ) ) { return i ; }
    }
    return null ;
  }
  /**
   *  Search a sequence defined by two iterators.
   *
   *  @param begin the beginning of the iterator defined sequence to be searched.
   *  @param end the end of the iterator defined sequence to be searched.
   *  @param o the <code>T</code> to be searched for.
   *  @param c the <code>Comparator</code> defining value equality.
   *  @return iterator positioned at the item or null if the item is not in <code>s</code>.
   *  @throws InvalidIteratorException if <code>begin</code> and <code>end</code> are not working the same
   *  data structure or <code>end</code> is before <code>begin</code>.
   */
  public static <T> InputIterator<T> search ( final InputIterator<T> begin , final InputIterator<T> end , final T o , final Comparator<? super T> c ) {
    if ( begin.distance ( end ) < 0 ) { throw new InvalidIteratorException ( ) ; }
    for ( InputIterator<T> i = begin.clone ( ) ; ! i.atEnd ( ) && ! i.equals ( end ) ; i.advance ( ) ) {
      if ( c.relation ( i.get ( ) , o ) ) { return i ; }
    }
    return null ;
  }
}
