/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import java.io.File ;
import java.io.FileNotFoundException ;
import java.io.IOException ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.InputIterator ;
/**
 *  A function object delivering a balanced merge sort of a file on the filestore.
 *
 *  @see SortableRecord
 *  @see SortableRecordReader
 *  @see SortableRecordWriter
 *  @see SortableRecordInformation
 *  @see SortableRecordCopyFile
 *  @author Russel Winder
 */
public final class BalancedMultiwayMergeSort {
  private BalancedMultiwayMergeSort ( ) { }
  /**
   *  A sort operation with some defaults.
   *
   *  @param fileName the <code>String</code> giving the name of the file to be sorted.
   *  @param r the <code>SortableRecordInformation</code> factory object for creating
   *  <code>SortableRecordReader</code>s and <code>SortableRecordWriters</code>s and able to deliver a
   *  <code>Comparator&lt;SortableRecord></code>.
   *  @exception FileNotFoundException if <code>fileName</code> cannot be found.
   *  @exception IOException if there are any errors opening, reading or writing <code>fileName</code>.
   */
  public static void sort ( final String fileName , final SortableRecordInformation r )
    throws FileNotFoundException, IOException {
    sort ( fileName , 20 , 2 , r ) ;
  }
  /**
   *  A sort operation.
   *
   *  @param fileName the <code>String</code> giving the name of the file to be sorted.
   *  @param blockSize the number of data items in the initial sorted blocks.
   *  @param numberOfFiles the number of files to use for initial dispersion.
   *  @param rInfo the <code>SortableRecordInformation</code> factory object for creating
   *  <code>SortableRecordReader</code>s and <code>SortableRecordWriters</code>s and able to deliver a
   *  <code>Comparator&lt;SortableRecord></code>.
   *  @exception FileNotFoundException if <code>fileName</code> cannot be found.
   *  @exception IOException if there are any errors opening, reading or writing <code>fileName</code>.
   */
  public static void sort (final String fileName ,
                           final int blockSize ,
                           final int numberOfFiles ,
                           final SortableRecordInformation rInfo )
    throws FileNotFoundException, IOException {
    //  Create all the files needed for the sorting.
    final File file = new File ( fileName ) ;
    final File[] f_A = new File [numberOfFiles] ;
    final File[] f_B = new File [numberOfFiles] ;
    for ( int i = 0 ; i < numberOfFiles ; ++i ) {
      f_A[i] = new File ( "tmp_A_" + i ) ;
      f_B[i] = new File ( "tmp_B_" + i ) ;
    }
    //  Perform the initial dispersion into the A files.
    distributeSortedBlocks ( file , f_A , blockSize , rInfo ) ;
    //  Undertake the number of merge loops required to guarantee that everything is sorted.  Remember
    //  whether we ended up with A_0 or B_0 containing the final sorted data.
    File[] from = f_A ;
    File[] to = f_B ;
    boolean isFinal = true ;
    for ( int i = 0 ; merge ( from , to , (int) Math.pow ( 2 , i ) * blockSize , rInfo ) ; ++i ) {
      final File[] temp = from ;
      from = to ;
      to = temp ;
      isFinal = ! isFinal ;
    }
    //  Copy the data to the final destination.
    final File fileToCopy = isFinal ? f_B[0] : f_A[0] ;
    SortableRecordCopyFile.copy ( fileToCopy , file , rInfo ) ;
    //  Delete all the files.
    for ( int i = 0 ; i < numberOfFiles ; ++i ) {
      f_A[i].delete ( ) ;
      f_B[i].delete ( ) ;
    }
  }
  /**
   *  Perform the initial dispersion of the data.
   *
   *  @param from the file from which to take data.
   *  @param to the array of files in which to distribute the data read.
   *  @param blockSize the size of block to use.
   *  @param rInfo the data about the records being sorted.
   *  @throws FileNotFoundException
   *  @throws IOException
   */
  private static void distributeSortedBlocks (final File from ,
                                              final File[] to ,
                                              final int blockSize ,
                                              final SortableRecordInformation rInfo )
    throws FileNotFoundException, IOException {
    //  Create a Reader for the original data and a set of Writers for the A files.
    final SortableRecordReader reader = rInfo.getSortableRecordReader ( from ) ;
    final SortableRecordWriter[] writers = new SortableRecordWriter[to.length] ;
    for ( int i = 0 ; i < to.length ; ++i ) {
      writers[i] = rInfo.getSortableRecordWriter ( to[i] ) ;
    }
    boolean allDone = false ;
    while ( ! allDone ) {
      for ( int i = 0 ; ! allDone && ( i < writers.length ) ; ++i ) {
        //  Pull in a few records, put them into an Array that is where we are performing the internal sort
        //  that creates us the sorted block.  Do the sort then write out the block.
        final Array<SortableRecord> a = new Array<SortableRecord> ( ) ;
        for ( int k = 0 ; k < blockSize ; ++k ) {
          SortableRecord r = reader.readSortableRecord ( ) ;
          if ( r == null )  { allDone = true ; break ; }
          a.add ( r ) ;
        }
        Quicksort.sort ( a , rInfo.getComparator ( ) , rInfo.getEqualToComparator ( ) ) ;
        for ( InputIterator<SortableRecord> iter = a.iterator ( ) ; ! iter.atEnd ( ) ; iter.advance ( ) ) {
          writers[i].writeSortableRecord ( iter.get ( ) ) ;
        }
      }
    }
    //  Be tidy and close all the files.  Actually this is essential to ensure we get a flush.
    for ( final SortableRecordWriter writer : writers ) { writer.close ( ) ; }
    reader.close ( ) ;
  }
  /**
   *  Undertake a round of merging.
   *
   *  @param from the file from which to take data.
   *  @param to the array of files in which to distribute the data read.
   *  @param blockSize the size of block to use.
   *  @param rInfo the data about the records being sorted.
   *  @throws FileNotFoundException
   *  @throws IOException
   */
  private static boolean merge (final File[] from ,
                                final File[] to ,
                                final int blockSize ,
                                final SortableRecordInformation rInfo )
    throws FileNotFoundException, IOException {
    //  Open up the set of Readers and the set of Writers.
    final SortableRecordReader[] readers = new SortableRecordReader[from.length] ;
    for ( int i = 0 ; i < readers.length ; ++i ) { readers[i] = rInfo.getSortableRecordReader ( from[i] ) ; }
    final SortableRecordWriter[] writers = new SortableRecordWriter [to.length] ;
    for ( int i = 0 ; i < writers.length ; ++i ) { writers[i] = rInfo.getSortableRecordWriter ( to[i] ) ; }
    //  We make us of an array which holds the next Record for each of the files -- we need to have the
    //  record in memory in order to compare the keys and so decide which Record to write to the output
    //  file.  Have another array which is keeping count of how many Records we take from each of the files
    //  so that we can cease drawing from a given file when we have taken the appropriate number yet there
    //  are still records left.
    boolean returnValue = false ;
    boolean allDone = false ;
    final SortableRecord[] items = new SortableRecord[readers.length] ;
    final int[] counts = new int[readers.length] ;
    while ( ! allDone ) {
      for ( int i = 0 ; i < writers.length ; ++i ) {
        //  Initialize the array holding the next record from each of the files.  Determine whether we are
        //  finished or not by whether there are any records left or not.
        allDone = true ;
        for ( int j = 0 ; j < readers.length ; ++j ) {
          counts[j] = 0 ;
          items[j] = readers[j].readSortableRecord ( ) ;
          if ( items[j] != null ) {
            counts[j] = 1 ;
            allDone = false ;
          }
        }
        if ( allDone ) { break ; }
        while ( true ) {
          //  Determine which is the next Record to add to the output stream.  If there isn't one then we
          //  get a negative index and we can terminate the loop.  If we do not terminate then there was a
          //  Record and we must write it out.
          final int index = findAppropriate ( items , rInfo.getComparator ( ) ) ;
          if ( index < 0 ) { break ; }
          writers[i].writeSortableRecord ( items[index] ) ;
          if ( i > 0 ) {
            //  We have not yet reduced the problem to only a single file so there must be at least one more
            //  iteration -- we know when we are finished when everything goes into a single file.
            returnValue = true ;
          }
          //  Draw a new Record from the file whose Record was chosen -- unless of course we have
          //  finished our quota from that file.
          if ( counts[index] < blockSize ) {
            items[index] = readers[index].readSortableRecord ( ) ;
            if ( items[index] != null ) { ++counts[index] ; }
          } else  { items[index] = null ; }
        }
      }
    }
    //  Be tidy, close all the files.  Actually this is essential to ensure there is a flush.
    for ( final SortableRecordWriter writer : writers ) { writer.close ( ) ; }
    for ( final SortableRecordReader reader : readers ) { reader.close ( ) ; }
    return returnValue ;
  }
  /**
   *  Determine which Record is the one to be output next.
   *
   *  @param items the array of <code>SortableRecords</code> from which to select the next according to the
   *  order relation defined by <code>c</code>.
   *  @param c the <code>Comparator&lt;SortableRecord></code> defining the required order relation on the
   *  <code>SortableRecord</code>s.
   *  @return the index in the array of the item that should be chosen next.
   */
  private static int findAppropriate ( final SortableRecord[] items , final Comparator<SortableRecord> c ) {
    //  Assume no output is to be done and then find the first non-empty entry.
    int index = -1 ;
    for ( int i = 0 ; i < items.length ; ++i ) {
      if ( items[i] != null ) { index = i ; break ; }
    }
    //  If there were no non-empty entries then do nothing, we are finshied.  Otherwise...
    if ( index >= 0 )  {
      //  ...do a linear search through the items to see which is the next one to select.
      SortableRecord value = items[index] ;
      for ( int i = index+1 ; i < items.length ; ++i ) {
        if ( items[i] != null ) {
          if ( c.relation ( items[i] , value ) ) {
            index = i ;
            value = items[i] ;
          }
        }
      }
    }
    return index ;
  }
}
