/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.BinaryTree ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.OrderedBinaryTree ;
import org.devjavasoft.ads.Sequence ;
/**
 *  Sort a sequence of <code>T</code>s using Binary Search Tree Sort.  This is an O(n.log(n)) sort in
 *  general but if the data is nearly sorted can be as bad as O(n^2).  This problem would be cured by using
 *  a Red-Black Tree or AVL Tree (i.e. a self-balancing tree) instead of the <code>OrderBinaryTree</code>.
 *
 *  @author Russel Winder
 */
public class BinaryTreeSort {
  private BinaryTreeSort ( ) { }
  /**
   *  The sort operation for primitive arrays.
   *
   *  @param a the primitive array to be sorted.
   *  @param c the <code>Comparator&lt;? super T></code> used to compare the <code>T</code> during the sort
   *  process.
   */
  public static <T> void sort ( final T[] a , final Comparator<? super T> c ) {
    final BinaryTree<T> tree = new OrderedBinaryTree<T> ( c ) ;
    for ( final T anA : a ) { tree.add ( anA ) ; }
    int index = 0 ;
    for ( InputIterator<T> i = tree.inorderTraversal ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) { a[index++] = i.get ( ) ; }
  }
  /**
   *  The sort operation for sequences.
   *
   *  @param s the <code>Sequence</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code> during the sort process.
   */
  public static <T> void sort ( final Sequence<T> s , final Comparator<? super T> c ) {
    final BinaryTree<T> tree = new OrderedBinaryTree<T> ( c ) ;
    for ( InputIterator<T> i = s.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) { tree.add ( i.get ( ) ) ; }
    s.makeEmpty ( ) ;
    for ( InputIterator<T> i = tree.inorderTraversal ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) { s.add ( i.get ( ) ) ; }
  }
}
