/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.BiDirectionalIterator ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.ForwardIterator ;
import org.devjavasoft.ads.InvalidOperationException ;
import org.devjavasoft.ads.Sequence ;
import org.devjavasoft.ads.SLList ;
/**
 *  Sort a sequence of <code>T</code>s using Bubble Sort.  This is an O(n^2) sort in general but, due
 *  to optimization, is extremely efficient on almost sorted data.
 *
 *  @author Russel Winder
 */
public class BubbleSort {
  private BubbleSort ( ) { }
  /**
   *  The sort operation for primitive arrays.
   *
   *  @param a the primitive array to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   */
  public static <T> void sort ( final T[] a , final Comparator<? super T> c ) {
    if ( a.length <= 1 ) { return ; }
    for ( int i = a.length-1 ; i > 0 ; --i ) {
      //  Iterate from the beginning of the array to the location of interest.  Keep track of whether we
      //  make any swaps.
      boolean swapped = false ;
      for ( int j = 0 ; j < i ; ++j ) {
        //  Check to see if this pair of elements is in the right order.  If they are not then swap them.
        if ( c.relation ( a[j+1] , a[j] ) ) {
          final T temp = a[j] ;
          a[j] = a[j+1] ;
          a[j+1] = temp ;
          swapped = true ;
        }
      }
      //  If we made no swaps then the data is already sorted so terminate.
      if ( ! swapped ) { break ; }
    }
  }
  /**
   *  The sort operation for sequences.
   *
   *  @param s the <code>Sequence</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code> during the sort process.
   */
  public static <T> void sort ( final Sequence<T> s , final Comparator<? super T> c ) {
    if ( ( s instanceof SLList<?> ) ) { throw new InvalidOperationException ( ) ; }
    if ( s.size ( ) <= 1 ) { return ; }
    //
    //  TODO : Use ReverseIterator here so as to be able to deal with SLList.
    //
    final BiDirectionalIterator<T> i = (BiDirectionalIterator<T>) s.end ( ) ;
    do {
      i.retreat ( ) ;
      boolean swapped = false ;
      final ForwardIterator<T> j = (ForwardIterator<T>) s.begin ( ) ;
      while ( ! j.equals ( i ) ) {
        final ForwardIterator<T> k = j.clone ( ) ;
        j.advance ( ) ;
        final T a = j.get ( ) ;
        final T b = k.get ( ) ;
        if ( c.relation ( a , b ) ) {
          j.set ( b ) ;
          k.set ( a ) ;
          swapped = true ;
        }
      }
      if ( ! swapped ) { break ; }
    } while ( ! i.atBegin ( ) ) ;
  }
}
