/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.Sequence ;
/**
 *  Sort a sequence of <code>Integer</code>s using Count Sort (also occasionally heard of as Pigeonhole Sort
 *  or even Bucket Sort).  This is an O(n) sort.
 *
 *  <p> The implementations are very simple and straighforward (they could do with being made more
 *  sophisticated).  They assume that the data are always positive integers.</p>
 *
 *  <p>These sort methods are not generic as it is only possible to apply this sort to integer types
 *  (<code>Integer</code>, <code>Long</code>, etc.) and these types are final and so cannot be subclassed.
 *  We provide here implementation only for <code>Integer</code>.  There would have to be a separate
 *  implementation for <code>Long</code> as there is no common superclass that is useful &mdash;
 *  <code>Number</code> is not sufficiently expressive and would include the floating point types which have
 *  to be excluded for this sort to work.</p>
 *
 *  @author Russel Winder
 */
public final class CountSort {
  private CountSort ( ) { }
  /**
   *  The sort operation for primitive arrays.
   *
   *  @param a the primitive array to be sorted.
   */
  public static  void sort ( final Integer[] a ) {
    //  Set up the counters.
    final int n = findMaximum ( a ) ;
    final int[] counts = new int [n + 1] ;
    //  For each datum in the primitive array increment the frequency count in the counter.
    for ( final Integer anA : a ) { ++counts[ anA ] ; }
    //  Create the sorted list by working through each counter and outputting into the array the number of
    //  of items of that value.
    int dataIndex = 0 ;
    for ( int i = 0 ; i < counts.length ; ++i ) {
      for ( int j = 0 ; j < counts[i] ; ++j ) { a[dataIndex++] = i ; }
    }
  }
  /**
   *  @param a the primitive array of value from which to find the largest.
   *  @return the maximum value in the primitive array that is the parameter.
   */
  private static int findMaximum ( final Integer[] a ) {
    int max = 0 ;
    for ( final Integer testValue : a ) {
      if ( testValue > max ) { max = testValue; }
    }
    return max ;
  }
  /**
   *  The sort operation for sequences.
   *
   *  @param s the <code>Sequence&lt:Integer></code> to be sorted.
   */
  public static void sort ( final Sequence<Integer> s ) {
    final int n = findMaximum ( s ) ;
    final int[] counts = new int [n + 1] ;
    for ( InputIterator<Integer> i = s.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) { ++counts[i.get ( )] ; }
    s.makeEmpty ( ) ;
    for ( int i = 0 ; i < counts.length ; ++i ) {
      for ( int j = 0 ; j < counts[i] ; ++j ) { s.add ( i ) ; }
    }
  }
  /**
   *  @param s the <code>Sequence&lt;Integer></code> of values from which to find the largest.
   *  @return the maximum value in the <code>Sequence&lt:Integer></code> that is the parameter.
   */
  private static int findMaximum ( final Sequence<Integer> s ) {
    int max = 0 ;
    for ( InputIterator<Integer> i = s.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) {
      final int testValue = i.get ( ) ;
      if ( testValue > max ) { max = testValue ; }
    }
    return max ;
  }
}
