/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.Heap ;
/**
 *  Sort an <code>Array</code> using Heap Sort.  This is guaranteed to be an O(n.log(n)) sort.
 *
 *  <p> NB This produces a result in reverse order to all the other sort algorithms given the same
 *  <code>Comparator&lt? super T></code>!</p>
 *
 *  @author Russel Winder
 */
public class HeapSortArray {
  private HeapSortArray ( ) { }
  /**
   *  The sort operation.
   *
   *  @param a the <code>Array</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   */
  public static <T> void sort ( final Array<T> a , final Comparator<? super T> c ) {
    // Apply the heap ordering property to the Array.
    Heap.makeHeap ( a , c ) ;
    //  For each item in the Array, move the top value on the Heap to the end position and percolate
    //  the value we swapped down the heap.  This guarantees that the appropriate value is at the very end
    //  of the heap so we can ignore it in the next iteration.
    for ( int i = a.size ( ) - 1 ; i > 0 ; --i ) {
      final T o = a.get ( 0 ) ;
      a.set ( 0 , a.get ( i ) ) ;
      a.set ( i , o ) ;
      Heap.percolateDown ( a , i , 0 , c ) ;
    }
    //  The sort order is the reverse of all the sort algorithms in this package at this point so we should
    //  reverse the Array<T> even though it takes time -- O(n).
    a.reverse ( ) ;
  }
}
