/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.DLList ;
import org.devjavasoft.ads.InvalidOperationException ;
/**
 *  Sort a sequence of <code>T</code>s using Insertion Sort.  This is an O(n^2) sort in general.
 *
 *  @author Russel Winder
 */
public class InsertionSort {
  private InsertionSort ( ) { }
  /**
   *  The sort operation for primitive arrays.  This version is optimized to cut down the number of
   *  assignments.
   *
   *  @param a the primitive array of <code>T</code>s to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   */
  public static <T> void sort ( final T[] a , final Comparator<? super T> c ) {
    for ( int i = 0 ; i < a.length - 1 ; ++i ) {
      //  If the current item is in the right place, do nothing, otherwise percolate it to the right place.
      if ( c.relation ( a[ i + 1 ] , a[i] ) ) {
        //  Remember the current value, and percolate the `gap' to the right place in the sorted part of the
        //  array.
        final T temp = a[ i + 1 ] ;
        int j = i ;
        while ( true ) {
          //  Move the space back down the sorted part of the array by one, by copying values.  If we reach
          //  the end of the array or find the right place, stop the loop.
          a[j+1] = a[j] ;
          if ( ( j == 0 ) || c.relation ( a[j-1] , temp ) ) { break ; }
          --j ;
        }
        //  We have got the `gap' to the right place so drop in the value.
        a[j] = temp ;
      }
    }
  }
  /**
   *  The sort operation for an <code>Array</code>.
   *
   *  @param a the <code>Array</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code> during the sort process.
   */
  public static <T> void sort ( final Array<T> a , final Comparator<? super T> c ) {
    if ( a.size ( ) <= 1 ) { throw new InvalidOperationException ( ) ; }
    for ( int i = 0 ; i < a.size ( ) - 1 ; ++i ) {
      if ( c.relation ( a.get ( i + 1 ) , a.get ( i ) ) ) {
        final T temp = a.get ( i + 1 ) ;
        int j = i ;
        while ( true ) {
          a.set ( j + 1 , a.get ( j ) ) ;
          if ( ( j == 0 ) || c.relation ( a.get ( j - 1 ) , temp ) ) { break ; }
          --j ;
        }
        a.set ( j , temp ) ;
      }
    }
  }
  /**
   *  The sort operation for a <code>DLList</code>.
   *
   *  @param d the <code>DLList</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code> during the sort process.
   */
  public static <T> void sort ( final DLList<T> d , final Comparator<? super T> c ) {
    if ( d.size ( ) <= 1 ) { throw new InvalidOperationException ( ) ; }
    final DLList<T>.Iterator i = d.iterator ( ) ;
    while ( true ) {
      final DLList<T>.Iterator j = i.clone ( ) ;
      i.advance ( ) ;
      if ( i.atEnd ( ) ) { break ; }
      if ( c.relation ( i.get ( ) , j.get ( ) ) ) {
        final T temp = i.get ( ) ;
        i.remove ( ) ;  // This moves the iterator on one which we...
        i.retreat ( ) ;  // ...have to undo.
        while ( true ) {
          if ( j.atBegin ( ) ) { j.insert ( temp ) ; break ; }
          j.retreat ( ) ;
          if ( c.relation ( j.get ( ) , temp ) ) { j.add ( temp ) ; break ; }
        }
      }
    }
  }
}
