/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.Sequence ;
import org.devjavasoft.ads.SLList ;
/**
 *  Sort a sequence of <code>T</code>s using Merge Sort.  This is an O(n.log(n)) sort.
 *
 *  @author Russel Winder
 */
public class MergeSort {
  private MergeSort ( ) { }
  /**
   *  The sort operation for a primitive array.
   *
   *  @param a the primitive array to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code> during the sort process.
   */
  @SuppressWarnings ( "unchecked" )
  public static <T> void sort ( final T[] a , final Comparator<? super T> c ) {
    if ( a.length > 1 ) {
      //  Find the length of the data.  Arbitrarily split the data into two arrays by simply splitting in
      //  two.  Then recursively sort the two arrays and when that has happened merge the two arrays into a
      //  single result array.
      final int xLength = ( a.length - 1 ) / 2 + 1 ;
      final int yLength = a.length - xLength ;
      //  Although this shortcut is frowned upon and is indeed can lead to situations that are not typesafe,
      //  we use it anyway as we know that the arrays are only ever used internally so the type unsafeness
      //  cannot be observed by client code.
      final T[] x = (T[]) new Object[xLength] ;
      final T[] y = (T[]) new Object[yLength] ;
      System.arraycopy ( a , 0 , x , 0 , xLength ) ;
      System.arraycopy ( a , xLength , y , 0 , yLength ) ;
      sort ( x , c ) ;
      sort ( y , c ) ;
      merge ( a , x , y , c ) ;
    }
  }
  /**
   *  Merge two source primitive arrays which are pre-sorted to a target array.  The <code>Comparator</code>
   *  is used to select which element from the two queues to take next.
   *
   *  @param target the array into which to copy things.
   *  @param a one of the two source arrays.
   *  @param b the second of the two source arrays.
   *  @param c the <code>Comparator</code> used to compare the objects to determine the correct selection
   *  policy.
   */
  private static <T> void merge ( final T[] target , final T[] a , final T[] b , final Comparator<? super T> c ) {
    int t_index = 0 ;
    int a_index = 0 ;
    int b_index = 0 ;
    while ( true ) {
      //  a is used up so just copy b and finish.
      if ( a_index >= a.length ) {
        System.arraycopy ( b , b_index , target , t_index , b.length - b_index ) ;
        break ;
      }
      //  b is used up so just copy a and finish.
      if ( b_index >= b.length ) {
        System.arraycopy ( a , a_index , target , t_index , a.length - a_index ) ;
        break ;
      }
      //  Select the appropriate element from the two arrays and go to the next iteration.
      target[t_index++] = c.relation ( a[a_index] , b[b_index] )  ? a[a_index++] : b[b_index++] ;
    }
  }
  /**
   *  The sort operation for <code>SLList</code>s.
   *
   *  @param s the <code>SLList</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code> during the sort process.
   */
  public static <T> void sort ( final SLList<T> s , final Comparator<? super T> c ) {
    if ( s.size() > 1 ) {
      final int splitIndex = ( s.size ( ) - 1) / 2 + 1 ;
      final SLList<T> s1 = new SLList<T> ( s , 0 , splitIndex ) ;
      final SLList<T> s2 = new SLList<T> ( s , splitIndex , s.size ( ) ) ;
      sort ( s1 , c ) ;
      sort ( s2 , c ) ;
      s.makeEmpty ( ) ;
      merge (s , s1 , s2 , c ) ;
    }
  }
  /**
   *  The sort operation for <code>Sequences</code>s other than <code>SLList</code>s.  Copies the elements
   *  into an <code>SLList</code>, does the sort and then copies things back.  The extra copying is a little
   *  inefficient but we make savings during the actual sorting working with <code>SLList</code>s.
   *
   *  @param s the <code>Sequence</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   */
  public static <T> void sort ( final Sequence<T> s , final Comparator<? super T> c ) {
    final SLList<T> sl = new SLList<T> ( s ) ;
    sort ( sl , c ) ;
    s.makeEmpty ( ) ;
    for ( InputIterator<T> i = sl.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) ) { s.add ( i.get ( ) ) ; }
  }
  /**
   *  Merge two source lists which are pre-sorted to a target list.  The <code>Comparator</code> is used to
   *  select which element from the two queues to take next.
   *
   *  @param target the <code>SLList</code> into which to copy things.
   *  @param a one of the two source <code>SLList</code>s.
   *  @param b the second of the two source <code>SLList</code>s.
   *  @param c the <code>Comparator</code> used to compare the objects to determine the correct selection
   *  policy.
   */
  private static <T> void merge (final SLList<T> target , final SLList<T> a , final SLList<T> b , final Comparator<? super T> c ) {
    while ( true ) {
      if ( a.isEmpty ( ) ) { target.merge ( b ) ; break ; }
      if ( b.isEmpty ( ) ) { target.merge ( a ) ; break ; }
      final T o_a = a.get ( 0 ) ;
      final T o_b = b.get ( 0 ) ;
      if ( c.relation ( o_a , o_b ) ) {
        target.add ( o_a ) ;
        a.remove ( 0 ) ;
      } else {
        target.add ( o_b ) ;
        b.remove ( 0 ) ;
      }
    }
  }
}
