/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import java.io.File ;
import java.io.FileNotFoundException ;
import java.io.IOException ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.InputIterator ;
/**
 *  A function object delivering a polyphase merge sort of a file on the filestore.
 *
 *  @see SortableRecord
 *  @see SortableRecordReader
 *  @see SortableRecordWriter
 *  @see SortableRecordInformation
 *  @see SortableRecordCopyFile
 *  @author Russel Winder
 */
public final class PolyphaseMergeSort {
  private PolyphaseMergeSort ( ) { }
  /**
   *  The sort operation with some defaults.
   *
   *  @param fileName the <code>String</code> giving the name of the file to be sorted.
   *  @param r the <code>SortableRecordInformation</code> factory object for creating
   *  <code>SortableRecordReader</code>s and <code>SortableRecordWriters</code>s and able to deliver a
   *  <code>Comparator&lt;SortableRecord></code>.
   *  @exception FileNotFoundException if <code>fileName</code> cannot be found.
   *  @exception IOException if there are any errors opening, reading or writing <code>fileName</code>.
   */
  public static void sort ( final String fileName, final SortableRecordInformation r )
    throws FileNotFoundException , IOException {
    sort ( fileName , 20 , 2 , r ) ;
  }
  /**
   *  The sort operation.
   *
   *  @param fileName the <code>String</code> giving the name of the file to be sorted.
   *  @param approximateBlockSize the number of data items in the initial sorted blocks.
   *  @param numberOfFiles the number of files to use for initial dispersion.
   *  @param rInfo the <code>SortableRecordInformation</code> factory object for creating
   *  <code>SortableRecordReader</code>s and <code>SortableRecordWriters</code>s and able to deliver a
   *  <code>Comparator&lt;SortableRecord></code>.
   *  @exception FileNotFoundException if <code>fileName</code> cannot be found.
   *  @exception IOException if there are any errors opening, reading or writing <code>fileName</code>.
   */
  public static void sort (final String fileName , final int approximateBlockSize , int numberOfFiles , final SortableRecordInformation rInfo )
    throws FileNotFoundException, IOException {
    //  Forceably stick to 3 files for now instead of just having one more than the number of input files.
    //
    //numberOfFiles++ ;
    numberOfFiles = 3 ;
    //  Create all the files needed for the sorting.
    final File file = new File ( fileName ) ;
    final File[] temp = new File [numberOfFiles] ;
    for ( int i = 0 ; i < numberOfFiles ; ++i ) { temp[i] = new File ( "tmp_" + i ) ; }
    //  Calculate the block size.  Must get things into the Fibonacci series for it to work
    //  properly.  Have to run through the file to find the number of records.  We need to copy the file
    //  anyway so this is not a wasted activity.  We use the array F to calculate the Fibonacci numbers as
    //  we go, calculating the initialBlockSize to best fit the nearest Fibonacci number.
    int indexOfNumberOfBlocks = 1 ;
    int initialBlockSize = 1 ;
    final int numberOfRecords = SortableRecordCopyFile.copy(file, temp[0], rInfo) ;
    final int[] F = new int [numberOfRecords] ;
    for ( int i = 0 ; i < numberOfFiles ; ++i ) { F[i] = 1 ; }
    for ( int i = numberOfFiles ; i < numberOfRecords ; ++i ) {
      F[i] = 0 ;
      for ( int j = i - 1 ; j > i-numberOfFiles ; --j ) { F[i] += F[j] ; }
      initialBlockSize = numberOfRecords / F[i] ;
      if ( initialBlockSize < approximateBlockSize ) {
        indexOfNumberOfBlocks = i - 1 ;
        break ;
      }
    }
    while ( true ) {
      if (++initialBlockSize * F[indexOfNumberOfBlocks] > numberOfRecords) { break ; }
    }
    //  Ceate the support arrays containing current block size and block count in the various files.
    final int[] blockSizes = new int [numberOfFiles] ;
    final int[] blockCounts = new int [numberOfFiles] ;
    blockSizes[0] = 0 ;
    blockCounts[0] = 0 ;
    for ( int i = 1, j = indexOfNumberOfBlocks-1 ; i < numberOfFiles ; ++i, --j ) {
      blockSizes[i] = initialBlockSize ;
      blockCounts[i] = F[j] ;
    }
    //  Create the files of blocks of sorted records.
    distributeSortedBlocks ( temp , 0 , initialBlockSize , blockCounts , rInfo ) ;
    //  Set up the file readers for all the files.
    final SortableRecordReader[] readers = new SortableRecordReader[numberOfFiles] ;
    for ( int i = 0 ; i < numberOfFiles ; ++i ) {
      readers[i] = rInfo.getSortableRecordReader ( temp[i] ) ;
    }
    while ( true ) {
      //  Check what work there is to do.  If there is, find out which is the empty file.
      int toIndex = -1 ;
      int numberOfNonEmptyFiles = 0 ;
      int indexOfNonEmptyFile = -1 ;
      for ( int i = 0 ; i < numberOfFiles ; ++i ) {
        if ( blockCounts[i] == 0 ) { toIndex = i ; }
        else {
          indexOfNonEmptyFile = i ;
          ++numberOfNonEmptyFiles ;
        }
      }
      //  Exit if everthing is done but close all the files and copy the result back before exiting.
      if ( numberOfNonEmptyFiles <= 1 ) {
        for ( int i = 0 ; i < numberOfFiles ; ++i ) { readers[i].close ( ) ; }
        SortableRecordCopyFile.copy ( temp[indexOfNonEmptyFile] , file , rInfo ) ;
        for ( int i = 0 ; i < numberOfFiles ; ++i ) { temp[i].delete ( ) ; }
        break ;
      }
      //  Perform the next round of merging.
      readers[toIndex].close ( ) ;
      final SortableRecordWriter writer = rInfo.getSortableRecordWriter ( temp[toIndex] ) ;
      merge ( readers , writer , toIndex , blockSizes , blockCounts , rInfo ) ;
      writer.close ( ) ;
      readers[toIndex] = rInfo.getSortableRecordReader ( temp[toIndex] ) ;
    }
  }
  /**
   *  Perform the initial dispersion of the data.
   *
   *  @param files the set of <code>Files</code>s being used for the merge sort.
   *  @param fromIndex  the index into <code>files</code> of the source, all the others are sinks.
   *  @param blockSize the size of a block.
   *  @param blockCounts the block counts.
   *  @param rInfo the <code>SortableRecordInformation</code> factory object for creating
   *  <code>SortableRecordReader</code>s and <code>SortableRecordWriters</code>s and able to deliver a
   *  <code>Comparator&lt;SortableRecord></code>.
   *  @exception FileNotFoundException if <code>fileName</code> cannot be found.
   *  @exception IOException if there are any errors opening, reading or writing <code>fileName</code>.
   */
  private static void distributeSortedBlocks (final File[] files , final int fromIndex , final int blockSize , final int[] blockCounts , final SortableRecordInformation rInfo )
    throws FileNotFoundException, IOException {
    //  Create a Reader for the original data and a set of Writers for the output files.
    final SortableRecordReader reader = rInfo.getSortableRecordReader ( files[fromIndex] ) ;
    final SortableRecordWriter[] writers = new SortableRecordWriter [files.length] ;
    for ( int i = 0 ; i < files.length ; ++i ) {
      writers[i] = i == fromIndex ? null : rInfo.getSortableRecordWriter ( files[i] ) ;
    }
    for ( int i = 0 ; i < writers.length ; ++i ) {
      if ( i != fromIndex ) {
        for ( int j = 0 ; j < blockCounts[i] ; ++j ) {
          //  Pull in a few records, put them into an Array that is where we are performing the internal
          //  sort that creates us the sorted block.  Do the sort then write out the block.
          final Array<SortableRecord> a = new Array<SortableRecord> () ;
          for ( int k = 0 ; k < blockSize ; ++k ) {
            final SortableRecord r = reader.readSortableRecord ( ) ;
            if ( r == null ) { break ; }
            a.add ( r ) ;
          }
          Quicksort.sort ( a , rInfo.getComparator ( ) , rInfo.getEqualToComparator ( ) ) ;
          for ( InputIterator<SortableRecord> iter = a.iterator ( ) ; ! iter.atEnd ( ) ; iter.advance ( ) ) {
            writers[i].writeSortableRecord ( iter.get ( ) ) ;
          }
        }
      }
    }
    //  Be tidy and close all the files.  Actually this is essential to ensure we get a flush.
    for ( int i = 0 ; i < writers.length ; ++i ) {
      if ( i != fromIndex ) { writers[i].close ( ) ; }
    }
    reader.close ( ) ;
  }
  /**
   *  Undertake a round of merging.
   *
   *
   *  @param readers the readers to take items from for merge.
   *  @param writer the writer to send merged items to.
   *  @param toIndex the index to...
   *  @param blockSizes the block sizes.
   *  @param blockCounts the block counts.
   *  @param rInfo the <code>SortableRecordInformation</code> factory object for creating
   *  <code>SortableRecordReader</code>s and <code>SortableRecordWriters</code>s and able to deliver a
   *  <code>Comparator&lt;SortableRecord></code>.
   *  @exception FileNotFoundException if <code>fileName</code> cannot be found.
   *  @exception IOException if there are any errors opening, reading or writing <code>fileName</code>.
   */
  private static void merge ( final SortableRecordReader[] readers , final SortableRecordWriter writer , final int toIndex , final int[] blockSizes , final int[] blockCounts , final SortableRecordInformation rInfo )
    throws FileNotFoundException, IOException {
    final SortableRecord[] items = new SortableRecord[readers.length] ;
    final int[] counts = new int[readers.length] ;
    int numberOfBlocksMerged = 0 ;
    while ( true ) {
      boolean allDone = false ;
      for ( int i = 0 ; i < readers.length ; ++i ) {
        counts[i] = 0 ;
        if ( i == toIndex ) { items[i] = null ; }
        else {
          readers[i].mark ( 64 ) ; // TODO:  Why 64?
          items[i] = readers[i].readSortableRecord ( ) ;
          if ( items[i] == null ) {
            for ( int j = 0 ; j < i ; ++j ) {
              if ( j != toIndex ) { readers[j].reset ( ) ; }
            }
            allDone = true ;
            break ;
          } else { counts[i] = 1 ; }
        }
      }
      if ( allDone ) { break ; }
      ++numberOfBlocksMerged ;
      while ( true ) {
        final int i = findAppropriate ( items , toIndex , rInfo.getComparator ( ) ) ;
        if ( i < 0 ) { break ; }
        writer.writeSortableRecord ( items[i] ) ;
        if ( counts[i] < blockSizes[i] ) {
          items[i] = readers[i].readSortableRecord ( ) ;
          if ( items[i] != null ) { ++counts[i] ; }
        } else { items[i] = null ; }
      }
    }
    blockSizes[toIndex] = 0 ;
    for ( int i = 0 ; i < readers.length ; ++i ) {
      if ( i != toIndex ) { blockSizes[toIndex] += blockSizes[i] ; }
    }
    for ( int i = 0 ; i < readers.length ; ++i ) { blockCounts[i] -= numberOfBlocksMerged ; }
    blockCounts[toIndex] = numberOfBlocksMerged ;
  }
  /**
   *  Determine which <code>SortableRecord</code> is the one to be output next.
   *
   *  @param items the array of <code>SortableRecords</code> from which to select the next according to the
   *  order relation defined by <code>c</code>.
   *  @param toIndex the index into the array of the target.  The others are assumed to be sources.
   *  @param c the <code>Comparator&lt;SortableRecord></code> defining the required order relation on the
   *  <code>SortableRecord</code>s.
   *  @return the index in the array of the item that should be chosen next.
   */
  private static int findAppropriate(final SortableRecord[] items , final int toIndex , final Comparator<SortableRecord> c ) {
    //  Assume no output is to be done and then find the first non-empty entry.
    int index = -1 ;
    for ( int i = 0 ; i < items.length ; ++i ) {
      if ( i != toIndex ) {
        if ( items[i] != null ) { index = i ; break ; }
      }
    }
    //  If there were no non-empty entries then do nothing, we are finshied.  Otherwise...
    if ( index >= 0 )  {
      //  ...do a linear search through the items to see which is the next one to select.
      SortableRecord value = items[index] ;
      for ( int i = index+1 ; i < items.length ; ++i ) {
        if ( i != toIndex ) {
          if ( items[i] != null ) {
            if ( c.relation ( items[i] , value ) ) {
              index = i ;
              value = items[i] ;
            }
          }
        }
      }
    }
    return index ;
  }
}
