/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.BiDirectionalIterator ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.DLList ;
/**
 *  Sort a sequence using Quicksort.  This is an O(n.log(n)) sort except when the data is sorted in which
 *  case it O(n^2).
 *
 *  @author Russel Winder
 */
public class Quicksort {
  private Quicksort ( ) { }
  /**
   *  The sort operation for primitive arrays.
   *
   *  @param a the array to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   */
  public static <T> void sort ( final T[] a , final Comparator<? super T> c ) { quicksort ( a , 0 , a.length - 1 , c ) ; }
  /**
   *  Given the array and two indices, swap the two items in the array.
   *
   *  @param a the array of values.
   *  @param x index of item to swap.
   *  @param y index of item to swap.
   */
  private static <T> void swap ( final T[] a , final int x , final int y ) {
    final T temp = a[x] ;
    a[x] = a[y] ;
    a[y] = temp ;
  }
  /**
   *  Partition a primitive array in two using the pivot value that is at the centre of the array being
   *  partitioned.
   *
   *  <p>This partition implementation is an invention of Russel Winder first published in Winder, R (1993)
   *  "Developing C++ Software", Wiley, p.395.  This implementation (unlike most others) does not guarantee
   *  that the split point contains the pivot value.  Unlike other implementations, it requires only < (or
   *  >) relation and not both < and <= (or > and >=).  Also, it seems easier to program and to
   *  comprehend.</p>
   *
   *  @param a the primitive array out of which to take a slice.
   *  @param lower the lower bound of this slice.
   *  @param upper the upper bound of this slice.
   *  @param c the <code>Comparator</code> to be used to define the order.
   *  @return the index of the pivot item.
   */
  private static <T> int partition ( final T[] a , int lower , int upper , final Comparator<? super T> c ) {
    final T pivotValue = a[( upper + lower + 1 ) / 2] ;
    while ( lower <= upper ) {
      while ( c.relation ( a[lower] , pivotValue ) ) { ++lower ; }
      while ( c.relation ( pivotValue , a[upper] ) ) { --upper ; }
      if ( lower <= upper ) {
        if ( lower < upper ) { swap ( a , lower , upper ) ; }
        ++lower ;
        --upper ;
      }
    }
    return upper ;
  }
  /**
   *  The recursive Quicksort function.
   *
   *  @param a the primitive array out of which to take a slice.
   *  @param lower the lower bound of this slice.
   *  @param upper the upper bound of this slice.
   *  @param c the <code>Comparator</code> to be used to define the order.
   */
  private static <T> void quicksort ( final T[] a , final int lower , final int upper , final Comparator<? super T> c ) {
    final int sliceLength = upper - lower + 1 ;
    if ( sliceLength > 1 ) {
      if ( sliceLength == 2 ) {
        if ( c.relation ( a[upper] , a[lower] ) ) { swap (a, lower, upper) ; }
      }
      else {
        //  This partition implementation does not guarantee that the split point contains the pivot value
        //  so we cannot assume that the pivot is between the two slices.
        final int pivotIndex = partition ( a , lower , upper , c ) ;
        quicksort ( a , lower , pivotIndex , c ) ;
        quicksort ( a , pivotIndex + 1 , upper , c ) ;
      }
    }
  }
  /**
   *  The sort operation on <code>Array</code>s.
   *
   *  @param a the <code>Array</code> to be sorted.
   *  @param cr the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   *  @param ce the <code>Comparator</code> defining value equality.
   */
  public static <T> void sort ( final Array<T> a , final Comparator<? super T> cr , final Comparator<? super T> ce ) {
    sort ( a , 0 , a.size ( ) - 1 , cr ) ;
  }
  /**
   *  The sort operation on <code>DLList</code>s.
   *
   *  @param d the <code>DLList</code> to be sorted.
   *  @param cr the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   *  @param ce the <code>Comparator</code> defining value equality.
   */
  public static <T> void sort ( final DLList<T> d , final Comparator<? super T> cr , final Comparator<? super T> ce ) {
    final BiDirectionalIterator<T> theEnd = d.end ( ) ;
    theEnd.retreat ( ) ;
    sort ( d.begin ( ) , theEnd , cr , ce ) ;
  }
  /**
   *  The sort operation for handling a slice of an <code>Array</code>.
   *
   *  @param a the <code>Array</code> out of which to take a slice.
   *  @param lower the lower bound of this slice.
   *  @param upper the upper bound of this slice.
   *  @param c the <code>Comparator</code> to be used to define the order.
   */
  private static <T> void sort ( final Array<T> a , final int lower , final int upper , final Comparator<? super T> c ) {
    final int sliceLength = upper-lower + 1 ;
    if ( sliceLength > 1 ) {
      if ( sliceLength == 2 ) {
        if ( c.relation ( a.get ( upper ) , a.get ( lower ) ) ) { swap ( a , lower , upper ) ; }
      }
      else {
        //  This partition implementation does not guarantee that the split point contains the pivot value
        //  so we cannot assume that the pivot is between the two slices.
        final int pivotIndex = partition ( a , lower , upper , c ) ;
        sort ( a , lower , pivotIndex , c ) ;
        sort ( a , pivotIndex + 1 , upper , c ) ;
      }
    }
  }
  /**
   *  The sort operation for <code>BiDirectionalIterator</code>s.
   *
   *  @param b iterator positioned at the bottom of the slice.
   *  @param t iterator positioned at the top of the slice.
   *  @param cr the <code>Comparator</code> to be used to define the order.
   *  @param ce the <code>Comparator</code> defining value equality.
   */
  private static <T> void sort ( final BiDirectionalIterator<T> b , final BiDirectionalIterator<T> t , final Comparator<? super T> cr , final Comparator<? super T> ce ) {
    if ( ! b.equals ( t ) ) {
      final BiDirectionalIterator<T> pivot = partition ( b , t , cr , ce ) ;
      if ( ! pivot.equals ( b ) ) {
        pivot.retreat ( ) ;
        sort ( b , pivot , cr , ce ) ;
        pivot.advance ( ) ;
      }
      if ( ! pivot.equals ( t ) ) {
        pivot.advance ( ) ;
        sort ( pivot , t , cr , ce ) ;
      }
    }
  }
  /**
   *  Given the array and two indices, swap the two items in the array.
   *
   *  @param a the <code>Array</code> of values.
   *  @param x index of item to swap.
   *  @param y index of item to swap.
   */
  private static <T> void swap ( final Array<T> a , final int x , final int y ) {
    final T temp = a.get ( x ) ;
    a.set ( x , a.get ( y ) ) ;
    a.set ( y , temp ) ;
  }
  /**
   *  Given two iterators swap the two items.
   *
   *  @param a the iterator positioned at one item to swap..
   *  @param b the iterator positioned at the other item to swap.
   */
  private static <T> void swap ( final BiDirectionalIterator<T> a , final BiDirectionalIterator<T> b ) {
    final T temp = a.get ( ) ;
    a.set ( b.get ( ) ) ;
    b.set ( temp ) ;
  }
  /**
   *  Partition an <code>Array</code> in two using the pivot value that is at the centre of the array being
   *  partitioned.
   *
   *  <p>This partition implementation is an invention of Russel Winder first published in Winder, R (1993)
   *  "Developing C++ Software", Wiley, p.395.  This implementation (unlike most others) does not guarantee
   *  that the split point contains the pivot value.  Unlike other implementations, it requires only < (or
   *  >) relation and not both < and <= (or > and >=).  Also, it seems easier to program and to
   *  comprehend.</p>
   *
   *  @param a the <code>Array</code> out of which to take a slice.
   *  @param lower the lower bound of this slice.
   *  @param upper the upper bound of this slice.
   *  @param c the <code>Comparator</code> to be used to define the order.
   *  @return the index of the pivot value.
   */
  private static <T> int partition ( final Array<T> a , int lower , int upper , final Comparator<? super T> c ) {
    final T pivotValue = a.get ( ( upper + lower + 1 ) / 2 ) ;
    while ( lower <= upper ) {
      while ( c.relation ( a.get ( lower ) , pivotValue ) ) { ++lower ; }
      while ( c.relation ( pivotValue , a.get ( upper ) ) ) { --upper ; }
      if ( lower <= upper ) {
        if ( lower < upper ) { swap ( a , lower , upper ) ; }
        ++lower ;
        --upper ;
      }
    }
    return upper ;
  }
  /**
   *  Partition an array in two using the pivot value that is at the beginning of the array being
   *  partitioned.
   *
   *  @param lower iterator positioned at the bottom of the slice.
   *  @param upper iterator positioned at the top of the slice.
   *  @param cr the <code>Comparator</code> to be used to define the order.
   *  @param ce the <code>Comparator</code> defining value equality.
   *  @return iterator positioned at the pivot value.
   */
  private static <T> BiDirectionalIterator<T> partition ( final BiDirectionalIterator<T> lower , final BiDirectionalIterator<T> upper , final Comparator<? super T> cr , final Comparator<? super T> ce) {
    //  **MUST** treat the parameters as immutable so make clones.  This has the advantage of preparing the
    //  object that we will return.  For some bizarre reason only Iterators and not subinterfaces can be
    //  cloned.  Is this a bug in Java or just in the interface inheritance hierarchy?
    final BiDirectionalIterator<T> b = lower.clone ( ) ;
    final BiDirectionalIterator<T> t = upper.clone ( ) ;
    final BiDirectionalIterator<T> pivot = b.clone ( ) ;
    final T pivotValue = pivot.get ( ) ;
    while ( true ) {
      while ( ( cr.relation ( b.get ( ) , pivotValue ) || ce.relation ( b.get ( ) , pivotValue ) ) && ! b.equals ( t ) ) { b.advance() ; }
      while ( cr.relation ( pivotValue , t.get ( ) ) && ! t.equals ( b ) ) { t.retreat ( ) ; }
      if ( b.equals ( t ) ) {
        while ( cr.relation ( pivotValue , t.get ( ) ) ) { t.retreat ( ) ; }
        break;
      }
      else { swap ( b , t ) ; }
    }
    swap ( pivot , t ) ;
    return t ;
  }
}
