/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.Comparator ;
/**
 *  Sort a primitive array of <code>T</code>s using Quicksort.  This is an O(n.log(n)) sort except when the
 *  data is almost sorted in which case it O(n^2).
 *
 *  @author Russel Winder
 */
public class Quicksort4 {
  private Quicksort4 ( ) { }
  /**
   *  The sort operation.
   *
   *  @param a the primitive array to be sorted.
   *  @param cr the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   *  @param ce the <code>Comparator</code> defining value equality.
   */
  public static <T> void sort ( final T[] a , final Comparator<T> cr , final Comparator<T> ce ) {
    quicksort ( a , 0 , a.length - 1 , cr , ce ) ;
  }
  /**
   *  Given the array and two indices, swap the two items in the array.
   *
   *  @param a the array of values.
   *  @param x index of item to swap.
   *  @param y index of item to swap.
   */
  private static <T> void swap ( final T[] a , final int x , final int y ) {
    final T temp = a[x] ;
    a[x] = a[y] ;
    a[y] = temp ;
  }
  /**
   *  Partition a primitive array in two using the pivot value that is at the centre of the array being
   *  partitioned.
   *
   *  <p> This partition implementation based on the algorithm in G Brassard & P Bratley (1988)
   *  "Algortihmics: Theory and Practice", Prentice Hall, p.117.</p>
   *
   *  @param a the array out of which to take a slice.
   *  @param lower the lower bound of this slice.
   *  @param upper the upper bound of this slice.
   *  @param cr the <code>Comparator</code> to be used to define the order.
   *  @param ce the <code>Comparator</code> defining value equality.
   *  @return the index of the pivot item.
   */
  private static <T> int partition ( final T[] a , final int lower , final int upper , final Comparator<T> cr , final Comparator<T> ce ) {
    int b = lower ;
    int t = upper + 1 ;
    final T pivotValue = a[lower] ;
    do { ++b ; } while ( (cr.relation ( a[b] , pivotValue ) || ce.relation ( a[b] , pivotValue ) ) && ( b < upper ) ) ;
    do { --t ; } while ( cr.relation ( pivotValue , a[t] ) ) ;
    while ( b < t ) {
      swap ( a , b , t ) ;
      do { ++b ; } while ( cr.relation ( a[b] , pivotValue ) || ce.relation ( a[b], pivotValue ) ) ;
      do { --t ; } while ( cr.relation ( pivotValue , a[t] ) ) ;
    }
    swap ( a , lower , t ) ;
    return t ;
  }
  /**
   *  The recursive Quicksort function.
   *
   *  @param a the primitive array out of which to take a slice.
   *  @param lower the lower bound of this slice.
   *  @param upper the upper bound of this slice.
   *  @param cr the <code>Comparator</code> to be used to define the order.
   *  @param ce the <code>Comparator</code> defining value equality.
   */
  private static <T> void quicksort ( final T[] a , final int lower , final int upper, final Comparator<T> cr, final Comparator<T> ce) {
    final int sliceLength = upper-lower + 1 ;
    if ( sliceLength > 1 ) {
      if ( sliceLength == 2 ) {
        if ( cr.relation ( a[upper] , a[lower] ) ) { swap ( a , lower , upper ) ; }
      } else {
        final int pivotIndex = partition ( a , lower , upper , cr , ce ) ;
        quicksort ( a , lower , pivotIndex - 1 , cr , ce ) ;
        quicksort ( a , pivotIndex + 1 , upper , cr , ce ) ;
      }
    }
  }
}
