/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.Sequence ;
/**
 *  Sort a <code>Sequence</code> using Quicksort.  This is an O(n.log(n)) sort except when the data is
 *  almost sorted in which case it O(n^2).
 *
 *  @author Russel Winder
 */
public class QuicksortSequence {
  private QuicksortSequence ( ) { }
  /**
   *  The sort operation.  Although the data is copied from <code>Sequence</code> to primitive array and
   *  back again, these are O(n) operations and therefore do not really give any penalty since the sort is
   *  O(n.log(n)).  It is an overhead though.  However, this overhead is probably more than compensated for
   *  by the much more efficient operations provided by primitive arrays compared to <code>Sequences</code>.
   *
   *  @param s the <code>Sequence</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code> during the sort process.
   */
  @SuppressWarnings ( "unchecked" )
  public static <T> void sort ( final Sequence<T> s, final Comparator<? super T> c ) {
    //  Cannot use s.toPrimitiveArray as we have no Class<T> or T[] object to use to access a correctly type
    //  Class object to act as a type discriminant for the toPrimitiveArray method so we do things
    //  ourselves.  The primitive array ia only ever used internally so cannot get the type errors that this
    //  technique could lead to.  Silly but this is what type erasure leads to in a language such as Java
    //  :-(
    final T[] data = (T[]) new Object [s.size ( )] ;
    int index = 0 ;
    for ( InputIterator<T> i = s.iterator ( ) ; ! i.atEnd ( ) ; i.advance ( ) , ++index ) { data[index] = i.get ( ) ; }
    Quicksort.sort ( data , c ) ;
    s.makeEmpty ( ) ;
    //  Do not use:
    //     for ( final T t : data ) { s.add ( t ) ; }
    //  in case it does not guarantee to iterate through the array in index order.
    for ( int i = 0 ; i < data.length ; ++i ) { s.add ( data[i] ) ; }
  }
}
