/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import java.math.BigInteger ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.Queue ;
import org.devjavasoft.ads.Sequence ;
import org.devjavasoft.ads.SLListQueue ;
/**
 *  Sort a sequence of items where each item is a sequence of n digits where there are m possibilities for
 *  each digit.  It is assumed that digits have a mapping to <code>int</code>s.  Use Radix Sort (also has
 *  been referred to as Digit Sort).  This is an O(q) sort, where q is the number of data items.
 *
 *  @author Russel Winder
 */
public final class RadixSort {
  private RadixSort ( ) { }
  /**
   *  The private implementation of the sort operation for primitive arrays.  This is not made public since
   *  we restrict the interface to integer types &ndash; <code>Number</code> is the common base class
   *  for all numeric types but the algorithm requires integral types, so we must exclude non-integral number
   *  types.  There is no common subclass of <code>Integer</code>, <code>Long</code>, and
   *  <code>BigInteger</code> other than <code>Number</code>.
   *
   *  @param a the primitive array to be sorted.
   *  @param n the number of digits in each sort key.
   *  @param m the number of different possible digits.
   */
	@SuppressWarnings ( "unchecked" )
  private static <T extends Number> void sortIt ( final T[] a , final int n , final int m ) {
    //  Set up the pigeonholes into which we will put numbers.  It is important that we treat the
    //  pigeonholes as FIFO data structures so choose to use SLListQueues.  Can't create an array of
    //  SLListQueue<T> can only use the raw type.
    final Queue<T>[] holes = new SLListQueue [m] ;
    for ( int i = 0 ; i < m ; ++i ) {
      holes[i] = new SLListQueue<T> ( ) ;
    }
    //  For each digit in the sort key do a pigeonhole stuffing exercise.
    for ( int i = 0 ; i < n ; ++i ) {
      final int divisor = (int) Math.pow ( m , i ) ;
      for ( T anA : a ) {
        //  Put the number in a pigeonhole determined by the value of the digit in the slot we are currently
        //  sorting on.
        holes[ ( anA.intValue () / divisor ) % m ].add ( anA ) ;
      }
      //  Merge all the entries in the pigeonholes in order into a single list ready for the next pass.
      //  Reset the pigeonholes ready for the next pass as well.
      int index = 0 ;
      for ( int j = 0 ; j < m ; ++j ) {
        while ( ! holes[j].isEmpty ( ) ) { a[index++] = holes[j].remove ( ) ; }
        holes[j].makeEmpty ( ) ;
      }
    }
  }
  /**
   *  The sort operation for primitive arrays of <code>int</code>s.
   *
   *  @param a the primitive array to be sorted.
   *  @param n the number of digits in each sort key.
   *  @param m the number of different possible digits.
   */
	//public static void sort ( final int[] a , final int n , final int m ) { sortIt ( a , n , m ) ;  }
  /**
   *  The sort operation for primitive arrays of <code>long</code>s.
   *
   *  @param a the primitive array to be sorted.
   *  @param n the number of digits in each sort key.
   *  @param m the number of different possible digits.
   */
  //public static void sort ( final long[] a , final int n , final int m ) { sortIt ( a , n , m ) ; }
	 /**
   *  The sort operation for primitive arrays of <code>Integer</code>s.
   *
   *  @param a the primitive array to be sorted.
   *  @param n the number of digits in each sort key.
   *  @param m the number of different possible digits.
   */
	public static void sort ( final Integer[] a , final int n , final int m ) { sortIt ( a , n , m ) ;	}
  /**
   *  The sort operation for primitive arrays of <code>Long</code>.
   *
   *  @param a the primitive array to be sorted.
   *  @param n the number of digits in each sort key.
   *  @param m the number of different possible digits.
   */
	public static void sort ( final Long[] a , final int n , final int m ) { sortIt ( a , n , m ) ; }
  /**
   *  The sort operation for primitive arrays of <code>BigInteger</code>s.
   *
   *  @param a the primitive array to be sorted.
   *  @param n the number of digits in each sort key.
   *  @param m the number of different possible digits.
   */
	public static void sort ( final BigInteger[] a , final int n , final int m ) { sortIt ( a , n , m ) ; }
  /**
   *  The implementation sort operation for sequences.  Because of type erasure it is not possible to
   *  overload on a parameterized type, i.e. it is not possible to distinguish Sequence&lt;Integer>,
   *  Sequence&lt;Long>, etc.  Currently there is no way of excluding non-integral types :-(
   *
   *  @param s the <code>Sequence</code> to be sorted.
   *  @param n the number of digits in each sort key.
   *  @param m the number of different possible digits.
   */
  public static <T extends Number> void sort ( final Sequence<T> s , final int n , final int m ) {
    //  Set up the numberOfDigits pigeonholes into which we will put numbers.  It is important that we treat
    //  the pigeonholes as FIFO data structures so we make the Queues.
    final Array<Queue<T>> holes = new Array<Queue<T>> ( ) ;
    for ( int i = 0 ; i < m ; ++i ) {
      holes.add ( new SLListQueue<T> ( ) ) ;
    }
    //  For each digit in the sort key do the pigeonhole stuffing followed by the merge.
    for ( int i = 0 ; i < n ; ++i ) {
      final int divisor = (int) Math.pow ( m , i ) ;
      for ( InputIterator<T> j = s.iterator ( ) ; ! j.atEnd ( ) ; j.advance ( ) ) {
        //  Put the number in a pigeonhole determined by the value of the digit in the slot we are currently
        //  sorting on.
        final T item = j.get ( ) ;
        holes.get ( ( item.intValue ( ) / divisor ) % m ).add ( item ) ;
      }
      //  Merge all the entries in the pigeonholes in order into a single sequence ready for the next pass.
      //  Reset the pigeonholes ready for the next pass as well.
      s.makeEmpty ( ) ;
      for ( int j = 0 ; j < m ; ++j ) {
        while ( ! holes.get ( j ).isEmpty ( ) ) {
          s.add ( holes.get ( j ).remove ( ) ) ;
        }
        holes.get ( j ).makeEmpty ( ) ;
      }
    }
  }
}
