/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import org.devjavasoft.ads.BiDirectionalIterator ;
import org.devjavasoft.ads.Comparator ;
import org.devjavasoft.ads.ForwardIterator ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.InvalidOperationException ;
import org.devjavasoft.ads.Sequence ;
import org.devjavasoft.ads.SLList ;
/**
 *  Sort using Selection Sort.  This is an O(n^2) sort.
 *
 *  @author Russel Winder
 */
public class SelectionSort {
  private SelectionSort ( ) { }
  /**
   *  The sort operation for a primitive array.
   *
   *  @param a the primitive array to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code>s during the sort process.
   */
  public static <T> void sort ( final T[] a , final Comparator<? super T> c ) {
    for ( int i = a.length-1 ; i > 0 ; --i ) {
      //  Find the next (maximum or minimum depending on the order relation) value from that part of the
      //  array that is as yet unsorted.  This is the value to put in the current location of interest.
      int indexOfValue = 0 ;
      T value = a[indexOfValue] ;
      for ( int j = 1 ; j <= i ; ++j ) {
        if ( c.relation ( value , a[j] ) ) {
          value = a[j] ;
          indexOfValue = j ;
        }
      }
      //  Exchange the value in the current location of interest with the "maximum" value from the rest of
      //  the unsorted array.  This increases by one the part of the array that is now fully sorted.
      if ( indexOfValue != i ) {
        a[indexOfValue] = a[i] ;
        a[i] = value ;
      }
    }
  }
  /**
   *  The sort operation for <code>Sequence</code>s.
   *
   *  @param s the <code>Sequence</code> to be sorted.
   *  @param c the <code>Comparator</code> used to compare the <code>T</code>s during the sort
   *  process.
   */
  public static <T> void sort ( final Sequence<T> s , final Comparator<? super T> c ) {
    if ( ( s instanceof SLList<?> ) || ( s.size ( ) <= 1 ) ) { throw new InvalidOperationException () ; }
    final BiDirectionalIterator<T> i = (BiDirectionalIterator<T>) s.end ( ) ;
    do {
      i.retreat ( ) ;
      T value = s.first ( ) ;
      final InputIterator<T> j = s.iterator ( ) ;
      ForwardIterator<T> swapItem = (ForwardIterator<T>) j.clone ( ) ;
      while ( true ) {
        if ( c.relation ( value , j.get ( ) ) ) {
          value = j.get ( ) ;
          swapItem = (ForwardIterator<T>) j.clone ( ) ;
        }
        if ( j.equals ( i ) ) { break ; }
        j.advance ( ) ;
      }
      if ( ! swapItem.equals ( i ) ) {
        swapItem.set ( i.get ( ) ) ;
        i.set ( value ) ;
      }
    } while ( ! i.atBegin ( ) ) ;
  }
}
