/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.sorting ;
import java.io.File ;
import java.io.IOException ;
import org.devjavasoft.ads.Comparator ;
/**
 *  An interface defining a factory object encapsulating <code>SortableRecord</code> management information.
 *
 *  <p> When dealing with files, there have to be records. This interface defines that which needs to be
 *  known in order to use any method object performing a file-based sort.  The user must supply a
 *  <code>SortableRecordInformation</code> conformant object in order to provide all the tools needed.</p>
 *
 *  @see SortableRecord
 *  @see SortableRecordReader
 *  @see SortableRecordWriter
 *  @author Russel Winder
 */
public interface SortableRecordInformation {
  /**
   *  Get the coparator that defines the total order.
   *
   *  @return the order relation <code>Comparator</code> so that is used to test the order of records.
   *  Usually this will be an ordering defined by some key in the record.
   */
  Comparator<SortableRecord> getComparator ( ) ;
  /**
   *  Get the comparator that implements the value equality test.
   *
   *  @return the <code>Comparator</code> that can test for value equality of the sort key.
   */
  Comparator<SortableRecord> getEqualToComparator ( ) ;
  /**
   *  Get a reader of records.
   *
   *  @param f the file to read records from.
   *  @return the <code>SortableRecordReader</code> that can read records from the file.
   *  @exception IOException if the <code>BufferedReader</code> cannot be created.
   */
  SortableRecordReader getSortableRecordReader ( File f ) throws IOException ;
  /**
   *  Get a writer of records.
   *
   *  @param f the file to write records to.
   *  @return the <code>SortableRecordWriter</code> that can write records to the file.
   *  @exception IOException if the <code>BufferedWriter</code> cannot be created.
   */
  SortableRecordWriter getSortableRecordWriter ( File f ) throws IOException ;
}
