/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.tests ;

import org.devjavasoft.ads.AccessEmptyContainerException ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.DLList ;
import org.devjavasoft.ads.ForwardIterator ;
import org.devjavasoft.ads.IndexOutOfBoundsException ;
import org.devjavasoft.ads.InvalidIteratorOperationException ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.ReverseIterator ;
import org.devjavasoft.ads.SLList ;
import org.devjavasoft.ads.Sequence ;

import org.testng.annotations.BeforeMethod ;
import org.testng.annotations.Test ;
import static org.testng.Assert.assertNotSame ;
import static org.testng.Assert.assertEquals ;
import static org.testng.Assert.assertTrue ;
import static org.testng.Assert.assertFalse ;


/**
 *  A class to test the class <code>Array&lt;Integer></code>
 *
 *  @author Russel Winder
 */
public class Array_Test {
  //  The following are variables we will use in the individual tests.  Not initialized here since, because
  //  we are using JUnit, we know that the method setUp will be run before each test method is run and we
  //  know that we correctly set up these variables in there.
  /**
   *  Reference to an empty <code>Array&lt;Integer></code>.
   */
  protected Array<Integer> empty = null ;
  /**
   *  Reference to an <code>Array&lt;Integer></code> which has an odd number of elements in it.
   */
  protected Array<Integer> nonEmptyOddSize = null ;
  /**
   *  Reference to an <code>Array&lt;Integer></code> which has an even number of elements in it.
   */
  protected Array<Integer> nonEmptyEvenSize = null ;
  /**
   *   The setup method which gets executed prior to any test method and must set up any variables needed by
   *   the tests.
   *
   *   <p>We are assuming that <code>Array<Integer>.add</code> completes execution even if it is not
   *   actually correct.</p>
   */
  @BeforeMethod public void setUp ( ) {
    empty = new Array<Integer> ( ) ;
    nonEmptyOddSize = new Array<Integer> ( ) ;
    fillOddSize ( nonEmptyOddSize ) ;
    nonEmptyEvenSize = new Array<Integer> ( ) ;
    fillEvenSize ( nonEmptyEvenSize ) ;
  }
  /**
   *   Fill a <code>Sequence&lt;Integer></code> with an even number of items.  Used to fill
   *   <code>SLList&lt;Integer></code>s and <code>DLList<Integer></code>s as well as
   *   <code>Array&lt;Integer></code>s. Requires <code>Sequence<Integer>.add</code> to work correctly.
   *
   *  <p>This procedure should add less items that would trigger automatic extension so as to avoid that
   *  method being axiomatic as well as add.</p>
   *
   *  @param s the sequence to add values to.
   */
  protected void fillEvenSize ( final Sequence<Integer> s ) {
    s.add ( 1 ) ;
    s.add ( 2 ) ;
    s.add ( 3 ) ;
    s.add ( 4 ) ;
  }
  /**
   *   Fill a <code>Sequence&lt;Integer></code> with an odd number of items.  Used to fill
   *   <code>SLList&lt;Integer></code>s and <code>DLList<Integer></code>s as well as
   *   <code>Array&lt;Integer></code>s. Requires <code>Sequence<Integer>.add</code> to work correctly.
   *
   *  <p>This procedure should add less items that would trigger automatic extension so as to avoid that
   *  method being axiomatic as well as add.</p>
   *
   *  @param s the sequence to add values to.
   */
  protected void fillOddSize ( final Sequence<Integer> s ) {
    s.add ( 1 ) ;
    s.add ( 2 ) ;
    s.add ( 3 ) ;
    s.add ( 4 ) ;
    s.add ( 5 ) ;
  }
  /**
   *  Constructor of a <code>SLList&lt;Integer></code> with an odd number of items.
   *
   *  @return a new, filled <code>SLList</code>.
   */
  protected SLList<Integer> makeSLListOddSize ( )  {
    final SLList<Integer> value = new SLList<Integer>  ( )  ;
    fillOddSize ( value ) ;
    return value ;
  }
  /**
   *  Constructor of a <code>DLList&lt;Integer></code> with an odd number of items.
   *
   *  @return a new, filled <code>DLList</code>
   */
  protected DLList<Integer> makeDLListOddSize ( )  {
    final DLList<Integer> value = new DLList<Integer>  ( )  ;
    fillOddSize ( value ) ;
    return value ;
  }
  //  ====================================================
  //
  //  The testing methods.
  //
  //  Each one's name starts test followed by something.  Hopefully the method names are
  //  self-explanatory -- they do get rather long.  The idea is for each method to perform a
  //  specific test, i.e. it develops the state of the variables in some way and then makes one or
  //  more assertions about the state.  If the assertions are satisfied then the test is passed.  It is up
  //  to the programmer to make sure the testing code is correct -- indeed that it performs some meaningful
  //  test.  The framework makes this easy though.  Most importantly it is the test can be run over and over
  //  and can be incrementally developed as the class being tested is developed.
  //
  /**
   *  Test <code>Array<Integer>.add</code>.  <code>Array<Integer>.add</code> is essentially axiomatic for
   *  the rest of the test system (since it is used in the <code>fillOddSize</code> and
   *  <code>fillEvenSize</code> methods) so we must be careful here to test with minimal use of untested
   *  features.
   */
  @Test public void add ( )  {
    //  Assuming that Array<Integer>.toString works correctly and that String.equals works correctly, the
    //  following tests as much as we can the basic Array<Integer> constructor.
    String s = empty.toString ( )  ;
    assertEquals ( s , "[ ]") ;
    //  The following two test construction and adding a few times.
    s = nonEmptyOddSize.toString ( )  ;
    assertEquals ( s , "[ 1 2 3 4 5 ]" ) ;
    s = nonEmptyEvenSize.toString ( )  ;
    assertEquals ( s , "[ 1 2 3 4 ]" ) ;
    //  The following adds Array<Integer>.equals to the list of things we hope work
    fillOddSize ( empty ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  /**
   *  Test the automatic extension of the <code>Array&lt;Integer></code> as items are added.
   *
   *  <p>We know that adding 100 items should trigger the extension of the underlying memory representation
   *  and so exercises that functionality.  It is also another test of add in some ways.<p>
   */
  @Test public void capacity ( )  {
    final int oldSize = nonEmptyOddSize.size ( )  ;
    for ( int i = 0 ; i < 100 ; ++i ) { nonEmptyOddSize.add ( i ) ; }
    assertEquals ( nonEmptyOddSize.size ( ) , 100 + oldSize ) ;
  }
  /**
   *  Test the cloning of an empty <code>Array&lt;Integer></code>.  This should result in a new object not
   *  two references to the same object.
   */
  @Test public void cloneEmpty ( )  {
    final Array<Integer> clone = empty.clone ( )  ;
    assertNotSame ( clone , empty ) ;
    assertEquals ( clone.size ( ) , empty.size ( ) ) ;
    assertEquals ( clone , empty ) ;
  }
  /**
   *  Test the cloning of a non-empty <code>Array&lt;Integer></code>.  This should result in a new object
   *  not two references to the same object.
   */
  @Test public void cloneNonEmpty ( )  {
    final Array<Integer> clone = nonEmptyOddSize.clone ( )  ;
    assertNotSame ( clone , nonEmptyOddSize ) ;
    assertEquals ( clone.size ( )  , nonEmptyOddSize.size ( ) ) ;
    assertEquals ( clone , nonEmptyOddSize ) ;
  }
  @Test public void containsEmpty ( )  {
    assertFalse ( empty.contains ( 1 ) ) ;
  }
  @Test public void containsNonEmpty ( )  {
    assertTrue ( nonEmptyOddSize.contains ( 1 ) ) ;
    assertTrue ( nonEmptyOddSize.contains ( 2 ) ) ;
    assertTrue ( nonEmptyOddSize.contains ( 3 ) ) ;
    assertTrue ( nonEmptyOddSize.contains ( 4 ) ) ;
    assertTrue ( nonEmptyOddSize.contains ( 5 ) ) ;
  }
  @Test public void copyConstructorsFromEmptyArray ( )  {
    final Array<Integer> copy = new Array<Integer> ( empty ) ;
    assertEquals ( copy , empty ) ;
    assertFalse ( nonEmptyOddSize.equals ( copy ) ) ;
  }
  @Test public void copyConstructorsFromNonEmptyArray ( )  {
    final Array<Integer> copy = new Array<Integer> ( nonEmptyOddSize ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
    assertFalse ( copy.equals ( empty ) ) ;
  }
  @Test public void copyConstructorsFromArraySlice ( )  {
    final Array<Integer> copy = new Array<Integer> ( nonEmptyOddSize , 0 , nonEmptyOddSize.size ( ) ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
  }
  @Test public void copyConstructorsFromArrayIterators ( )  {
    final Array<Integer> copy = new Array<Integer> ( nonEmptyOddSize.begin ( )  , nonEmptyOddSize.end ( ) ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
  }
  @Test public void copyConstructorsFromEmptySLList ( )  {
    final SLList<Integer> list = new SLList<Integer>  ( )  ;
    final Array<Integer> copy = new Array<Integer> ( list ) ;
    assertEquals ( copy , empty ) ;
    assertFalse ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromNonEmptySLList ( )  {
    final SLList<Integer> list = makeSLListOddSize ( )  ;
    final Array<Integer> copy = new Array<Integer> ( list ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
    assertFalse ( copy.equals ( empty ) ) ;
  }
  @Test public void copyConstructorsFromSLListSlice ( )  {
    final SLList<Integer> list = makeSLListOddSize ( )  ;
    final Array<Integer> copy = new Array<Integer> ( list , 0 , list.size ( ) ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
  }
  @Test public void copyConstructorsFromSLListIterators ( )  {
    final SLList<Integer> list = makeSLListOddSize ( )  ;
    final Array<Integer> copy = new Array<Integer> ( list.begin ( ) , list.end ( ) ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
  }
  @Test public void copyConstructorsFromEmptyDLList ( )  {
    final DLList<Integer> list = new DLList<Integer>  ( )  ;
    final Array<Integer> copy = new Array<Integer> ( list ) ;
    assertEquals ( copy , empty ) ;
    assertFalse ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromNonEmptyDLList ( )  {
    final DLList<Integer> list = makeDLListOddSize ( )  ;
    final Array<Integer> copy = new Array<Integer> ( list ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
    assertFalse ( copy.equals ( empty ) ) ;
  }
  @Test public void copyConstructorsFromDLListSlice ( )  {
    final DLList<Integer> list = makeDLListOddSize ( )  ;
    final Array<Integer> copy = new Array<Integer> ( list , 0 , list.size ( ) ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
  }
  @Test public void copyConstructorsFromDLListIterators ( )  {
    final DLList<Integer> list = makeDLListOddSize ( )  ;
    final Array<Integer> copy = new Array<Integer> ( list.begin ( ) , list.end ( ) ) ;
    assertEquals ( copy , nonEmptyOddSize ) ;
  }
  @Test ( expectedExceptions = { IndexOutOfBoundsException.class } )
  public void get ( )  {
    Integer i = nonEmptyOddSize.get ( 0 ) ;
    assertEquals ( i.intValue ( ) , 1 ) ;
    i = nonEmptyOddSize.get ( 1 ) ;
    assertEquals ( i.intValue ( ) , 2 ) ;
    nonEmptyOddSize.get ( nonEmptyOddSize.size ( ) ) ;
  }
  @Test ( expectedExceptions = { IndexOutOfBoundsException.class } )
  public void insert ( )  {
    empty.insert ( 0,  5 ) ;
    empty.insert ( 0,  1 ) ;
    empty.insert ( 1,  2 ) ;
    empty.insert ( 2,  3 ) ;
    empty.insert ( 3,  4 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
    empty.insert ( 5 ,  6 ) ;
  }
  @Test public void iteratorGetFirst ( )  {
    final InputIterator<Integer> i = nonEmptyOddSize.iterator ( )  ;
    assertEquals ( i.get ( ).intValue ( ) , 1 ) ;
  }
  @Test public void iteratorGetMiddle ( )  {
    final InputIterator<Integer> i = nonEmptyOddSize.iterator ( )  ;
    i.advance ( )  ;
    assertEquals ( i.get ( ).intValue ( ) , 2 ) ;
  }
  @Test public void iteratorSetFirst ( )  {
    final ForwardIterator<Integer> i = nonEmptyOddSize.iterator ( )  ;
    i.set ( 3 ) ;
    empty.add ( 3 ) ;
    empty.add ( 2 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test public void iteratorSetMiddle ( )  {
    final ForwardIterator<Integer> i = nonEmptyOddSize.iterator ( )  ;
    i.advance ( )  ;
    i.set ( 3 ) ;
    empty.add ( 1 ) ;
    empty.add ( 3 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test public void makeEmptyFromEmpty ( )  {
    empty.makeEmpty ( )  ;
    assertTrue ( empty.isEmpty ( ) ) ;
  }
  @Test public void makeEmptyFromNonEmpty ( )  {
    nonEmptyOddSize.makeEmpty ( )  ;
    assertTrue ( nonEmptyOddSize.isEmpty ( ) ) ;
  }

  //  NB Array<Integer>.remove ( int ) exists and is remove by index.  To remove by value we cannot use
  //  autoboxing as it is never invoked so we must use x.remove ( new Integer ( int ) ) for remove by value.

  @Test ( expectedExceptions = { AccessEmptyContainerException.class } )
  public void removeByIndexFromEmpty ( )  {
     empty.remove ( 0 ) ;
  }
  @Test public void removeByIndexFromIndexZero ( )  {
    nonEmptyOddSize.remove ( 0 ) ;
    assertFalse ( nonEmptyOddSize.contains ( 1 ) ) ;
    empty.add ( 2 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test public void removeByIndexFromMiddle ( )  {
    nonEmptyOddSize.remove ( 1 ) ;
    assertFalse ( nonEmptyOddSize.contains ( 2 ) ) ;
    empty.add ( 1 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test public void removeByIndexFromEnd ( )  {
    nonEmptyOddSize.remove ( 2 ) ;
    assertFalse ( nonEmptyOddSize.contains ( 3 ) ) ;
    empty.add ( 1 ) ;
    empty.add ( 2 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test ( expectedExceptions = { IndexOutOfBoundsException.class } )
  public void removeByIndexWithIllegalIndex ( )  {
    nonEmptyOddSize.remove ( 5 ) ;
  }
  @Test ( expectedExceptions = { AccessEmptyContainerException.class } )
  public void removeByValueFromEmpty ( )  {
    empty.remove ( 1 ) ;
  }
  @Test public void removeByValueFromIndexZero ( )  {
    nonEmptyOddSize.remove ( new Integer ( 1 ) ) ; //  Must force an Integer object to avoid calling remove ( int )
    assertFalse ( nonEmptyOddSize.contains ( 1 ) ) ;
    empty.add ( 2 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test public void removeByValueFromMiddle ( )  {
    nonEmptyOddSize.remove ( new Integer ( 2 ) ) ; //  Must force an Integer object to avoid calling remove ( int )
    assertFalse ( nonEmptyOddSize.contains ( 2 ) ) ;
    empty.add ( 1 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test public void removeByValueFromEnd ( )  {
    nonEmptyOddSize.remove ( new Integer ( 3 ) ) ; //  Must force an Integer object to avoid calling remove ( int )
    assertFalse ( nonEmptyOddSize.contains ( 3 ) ) ;
    empty.add ( 1 ) ;
    empty.add ( 2 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test public void removeByValueWithValueNotContained ( )  {
    final Array<Integer> clone = nonEmptyOddSize.clone ( )  ;
    clone.remove ( new Integer ( 6 ) ) ; //  Must force an Integer object to avoid calling remove ( int )
    assertEquals ( clone , nonEmptyOddSize ) ;
  }
  @Test ( expectedExceptions = { InvalidIteratorOperationException.class } )
  public void reverseEmpty ( )  {
    empty.reverse ( )  ;
  }
  @Test public void reverseEvenSize ( )  {
    final Array<Integer> clone = nonEmptyEvenSize.clone ( )  ;
    clone.reverse ( )  ;
    final Array<Integer> reversed = new Array<Integer>  ( )  ;
    reversed.add ( 4 ) ;
    reversed.add ( 3 ) ;
    reversed.add ( 2 ) ;
    reversed.add ( 1 ) ;
    assertEquals ( reversed , clone ) ;
    clone.reverse ( )  ;
    assertEquals ( clone , nonEmptyEvenSize ) ;
  }
  @Test public void reverseOddSize ( )  {
    final Array<Integer> clone = nonEmptyOddSize.clone ( )  ;
    clone.reverse ( )  ;
    final Array<Integer> reversed = new Array<Integer>  ( )  ;
    reversed.add ( 5 ) ;
    reversed.add ( 4 ) ;
    reversed.add ( 3 ) ;
    reversed.add ( 2 ) ;
    reversed.add ( 1 ) ;
    assertEquals ( reversed , clone ) ;
    clone.reverse ( )  ;
    assertEquals ( clone , nonEmptyOddSize ) ;
  }
  @Test public void reverseIterator ( )  {
    final Array<Integer> a = new Array<Integer>  ( )  ;
    //  Big problem with reverse iterators is that the semantics of the operations are different from
    //  bi-directional iterators.
    final ReverseIterator<Integer> r = new ReverseIterator<Integer> ( nonEmptyOddSize.end ( ) ) ;
    do {
      r.advance ( )  ;
      a.add ( r.get ( ) ) ;
    } while ( ! r.atEnd ( )  ) ;
    nonEmptyOddSize.reverse ( )  ;
    assertEquals ( a , nonEmptyOddSize ) ;
  }
  @Test public void set ( )  {
    empty.add ( 2 ) ;
    empty.add ( 2 ) ;
    empty.add ( 2 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    nonEmptyOddSize.set ( 0 , 2 ) ;
    nonEmptyOddSize.set ( 2 , 2 ) ;
    assertEquals ( empty , nonEmptyOddSize ) ;
  }
  @Test public void toPrimitiveArrayClass ( ) {
    final Integer [] x = nonEmptyEvenSize.toPrimitiveArray ( Integer.class ) ;
    for ( int i = 0 ; i < x.length ; ++i ) { assertEquals ( x[i].intValue ( ) , i + 1 ) ; }
  }
  @Test public void toPrimitiveArrayArray ( ) {
    final Integer [] x = new Integer [nonEmptyEvenSize.size ( )] ;
    final Integer [] y = nonEmptyEvenSize.toPrimitiveArray ( x ) ;
    assertEquals ( x , y ) ;
    for ( int i = 0 ; i < x.length ; ++i ) { assertEquals ( x[i].intValue ( ) , i + 1 ) ; }
  }
}
