/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.tests ;

import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.DLList ;
import org.devjavasoft.ads.Sequence ;
import org.devjavasoft.ads.SLList ;

import org.testng.annotations.Test ;
import static org.testng.Assert.assertTrue ;

/**
 *  A class to test that the <code>for ( E e : c ) { ... }</code> style loop works for ADS containers.  The
 *  issue here is that containers must implement <code>java.lang.Iterable</code> and the iterators must
 *  implement <code>java.lang.SimpleIterator</code>.  For the Collection Framework this is not a problem
 *  because <code>java.lanag.SimpleIterator</code> follows the same iterator philosophy that
 *  <code>java.util.Iterator</code> has always followed.  FOr ADS though there has been a bit of an issue
 *  integrating with this feature new in J2SEv1.5 (Tiger) as the iterator structure is very different.
 *
 *  <p>Test the iterator by making a copy of a data structure.  These tests should be integrated into the
 *  various data structure tests as a test of clone (which should use the for loop form) but currently there
 *  is a compiler problem which means these for loop forms cannot be used in the clone method of the data
 *  structures.<p>
 *
 *  @author Russel Winder
 */
public class ContainerIteration_Test {
  //  Using iterators explicitly the copy can be undertaken using:
  //
  //    for ( ForwardIterator<Integer> i = source.iterator() ; ! i.atEnd() ; i.advance() ) {
  //      target.add(i.get()) ;
  //    }
  //
  //  But now we have the smart new way of writing this without having an explicit iterator.
  //
  //    for ( Integer i : source ) {
  //      target.add(i) ;
  //    }
  //
  //  and it is this that is being tested.  The compiler creates an implicit SimpleIterator to
  //  perform the action.  This is somewhat less efficient thanthe explicit iterator form but so much easier
  //  to read and comprehend.
  //
  //  Of course in production code you would just use clone!

  /**
   *  Fill up  a <code>Sequence&lt;Integer></code>.
   */
  private static void fillSequence ( final Sequence<Integer> s ) {
    s.add ( new Integer ( 1 ) ) ;
    s.add ( new Integer ( 2 ) ) ;
    s.add ( new Integer ( 3 ) ) ;
    s.add ( new Integer ( 4 ) ) ;
    s.add ( new Integer ( 5 ) ) ;
    s.add ( new Integer ( 6 ) ) ;
    s.add ( new Integer ( 7 ) ) ;
    s.add ( new Integer ( 8 ) ) ;
  }

  /**
   *  Try thing with <code>Array</code>.
   */
  @Test public void array ( ) {
    final Array<Integer> source = new Array<Integer>  ( ) ;
    fillSequence ( source ) ;
    final Array<Integer> target = new Array<Integer>  ( ) ;
    for ( Integer i : source ) { target.add ( i ) ; }
    assertTrue ( source.equals ( target ) ) ;
  }
  /**
   *  Try thing with <code>DLList</code>.
   */
  @Test public void dLList ( ) {
    final DLList<Integer> source = new DLList<Integer>  ( ) ;
    fillSequence ( source ) ;
    final DLList<Integer> target = new DLList<Integer>  ( ) ;
    for ( Integer i : source ) { target.add ( i ) ; }
    assertTrue ( source.equals ( target ) ) ;
  }
  /**
   *  Try thing with <code>SLList</code>.
   */
  @Test public void sLList ( ) {
    final SLList<Integer> source = new SLList<Integer>  ( ) ;
    fillSequence ( source ) ;
    final SLList<Integer> target = new SLList<Integer>  ( ) ;
    for ( Integer i : source ) { target.add ( i ) ; }
    assertTrue ( source.equals ( target ) ) ;
  }
}
