/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.tests ;

import org.devjavasoft.ads.EqualToComparator ;
import org.devjavasoft.ads.Pair ;
import org.devjavasoft.ads.PairAComparator ;
import org.devjavasoft.ads.PairBComparator ;
import org.devjavasoft.ads.PairABComparator ;

import org.testng.annotations.Test ;
import static org.testng.Assert.assertEquals ;
import static org.testng.Assert.assertFalse ;
import static org.testng.Assert.assertTrue ;

/**
 *  A class to test the classes associatewith <code>Pair</code>.
 *
 *  @author Russel Winder
 */
public class Pair_Test {
  private final Integer i = 10 ;
  private final String s = "flobadob" ;
  private final Pair<Integer, String> p = new Pair<Integer, String> ( i , s ) ;
  @Test public void pairAccessors ( ) {
    assertEquals ( p.getA ( ) , i ) ;
    assertEquals ( p.getB ( ) , s ) ;
  }
  @Test public void pairClone ( ) {
    final Pair<Integer, String>q = p.clone ( ) ;
    assertEquals ( q.getA ( ) , i ) ;
    assertEquals ( q.getB ( ) , s ) ;
  }
  @Test public void pairEqualToTrueTrivial ( ) {
    assertTrue ( p.equals ( p ) ) ;
  }
  @Test public void pairEqualToTrue ( ) {
    final Pair<Integer, String>q = p.clone ( ) ;
    assertTrue ( p.equals ( q ) ) ;
  }
  @Test public void pairEqualToTrueNulls ( ) {
    final Pair<Integer, String>q = new Pair<Integer, String> ( null , null ) ;
    final Pair<Integer, String>r = new Pair<Integer, String> ( null , null ) ;
    assertTrue ( q.equals ( r ) ) ;
  }
  @Test public void pairEqualToFalse ( ) {
    final Pair<Integer, String>q = new Pair<Integer, String> ( 20 , "blah" ) ;
    assertFalse ( p.equals ( q ) ) ;
  }
  @Test public void pairEqualToFalseNulls ( ) {
    final Pair<Integer, String>q = new Pair<Integer, String> ( null , null ) ;
    assertFalse ( p.equals ( q ) ) ;
  }
  @Test public void pairToString ( ) {
    assertEquals ( p.toString ( ) , "(" + i + ", " + s + ")" ) ;
  }
  @Test public void pairAComparator ( ) {
    PairAComparator<Integer, String, Pair<Integer, String>> pac = new PairAComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<Integer> ( ) ) ;
     final Pair<Integer, String>q = p.clone ( ) ;
     assertTrue ( pac.relation ( p, q ) ) ;
  }
  @Test public void pairAComparatorEqualsTrue ( ) {
    PairAComparator<Integer, String, Pair<Integer, String>> pac_A = new PairAComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<Integer> ( ) ) ;
    PairAComparator<Integer, String, Pair<Integer, String>> pac_B = new PairAComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<Integer> ( ) ) ;
    assertTrue ( pac_A.equals ( pac_B ) ) ;
  }
  @Test public void pairBComparator ( ) {
    PairBComparator<Integer, String, Pair<Integer, String>> pbc = new PairBComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<String> ( ) ) ;
     final Pair<Integer, String>q = p.clone ( ) ;
     assertTrue ( pbc.relation ( p, q ) ) ;
  }
  @Test public void pairBComparatorEqualsTrue ( ) {
    PairBComparator<Integer, String, Pair<Integer, String>> pbc_A = new PairBComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<String> ( ) ) ;
    PairBComparator<Integer, String, Pair<Integer, String>> pbc_B = new PairBComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<String> ( ) ) ;
    assertTrue ( pbc_A.equals ( pbc_B ) ) ;
  }
 @Test public void pairABComparator ( ) {
    PairABComparator<Integer, String, Pair<Integer, String>> pabc = new PairABComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<Integer> ( ) , new EqualToComparator<String> ( ) ) ;
     final Pair<Integer, String>q = p.clone ( ) ;
     assertTrue ( pabc.relation ( p, q ) ) ;
  }
  @Test public void pairABComparatorEqualsTrue ( ) {
    PairABComparator<Integer, String, Pair<Integer, String>> pabc_A = new PairABComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<Integer> ( ) , new EqualToComparator<String> ( ) ) ;
    PairABComparator<Integer, String, Pair<Integer, String>> pabc_B = new PairABComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<Integer> ( ) , new EqualToComparator<String> ( ) ) ;
    assertTrue ( pabc_A.equals ( pabc_B ) ) ;
  }
  /*
   *  Due to erasure, X<A> and X<B> are the same type at run time in Java (unlike any other programming
   *  language with parameterized types).  Given that they are different types at compile time this leads to
   *  some seriously confused programmers.  The upshot is that the assertions below seem to be the wrong way
   *  round from a type theory perspective but they are the right way round from a Java perspective.
   */
  @Test public void pairAComparatorEqualsFalse ( ) {
    PairAComparator<Integer, String, Pair<Integer, String>> pac_A = new PairAComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<Integer> ( ) ) ;
    PairAComparator<String, Integer, Pair<String, Integer>> pac_B = new PairAComparator<String, Integer, Pair<String, Integer>> ( new EqualToComparator<String> ( ) ) ;
    assertTrue ( pac_A.equals ( pac_B ) ) ;
  }
 @Test public void pairBComparatorEqualsFalse ( ) {
    PairBComparator<Integer, String, Pair<Integer, String>> pbc_A = new PairBComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<String> ( ) ) ;
    PairBComparator<String, Integer, Pair<String, Integer>> pbc_B = new PairBComparator<String, Integer, Pair<String, Integer>> ( new EqualToComparator<Integer> ( ) ) ;
    assertTrue ( pbc_A.equals ( pbc_B ) ) ;
  }
 @Test public void pairABComparatorEqualsFalse ( ) {
    PairABComparator<Integer, String, Pair<Integer, String>> pabc_A = new PairABComparator<Integer, String, Pair<Integer, String>> ( new EqualToComparator<Integer> ( ) , new EqualToComparator<String> ( ) ) ;
    PairABComparator<String, Integer, Pair<String, Integer>> pabc_B = new PairABComparator<String, Integer, Pair<String, Integer>> ( new EqualToComparator<String> ( ) , new EqualToComparator<Integer> ( ) ) ;
    assertTrue ( pabc_A.equals ( pabc_B ) ) ;
  }
}
