/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.tests ;

import org.devjavasoft.ads.PriorityQueue ;
import org.devjavasoft.ads.SortedListPriorityQueue ;
import org.devjavasoft.ads.HeapPriorityQueue ;
import org.devjavasoft.ads.SequenceQueuePriorityQueue ;

import org.testng.annotations.Test ;
import static org.testng.Assert.assertTrue ;

/**
 *  A unit test of the <code>PriorityQueue&lt;T></code> using JUnit.
 *
 *  <p>Exercises some but not all of the methods in some but not all of the situations.  This is certainly
 *  not exhaustive.  Its main purpose is to test the clone method and the equality testing.</p>
 *
 *  @author Russel Winder
 */
public class PriorityQueue_Test {
  @Test public void emptySortedListPriorityQueue ( ) {
    PriorityQueue<Integer> q = new SortedListPriorityQueue<Integer> ( ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptyHeapPriorityQueue ( ) {
    PriorityQueue<Integer> q = new HeapPriorityQueue<Integer> ( ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptySequenceQueuePriorityQueue ( ) {
    PriorityQueue<Integer> q = new SequenceQueuePriorityQueue<Integer> ( ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptySynchronizedSortedListPriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new SortedListPriorityQueue<Integer> ( ) ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptySynchronizedHeapPriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new HeapPriorityQueue<Integer> ( ) ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptySynchronizedSequenceQueuePriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new SequenceQueuePriorityQueue<Integer> ( ) ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  private void runDestructiveAssertions ( final PriorityQueue<Integer> q ) {
    assertTrue ( q.peek ( ).equals ( new Integer ( 1 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 1 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 2 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 3 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 4 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 5 ) ) ) ;
  }
  @Test public void fillingSortedListPriorityQueue ( ) {
    PriorityQueue<Integer> q = new SortedListPriorityQueue<Integer> ( ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingHeapPriorityQueue ( ) {
    PriorityQueue<Integer> q = new HeapPriorityQueue<Integer> ( ) ;
    fillUp ( q ) ;
    //  NB HeapPriorityQueue does not preserve nsertion order at a given priority!
    assertTrue ( q.contains ( new Integer ( 1 ) ) ) ;
    assertTrue ( q.contains ( new Integer ( 2 ) ) ) ;
    assertTrue ( q.contains ( new Integer ( 3 ) ) ) ;
    assertTrue ( q.contains ( new Integer ( 4 ) ) ) ;
    assertTrue ( q.contains ( new Integer ( 5 ) ) ) ;
  }
  @Test public void fillingSequenceQueuePriorityQueue ( ) {
    PriorityQueue<Integer> q = new SequenceQueuePriorityQueue<Integer> ( ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingSynchronizedSortedListPriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new SortedListPriorityQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingSynchronizedHeapPriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new HeapPriorityQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    //  NB HeapPriorityQueue does not preserve nsertion order at a given priority!
    assertTrue ( q.contains ( new Integer ( 1 ) ) ) ;
    assertTrue ( q.contains ( new Integer ( 2 ) ) ) ;
    assertTrue ( q.contains ( new Integer ( 3 ) ) ) ;
    assertTrue ( q.contains ( new Integer ( 4 ) ) ) ;
    assertTrue ( q.contains ( new Integer ( 5 ) ) ) ;
  }
  @Test public void fillingSynchronizedSequenceQueuePriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new SequenceQueuePriorityQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingAndEquality ( ) {
    PriorityQueue<Integer> a = new SortedListPriorityQueue<Integer> ( ) ;
    PriorityQueue<Integer> b = new HeapPriorityQueue<Integer> ( ) ;
    PriorityQueue<Integer> c = new SequenceQueuePriorityQueue<Integer> ( ) ;
    PriorityQueue<Integer> d = new PriorityQueue.Synchronized<Integer> ( new SortedListPriorityQueue<Integer> ( ) ) ;
    PriorityQueue<Integer> e = new PriorityQueue.Synchronized<Integer> ( new HeapPriorityQueue<Integer> ( ) ) ;
    PriorityQueue<Integer> f = new PriorityQueue.Synchronized<Integer> ( new SequenceQueuePriorityQueue<Integer> ( ) ) ;
    fillUp ( a ) ;
    fillUp ( b ) ;
    fillUp ( c ) ;
    fillUp ( d ) ;
    fillUp ( e ) ;
    fillUp ( f ) ;
    assertTrue ( a.equals ( a ) ) ;
    assertTrue ( a.equals ( b ) ) ;
    assertTrue ( a.equals ( c ) ) ;
    assertTrue ( a.equals ( d ) ) ;
    assertTrue ( a.equals ( e ) ) ;
    assertTrue ( a.equals ( f ) ) ;
    assertTrue ( b.equals ( a ) ) ;
    assertTrue ( b.equals ( b ) ) ;
    assertTrue ( b.equals ( c ) ) ;
    assertTrue ( b.equals ( d ) ) ;
    assertTrue ( b.equals ( e ) ) ;
    assertTrue ( b.equals ( f ) ) ;
    assertTrue ( c.equals ( a ) ) ;
    assertTrue ( c.equals ( b ) ) ;
    assertTrue ( c.equals ( c ) ) ;
    assertTrue ( c.equals ( d ) ) ;
    assertTrue ( c.equals ( e ) ) ;
    assertTrue ( c.equals ( f ) ) ;
    assertTrue ( d.equals ( a ) ) ;
    assertTrue ( d.equals ( b ) ) ;
    assertTrue ( d.equals ( c ) ) ;
    assertTrue ( d.equals ( d ) ) ;
    assertTrue ( d.equals ( e ) ) ;
    assertTrue ( d.equals ( f ) ) ;
    assertTrue ( e.equals ( a ) ) ;
    assertTrue ( e.equals ( b ) ) ;
    assertTrue ( e.equals ( c ) ) ;
    assertTrue ( e.equals ( d ) ) ;
    assertTrue ( e.equals ( e ) ) ;
    assertTrue ( e.equals ( f ) ) ;
    assertTrue ( f.equals ( a ) ) ;
    assertTrue ( f.equals ( b ) ) ;
    assertTrue ( f.equals ( c ) ) ;
    assertTrue ( f.equals ( d ) ) ;
    assertTrue ( f.equals ( e ) ) ;
    assertTrue ( f.equals ( f ) ) ;
  }
  @Test public void cloneSSListPriorityQueue ( ) {
    PriorityQueue<Integer> q = new SortedListPriorityQueue<Integer> ( ) ;
    fillUp ( q ) ;
    PriorityQueue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneHeapPriorityQueue ( ) {
    PriorityQueue<Integer> q = new HeapPriorityQueue<Integer> ( ) ;
    fillUp ( q ) ;
    PriorityQueue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneSequenceQueuePriorityQueue ( ) {
    PriorityQueue<Integer> q = new SequenceQueuePriorityQueue<Integer> ( ) ;
    fillUp ( q ) ;
    PriorityQueue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneSynchronizedSSListPriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new SortedListPriorityQueue<Integer> ( ) );
    fillUp ( q ) ;
    PriorityQueue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneSynchronizedHeapPriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new HeapPriorityQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    PriorityQueue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneSynchronizedSequenceQueuePriorityQueue ( ) {
    PriorityQueue<Integer> q = new PriorityQueue.Synchronized<Integer> ( new SequenceQueuePriorityQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    PriorityQueue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  /**
   *  Fill up a queue.
   */
  private void fillUp ( final PriorityQueue<Integer> s ) {
    s.add ( new Integer ( 1 ) ) ;
    s.add ( new Integer ( 2 ) ) ;
    s.add ( new Integer ( 3 ) ) ;
    s.add ( new Integer ( 4 ) ) ;
    s.add ( new Integer ( 5 ) ) ;
  }
}
