/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.tests ;

import org.devjavasoft.ads.Queue ;
import org.devjavasoft.ads.ArrayQueue ;
import org.devjavasoft.ads.SLListQueue ;
import org.devjavasoft.ads.DLListQueue ;

import org.testng.annotations.Test ;
import static org.testng.Assert.assertTrue ;

/**
 *  A unit test of the <code>Queue&lt;T></code>.
 *
 *  <p>Exercises some but not all of the methods in some but not all of the situations.  This is certainly
 *  not exhaustive.  Its main purpose is to test the clone method and the equality testing.</p>
 *
 *  @author Russel Winder
 */
public class Queue_Test {
  @Test public void emptySLListQueue ( ) {
    Queue<Integer> q = new SLListQueue<Integer> ( ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptyDLListQueue ( ) {
    Queue<Integer> q = new DLListQueue<Integer> ( ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptyArrayQueue ( ) {
    Queue<Integer> q = new ArrayQueue<Integer> ( ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptySynchronizedSLListQueue ( ) {
    Queue<Integer> q = new Queue.Synchronized<Integer> ( new SLListQueue<Integer> ( ) ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptySynchronizedDLListQueue ( ) {
    Queue<Integer> q = new Queue.Synchronized<Integer> ( new DLListQueue<Integer> ( ) ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  @Test public void emptySynchronizedArrayQueue ( ) {
    Queue<Integer> q = new Queue.Synchronized<Integer> ( new ArrayQueue<Integer> ( ) ) ;
    assertTrue ( q.isEmpty ( ) ) ;
  }
  private void runDestructiveAssertions ( final Queue<Integer> q ) {
    assertTrue ( q.peek ( ).equals ( new Integer ( 1 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 1 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 2 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 3 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 4 ) ) ) ;
    assertTrue ( q.remove ( ).equals ( new Integer ( 5 ) ) ) ;
  }
  @Test public void fillingSLListQueue ( ) {
    Queue<Integer> q = new SLListQueue<Integer> ( ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingDLListQueue ( ) {
    Queue<Integer> q = new DLListQueue<Integer> ( ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingArrayQueue ( ) {
    Queue<Integer> q = new ArrayQueue<Integer> ( ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingSynchronizedSLListQueue ( ) {
    Queue<Integer> q = new Queue.Synchronized<Integer> ( new SLListQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingSynchronizedDLListQueue ( ) {
    Queue<Integer> q = new Queue.Synchronized<Integer> ( new DLListQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingSynchronizedArrayQueue ( ) {
    Queue<Integer> q = new Queue.Synchronized<Integer> ( new ArrayQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    runDestructiveAssertions ( q ) ;
  }
  @Test public void fillingAndEquality ( ) {
    Queue<Integer> a = new SLListQueue<Integer> ( ) ;
    Queue<Integer> b = new DLListQueue<Integer> ( ) ;
    Queue<Integer> c = new ArrayQueue<Integer> ( ) ;
    Queue<Integer> d = new Queue.Synchronized<Integer> ( new SLListQueue<Integer> ( ) ) ;
    Queue<Integer> e = new Queue.Synchronized<Integer> ( new DLListQueue<Integer> ( ) ) ;
    Queue<Integer> f = new Queue.Synchronized<Integer> ( new ArrayQueue<Integer> ( ) ) ;
    fillUp ( a ) ;
    fillUp ( b ) ;
    fillUp ( c ) ;
    fillUp ( d ) ;
    fillUp ( e ) ;
    fillUp ( f ) ;
    assertTrue ( a.equals ( a ) ) ;
    assertTrue ( a.equals ( b ) ) ;
    assertTrue ( a.equals ( c ) ) ;
    assertTrue ( a.equals ( d ) ) ;
    assertTrue ( a.equals ( e ) ) ;
    assertTrue ( a.equals ( f ) ) ;
    assertTrue ( b.equals ( a ) ) ;
    assertTrue ( b.equals ( b ) ) ;
    assertTrue ( b.equals ( c ) ) ;
    assertTrue ( b.equals ( d ) ) ;
    assertTrue ( b.equals ( e ) ) ;
    assertTrue ( b.equals ( f ) ) ;
    assertTrue ( c.equals ( a ) ) ;
    assertTrue ( c.equals ( b ) ) ;
    assertTrue ( c.equals ( c ) ) ;
    assertTrue ( c.equals ( d ) ) ;
    assertTrue ( c.equals ( e ) ) ;
    assertTrue ( c.equals ( f ) ) ;
    assertTrue ( d.equals ( a ) ) ;
    assertTrue ( d.equals ( b ) ) ;
    assertTrue ( d.equals ( c ) ) ;
    assertTrue ( d.equals ( d ) ) ;
    assertTrue ( d.equals ( e ) ) ;
    assertTrue ( d.equals ( f ) ) ;
    assertTrue ( e.equals ( a ) ) ;
    assertTrue ( e.equals ( b ) ) ;
    assertTrue ( e.equals ( c ) ) ;
    assertTrue ( e.equals ( d ) ) ;
    assertTrue ( e.equals ( e ) ) ;
    assertTrue ( e.equals ( f ) ) ;
    assertTrue ( f.equals ( a ) ) ;
    assertTrue ( f.equals ( b ) ) ;
    assertTrue ( f.equals ( c ) ) ;
    assertTrue ( f.equals ( d ) ) ;
    assertTrue ( f.equals ( e ) ) ;
    assertTrue ( f.equals ( f ) ) ;
  }

  @Test public void cloneSSListQueue ( ) {
    Queue<Integer> q = new SLListQueue<Integer> ( ) ;
    fillUp ( q ) ;
    Queue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneDLListQueue ( ) {
    Queue<Integer> q = new DLListQueue<Integer> ( ) ;
    fillUp ( q ) ;
    Queue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneArrayQueue ( ) {
    Queue<Integer> q = new ArrayQueue<Integer> ( ) ;
    fillUp ( q ) ;
    Queue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneSynchronizedSSListQueue ( ) {
    Queue<Integer> q = new Queue.Synchronized<Integer> ( new SLListQueue<Integer> ( ) );
    fillUp ( q ) ;
    Queue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneSynchronizedDLListQueue ( ) {
      Queue<Integer> q = new Queue.Synchronized<Integer> ( new DLListQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    Queue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  @Test public void cloneSynchronizedArrayQueue ( ) {
      Queue<Integer> q = new Queue.Synchronized<Integer> ( new ArrayQueue<Integer> ( ) ) ;
    fillUp ( q ) ;
    Queue<Integer> c = q.clone ( ) ;
    assertTrue ( q.equals ( c ) ) ;
    assertTrue ( c.equals ( q ) ) ;
  }
  /**
   *  Fill up a queue.
   */
  private void fillUp ( final Queue<Integer> s ) {
    s.add ( new Integer ( 1 ) ) ;
    s.add ( new Integer ( 2 ) ) ;
    s.add ( new Integer ( 3 ) ) ;
    s.add ( new Integer ( 4 ) ) ;
    s.add ( new Integer ( 5 ) ) ;
  }
}
