/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.tests ;

import org.devjavasoft.ads.AccessEmptyContainerException ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.DLList ;
import org.devjavasoft.ads.ForwardIterator ;
import org.devjavasoft.ads.IndexOutOfBoundsException ;
import org.devjavasoft.ads.InvalidIteratorOperationException ;
import org.devjavasoft.ads.InputIterator ;
import org.devjavasoft.ads.SLList ;
import org.devjavasoft.ads.Sequence ;

import org.testng.annotations.BeforeMethod ;
import org.testng.annotations.Test ;
import static org.testng.Assert.assertTrue ;
import static org.testng.Assert.assertFalse ;

/**
 *  A class to test the class <code>SLList&lt;Integer></code>.
 *
 *  <p>It is more or less assumed that creation of an empty <code>SLList&lt;Integer></code>,
 *  <code>SLList<Integer>.add</code>, <code>SLList<Integer>.equals</code> and
 *  <code>SLList<Integer>.toString</code> work correctly.</p>
 *
 *  @author Russel Winder
 */
public class SLList_Test {
  protected SLList<Integer> empty ;
  protected SLList<Integer> nonEmptyOddSize ;
  protected SLList<Integer> nonEmptyEvenSize ;

  @BeforeMethod  public void setUp ( ) {
    empty = new SLList<Integer>  ( ) ;
    nonEmptyOddSize = new SLList<Integer>  ( ) ;
    fillOddSize ( nonEmptyOddSize ) ;
    nonEmptyEvenSize = new SLList<Integer>  ( ) ;
    fillEvenSize ( nonEmptyEvenSize ) ;
  }

  protected void fillEvenSize ( final Sequence<Integer> s ) {
    s.add ( 1 ) ;
    s.add ( 2 ) ;
    s.add ( 3 ) ;
    s.add ( 4 ) ;
  }
  protected void fillOddSize ( final Sequence<Integer> s ) {
    s.add ( 1 ) ;
    s.add ( 2 ) ;
    s.add ( 3 ) ;
    s.add ( 4 ) ;
    s.add ( 5 ) ;
  }
  protected Array<Integer> makeArrayOddSize ( ) {
    Array<Integer> value = new Array<Integer>  ( ) ;
    fillOddSize ( value ) ;
    return value ;
  }
  protected DLList<Integer> makeDLListOddSize ( ) {
    DLList<Integer> value = new DLList<Integer>  ( ) ;
    fillOddSize ( value ) ;
    return value ;
  }
  //  ====================================================
  //
  @Test public void add ( ) {
    String s = empty.toString ( ) ;
    assertTrue ( s.equals ( "[ ]" ) ) ;
    s = nonEmptyOddSize.toString ( ) ;
    assertTrue ( s.equals ( "[ 1 2 3 4 5 ]" ) ) ;
    s = nonEmptyEvenSize.toString ( ) ;
    assertTrue ( s.equals ( "[ 1 2 3 4 ]" ) ) ;
    fillOddSize(empty) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void capacity ( ) {
    int oldSize = nonEmptyOddSize.size ( ) ;
    for ( int i = 0 ; i < 100 ; ++i ) {
      nonEmptyOddSize.add ( i ) ;
    }
    assertTrue ( nonEmptyOddSize.size ( ) == ( 100 + oldSize ) ) ;
  }
  @Test public void cloneEmpty ( ) {
    SLList<Integer> clone = empty.clone ( ) ;
    assertTrue ( clone.size ( ) == empty.size ( ) ) ;
    assertTrue ( clone.equals ( empty ) ) ;
  }
  @Test public void cloneNonEmpty ( ) {
    SLList<Integer> clone = nonEmptyOddSize.clone ( ) ;
    assertTrue ( clone.size ( ) == nonEmptyOddSize.size ( ) ) ;
    assertTrue ( clone.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void containsEmpty ( ) {
    assertFalse ( empty.contains ( 1 ) ) ;
  }
  @Test public void containsNonEmpty ( ) {
    assertTrue ( nonEmptyOddSize.contains ( 1 ) ) ;
    assertTrue ( nonEmptyOddSize.contains ( 2 ) ) ;
    assertTrue ( nonEmptyOddSize.contains ( 3 ) ) ;
    assertTrue ( nonEmptyOddSize.contains ( 4 ) ) ;
    assertTrue ( nonEmptyOddSize.contains ( 5 ) ) ;
  }
  @Test public void copyConstructorsFromEmptyArray ( ) {
    Array<Integer> array = new Array<Integer>  ( ) ;
    SLList<Integer> copy = new SLList<Integer> ( array ) ;
    assertTrue ( copy.equals ( empty ) ) ;
    assertFalse ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromNonEmptyArray ( ) {
    Array<Integer> array = makeArrayOddSize ( ) ;
    SLList<Integer> copy = new SLList<Integer> ( array ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
    assertFalse ( copy.equals ( empty ) ) ;
  }
  @Test public void copyConstructorsFromArraySlice ( ) {
    Array<Integer> array = makeArrayOddSize ( ) ;
    SLList<Integer> copy = new SLList<Integer> ( array , 0 , array.size ( ) ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromArrayIterators ( ) {
    Array<Integer> array = makeArrayOddSize ( ) ;
    SLList<Integer> copy = new SLList<Integer> ( array.begin ( ) , array.end ( ) ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromEmptySLList ( ) {
    SLList<Integer> copy = new SLList<Integer> ( empty ) ;
    assertTrue ( copy.equals ( empty ) ) ;
    assertFalse ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromNonEmptySLList ( ) {
    SLList<Integer> copy = new SLList<Integer> ( nonEmptyOddSize ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
    assertFalse ( copy.equals ( empty ) ) ;
  }
  @Test public void copyConstructorsFromSLListSlice ( ) {
    SLList<Integer> copy = new SLList<Integer> ( nonEmptyOddSize , 0 , nonEmptyOddSize.size ( ) ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromSLListIterators ( ) {
    SLList<Integer> copy = new SLList<Integer> ( nonEmptyOddSize.begin ( ) , nonEmptyOddSize.end ( ) ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromEmptyDLList ( ) {
    DLList<Integer> list = new DLList<Integer>  ( ) ;
    SLList<Integer> copy = new SLList<Integer> ( list ) ;
    assertTrue ( copy.equals ( empty ) ) ;
    assertFalse ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromNonEmptyDLList ( ) {
    DLList<Integer> list = makeDLListOddSize ( ) ;
    SLList<Integer> copy = new SLList<Integer> ( list ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
    assertFalse ( copy.equals ( empty ) ) ;
  }
  @Test public void copyConstructorsFromDLListSlice ( ) {
    DLList<Integer> list = makeDLListOddSize ( ) ;
    SLList<Integer> copy = new SLList<Integer> ( list , 0 , list.size ( ) ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void copyConstructorsFromDLListIterators ( ) {
    DLList<Integer> list = makeDLListOddSize ( ) ;
    SLList<Integer> copy = new SLList<Integer> ( list.begin ( ) , list.end ( ) ) ;
    assertTrue ( copy.equals ( nonEmptyOddSize ) ) ;
  }
  @Test ( expectedExceptions = { IndexOutOfBoundsException.class } )
  public void get ( ) {
    Integer i = nonEmptyOddSize.get ( 0 ) ;
    assertTrue ( i == 1 ) ;
    i = nonEmptyOddSize.get ( 1 ) ;
    assertTrue ( i == 2 ) ;
    i = nonEmptyOddSize.get ( nonEmptyOddSize.size ( ) ) ;
    assertTrue ( i == 3 ) ;
  }
  @Test ( expectedExceptions = { IndexOutOfBoundsException.class } )
  public void insert ( ) {
    empty.insert ( 0 ,  5 ) ;
    empty.insert ( 0 ,  1 ) ;
    empty.insert ( 1 ,  2 ) ;
    empty.insert ( 2 ,  3 ) ;
    empty.insert ( 3 ,  4 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
    empty.insert ( 5 ,  6 ) ;
  }
  @Test public void iteratorGetFirst ( ) {
    InputIterator<Integer> i = nonEmptyOddSize.iterator ( ) ;
    assertTrue ( i.get ( ) == 1 ) ;
  }
  @Test public void iteratorGetMiddle ( ) {
    InputIterator<Integer> i = nonEmptyOddSize.iterator ( ) ;
    i.advance ( ) ;
    assertTrue ( i.get ( ) == 2 ) ;
  }
  @Test public void iteratorSetFirst ( ) {
    ForwardIterator<Integer> i = nonEmptyOddSize.iterator ( ) ;
    i.set ( 3 ) ;
    empty.add ( 3 ) ;
    empty.add ( 2 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void iteratorSetMiddle ( ) {
    ForwardIterator<Integer> i = nonEmptyOddSize.iterator ( ) ;
    i.advance ( ) ;
    i.set ( 3 ) ;
    empty.add ( 1 ) ;
    empty.add ( 3 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void makeEmptyFromEmpty ( ) {
    empty.makeEmpty ( ) ;
    assertTrue ( empty.isEmpty ( ) ) ;
  }
  @Test public void makeEmptyFromNonEmpty ( ) {
    nonEmptyOddSize.makeEmpty ( ) ;
    assertTrue ( nonEmptyOddSize.isEmpty ( ) ) ;
  }
  @Test ( expectedExceptions = { AccessEmptyContainerException.class } )
  public void removeByIndexFromEmpty ( ) {
     empty.remove( 0 ) ;
  }
  @Test public void removeByIndexFromIndexZero ( ) {
    nonEmptyOddSize.remove ( 0 ) ;
    assertFalse ( nonEmptyOddSize.contains ( 1 ) ) ;
    empty.add ( 2 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void removeByIndexFromMiddle ( ) {
    nonEmptyOddSize.remove ( 1 ) ;
    assertFalse ( nonEmptyOddSize.contains ( 2 ) ) ;
    empty.add ( 1 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void removeByIndexFromEnd ( ) {
    nonEmptyOddSize.remove ( 2 ) ;
    assertFalse ( nonEmptyOddSize.contains ( 3 ) ) ;
    empty.add ( 1 ) ;
    empty.add ( 2 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test ( expectedExceptions = { IndexOutOfBoundsException.class } )
  public void removeByIndexWithIllegalIndex ( ) {
    nonEmptyOddSize.remove ( 5 ) ;
  }
  @Test ( expectedExceptions = { AccessEmptyContainerException.class } )
  public void removeByValueFromEmpty ( ) {
     empty.remove ( new Integer ( 1 ) ) ;
  }
  @Test public void removeByValueFromIndexZero ( ) {
    nonEmptyOddSize.remove ( new Integer ( 1 ) ) ;
    assertFalse ( nonEmptyOddSize.contains ( 1 ) ) ;
    empty.add ( 2 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void removeByValueFromMiddle ( ) {
    nonEmptyOddSize.remove ( new Integer ( 2 ) ) ;
    assertFalse ( nonEmptyOddSize.contains ( 2 ) ) ;
    empty.add ( 1 ) ;
    empty.add ( 3 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void removeByValueFromEnd ( ) {
    nonEmptyOddSize.remove ( new Integer ( 3 ) ) ;
    assertFalse ( nonEmptyOddSize.contains ( 3 ) ) ;
    empty.add ( 1 ) ;
    empty.add ( 2 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void removeByValueWithValueNotContained ( ) {
    SLList<Integer> clone = nonEmptyOddSize.clone ( ) ;
    clone.remove ( new Integer ( 6 ) ) ;
    assertTrue ( clone.equals ( nonEmptyOddSize ) ) ;
  }
  @Test ( expectedExceptions = { InvalidIteratorOperationException.class } )
  public void reverseEmpty ( ) {
    empty.reverse ( ) ;
  }
  @Test public void reverseEvenSize ( ) {
    SLList<Integer> clone = nonEmptyEvenSize.clone ( ) ;
    clone.reverse ( ) ;
    SLList<Integer> reversed = new SLList<Integer>  ( ) ;
    reversed.add ( 4 ) ;
    reversed.add ( 3 ) ;
    reversed.add ( 2 ) ;
    reversed.add ( 1 ) ;
    assertTrue ( reversed.equals ( clone ) ) ;
    clone.reverse ( ) ;
    assertTrue ( clone.equals ( nonEmptyEvenSize ) ) ;
  }
  @Test public void reverseOddSize ( ) {
    SLList<Integer> clone = nonEmptyOddSize.clone ( ) ;
    clone.reverse ( ) ;
    SLList<Integer> reversed = new SLList<Integer>  ( ) ;
    reversed.add ( 5 ) ;
    reversed.add ( 4 ) ;
    reversed.add ( 3 ) ;
    reversed.add ( 2 ) ;
    reversed.add ( 1 ) ;
    assertTrue ( reversed.equals ( clone ) ) ;
    clone.reverse ( ) ;
    assertTrue ( clone.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void set ( ) {
    empty.add ( 2 ) ;
    empty.add ( 2 ) ;
    empty.add ( 2 ) ;
    empty.add ( 4 ) ;
    empty.add ( 5 ) ;
    nonEmptyOddSize.set ( 0 ,  2 ) ;
    nonEmptyOddSize.set ( 2 ,  2 ) ;
    assertTrue ( empty.equals ( nonEmptyOddSize ) ) ;
  }
  @Test public void toPrimitiveArrayClass ( ) {
    Integer[] x = nonEmptyEvenSize.toPrimitiveArray ( Integer.class ) ;
    for ( int i = 0 ; i < x.length ; ++i ) {
      assertTrue ( x[i] == ( i + 1 ) ) ;
    }
  }
  @Test public void toPrimitiveArrayArray ( ) {
    Integer[] x = new Integer [nonEmptyEvenSize.size ( )] ;
    Integer[] y = nonEmptyEvenSize.toPrimitiveArray ( x ) ;
    assertTrue ( x== y ) ;
    for ( int i = 0 ; i < x.length ; ++i ) {
      assertTrue ( x[i] == ( i + 1 ) ) ;
    }
  }
}
