/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.tests ;

import org.devjavasoft.ads.Stack ;
import org.devjavasoft.ads.SequenceStack ;
import org.devjavasoft.ads.PrimitiveArrayStack ;
import org.devjavasoft.ads.Array ;
import org.devjavasoft.ads.SLList ;
import org.devjavasoft.ads.DLList ;

import org.testng.annotations.BeforeMethod ;
import org.testng.annotations.Test ;
import static org.testng.Assert.assertTrue ;

/**
 *  A unit test of the <code>Stack&lt;T></code>.
 *
 *  Exercises some but not all of the methods in some but not all of the situations.  This is certainly not
 *  exhaustive.  Its main purpose is to test the clone method and the equality testing.
 *
 *  @author Russel Winder
 */
public class Stack_Test {
  /**
   *  A synchronized <code>Stack&lt;Integer></code> with default representation.
   */
  Stack<Integer> a ;
  /**
   *  A non-synchronized <code>Stack&lt;Integer></code> with <code>Array&lt;Integer></code> representation.
   */
  Stack<Integer> b ;
  /**
   *  A non-synchronized <code>Stack&lt;Integer></code>with <code>SLList&lt;Integer></code> representation.
   */
  Stack<Integer> c ;
  /**
   *  A non-synchronized <code>Stack&lt;Integer></code> with <code>DLList&lt;Integer></code> representation.
   */
  Stack<Integer> d ;
  /**
   *  A synchronized <code>Stack&lt;Integer></code> with default representation.
   */
  Stack<Integer> e ;
  /**
   *   The setup method which gets executed prior to any test method and must set up any variables needed by
   *   the tests.
   *
   *   <p>We are assuming that <code>Stack&lt;Integer>.add</code> completes execution even if it is not
   *   actually correct.</p>
   */
  @BeforeMethod  public void setUp ( ) {
    a = new Stack.Synchronized<Integer> ( new SequenceStack<Integer>  ( ) ) ;
    b = new SequenceStack<Integer> ( new Array<Integer>  ( ) ) ;
    c = new SequenceStack<Integer> ( new SLList<Integer>  ( ) ) ;
    d = new SequenceStack<Integer> ( new DLList<Integer>  ( ) ) ;
    e = new Stack.Synchronized<Integer> ( new PrimitiveArrayStack<Integer>  ( ) ) ;
  }
  /**
   *  Ensure the <code>setUp</code> method creates empty <code>Stack&lt;Integer></code>.
   */
  @Test public void empty ( ) {
    assertTrue ( a.isEmpty ( ) ) ;
    assertTrue ( b.isEmpty ( ) ) ;
    assertTrue ( c.isEmpty ( ) ) ;
    assertTrue ( d.isEmpty ( ) ) ;
    assertTrue ( e.isEmpty ( ) ) ;
  }
  /**
   *  Test that all types of stack fill the same.
   */
  @Test public void filling ( ){
    fillUp ( a ) ;
    fillUp ( b ) ;
    fillUp ( c ) ;
    fillUp ( d ) ;
    fillUp ( e ) ;
    assertTrue ( a.peek ( ).equals ( new Integer ( 5 ) ) ) ;
    assertTrue ( b.peek ( ).equals ( new Integer ( 5 ) ) ) ;
    assertTrue ( c.peek ( ).equals ( new Integer ( 5 ) ) ) ;
    assertTrue ( d.peek ( ).equals ( new Integer ( 5 ) ) ) ;
    assertTrue ( e.peek ( ).equals ( new Integer ( 5) ) ) ;
    assertTrue ( a.equals ( a ) ) ;
    assertTrue ( a.equals ( b ) ) ;
    assertTrue ( a.equals ( c ) ) ;
    assertTrue ( a.equals ( d ) ) ;
    assertTrue ( a.equals ( e ) ) ;
    assertTrue ( b.equals ( a ) ) ;
    assertTrue ( b.equals ( b ) ) ;
    assertTrue ( b.equals ( c ) ) ;
    assertTrue ( b.equals ( d ) ) ;
    assertTrue ( b.equals ( e ) ) ;
    assertTrue ( c.equals ( a ) ) ;
    assertTrue ( c.equals ( b ) ) ;
    assertTrue ( c.equals ( c ) ) ;
    assertTrue ( c.equals ( d ) ) ;
    assertTrue ( c.equals ( e ) ) ;
    assertTrue ( d.equals ( a ) ) ;
    assertTrue ( d.equals ( b ) ) ;
    assertTrue ( d.equals ( c ) ) ;
    assertTrue ( d.equals ( d ) ) ;
    assertTrue ( d.equals ( e ) ) ;
    assertTrue ( e.equals ( a ) ) ;
    assertTrue ( e.equals ( b ) ) ;
    assertTrue ( e.equals ( c ) ) ;
    assertTrue ( e.equals ( d ) ) ;
    assertTrue ( e.equals ( e ) ) ;
  }
  /**
   *  Test that all types of stack clone the same.  Cannot call this <code>clone</code> of course.
   */
  @Test public void testClone ( ) {
    fillUp ( a ) ;
    fillUp ( b ) ;
    fillUp ( c ) ;
    fillUp ( d ) ;
    fillUp ( e ) ;
    Stack<Integer> aa = a.clone ( ) ;
    Stack<Integer> bb = b.clone ( ) ;
    Stack<Integer> cc = c.clone ( ) ;
    Stack<Integer> dd = d.clone ( ) ;
    Stack<Integer> ee = e.clone ( ) ;
    assertTrue ( aa.equals ( a ) ) ;
    assertTrue ( bb.equals ( b ) ) ;
    assertTrue ( cc.equals ( c ) ) ;
    assertTrue ( dd.equals ( d ) ) ;
    assertTrue ( ee.equals ( e ) ) ;
    assertTrue ( a.equals ( aa ) ) ;
    assertTrue ( b.equals ( bb ) ) ;
    assertTrue ( c.equals ( cc ) ) ;
    assertTrue ( d.equals ( dd ) ) ;
    assertTrue ( e.equals ( ee ) ) ;
    assertTrue ( aa.peek ( ).equals ( new Integer ( 5 ) ) ) ;
    assertTrue ( bb.peek ( ).equals ( new Integer ( 5 ) ) ) ;
    assertTrue ( cc.peek ( ).equals ( new Integer ( 5 ) ) ) ;
    assertTrue ( dd.peek ( ).equals ( new Integer ( 5 ) ) ) ;
    assertTrue ( ee.peek ( ).equals ( new Integer ( 5 ) ) ) ;
  }
  /**
   *  Fill up a stack.
   */
  private void fillUp ( final Stack<Integer> s ) {
    s.push ( new Integer ( 1 ) ) ;
    s.push ( new Integer ( 2 ) ) ;
    s.push ( new Integer ( 3 ) ) ;
    s.push ( new Integer ( 4 ) ) ;
    s.push ( new Integer ( 5 ) ) ;
  }
}
