/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2010 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads.tests.filesort ;

import java.io.File ;
import java.io.FileNotFoundException ;
import java.io.IOException ;
import java.io.PrintWriter ;
import java.util.Arrays ;
import java.util.ArrayList ;
import java.util.Scanner ;

import org.devjavasoft.ads.sorting.BalancedMultiwayMergeSort ;
import org.devjavasoft.ads.sorting.PolyphaseMergeSort ;

import org.testng.annotations.BeforeMethod ;
import org.testng.annotations.AfterSuite ;
import org.testng.annotations.Test ;
import static org.testng.Assert.assertTrue ;
import static org.testng.Assert.fail ;

public class FileSort_Test {
  private final Integer[] unsorted = { 123 , 23 , 34 , 56 , 98 , 6 , 5 , 7 , 1234 , 5 , 64 , 876 , 54 , 65 ,
                                       34 , 65 , 87 , 34 , 56 , 234 , 876 , 777 , 6 , 5 , 4 , 5 , 56556 , 45 ,
                                       4 , 5 , 23 , 444 , 65 , 783 , 63 , 8343 , 63 , 36 , 3788 , 37 , 3 , 9 ,
                                       8 , 56 , 4 , 6 , 3 , 7 , 36 , 84 , 36 , 7 , 4 , 643 , 7 , 36 , 7 , 367 ,
                                       73 , 3 , 3 , 7 , 8 , 4 , 446 , 437 , 7 , 437 , 6 , 756 , 8 , 8658 , 58 ,
                                       586 , 8 , 856 , 8 , 548 , 9 , 49 , 33 , 48 , 123 , 23 , 34 , 56 , 98 ,
                                       6 , 5 , 7 , 1234 , 5 , 64 , 876 , 54 , 65 , 34 ,  65 , 87 , 34 , 56 ,
                                       234 , 876 , 777 , 6 , 5 , 4 , 5 , 56556 , 45 , 4 , 5 , 23 , 444 , 65 ,
                                       783 , 63 , 8343 , 63 , 36 , 3788 , 37 , 3 , 9 , 8 , 56 , 4 , 6 , 3 , 7 ,
                                       36 , 84 , 36 , 7 , 4 , 643 , 7 , 36 , 7 , 367 , 73 , 3 , 3 , 7 , 8 , 4 ,
                                       446 , 437 , 7 , 437 , 6 , 756 , 8 , 8658 , 58 , 586 , 8 , 856 , 8 ,
                                       548 , 9 , 49 , 33 , 48 } ;
  private final String filename = "__temporary__.txt" ;
  private final File file = new File ( filename ) ;
  /**
   *  Construct the ASCII encoded data file that is to be sorted.  Do this prior to every test so that the
   *  same uinsorted fiel is used for all tests.  Opening an existing file truncates the file to zero length
   *  which is the necessary behaviour.
   *
   *  @throws FileNotFoundException if the file cannot be created.
   */
  @BeforeMethod public void writeDataFile ( ) throws FileNotFoundException {
    final PrintWriter output = new PrintWriter ( filename ) ;
    for ( Integer i : unsorted ) { output.println ( i ) ; }
    output.close ( ) ;
  }
  @AfterSuite public void removeTemporaryFile ( ) { file.delete ( ) ; }
  @Test ( expectedExceptions = { FileNotFoundException.class } )
  public void balancedMultiWayMergeSortFNFE ( ) throws FileNotFoundException , IOException {
    BalancedMultiwayMergeSort.sort ( "blah" , 10 , 2 , new TestRecordInformation ( new TestRecordComparator ( ) , new TestRecordEqualToComparator ( ) ) ) ;
  }
  @Test public void balancedMultiWayMergeSort ( ) throws FileNotFoundException , IOException {
    BalancedMultiwayMergeSort.sort ( filename , 10 , 2 , new TestRecordInformation ( new TestRecordComparator ( ) , new TestRecordEqualToComparator ( ) ) ) ;
    final Integer[] sorted = unsorted.clone ( ) ;
    Arrays.sort ( sorted ) ;
    assertTrue ( Arrays.equals ( sorted , getData ( filename ) ) ) ;
  }
  @Test ( expectedExceptions = { FileNotFoundException.class } )
  public void polyphaseMergeSortFNFE ( ) throws FileNotFoundException , IOException {
    PolyphaseMergeSort.sort ( "blah" , 10 , 2 , new TestRecordInformation ( new TestRecordComparator ( ) , new TestRecordEqualToComparator ( ) ) ) ;
  }
  @Test public void polyphaseMergeSort ( ) throws FileNotFoundException , IOException {
    PolyphaseMergeSort.sort ( filename , 10 , 2 , new TestRecordInformation ( new TestRecordComparator ( ) , new TestRecordEqualToComparator ( ) ) ) ;
    final Integer[] sorted = unsorted.clone ( ) ;
    Arrays.sort ( sorted ) ;
    assertTrue ( Arrays.equals ( sorted , getData ( filename ) ) ) ;
  }
  /**
   *  Read in all the integer data from the named file and return an array of the values read.
   *
   *  @param filename name of the file to read.
   *  @return the array of integer values read.
   */
  private Integer[] getData ( final String f ) {
    Scanner input = null ;
    try { input = new Scanner ( new File ( f ) ) ; }
    catch ( FileNotFoundException fnfe ) { fail ( "File " + f + " could not be opened." ) ; }
    final ArrayList<Integer> buffer = new ArrayList<Integer> ( ) ;
    assert input != null ;
    while ( input.hasNextInt ( ) ) { buffer.add ( input.nextInt ( ) ) ; }
    return buffer.toArray ( new Integer[0] ) ;
  }
}
