/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2011 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;

/**
 *  A mapping implemented using an <code>ChainedHashTable&lt;K, C></code>.
 *
 *  @author Russel Winder
 */
public class ChainedHashTableRelation<D, R> extends AbstractRelation<D, R> {
  /**
   *  The <code>ChainedHashTable&lt;D, C></code> that acts as our storage.
   */
  protected ChainedHashTable<D, R> storage ;
  /**
   *  Construct an <code>ChainedHashTableRelation&lt;D, C></code>.
   */
  public ChainedHashTableRelation ( ) { storage = new ChainedHashTable<D, R> ( ) ; }
  @Override public void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public int size ( ) { return storage.size ( ) ; }
  @Override public void add ( final Pair<D, R> o ) { storage.add ( o ) ; }
  @Override public boolean remove ( final Pair<D, R> o) { return storage.remove ( o ) ; }
  @Override public ChainedHashTableRelation<D, R> clone ( ) {
    final ChainedHashTableRelation<D, R> t = (ChainedHashTableRelation<D, R>) super.clone ( ) ;
    if ( ! isEmpty ( ) ) { t.storage = storage.clone ( ) ; }
    return t ;
  }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public Iterator iterator ( ) { return new Iterator ( ) ; }
  /**
   *  Define the iterator over <code>ChainedHashTableRelation&lt;D, C></code>s.
   *
   *  @author Russel Winder
   */
  public class Iterator extends LocalIterator<Pair<D, R>> {
  	@Override public Pair<D, R> get ( ) { return cursor.get ( ) ; }
		@Override public ChainedHashTableRelation<D, R> getContainer ( ) { return ChainedHashTableRelation.this ; }
  }
  /**
   *  A class to support creating iterator types.  The standard iterator (which iterates over
   *  <code>Pair</code>s) and the domain and range iterators all have the same basic structure so we use
   *  this class as a superclass for all of them.
   *
   *  @author Russel Winder
   */
  public abstract class LocalIterator<T> extends AbstractIterator<T> {
    /**
     *  The cursor is just an iterator over the <code>ChainedHashTableRelation&lt;D, C></code>. Use the
     *  default of in-order traversal.
     */
    protected InputIterator<Pair<D, R>> cursor = storage.iterator ( ) ;
    @Override public abstract T get ( ) ;
    @Override public void advance ( ) { cursor.advance ( ) ; }
    @Override public void advance ( final int increment ) { cursor.advance ( increment ) ; }
    @Override public boolean atBegin ( ) { return cursor.atBegin ( ) ; }
    @Override public boolean atEnd ( ) { return cursor.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
    	if ( this == o ) { return true ; }
    	if ( ( o == null ) || ! ( o instanceof LocalIterator ) ) { return false ; }
    	return cursor.equals( ( (LocalIterator<T>) o ).cursor ) ;
    }
    @Override public Container<T> getContainer ( ) { throw new InvalidIteratorOperationException ( ) ; }
    @Override public LocalIterator<T> clone ( ) {
      final LocalIterator<T> i = (LocalIterator<T>) super.clone ( ) ;
      i.cursor = cursor.clone ( ) ;
      return i ;
    }
  }
  @Override public InputIterator<D> domain ( ) {
    return new LocalIterator<D> ( ) { @Override public D get ( ) { return cursor.get ( ).getA ( ) ; } } ;
  }
  // TODO: Eclipse marks this warning suppression as unneeded, but javac complains.
  @SuppressWarnings ( "unchecked" )
  @Override public InputIterator<D> domain ( final R r ) {
    //  Use a local class as we have to initialize it and so need a constructor so we have to name it,
    //  cannot use an anonymous class.
    class DomainIterator extends LocalIterator<D> {
      public DomainIterator ( ) { if ( ! cursor.get ( ).getB ( ).equals ( r ) ) { moveOn ( ) ; } }
      @Override public D get ( ) { return cursor.get ( ).getA ( ) ; }
      @Override public void advance ( ) { moveOn ( ) ; }
      private void moveOn ( ) {
        while ( ! atEnd ( ) ) {
          super.advance ( ) ;
          if ( atEnd ( ) || cursor.get ( ).getB ( ).equals ( r ) ) { break ; }
        }
      }
    }
    return new DomainIterator ( ) ;
  }
  @Override public InputIterator<R> range ( ) {
    return new LocalIterator<R> ( ) { @Override public R get() { return cursor.get ( ).getB ( ) ; } } ;
  }
  // TODO: Eclipse marks this warning suppression as unneeded but javac complains.
  @SuppressWarnings ( "unchecked" )
  @Override public InputIterator<R> range ( final D d ) {
    class RangeIterator extends LocalIterator<R> {
      public RangeIterator ( ) { if ( ! cursor.get ( ).getA ( ).equals ( d ) ) { moveOn ( ) ; } }
      @Override public R get ( ) { return cursor.get ( ).getB ( ) ; }
      @Override public void advance ( ) { moveOn ( ) ; }
      private void moveOn ( ) {
        while ( ! atEnd ( ) ) {
          super.advance ( ) ;
          if ( atEnd ( ) || cursor.get ( ).getA ( ).equals ( d ) ) { break ; }
        }
      }
    }
    return new RangeIterator ( ) ;
  }
}
