/*
 *  ADS – A Java package of algorithms and data structures for educational purposes.  This software was
 *  originally created for the book Winder & Roberts, "Developing Java Software", Wiley.
 *
 *  Copyright © 1997–2008,2011 Russel Winder
 *
 *  This file is part of ADS.
 *
 *  ADS is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
 *  Public License as published by the Free Software Foundation, either version 3 of the License, or (at
 *  your option) any later version.
 *
 *  ADS is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License along with ADS.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */
package org.devjavasoft.ads ;
/**
 *  A mapping implemented using an <code>OpenHashTable&lt;D, R></code>.
 *
 *  @author Russel Winder
 */
public class OpenHashTableRelation<D extends Comparable<D>, R> extends AbstractRelation<D, R> {
  /**
   *  The <code>OpenHashTable</code> that acts as our storage.
   */
  protected OpenHashTable<D, R> storage ;
  /**
   *  Construct an <code>OpenHashTableRelation&lt;D, R></code>.
   */
  public OpenHashTableRelation ( ) { storage = new OpenHashTable<D, R> ( ) ; }
  @Override public void makeEmpty ( ) { storage.makeEmpty ( ) ; }
  @Override public int size ( ) { return storage.size ( ) ; }
  @Override public void add ( final Pair<D, R> o ) { storage.add ( o ) ; }
  @Override public boolean remove ( final Pair<D, R> o ) { return storage.remove ( o ) ; }
  @Override public OpenHashTableRelation<D, R> clone ( ) {
    final OpenHashTableRelation<D, R> t = (OpenHashTableRelation<D, R>) super.clone ( ) ;
    if ( ! isEmpty ( ) ) { t.storage = storage.clone ( ) ; }
    return t ;
  }
  @Override public boolean contains ( final Pair<D, R> o ) { return storage.contains ( o ) ; }
  @Override public final Iterator begin ( ) { return new Iterator ( ) ; }
  @Override public final Iterator end ( ) {
    final Iterator i = new Iterator ( ) ;
    i.advance ( size ( ) - 1 ) ;
    return i ;
  }
  @Override public Iterator iterator ( ) { return new Iterator ( ) ; }
 	/**
   *  Define the iterator over <code>OpenHashTableRelation</code>s.
   *
   *  @author Russel Winder
   */
  public class Iterator extends LocalIterator<Pair<D, R>> {
  	@Override public Pair<D, R> get ( ) { return cursor.get ( ) ; }
		@Override public OpenHashTableRelation<D, R> getContainer ( ) { return OpenHashTableRelation.this ; }
  }
  /**
   *  A class to support creating iterator types.  The standard iterator (which iterates over
   *  <code>Pair</code>s) and the domain and range iterators all have the same basic structure so we use
   *  this class as a superclass for all of them.
   *
   *  @author Russel Winder
   */
  private abstract class LocalIterator<T> extends AbstractIterator<T> {
    /**
     *  The cursor is just an iterator over the <code>RelationTree</code>. Use the default of
     *  in-order traversal.
     */
    protected InputIterator<Pair<D, R>> cursor = storage.iterator ( ) ;
    @Override public abstract T get ( ) ;
    @Override public void advance ( ) { cursor.advance ( ) ; }
    @Override public void advance ( final int increment ) { cursor.advance ( increment ) ; }
    @Override public boolean atBegin ( ) { return cursor.atBegin ( ) ; }
    @Override public boolean atEnd ( ) { return cursor.atEnd ( ) ; }
    @SuppressWarnings ( "unchecked" )
    @Override public boolean equals ( final Object o ) {
      if ( this == o ) { return true ; }
      if ( ( o == null ) || ! ( o instanceof LocalIterator ) ) { return false ; }
      return cursor.equals ( ( (LocalIterator<T>) o ).cursor ) ;
    }
     //  When the type parameter T is consistent with ComparablePair<D, R> this method is type correct.  If
     //  the T has any other value then we cannot return the correct type.  We could without generics but
     //  this wasn't really type correct though in the context of typeless containers it was fine.  The
     //  upshot is that we now enforce type correctness and so we throw an exception here and let the
     //  sub-classes do the right thing if they can.</p>
    @Override public  Container<T> getContainer ( ) { throw new InvalidIteratorOperationException ( ) ; }
    @Override public final LocalIterator<T> clone ( ) {
      final LocalIterator<T> i = (LocalIterator<T>) super.clone ( ) ;
      i.cursor = cursor.clone ( ) ;
      return i ;
    }
  }
  @Override public InputIterator<D> domain ( ) {
    return new LocalIterator<D> ( ) { @Override public D get ( ) { return cursor.get ( ).getA ( ) ; } } ;
  }
  // TODO: Eclipse marks this warning suppression as unneeded, but javac complains.
  @SuppressWarnings ( "unchecked" )
  @Override public InputIterator<D> domain ( final R r ) {
    class DomainIterator extends LocalIterator<D> {
      public DomainIterator ( ) { if ( ! cursor.get ( ).getB ( ).equals ( r ) ) { moveOn ( ) ; } }
      @Override public D get ( ) { return cursor.get ( ).getA ( ) ; }
      @Override public void advance ( ) { moveOn ( ) ; }
      private void moveOn ( ) {
        while ( ! atEnd ( ) ) {
          super.advance ( ) ;
          if ( atEnd ( ) || cursor.get ( ).getB ( ).equals ( r ) ) { break ; }
        }
      }
    }
    return new DomainIterator ( )  ;
  }
  @Override public InputIterator<R> range ( ) {
    return new LocalIterator<R> ( ) { @Override public R get ( ) { return cursor.get ( ).getB ( ) ; } } ;
  }
  // TODO: Eclipse marks this warning suppression as unneeded, but javac complains.
  @SuppressWarnings ( "unchecked" )
  @Override public InputIterator<R> range ( final D d) {
    class RangeIterator extends LocalIterator<R> {
      public RangeIterator ( ) { if (! cursor.get ( ).getA ( ).equals ( d ) ) { moveOn ( ) ; } }
      @Override public R get ( ) { return cursor.get ( ).getB ( ) ; }
      @Override public void advance ( ) { moveOn ( ) ; }
      private void moveOn ( ) {
        while ( ! atEnd ( ) ) {
          super.advance ( ) ;
          if ( atEnd ( ) || cursor.get ( ).getA ( ).equals ( d ) ) { break ; }
        }
      }
    }
    return new RangeIterator ( ) ;
  }
}
